/*
bhmsqq.cpp - an implementation of a template class
Copyright (C) 2001  Al Kingsley


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos version 0.9.4, Copyright (C) 2001 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY, or MASONARY DRILL BITS. 
	The author takes no responsibility for anything at all, even the heat 
	death of the Universe.
    
*/
#include <pthread.h>
#include "msgq.h"

template<class BhPosMsg>
CMessageQueue<BhPosMsg>::CMessageQueue()
{
	int ret;
	
	bFull=false;
	bReorderArray=false;
	bClosedForSeason=false;
	bHasMsg=false;
	Error=0;
	Count=0;

	ret=pthread_mutex_init(&busy, NULL);
	if(ret!=0){
		printf("mutex failed to initialize\n");
		Error=ret;
		return;
	}
	
	ret=pthread_cond_init(&cond, NULL);
	if(ret!=0){
		printf("cond failed to initialize\n");
		Error=ret;
		return;
	}
	
}
template <class BhPosMsg>
CMessageQueue<BhPosMsg>::~CMessageQueue()
{
	if(!MsgQueue.empty())
		RemoveAll();
		
	pthread_mutex_destroy(&busy);
	pthread_cond_destroy(&cond);
}

template <class BhPosMsg>
int CMessageQueue<BhPosMsg>::GetError() 
{ 
   return Error; 
}

template <class BhPosMsg>
void CMessageQueue<BhPosMsg>::GoToPub()
{ 
   bClosedForSeason=true; 
}

template <class BhPosMsg>
bool CMessageQueue<BhPosMsg>::HasMsg()
{ 
   return bHasMsg; 
}

template <class BhPosMsg>
int CMessageQueue<BhPosMsg>::WaitOnMsg()
{
   if(bHasMsg) return 0;
	return pthread_cond_wait(&cond, &busy);
}

template <class BhPosMsg>
int CMessageQueue<BhPosMsg>::WaitOnMsgTO(const timespec *to)
{
	return pthread_cond_timedwait(&cond, &busy, to);
}

template <class BhPosMsg>
int CMessageQueue<BhPosMsg>::CanGrab()
{
	return pthread_mutex_trylock(&busy);
}

template <class BhPosMsg>
int CMessageQueue<BhPosMsg>::GrabOnMsg()
{
   if(bHasMsg) return 0;
	return pthread_mutex_lock(&busy);
}

template <class BhPosMsg>
int CMessageQueue<BhPosMsg>::Grab()
{
	return pthread_mutex_lock(&busy);
}
template <class BhPosMsg>
int CMessageQueue<BhPosMsg>::Release()
{
	return pthread_mutex_unlock(&busy);
}
template <class BhPosMsg>
bool CMessageQueue<BhPosMsg>::IsValidMsg(unsigned int n)
{
	if(n < 0 || n > MsgQueue.count()) return false;
	return true;
}

template <class BhPosMsg>
int CMessageQueue<BhPosMsg>::Add(BhPosMsg *arg)
{
   pthread_mutex_lock(&busy);
   MsgQueue.push_back(*arg);
	pthread_mutex_unlock(&busy);
	
	return SetHasMsgFlag(true);
}

template <class BhPosMsg>
int CMessageQueue<BhPosMsg>::AddUnique(BhPosMsg *arg)
{
	/*
   pthread_mutex_lock(&busy);
	std::list<BhPosMsg>::iterator fi;
	for(fi=MsgQueue.begin(); fi!=MsgQueue.end(); fi++){
		if(memcmp(fi, arg, sizeof(arg))==0) return 1;
	}
	MsgQueue.push_back(*arg);
	pthread_mutex_unlock(&busy);
	return SetHasMsgFlag(true);
	*/
	return 0;
}

template <class BhPosMsg>
pthread_cond_t *CMessageQueue<BhPosMsg>::GetCondition()
{
	return &cond;
}

template <class BhPosMsg>
int CMessageQueue<BhPosMsg>::SetHasMsgFlag(bool b)
{
	if(b){
	   if(!bHasMsg){
   		bHasMsg=true;
   	}
   	return pthread_cond_signal(&cond);
	}
	else
		bHasMsg=b;
	return 0;	 
}

template <class BhPosMsg>
int CMessageQueue<BhPosMsg>::Insert(BhPosMsg *arg, unsigned int pos)
{
   pthread_mutex_lock(&busy);
   
   int ret=0;
  
	MsgQueue.insert(MsgQueue.begin()+pos, arg);
	ret=SetHasMsgFlag(true);
	pthread_mutex_unlock(&busy);
	return ret;
}

template <class BhPosMsg>
bool CMessageQueue<BhPosMsg>::GetAt(unsigned int pos, BhPosMsg *d)
{
	//if(!IsMsgValid(pos))return true;

	BhPosMsg msg=MsgQueue.at(pos);
	memcpy(d, &msg, sizeof(BhPosMsg));
	LastMsg=pos;
	return false;
}
template <class BhPosMsg>
bool CMessageQueue<BhPosMsg>::Peek(unsigned int pos, void *d, int len)
{
	//if(!IsMsgValid(pos)) return true;
	BhPosMsg msg=MsgQueue.at(pos);
	memcpy(d, &msg, len);
	return false;
}
template <class BhPosMsg>
bool CMessageQueue<BhPosMsg>::Peek(unsigned int pos, BhPosMsg *d)
{
	//if(!IsMsgValid(pos)) return true;
	
	BhPosMsg msg=MsgQueue.at(pos);
	memcpy(d, &msg, sizeof(BhPosMsg));
	return false;
}
template <class BhPosMsg>
unsigned int CMessageQueue<BhPosMsg>::GetCount()
{
	unsigned int l = MsgQueue.size();
	if(l==0) SetHasMsgFlag(false);
	return l;
}
template <class BhPosMsg>
void CMessageQueue<BhPosMsg>::RemoveAll()
{
	MsgQueue.clear();
	SetHasMsgFlag(false);
}
template <class BhPosMsg>
void CMessageQueue<BhPosMsg>::DeleteLast()
{
	MsgQueue.pop_front();
	if(MsgQueue.empty()) bHasMsg=false;
}
template <class BhPosMsg>
bool CMessageQueue<BhPosMsg>::GetNext(BhPosMsg *t)
{
	if(MsgQueue.empty()){
		bHasMsg=false;
		return true;
	}
	
	if(bReorderArray){
	}
	BhPosMsg &m=MsgQueue.front();
	memcpy(t, &m, sizeof(BhPosMsg));
	return false;
}
