/*
clients.cpp - more space 
Copyright (C) 2001  Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos, Copyright (C) 2001-2004 Al Kingsley
	bhpos:: comes with ABSOLUTELY NO WARRANTY, or CLIENTS.
*/

#include <pthread.h>
#include <cstdio>

#include "plusd.h"
#include "ctransitems.h"
#include "clientinfo.h"
#include "msgq.h"

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

bool PLUServer::AddClient(bhpos::CClientInfo& c, BhPosMsg *msg)
{
	std::list<bhpos::CClientInfo>::iterator fi;
	BhPosConnect *conn=(BhPosConnect *) msg->data;
	int session;

#ifdef _DEBUGGERY
	if(DebugLevel > 2)
		printf("PluServer::AddClient\n");
#endif
	
	if(!clients.empty()){
		for(fi=clients.begin(); fi!=clients.end(); fi++){
			fi->GetParameter(BHCI_PARAM_SESSID, &session);
			if(session==msg->hdr.sessionID){
				fi->UpdateHdr(&msg->hdr);
				c=*fi;
				return false;
			}
		}
	}
	// Add a client
	sessionStatus++;
	msg->hdr.sessionID=sessionStatus;
	msg->hdr.msgID=0;
	c.SetParameter(BHCI_PARAM_SESSID, &sessionStatus);
	c.SetParameter(BHCI_PARAM_MAC, conn->mac);
	c.UpdateHdr(&msg->hdr);
	clients.push_back(c);
	
#ifdef _DEBUGGERY
	if(DebugLevel > 1)	
		printf("Adding client, sessionID=%d\n", sessionStatus);
#endif

	return false;
}

bool PLUServer::GetClient(int who, bhpos::CClientInfo& c)
{
	std::list<bhpos::CClientInfo>::iterator fi=clients.begin();
	
	for(int i=0; i!=(who-1); i++) fi++;
	
	if(fi==clients.end()) return true;
	c=*fi;
	return false;
}

bool PLUServer::DeleteClient(bhpos::CClientInfo& c)
{
	std::list<bhpos::CClientInfo>::iterator fi=find(clients.begin(), clients.end(), c);		

	if(fi==clients.end()) return true;

#ifdef _DEBUGGERY	
	if(DebugLevel > 1)
		printf("Removing client, sessionID=%d", sessionStatus);
#endif

	clients.erase(fi);
	return false;
}

bool PLUServer::UpdateClient(bhpos::CClientInfo& c)
{
	std::list<bhpos::CClientInfo>::iterator fi=find(clients.begin(), clients.end(), c);
	
	if(fi==clients.end()) return true;
	*fi=c;
	return false;
}

bool PLUServer::ClientExists(bhpos::CClientInfo& c, BhPosMsg *msg)
{
	std::list<bhpos::CClientInfo>::iterator fi;
	int session=0;
	unsigned char mac[6];
	
	if(clients.empty()){
#ifdef _DEBUGGERY	
		if(DebugLevel > 2)		
			printf("Client list is empty\n");
#endif	   
	   return false;
   }
	for(fi=clients.begin(); fi!=clients.end(); fi++){
	   fi->GetParameter(BHCI_PARAM_SESSID, &session);
	   if(session==msg->hdr.sessionID){
#ifdef _DEBUGGERY	
			if(DebugLevel > 2)		
				printf("found matching sessionID\n");
#endif			
			c=*fi;
			return true;
		}
	   
		memset(&mac, 0, sizeof(mac));
#ifdef _DEBUGGERY	
		if(DebugLevel > 2)		
			printf("checking client for %d - %d\n", session, msg->hdr.sessionID);
#endif
		if(msg->hdr.sessionID==0){
			BhPosConnect *conn=(BhPosConnect *) msg->data;
			fi->GetParameter(BHCI_PARAM_MAC, &mac);
			
#ifdef _DEBUGGERY	
			if(DebugLevel > 2)		
				printf("checking client mac %02X:%02X:%02X:%02X:%02X:%02X - %02X:%02X:%02X:%02X:%02X:%02X\n", 
				conn->mac[0], conn->mac[1], conn->mac[2], conn->mac[3], conn->mac[4], conn->mac[5],
				mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
#endif			
			
			if(memcmp(conn->mac, &mac, 6)==0){
#ifdef _DEBUGGERY	
			if(DebugLevel > 2)		
				printf("found matching mac\n");
#endif			   
				msg->hdr.sessionID=session;
				//fi->FlagKeepalive(false);
				return true;
			}
		}
		else{
			if(msg->hdr.type==BHPOS_TYPE_SYSTEM && msg->hdr.cmd==BHMSG_SYSTEM_CONNECTION){
				BhPosConnect *conn=(BhPosConnect *) msg->data;
				fi->GetParameter(BHCI_PARAM_MAC, &mac);
				if(memcmp(conn->mac, &mac, 6)==0){
					//fi->FlagKeepalive(false);
					msg->hdr.sessionID=session;
					c=*fi;
					return true;
				}
			}
		}
	}
	return false;
}

bool PLUServer::FillClientList(bhpos::service::CClientAckCollector& ac)
{
   std::list<bhpos::CClientInfo>::iterator ci;
   bhpos::service::CClientAck ack;
   bhpos::CStation   station;
   
   for(ci=clients.begin(); ci!=clients.end(); ci++){
      station=ci->GetStation();
      
      if(ac.GetSite()==0){             // all sites
         if(ac.GetDepartment()==0){    // all departments
            if(ac.GetGroup()==0){      // all groups
               if(ac.GetPost()==0){    // all posts
                  // it's everyone
                  for(ci=clients.begin(); ci!=clients.end(); ci++){
                     ac.Add(ack);
                  }
               }
               else{
                  // it's
               }
            }
            else{
            }
         }
         else{
         }
     
      }
      else{   // not all sites
      }
      ac.GetDepartment();
      ac.GetGroup();
      ac.GetPost();
   }
   
   return true;
}
