/*
tcp.cpp - a tcp thing 
Copyright (C) 2001 Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos version, Copyright (C) 2001-2004 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY.
*/
#include <pthread.h>
#include <cstdio>
#include <cstring>
#include <unistd.h>

#include "pos.h"
#include "msgq.h"
#include "ctcp.h"

#include "bhmsgq.cpp"

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

extern CMessageQueue<BhPosMsg>   *InQ;
extern CMessageQueue<BhPosMsg>   *CtrlQ;

extern pthread_t InQThread, CtrlQThread, OutQThread, SqlQThread, ErrorQThread;

void *TcpWorker(void *arg)
{
	int			i, status=0;
	CTCPThread	*t=(CTCPThread	*) arg;
	int			lsock=t->GetListenerSock();
	fd_set		ad, rd;
	BhPosMsg		pmsg;
	
	t->GetSockSet(&ad);
	
	while(1){
		memcpy(&rd, &ad,sizeof(fd_set));
		status=select(FD_SETSIZE, &rd, NULL, NULL, NULL);
		if(status < 0){
         printf("FAILURE\n");
		}
		for(i=0; i < FD_SETSIZE; ++i){
			if(FD_ISSET(i, &rd)){
				if(i==lsock){
					t->Add();
					// connection request, so reload the fd_set
					t->GetSockSet(&ad);
					continue;
				}
				else{
					if(t->IsValidSock(i)){
						status=recv(i, (void *) &pmsg.hdr, sizeof(pmsg), 0);
						if(status==-1 || status==0){
							// error rxing hdr or closed
                     t->Remove(i);
                     continue;
						}
						else{
							pmsg.flags=1;
							pmsg.sd=i;
							//plus->msg_count++;
							//plus->total_msg++;
							if(InQ)
								InQ->Add(&pmsg);
							
						}
					}
				}
			}
		}
	}
	t->Close();
	return NULL;

}

CTCPThread::CTCPThread()
{
}
CTCPThread::~CTCPThread()
{
	Close();
}

void CTCPThread::Remove(int i)
{
   pthread_mutex_lock(&mutex);
   
   std::list<int>::iterator si=find(sd.begin(), sd.end(), i);
	
	if(si==sd.end()) return;
	
	sd.erase(si);

	pthread_mutex_unlock(&mutex);
}

int CTCPThread::Add()
{
	sockaddr_in from;
	socklen_t	len;

	int nsd=accept(lsock, (sockaddr *)&from, &len);
	if(nsd < 0){
		return -1;
	}
	else{
		sd.push_back(nsd);
      printf("connect: from host %s, port %hd.\n", inet_ntoa(from.sin_addr), ntohs(from.sin_port));
	}
	return 0;
}

int CTCPThread::Create(void *p, int s)
{
	int			status;
	sched_param sp;

	lsock=s;
   
	plus= (PLUServer *) p;
	sd.push_back(lsock);
	sp.sched_priority=-20;

	status=::pthread_attr_init(&attributes);
	if(status!=0){
#ifdef _DEBUGGERY		
		if(DebugLevel>2) printf("ctcpthread : No Thread attrib init\n");
#endif
		return status;
	}
	pthread_attr_setschedpolicy(&attributes, SCHED_RR);
	pthread_attr_setschedparam(&attributes, &sp);
	
	status=::pthread_mutex_init(&mutex, NULL);

	if(status!=0){
		::pthread_attr_destroy(&attributes);
#ifdef _DEBUGGERY
		if(DebugLevel>2) printf("ctcpthread : No mutex init\n");
#endif
		return status;
	}

	status=::pthread_create(&tid, &attributes, TcpWorker, this);
	if(status!=0){
		::pthread_mutex_destroy(&mutex);
		::pthread_attr_destroy(&attributes);
#ifdef _DEBUGGERY
		if(DebugLevel>2){
			printf("ctcpthread : No Thread Create - error: %d - %s\n", status, strerror(status));
		}
#endif
		return status;
	}

	return 0;
}

int CTCPThread::GetListenerSock()
{
	return lsock;
}
void CTCPThread::Close()
{
	std::list<int>::iterator	si;
	int	s;

	pthread_mutex_lock(&mutex);

	for(si=sd.begin(); si!=sd.end(); si++){
		s=*si;
		if(s) close(s);
	}
	pthread_mutex_unlock(&mutex);

	::pthread_cancel(tid);
	::pthread_join(tid, NULL);

	::pthread_mutex_destroy(&mutex);
	::pthread_attr_destroy(&attributes);
}

int CTCPThread::GetSockSet(fd_set *set)
{
	std::list<int>::iterator	si;
	int	s;

	pthread_mutex_lock(&mutex);

	FD_ZERO(set);

	for(si=sd.begin(); si!=sd.end(); si++){
		s=*si;
		FD_SET(s, set);
	}
	pthread_mutex_unlock(&mutex);
	return 0;
}

bool CTCPThread::IsValidSock(int i)
{
	bool	ret=true;
	/*
	list<int>::iterator	si;

	pthread_mutex_lock(&mutex);
	
	find(si=sd.begin(); si!=sd.end(); si++);
	if(si==sd.end()) ret=false;
	
	pthread_mutex_unlock(&mutex);*/

	return ret;
}
