/*
ctrl.cpp - a multicast stack.
Copyright (C) 2001 Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos, Copyright (C) 2001-2005 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY.
*/
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <ctime>
#include <cerrno>

#include <unistd.h>
#include <pthread.h>

#include "msgq.h"
#include "plusd.h"

#include "bhmsgq.cpp"

extern CMessageQueue<BhPosMsg>   *InQ;
extern CMessageQueue<BhPosMsg>   *OutQ;
extern CMessageQueue<BhPosMsg>   *SqlQ;
extern CMessageQueue<BhPosMsg>   *ErrorQ;

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

CMessageQueue<BhPosMsg>   *CtrlQ;

void *CtrlQMonitor(void *arg)
{
	PLUServer *plus=(PLUServer *) arg;
	BhMcPosMsg  msg;

	int		  McSize=0;
	int		  status;
//	bool		 bNoLoopback=true;
	char		 temp[32];
	unsigned short port;

	int 			sock;
	socklen_t   opt=1;

	CtrlQ=new CMessageQueue<BhPosMsg>();
	if(!CtrlQ){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("No CtrlQ Allocation\n");
#endif			
		status=THREAD_DEAD;
		plus->SetThread(CTRLQ_THREADID, THREAD_PARAM_STATE, &status);
		return NULL;
	}

	status=::pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	if(status!=0){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("ctrl thread: setcanceltype failed.\n");
#endif			
	}
	
	// get ip for control multicast socket from PLUS...
	sockaddr_in sin;
	memset(&sin, 0, sizeof(sin));
	sin.sin_family=AF_INET;
	plus->GetOption(BH_PLUS_OPTION_MCHOST, temp);
	plus->GetOption(BH_PLUS_OPTION_MCPORT, &port);
	sin.sin_addr.s_addr=inet_addr(temp);
	sin.sin_port=htons(port);
	
	sock=socket(AF_INET, SOCK_DGRAM, 0);
	if(sock < 0){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("no ctrl socket creation\n");
#endif	
		delete CtrlQ;
		status=THREAD_DEAD;
		plus->SetThread(CTRLQ_THREADID, THREAD_PARAM_STATE, &status);		
		return NULL;
	}
	
	if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(int)) < 0){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("no ctrl sock socket reuse\n");
#endif			
	}

	if(bind(sock, (sockaddr *) &sin, sizeof(sin)) < 0){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("no ctrl socket bind\n");
#endif		
		delete CtrlQ;
		close(sock);
		status=THREAD_DEAD;
		plus->SetThread(CTRLQ_THREADID, THREAD_PARAM_STATE, &status);	
		return NULL;
	}

	plus->GetOption(BH_PLUS_OPTION_ADDRESSRANGE, &opt);
	if(setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, &opt, sizeof(int)) < 0){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("no ctrl socket ttl set\n");
#endif			
	}
	// no loopback
	opt=0;
	if(setsockopt(sock, IPPROTO_IP, IP_MULTICAST_LOOP, &opt, sizeof(int)) < 0){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("no ctrl socket mc loop set\n");
#endif			
	}

	status=CtrlQ->Grab();

	if(status!=0){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("CtrlOutQ failed to Grab Mutex, %d (%s)\n", status, strerror(status));
#endif			
		close(sock);
		delete CtrlQ;
		status=THREAD_DEAD;
		plus->SetThread(CTRLQ_THREADID, THREAD_PARAM_STATE, &status);
		return NULL;
	}
	
	status=THREAD_RUNNING;
	plus->SetThread(CTRLQ_THREADID, THREAD_PARAM_STATE, &status);

	while(1){
		status=CtrlQ->WaitOnMsg();
		if(status != 0) {
#ifdef _DEBUGGERY
			if(DebugLevel)		
				printf("CtrlOutQ thread wait failed, %d (%s)\n", status, strerror(status));
#endif				
			CtrlQ->Release();
			delete CtrlQ;
			close(sock);
			status=THREAD_DEAD;
			plus->SetThread(CTRLQ_THREADID, THREAD_PARAM_STATE, &status);
			return NULL;
		}
		if(!CtrlQ->HasMsg()){
			continue;
		}
		CtrlQ->GetNext((BhPosMsg *) &msg);
		CtrlQ->DeleteLast();
		CtrlQ->Release();
		
		McSize=sizeof(BhHdr) + 16 + msg.hdr.size;
		
		if((status=sendto(sock, (void *)&msg, McSize, 0, (sockaddr *) &sin, sizeof(sin)))<0){
#ifdef _DEBUGGERY
			if(DebugLevel > 2)		
				printf("\nMulticast tx error: %d\n\n", status);
#endif				
		}
		
		plus->GetThread(CTRLQ_THREADID, THREAD_PARAM_DEATH, &status);
		if(status && !CtrlQ->HasMsg()) break;
		
		status=CtrlQ->Grab();
		if(status!=0){
#ifdef _DEBUGGERY
			if(DebugLevel)	
				printf("CtrlOutQ failed to Grab Mutex, %d (%s)\n", status, strerror(status));
#endif			
			close(sock);
			delete CtrlQ;
			status=THREAD_DEAD;
			plus->SetThread(CTRLQ_THREADID, THREAD_PARAM_STATE, &status);
			return NULL;
		}
	}
	CtrlQ->Release();
	delete CtrlQ;
	close(sock);
	status=THREAD_DEAD;
	plus->SetThread(CTRLQ_THREADID, THREAD_PARAM_STATE, &status);
	return NULL;
}
