/*
error.cpp - an error queue
Copyright (C) 2001 Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Part of:
	BhPos, Copyright (C) 2001-2005 Al Kingsley
*/
#include <pthread.h>
#include <cstdio>
#include <cstring>
#include <cerrno>
#include <fcntl.h>
#include <unistd.h>

#include "pos.h"
#include "msgq.h"
#include "plusd.h"
#include "sql.h"

#include "bhmsgq.cpp"

extern CMessageQueue<BhPosMsg> *SqlQ;
extern CMessageQueue<BhPosMsg> *InQ;
extern CMessageQueue<BhPosMsg> *OutQ;
extern CMessageQueue<BhPosMsg> *CtrlQ;

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

CMessageQueue<BhPosMsg> *ErrorQ;

static int ErrorSent(PLUServer *p, bhpos::BhSQL *errdb, BhPosMsg *msg);
static int ErrorSystem(PLUServer *p, bhpos::BhSQL *errdb, BhPosMsg *msg);
static int ErrorLogSent(PLUServer *p, bhpos::BhSQL *errdb, BhPosMsg *msg);
static int ErrorLogSystem(PLUServer *p, bhpos::BhSQL *errdb, BhPosMsg *msg);

void *ErrorQMonitor(void *arg)
{
	PLUServer	*plu=(PLUServer *) arg;
   bhpos::BhSQL			         *errdb;
   
	BhPosMsg	 msg;
   int status, ret;
	int sock;
	sockaddr_in esin;
	bool bErrServ;
	bool bConnected=false;
	int errserv_retries=0;
	int errserv_timeout=0;	
	int errserv_retrtcnt=0;
	timespec	 to;
	socklen_t   opt=1;
	unsigned short port;
	char host[33];
   
   char dbhost[32];
	char dbname[32];
	char dbuser[32];
	char dbpass[32];
		
	ErrorQ=new CMessageQueue<BhPosMsg>();
	if(!ErrorQ){
#ifdef _DEBUGGERY
if(DebugLevel)
			printf("No ErrorQ Allocation\n");
#endif 
      return NULL;
	}
   
   plu->GetOption(BH_PLUS_OPTION_DBHOSTPRIME, dbhost);
	plu->GetOption(BH_PLUS_OPTION_DBDATABASE, dbname);
	plu->GetOption(BH_PLUS_OPTION_DBUSER, dbuser);
	plu->GetOption(BH_PLUS_OPTION_DBPASSWORD, dbpass);
   
   errdb=new bhpos::BhSQL(dbhost, dbname, 0, dbuser, dbpass);
	
	if(!errdb){
#ifdef _DEBUGGERY
		if(DebugLevel)		
			printf("PLUServer - Error Thread : failed to create SQL object\n");
#endif
		delete ErrorQ;
		ErrorQ=NULL;	
		status=THREAD_DEAD;
		plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &status);			
		return NULL;
	}

	if(errdb->Connect()){
#ifdef _DEBUGGERY
		if(DebugLevel)		
			printf("PLUServer - Error Thread : failed to connect to SQL database - \n%s\n", errdb->ErrorMsg());
#endif			
		delete ErrorQ;
		ErrorQ=NULL;	
		status=THREAD_DEAD;
		plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &status);
		return NULL;
	}
	
	plu->GetOption(BH_PLUS_OPTION_ERRREDIRECT, &bErrServ);
	plu->GetOption(BH_PLUS_OPTION_ERRHOST, host);
	plu->GetOption(BH_PLUS_OPTION_ERRPORT, &port);
	plu->GetOption(BH_PLUS_OPTION_ERRRETRIES, &errserv_retries);
	plu->GetOption(BH_PLUS_OPTION_ERRTIMEOUT, &errserv_timeout);

#ifdef _DEBUGGERY
   if(DebugLevel)
		printf("ErrorQ server - %s:%d\n", host, port);
#endif 

   if(bErrServ && errserv_timeout!=0){
		to.tv_sec=(int) (errserv_timeout/1000);
		to.tv_nsec=(errserv_timeout-(to.tv_sec * 10000)) * 1000000;
	}
	
	memset(&esin, 0, sizeof(esin));
	esin.sin_family=AF_INET;
	esin.sin_addr.s_addr=inet_addr(host);
	esin.sin_port=htons(port);
	
	sock=socket(PF_INET, SOCK_STREAM, 0);
	if(sock < 0){
#ifdef _DEBUGGERY
      printf("error thread: no socket creation\n");
#endif 
      ret=THREAD_DEAD;
		plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
		return NULL;
	}
	
	if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(int))<0){
#ifdef _DEBUGGERY
      if(DebugLevel>1)
			printf("error thread: no socket reuse\n");
#endif
   }

	// set non blocking
   ret=fcntl(sock, F_GETFL, 0);
   if(ret == -1){
	   // buggerfish
      ret=THREAD_DEAD;
	   plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
	   close(sock);
#ifdef _DEBUGGERY
      if(DebugLevel>1)
			printf("error thread: can't set non-block - quitting\n");
#endif
      return NULL;
  }
  if(ret!=0) ret |= O_NONBLOCK;
  else ret &= ~O_NONBLOCK;
  fcntl(sock, F_SETFL, ret);

	
	status=::pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	if(status!=0){
#ifdef _DEBUGGERY
      if(DebugLevel)	
			printf("error thread: setcanceltype failed.\n");
#endif 
   }
		
	ret=THREAD_RUNNING;
	plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
	
	while(42){
      if(!bErrServ){ // let's keep this simple
         if((status=ErrorQ->Grab()) == 0)
			   status=ErrorQ->WaitOnMsg();
		
		   if(status != 0){
#ifdef _DEBUGGERY
            if(DebugLevel)		
			      printf("ErrorQ thread wait failed, %d (%s)\n", status, strerror(status));
#endif 
            ErrorQ->Release();
			   if(bConnected && bErrServ) close(sock);
			   ret=THREAD_DEAD;
			   plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
			   return NULL;
		   }
		   if(!ErrorQ->HasMsg()){
            ErrorQ->Release();
			   continue;
		   }
#ifdef _DEBUGGERY
         if(DebugLevel>1)
			   printf("error thread: popped error msg\n");
#endif 
         ErrorQ->GetNext(&msg);
         ErrorQ->DeleteLast();
         ErrorQ->Release();
		
   		if(msg.hdr.type==BHPOS_TYPE_ERROR){	// msg was sent
            ErrorLogSent(plu, errdb, &msg);
		   }
		   else{	// error occured in system
            ErrorLogSystem(plu, errdb, &msg);
		   }
         plu->GetThread(ERRQ_THREADID, THREAD_PARAM_DEATH, &status);
		   if(status){
            ErrorQ->Release();
	         ret=THREAD_DEAD;
	         plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
	         return NULL;
         }
         continue;
      }
      
      
		if(!bConnected && bErrServ){
			ErrorQ->Release();
			// attempt a connect
#ifdef _DEBUGGERY
         if(DebugLevel)	
				printf("error thread: connecting.\n");
#endif 
         if(connect(sock, (sockaddr *) &esin, sizeof(esin)) < 0){
				switch(errno){
				case EBADF:				// The socket socket is not a valid file descriptor.
            case ENOTSOCK:			// File descriptor socket is not a socket.
#ifdef _DEBUGGERY
               if(DebugLevel)	
						printf("error thread: connect ENOTSOCK.\n");
#endif 
               // kill thread for restart...
               break;

				case EADDRNOTAVAIL:	// The specified address is not available on the remote machine.
            case EAFNOSUPPORT:	// The namespace of the addr is not supported by this socket
#ifdef _DEBUGGERY
               if(DebugLevel)	
						printf("error thread: connect EAFNOSUPPORT.\n");
#endif 
               break;

				case EISCONN:			// The socket socket is already connected. 
#ifdef _DEBUGGERY
               if(DebugLevel)	
						printf("error thread: connect EISCONN.\n");
#endif 
               break;

				case ETIMEDOUT:		// The attempt to establish the connection timed out. 
#ifdef _DEBUGGERY
               if(DebugLevel)	
						printf("error thread: connect ETIMEDOUT.\n");
#endif 
               break;

				case ECONNREFUSED:	// The server has actively refused to establish the connection. 
            case ENETUNREACH:		// The network of the given addr isn't reachable from this host.
#ifdef _DEBUGGERY
               if(DebugLevel)	
						printf("error thread: connect ENETUNREACH.\n");
#endif 
               break;

				case EADDRINUSE:		// The socket address of the given addr is already in use. 
#ifdef _DEBUGGERY
               if(DebugLevel)	
						printf("error thread: connect EADDRINUSE.\n");
#endif 
               break;

				case EINPROGRESS:		// The socket socket is non-blocking and the connection could not be established immediately. You can determine when the connection is completely established with select; see section Waiting for Input or Output. Another connect call on the same socket, before the connection is completely established, will fail with EALREADY. 
            case EALREADY:			// The socket socket is non-blocking and already has a pending connection in progress
#ifdef _DEBUGGERY
               if(DebugLevel)	
						printf("error thread: connect EINPROGRESS.\n");
#endif 
               break;
				}
				if(errserv_retries){	// 0 for forever
               errserv_retrtcnt--;
					if(errserv_retrtcnt <=0){
						ret=THREAD_DEAD;
						plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
						return NULL;
					}
				}
				else{	// this is await forever, so discard messages to stop queue filling up
               if(ErrorQ->HasMsg()){
						ErrorQ->DeleteLast();
					}
				}
				sleep(5);
				continue;
			}
			else{			
				// select test it
            bConnected=true;
			}
		}
		
		plu->GetThread(ERRQ_THREADID, THREAD_PARAM_DEATH, &status);
		if(status && !ErrorQ->HasMsg()) break;
		
		if((status=ErrorQ->Grab()) == 0)
			status=ErrorQ->WaitOnMsg();
		
		if(status != 0){
#ifdef _DEBUGGERY
         if(DebugLevel)		
			   printf("ErrorQ thread wait failed, %d (%s)\n", status, strerror(status));
#endif 
         ErrorQ->Release();
			if(bConnected && bErrServ) close(sock);
			ret=THREAD_DEAD;
			plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
			return NULL;
		}
		if(!ErrorQ->HasMsg()){
			continue;
		}
#ifdef _DEBUGGERY
      if(DebugLevel>1)
			printf("error thread: popped error msg\n");
#endif 
      ErrorQ->GetNext(&msg);
		
		// redirect some crap
      if(bConnected && bErrServ){
			if((ret=send(sock, &msg, sizeof(msg.hdr)+msg.hdr.size, 0)) < 0){
#ifdef _DEBUGGERY
            if(DebugLevel>1)
						printf("error thread: failed to send to error server\n");
#endif 
            if(!errserv_retries){	// 0 for forever
               errserv_retrtcnt=errserv_retries;
				}
				ErrorQ->Release();
				bConnected=false;
			}
			else{
				ErrorQ->DeleteLast();
				ErrorQ->Release();
			}
#ifdef _DEBUGGERY
         if(DebugLevel>1)
				printf("error thread: successfully send to error server\n");
#endif
}
		else{
#ifdef _DEBUGGERY
         if(DebugLevel>1)
				printf("error thread: no error server: dumping\n");
#endif
         ErrorQ->DeleteLast();
			ErrorQ->Release();
		}

		if(msg.hdr.type==BHPOS_TYPE_ERROR){	// msg was sent
         ErrorSent(plu, errdb, &msg);
		}
		else{	// error occured in system
         ErrorSystem(plu, errdb, &msg);
		}
						
		//printf("ErrorQ System Message\ntype:%02d\ncmd:%02d", msg.hdr.type, msg.hdr.cmd);
   }
	ErrorQ->Release();
	if(bConnected) close(sock);
	ret=THREAD_DEAD;
	plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
	return NULL;
}

int ErrorSent(PLUServer *p, bhpos::BhSQL *errdb, BhPosMsg *msg)
{
   bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow   *sql_row;
	char				    buff[128];
   int                status;
   BhErrorMsg        *err=(BhErrorMsg *) msg->data;
   
   sprintf(buff, "INSERT INTO errorlog VALUES(NULL, NOW(), '%s', '%d', '%s', '%d', 0, '%s')", err->enteredAt, err->userID, "0000-00-00 00:00:00",err->type, err->more);
   status=errdb->SimpleQuery(buff);
   if(status!=0){
#ifdef _DEBUGGERY          
      if(DebugLevel > 2){
         printf("ErrorSent: Failed to log error message for post %s\n", err->enteredAt);
      }
#endif
   }
	return 0;
}

int ErrorSystem(PLUServer *p, bhpos::BhSQL *errdb, BhPosMsg *msg)
{
	return 0;
}

int ErrorLogSent(PLUServer *p, bhpos::BhSQL *errdb, BhPosMsg *msg)
{
   bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow   *sql_row;
	char				    buff[128];
   int                status;
   BhErrorMsg        *err=(BhErrorMsg *) msg->data;
   
   sprintf(buff, "INSERT INTO errorlog VALUES(0, NOW(), '%s', '%d', '%s', '%d', 0, '%s')", err->enteredAt, err->userID, "0000-00-00 00:00:00",err->type, err->more);
   status=errdb->SimpleQuery(buff);
   if(status!=0){
#ifdef _DEBUGGERY          
      if(DebugLevel > 2){
         printf("ErrorLogSent: Failed to log error message for post: %s\n", err->enteredAt);
      }
#endif
   }
	return 0;
}

int ErrorLogSystem(PLUServer *p, bhpos::BhSQL *errdb, BhPosMsg *msg)
{
   bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow   *sql_row;
	char				    buff[128];
   int                status;
   BhErrorMsg        *err=(BhErrorMsg *) msg->data;
   
   sprintf(buff, "INSERT INTO errorlog VALUES(0, NOW(), '%s', '%d', '%s', '%d', 0, '%s')", err->enteredAt, err->userID, "0000-00-00 00:00:00", err->type, err->more);
   status=errdb->SimpleQuery(buff);
   if(status!=0){
#ifdef _DEBUGGERY          
      if(DebugLevel > 2){
         printf("ErrorLogSystem: Failed to log error message %s, for post: %s\n", buff, err->enteredAt);
      }
#endif
   }
	return 0;
}
