/*
idleproc.cpp - an idle processing thing, though not in essence an idle proc.
Copyright (C) 2001  Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Don't write to me at: kingsley@bananahead.com 

   Part of:
   BhPos, Copyright (C) 2001-2004 Al Kingsley
   BhPos comes with ABSOLUTELY NO WARRANTY.
*/
#include <pthread.h>

#include <cstdio>
#include <unistd.h>
#include <time.h>

#include "pos.h"
#include "msgq.h"
#include "plusd.h"
#include "cevent.h"

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

void *IdleThread(void *arg)
{
   PLUServer    *plus=(PLUServer *) arg;
   double        temp, curr_load, max_load;
   unsigned int  curr_time, start_time, up_time, clresettime;
   unsigned long msg_count=0;
   unsigned long msg_elapsed_time=0;
   time_t        tossertime, rtime;
   struct tm    *tickletime;
   bool          bClientReset=false, bkeepalive=false;

   int idleupdate, statupdate, upsupdate, threadcheck, procwriter, keepalive, recommit;
   //int nowtime=1;

   bhpos::CTEShed    events;
   bhpos::CTimeEvent event;
   
//   bool        bNetworkUPS;
   int         status;
   
   plus->GetParameter(BH_PLUS_PARM_STARTTIME, &start_time);
   plus->GetParameter(BH_PLUS_PARM_UPTIME, &up_time);
   plus->GetParameter(BH_PLUS_PARM_MAXLOAD, &max_load);
// plus->GetOption(BH_PLUS_OPTION_NETWORKUPS, &bNetworkUPS);
   plus->GetOption(BH_PLUS_OPTION_UPSUPDATETIME, &upsupdate);
   plus->GetOption(BH_PLUS_OPTION_IDLEUPDATEPERIOD, &idleupdate);
   plus->GetOption(BH_PLUS_OPTION_STATSUPDATE, &statupdate);
   plus->GetOption(BH_PLUS_OPTION_THREADCHECK, &threadcheck);
   plus->GetOption(BH_PLUS_OPTION_PROCWRITER, &procwriter);
   plus->GetOption(BH_PLUS_OPTION_KEEPALIVE, &bkeepalive);
   plus->GetOption(BH_PLUS_OPTION_RECOMMIT, &recommit);
   plus->GetOption(BH_PLUS_OPTION_CLIENTRESETTIME, &clresettime);
   
   if(bkeepalive){
	   plus->GetOption(BH_PLUS_OPTION_ALIVETIME, &keepalive);
	}

#ifdef _DEBUGGERY
   if(DebugLevel>2) printf("IdleThread: update period: %0.3f mS\n", ((double) (idleupdate /1000)));
#endif

   if(statupdate!=0){
#ifdef _DEBUGGERY
      if(DebugLevel>2) printf("IdleThread: Adding Event stats: %0.3f mS\n", ((double) (statupdate /1000)));
#endif
      event.Set(statupdate, bhpos::EVENT_TYPE_PERIODIC, 0);
      statupdate=events.Add(event);
   }
   if(upsupdate!=0){
#ifdef _DEBUGGERY
      if(DebugLevel>2) printf("IdleThread: Adding Event ups: %0.3f mS\n", ((double) (upsupdate /1000)));
#endif
      event.Set(upsupdate, bhpos::EVENT_TYPE_PERIODIC, 0);
      upsupdate=events.Add(event);
   }
   if(threadcheck!=0){
#ifdef _DEBUGGERY
      if(DebugLevel>2) printf("IdleThread: Adding Event thread: %0.3f mS\n", ((double) (threadcheck /1000)));
#endif      
      event.Set(threadcheck, bhpos::EVENT_TYPE_PERIODIC, 0);
      threadcheck=events.Add(event);
   }
   if(procwriter!=0){
#ifdef _DEBUGGERY
      if(DebugLevel>2) printf("IdleThread: Adding Event procwriter: %0.3f mS\n", ((double) (procwriter /1000)));
#endif
      event.Set(procwriter, bhpos::EVENT_TYPE_PERIODIC, 0);
      procwriter=events.Add(event);
   }
   if(keepalive!=0 && bkeepalive){
#ifdef _DEBUGGERY
      if(DebugLevel>2) printf("IdleThread: Adding Event keepalive: %0.3f mS\n", ((double) (keepalive /1000)));
#endif      
      event.Set(keepalive, bhpos::EVENT_TYPE_PERIODIC, 0);
      keepalive=events.Add(event);
   }
   if(recommit!=0){
#ifdef _DEBUGGERY
      if(DebugLevel>2) printf("IdleThread: Adding Event recommit: %0.3f mS\n", ((double) (recommit /1000)));
#endif
      event.Set(recommit, bhpos::EVENT_TYPE_PERIODIC, 0);
      recommit=events.Add(event);
   }

   events.Reset();

   status=::pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
   status=THREAD_RUNNING;
   plus->SetThread(IDLE_THREADID, THREAD_PARAM_STATE, &status);

   events.Tick();
   
   while(42){
      if((status=events.hasEvent(event))){
         if(status==statupdate){
            //printf("Idelproc: stats\n");
            curr_time=clock();
            temp=((double) curr_time - start_time)/CLOCKS_PER_SEC;
            up_time=(unsigned long) temp;
            temp=((double) curr_time - msg_elapsed_time)/CLOCKS_PER_SEC;
            curr_load=msg_count / temp;
            msg_count=0;
            msg_elapsed_time=curr_time;
      
            if(curr_load > max_load){
               max_load=curr_load;
               plus->SetParameter(BH_PLUS_PARM_MAXLOAD, (void *) &max_load);
            }
            continue;
         }
         else if(status==upsupdate){
            //printf("Idelproc: ups update\n");
            continue;
         }
         else if(status==threadcheck){
            //printf("Idelproc: thread check\n");
            plus->CheckThreads(ALL_THREADS);
            continue;
         }
         else if(status==procwriter){
            //printf("Idelproc: procwriter\n");
            continue;
         }
         else if(status==keepalive){
#ifdef _DEBUGGERY         	
            printf("Idelproc: keep alive\n");
#endif
            plus->KeepAlive();
            continue;
         }
         else if(status==recommit){
            //printf("Idelproc: recommit\n");
            continue;
         }
         else if(status==0){  // no more events left
            //printf("Idelproc: sleeping\n");            
         }
         else{ // error condition
         }
      }
      usleep(idleupdate);
#ifdef _DEBUGGERY
      //if(DebugLevel>2) printf("IdleThread: Tick!\n");
#endif      
      events.Tick();
      
      // check the number of seconds past midnight
      tossertime=time(NULL);
      tickletime=localtime(&tossertime);
      curr_time=tickletime->tm_sec + (tickletime->tm_min * 60) + (tickletime->tm_hour * 60 * 60);
      if(!bClientReset){
         if(curr_time > clresettime){
            plus->ResetClients();
            bClientReset=true;
            memcpy(&rtime, &tossertime, sizeof(time_t));
         }
      }
      else{   // check for new day
         if(tossertime < rtime){
            bClientReset=false;
         }
      }
      
      plus->GetThread(IDLE_THREADID, THREAD_PARAM_DEATH, &status);
      if(status) break;
   }

   status=THREAD_DEAD;
   plus->SetThread(IDLE_THREADID, THREAD_PARAM_STATE, &status);
   
   return NULL;
}
