/*
inq.cpp - a queue thing
Copyright (C) 2001  Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos, Copyright (C) 2001-2005 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY. 
*/
#include <pthread.h>
#include <cstdio>
#include <cstring>
#include <unistd.h>

#include "pos.h"
#include "msgq.h"
#include "plusd.h"

#include "bhmsgq.cpp"

CMessageQueue<BhPosMsg>   *InQ;

extern CMessageQueue<BhPosMsg>   *OutQ;
extern CMessageQueue<BhPosMsg>   *SqlQ;
extern CMessageQueue<BhPosMsg>   *ErrorQ;
extern CMessageQueue<BhPosMsg>   *TransQ;
extern CMessageQueue<BhPosMsg>   *CtrlQ;

#ifdef _CCVENGINEMCVE
extern CMessageQueue<BhPosMsg>   *CcvQ;
#endif

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

extern bool	bgoingtodie;

static int SystemMessage(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int Connection(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int Login(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SysService(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int Authenticate(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int Version(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int Shutdown(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int Message(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int Query(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int Time(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int Statistics(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int Process(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int ProcInfo(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int ProcKill(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int Debug(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);

static int PostMessage(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int PLU(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int Transaction(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int PostService(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int PostQuery(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int PostPrinter(PLUServer *p, BhPosMsg *msg);
static int PostList(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int PostDelivery(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int PostRental(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int PostCustomer(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int PostSignature(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int QueryStatus(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int QueryMode(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int QueryWho(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);

int ServiceMessage(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);

int GoodsMessage(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);

int AttendanceMessage(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
int EntryMessage(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
int InfoMessage(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);

int MenuMessage(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);

void *InQMonitor(void *arg)
{
	PLUServer *plu=(PLUServer *) arg;
	BhPosMsg   msg;
	int		status;
	
	bhpos::CClientInfo c;
	
	InQ=new CMessageQueue<BhPosMsg>();
	if(!InQ){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("No InQ Allocation\n");
#endif
		status=THREAD_DEAD;
		plu->SetThread(INQ_THREADID, THREAD_PARAM_STATE, &status);
		return NULL;
	}
	/*
	status=::pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	if(status!=0){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("in thread: setcanceltype failed.\n");
#endif		
	}*/
	
	// inital lock out
	status=InQ->Grab();
	if(status != 0) {
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("InQ failed to Grab Mutex, %d (%s)\n", status, strerror(status));
#endif		
		delete InQ;
		InQ=NULL;	
		status=THREAD_DEAD;
		plu->SetThread(INQ_THREADID, THREAD_PARAM_STATE, &status);
		return NULL;
	}
	
	// wibble wibble, toil and flibble let's check it out...
	status=THREAD_RUNNING;
	plu->SetThread(INQ_THREADID, THREAD_PARAM_STATE, &status);
	while(1){
		// wait on condition - q has msg
		status=InQ->WaitOnMsg();
		if(status != 0) {
#ifdef _DEBUGGERY
			if(DebugLevel)		
				printf("InQ thread wait failed, %d (%s)\n", status, strerror(status));
#endif			
			InQ->Release();
			delete InQ;
			InQ=NULL;
			status=THREAD_DEAD;
			plu->SetThread(INQ_THREADID, THREAD_PARAM_STATE, &status);
			return NULL;
		}
		// mutex locked, stocked and err, somethingorother...
		
		// double check queue has msg
		if(!InQ->HasMsg()){
#ifdef _DEBUGGERY
			if(DebugLevel > 2)		
				printf("InQ thread: report no messages\n");
#endif			
			continue;
		}

		InQ->GetNext(&msg);
		InQ->DeleteLast();
		InQ->Release();
		
		if(bgoingtodie){
		   
		}
		
#ifdef _DEBUGGERY
		if(DebugLevel > 2)
			printf("\nInQpacket from:%s:%d:%d\n\n", inet_ntoa((struct in_addr) msg.return_addr.sin_addr), ntohs(msg.return_addr.sin_port),msg.hdr.msgID);
#endif
      // check msg number 
      if(!plu->ClientExists(c, &msg)){
#ifdef _DEBUGGERY
         if(DebugLevel)
            printf("InqMonitor : No Client ID\n");
#endif
			if(msg.hdr.type!=BHPOS_TYPE_SYSTEM && msg.hdr.cmd!=BHMSG_SYSTEM_CONNECTION){
#ifdef _DEBUGGERY
            if(DebugLevel)
               printf("InqMonitor : Client refuse - not a connect msg:%d-%d\n",msg.hdr.type,msg.hdr.cmd);
#endif
				msg.hdr.Ex=BHPOS_REFUSE;
				if(!OutQ){
#ifdef _DEBUGGERY
					if(DebugLevel)
						printf("No OutQ in InQ... quitting thread\n");
#endif
					delete InQ;
					InQ=NULL;
					status=THREAD_DEAD;
					plu->SetThread(INQ_THREADID, THREAD_PARAM_STATE, &status);
					return NULL;
				}
				OutQ->Add(&msg);
				status=InQ->Grab();
				continue;
			}
			else{
				if(msg.hdr.type==BHMSG_SYSTEM_CONNECTION && (msg.hdr.subcmd==BHMSG_SYSTEM_CONNECT || msg.hdr.subcmd==BHMSG_SYSTEM_SCONNECT)){
					if(plu->AddClient(c, &msg)){
#ifdef _DEBUGGERY
						if(DebugLevel)
							printf("InqMonitor : Failed to AddClient\n");
#endif							
					}
				}
			}
      }
      if(msg.hdr.msgID==0){
      	c.SetLastMsgID(1);
	   	plu->UpdateClient(c);
      }
	   if(msg.hdr.msgID > 0 && c.GetLastMsgID()==1){
#ifdef _DEBUGGERY
			if(DebugLevel)
				printf("InqMonitor : Dropping Msg - bad ID\n");
#endif
			msg.hdr.msgID=1;
			msg.hdr.size=0;
			if(OutQ) OutQ->Add(&msg);
         status=InQ->Grab();
			continue;      
      }
      else{
         c.SetLastMsgID(0);
	   	plu->UpdateClient(c);
      }
		
		// Setup a client for the following functions
		//if(msg.hdr.type!=BHPOS_TYPE_SERVICE){
			//if(msg.hdr.type==BHPOS_TYPE_SYSTEM){
				// silent granting stuff...
				// in the case of quick drop outs and connects between keep alives
				//if(msg.hdr.cmd==BHMSG_SYSTEM_CONNECTION){
#ifdef _DEBUGGERY
					//if(DebugLevel > 2)
						//printf("InQMonitor: Client Exists: re-connection:%d\n", msg.hdr.sessionID);
#endif
					// this requires working multicast
               //c.SetLastMsgID(msg.hdr.msgID);
					//c.FlagKeepalive(false);
      		   //plu->UpdateClient(c);
         		   
      		   // and here is where we restore the last client state
      		   // if the option is set
      		   // we'll respond with a transrestore and wait for either:
      		   // trans restore msg or
      		   // a restorerefuse msg
         		   
      		   //if(SqlQ) SqlQ->Add(&msg);
      		   //else printf("InQMonitor: no sql q\n");

#ifdef _DEBUGGERY
					//if(DebugLevel > 2)         		   
      		   	//printf("InQMonitor: SqlQ::NumMsgs=%d\n", SqlQ->NumInQueue());
#endif         		   
      		   //Connection(plu, c, &msg);
					//status=InQ->Grab();
					//continue;
				//}
/*					
				if(msg.hdr.cmd==BHMSG_SYSTEM_LOGIN){
#ifdef _DEBUGGERY
					if(DebugLevel > 2)
						printf("InQMonitor: Client Exists: re-login\n");
#endif							
					Login(plu, c, &msg);
					//c.FlagKeepalive(false);
					status=InQ->Grab();
					continue;
				}
*/					
			//}
		//}
      
		switch(msg.hdr.type){
		case BHPOS_TYPE_NULL:
			break;
		case BHPOS_TYPE_SYSTEM: 
			SystemMessage(plu, c, &msg); 
			break;

		case BHPOS_TYPE_POST:
			PostMessage(plu, c, &msg); 
			break;
			
		case BHPOS_TYPE_SERVICE:
			ServiceMessage(plu, c, &msg);
			break;
			
		case BHPOS_TYPE_ATTENDANCE:
			AttendanceMessage(plu, c, &msg);
			break;

		case BHPOS_TYPE_ENTRY:
			EntryMessage(plu, c, &msg);
			break;
			
		case BHPOS_TYPE_SCALES:
			if(SqlQ) SqlQ->Add(&msg);
			break;

      case BHPOS_TYPE_MENU:
         MenuMessage(plu, c, &msg);
         break;
      
		case BHPOS_TYPE_INFO:
		case BHPOS_TYPE_RPRINTER:
		case BHPOS_TYPE_CUSTOM:
			break;
      
      //case BHPOS_TYPE_GOODS:
         //GoodsMessage(plu, c, &msg);
         //break;
      
		case BHPOS_TYPE_ERROR:
			ErrorQ->Add(&msg);
			break;

		default:
#ifdef _DEBUGGERY
         if(DebugLevel > 2)
            printf("InQ: Module Check\n");
#endif
			if(plu->HasModules()){
#ifdef _DEBUGGERY
         if(DebugLevel > 2)
            printf("InQ: Has Modules\n");
#endif
			   /*bhpos::bhm mt;
   
            mt.bmsg[0]=m->hdr.type;
            mt.bmsg[1]=m->hdr.cmd;
            mt.bmsg[2]=m->hdr.subcmd;
            mt.bmsg[3]=m->hdr.Ex=0;
				if(plu->ModuleHandlesMessages(mt)){*/
#ifdef _DEBUGGERY
         if(DebugLevel > 2)
            printf("InQ: Module Processing\n");
#endif
					plu->HandleMsg(&msg);
				}
			//}
			break;
		}

		// check cancellation
		plu->GetThread(INQ_THREADID, THREAD_PARAM_DEATH, &status);
		if(status && !InQ->HasMsg()) break;
		
		// lock out mutex again
		status=InQ->Grab();
		if(status != 0) {
			printf("InQ failed to Mutex, %d (%s)\n", status, strerror(status));
			delete InQ;
			InQ=NULL;
			status=THREAD_DEAD;
			plu->SetThread(INQ_THREADID, THREAD_PARAM_STATE, &status);
			return NULL;
		}
	}
	InQ->Release();
	delete InQ;
	InQ=NULL;
	status=THREAD_DEAD;
	plu->SetThread(INQ_THREADID, THREAD_PARAM_STATE, &status);
	return NULL;
}

int SystemMessage(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	int	ret=0;

#ifdef _DEBUGGERY
	if(DebugLevel > 2)
		printf("System Message\n");
#endif
	switch(msg->hdr.cmd){
	case BHMSG_SYSTEM_CONNECTION: ret=Connection(p, c, msg); break;
	case BHMSG_SYSTEM_LOGIN: ret=Login(p, c, msg); break;
   case BHMSG_SYSTEM_INTERFACE: if(SqlQ) SqlQ->Add(msg); break;
	case BHMSG_SYSTEM_AUTHENTICATE: ret=Authenticate(p, c, msg); break;
	case BHMSG_SYSTEM_SERVICE: ret=SysService(p, c, msg); break;
	case BHMSG_SYSTEM_VERSION: ret=Version(p, c, msg); break;
	case BHMSG_SYSTEM_SHUTDOWN: ret=Shutdown(p, c, msg); break;
	case BHMSG_SYSTEM_MESSAGE: ret=Message(p, c, msg); break;
	case BHMSG_SYSTEM_QUERY: ret=Query(p, c, msg); break;
	case BHMSG_SYSTEM_TIME: ret=Time(p, c, msg); break;
	case BHMSG_SYSTEM_STATISTICS: ret=Statistics(p, c, msg); break;
	case BHMSG_SYSTEM_PROCESS: ret=Process(p, c, msg); break;
	case BHMSG_SYSTEM_DEBUG: ret=Debug(p, c, msg); break;

	default:	break;
	}
	return ret;
}

int Connection(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	BhPosMsg rmsg;
	int		status;

#ifdef _DEBUGGERY
	if(DebugLevel > 2)
		printf("System Message: Connection\n");
#endif

	memcpy(&rmsg, msg, sizeof(rmsg));
	
	c.GetParameter(BHCI_PARAM_STATUS, &status);
	
	switch(msg->hdr.subcmd){
	case BHMSG_SYSTEM_CONNECT:
   case BHMSG_SYSTEM_SCONNECT:
#ifdef _DEBUGGERY
	if(DebugLevel > 2)
		printf("System Message: Connection\n");
#endif	
		if(SqlQ)
			SqlQ->Add(msg);
		
		break;

	case BHMSG_SYSTEM_DISCONNECT:
#ifdef _DEBUGGERY
	   if(DebugLevel > 2)
		   printf("System Message: Disconnection ");
#endif
		if(p->DeleteClient(c)){
#ifdef _DEBUGGERY
   	   if(DebugLevel > 2)
	   	   printf(" - Refused\n");
#endif         
			rmsg.hdr.Ex=BHPOS_REFUSE;
			status=POST_STATUS_DISCONNECTED;
			c.SetParameter(BHCI_PARAM_STATUS, &status);
			status=POST_MODE_NOUSER;
			c.SetParameter(BHCI_PARAM_MODE, &status);
			if(OutQ)
				OutQ->Add(&rmsg);
			return 0;
		}
		else{
#ifdef _DEBUGGERY
	      if(DebugLevel > 2)
		      printf(" - Granted\n");
#endif         
			rmsg.hdr.Ex=BHPOS_GRANT;
			if(OutQ)
				OutQ->Add(&rmsg);
			return 0;
		}
		break;
	}
	return 0;
}

int Login(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   
#ifdef _DEBUGGERY
	if(DebugLevel > 2)
		printf("System Message: Login\n");
#endif

	if(SqlQ)
		SqlQ->Add(msg);
#ifdef _DEBUGGERY
	else
	   if(DebugLevel > 2)
		printf("System Message: Login - No Sql Queue...\n");
#endif

	return 0;
}

int Authenticate(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	switch(msg->hdr.subcmd){
	case BHMSG_SYSTEM_AUTHORIZE_1:
		// decode crypt text
	case BHMSG_SYSTEM_AUTHORIZE_2:
		// encode plain text
	case BHMSG_SYSTEM_AUTHORIZE_3:
		// grant
		break;
	}
	return 0;
}

int Version(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	BhPosMsg	 rmsg;
	BhVersion	ver;
	
#ifdef _DEBUGGERY
		if(DebugLevel > 2)
			printf("System Message: version");
#endif			

	memcpy(&rmsg, msg, sizeof(rmsg));
	rmsg.hdr.size=sizeof(BhVersion);

	switch(msg->hdr.subcmd){
	case BHMSG_SYSTEM_VERSION_SERVER:
		ver.MajorVersion=BH_PLUS_SERVER_MAJVERSION;
		ver.MinorVersion=BH_PLUS_SERVER_MINVERSION;
		ver.BuildVersion=BH_PLUS_SERVER_BUILDVERSION;
		ver.packermack=0;
		memcpy(rmsg.data, &ver, sizeof(BhVersion));
		break;
	
	case BHMSG_SYSTEM_VERSION_CLIENT:
		//strcpy(rmsg.data, CLIENT_VERSION);
		break;

	case BHMSG_SYSTEM_VERSION_PROTOCOL:
		ver.MajorVersion=BH_PLUS_PROTOCOL_MAJVERSION;
		ver.MinorVersion=BH_PLUS_PROTOCOL_MINVERSION;
		ver.BuildVersion=0;
		ver.packermack=0;
		memcpy(rmsg.data, &ver, sizeof(BhVersion));
		break;
	}
	if(OutQ)
		OutQ->Add(&rmsg);
	return 0;
}

int Shutdown(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	unsigned int v=0;
	
	switch(msg->hdr.subcmd){
	case BHMSG_SYSTEM_SHUTDOWN_SHUTDOWN:
		c.GetParameter(BHCI_PARAM_STATUS, (void *) &v);
		if(v==POST_STATUS_SHUTDOWNINPROG){
			p->ClientAck(msg->hdr.sessionID);
		}
		// drop through for logging
	case BHMSG_SYSTEM_SHUTDOWN_REBOOT:
	case BHMSG_SYSTEM_SHUTDOWN_RESTART:
	case BHMSG_SYSTEM_SHUTDOWN_SLEEP:
		if(SqlQ)
			SqlQ->Add(msg);
		break;
	}
	return 0;
}
int SysService(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	return ServiceMessage(p, c, msg);
}

int Message(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	return 0;
}
int Query(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	return 0;
}
int Time(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	return 0;
}
int Statistics(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	return 0;
}
int Process(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	switch(msg->hdr.subcmd){
	case BHMSG_SYSTEM_PROCESS_INFO: return ProcInfo(p, c, msg); break;
	case BHMSG_SYSTEM_PROCESS_KILL: return ProcKill(p, c, msg); break;
	}
	return 0;
}
int ProcInfo(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	return 0;
}
int ProcKill(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	// no

	return 0;
}
int Debug(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	return 0;
}



//////////////////////////////////////////////////////////////////////////////////////////////////


int PostMessage(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
#ifdef _DEBUGGERY
		if(DebugLevel > 2)
			printf("Inq:PostMsg\n");
#endif

	switch(msg->hdr.cmd){
	case BHMSG_POST_PLU: return PLU(p, c, msg);
	case BHMSG_POST_TRANSACTION: return Transaction(p, c, msg);
	case BHMSG_POST_SERVICE: return PostService(p, c, msg);
	case BHMSG_POST_QUERY: return PostQuery(p, c, msg); 
	case BHMSG_POST_ADJUST: if(SqlQ) SqlQ->Add(msg); break;
	case BHMSG_POST_PRINTER: 
#ifdef _DEBUGGERY
		if(DebugLevel > 2)
			printf("Inq:PostPrintMsg\n");
#endif		
		return PostPrinter(p, msg);
	case BHMSG_POST_STATE:
		break;
	case BHMSG_POST_LIST: return PostList(p, c, msg);
	case BHMSG_POST_STOCKTAKE:
		break;
   case BHMSG_POST_DELIVERY: return PostDelivery(p, c, msg);
   case BHMSG_POST_RENTAL:   return PostRental(p, c, msg);
   case BHMSG_POST_CUSTOMER: return PostCustomer(p, c, msg);
   case BHMSG_POST_SIGNATURE: return PostSignature(p, c, msg);
   case BHMSG_POST_BLINDBALANCE: if(SqlQ) SqlQ->Add(msg); break;
   case BHMSG_POST_DISCREPANCY: if(SqlQ) SqlQ->Add(msg); break;
	case BHMSG_POST_EPHEMERAL: if(SqlQ) SqlQ->Add(msg); break;
   case BHMSG_POST_ATTENDANCE: if(SqlQ) SqlQ->Add(msg); break;
	default:	break;
	}
	return 0;
}

int PLU(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
#ifdef _DEBUGGERY
	if(DebugLevel > 2)
		printf("InQ: Post Message: PLU\n");
#endif

	if(SqlQ)
		SqlQ->Add(msg);
		
	return 0;
}

int Transaction(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	BhPosMsg 	     rmsg;
	bhpos::CSaleItem si;
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	switch(msg->hdr.subcmd){
	case BHMSG_POST_TRANSACTION_NOSALE:
	   if(SqlQ) SqlQ->Add(msg);
	   return 0;
	   
	case BHMSG_POST_TRANSACTION_PAYAUTH:
	case BHMSG_POST_TRANSACTION_NEW:
	case BHMSG_POST_TRANSACTION_NEWEX:
	case BHMSG_POST_TRANSACTION_RETURNSGET:
	case BHMSG_POST_TRANSACTION_RETURNSNEW:
		if(SqlQ) SqlQ->Add(msg); 
		return 0;
	
	case BHMSG_POST_TRANSACTION_END:
	case BHMSG_POST_TRANSACTION_STORE:
	case BHMSG_POST_TRANSACTION_RETRIEVE:
	case BHMSG_POST_TRANSACTION_RESTORE:
	case BHMSG_POST_TRANSACTION_RECALL:
	case BHMSG_POST_TRANSACTION_CANCEL:
	case BHMSG_POST_TRANSACTION_RET_PLU:
	case BHMSG_POST_TRANSACTION_RET_UPLU:
	case BHMSG_POST_TRANSACTION_RET_PT:
	case BHMSG_POST_TRANSACTION_RET_END:
	case BHMSG_POST_TRANSACTION_RETURNS_END:
	case BHMSG_POST_TRANSACTION_PRETURNS_END:
	case BHMSG_POST_TRANSACTION_CUSTSUMMARY:
	case BHMSG_POST_TRANSACTION_ADDITEM:
	case BHMSG_POST_TRANSACTION_ADDPAYITEM:
	case BHMSG_POST_TRANSACTION_MICRCOMMIT:
	case BHMSG_POST_TRANSACTION_CREDITCOMMIT:
		if(TransQ) TransQ->Add(msg); 
		return 0;

	case BHMSG_POST_TRANSACTION_COMMIT_PLU: 	
	{
		BhTransPluCommit *pc=(BhTransPluCommit *) msg->data;
		int tmp_numplu=pc->num_items;
		BhTransItem *plu_c=pc->items;

#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Inq: TransCommit commiting %d PLUs\n", tmp_numplu);
			printf("InQ:: Num Trans: %d this trans items: %d\n", c.NumTrans(), c.NumItems(pc->transID, bhpos::BHITEM_TYPE_ALL));
		}
#endif

		while(tmp_numplu){
			if(c.TransPLU(pc->transID, plu_c)){
#ifdef _DEBUGGERY				
				if(DebugLevel > 2){
					printf("Inq: TransCommit clientinfo refused PLU\n");
				}
#endif
				rmsg.hdr.Ex=BHPOS_REFUSE;
				break;
			}
#ifdef _DEBUGGERY				
			if(DebugLevel > 2){
				printf("Inq: PLU Commit:\nPLU\t%s\nflags\t%d\nqty\t%d\n", plu_c->PLU, plu_c->flags, plu_c->quantity);
			}
#endif
			plu_c++;
			tmp_numplu--;
		}
		// ack this one
		rmsg.hdr.Ex=BHPOS_GRANT;
	}
		break;

	case BHMSG_POST_TRANSACTION_RETURNS_PLU:
	{
		BhTransPluCommit *pc=(BhTransPluCommit *) msg->data;
		int tmp_numplu=pc->num_items;
		BhTransItem *plu_c=(BhTransItem *) pc->items;

#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Inq: TransCommit commiting %d return PLUs\n", tmp_numplu);
			printf("InQ:: Num Trans: %d this trans items: %d\n", c.NumTrans(), c.NumItems(pc->transID, bhpos::BHITEM_TYPE_ALL));
		}
#endif

		while(tmp_numplu){
		   si.SetPLU(plu_c->PLU);
         si.SetQuantity(plu_c->quantity);
         si.SetReason(plu_c->flags);
      	si.SetType(bhpos::BHSI_SITYPE_RETURN);
			if(c.TransPLU(pc->transID, si)){
#ifdef _DEBUGGERY				
				if(DebugLevel > 2){
					printf("Inq: TransCommit clientinfo refused PLU\n");
				}
#endif
				rmsg.hdr.Ex=BHPOS_REFUSE;
				break;
			}
#ifdef _DEBUGGERY				
			if(DebugLevel > 2){
				printf("Inq: Return PLU Commit:\nPLU\t%s\nflags\t%d\nqty\t%d\n", plu_c->PLU, plu_c->flags, plu_c->quantity);
			}
#endif
			plu_c++;
			tmp_numplu--;
		}
		// ack this one
		rmsg.hdr.Ex=BHPOS_GRANT;
	}
	   break;
	
	case BHMSG_POST_TRANSACTION_RENTAL_PLU:
	{
		BhTransPluCommit *pc=(BhTransPluCommit *) msg->data;
		int tmp_numplu=pc->num_items;
		BhTransItem *plu_c=pc->items;

#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Inq: TransCommit commiting %d rental PLUs\n", tmp_numplu);
			printf("InQ:: Num Trans: %d this trans items: %d\n", c.NumTrans(), c.NumItems(pc->transID, bhpos::BHITEM_TYPE_ALL));
		}
#endif

		while(tmp_numplu){
		   si.SetPLU(plu_c->PLU);
         si.SetQuantity(plu_c->quantity);
      	si.SetFlags(plu_c->flags);
      	si.SetType(bhpos::BHSI_SITYPE_RENTAL);
			if(c.TransPLU(pc->transID, si)){
#ifdef _DEBUGGERY				
				if(DebugLevel > 2){
					printf("Inq: TransCommit clientinfo refused PLU\n");
				}
#endif
				rmsg.hdr.Ex=BHPOS_REFUSE;
				break;
			}
#ifdef _DEBUGGERY				
			if(DebugLevel > 2){
				printf("Inq: Rental PLU Commit:\nPLU\t%s\nflags\t%d\nqty\t%d\n", plu_c->PLU, plu_c->flags, plu_c->quantity);
			}
#endif
			plu_c++;
			tmp_numplu--;
		}
		// ack this one
		rmsg.hdr.Ex=BHPOS_GRANT;
	}
	   break;
	
	case BHMSG_POST_TRANSACTION_LAYBY_PLU:
	{
		BhTransPluCommit *pc=(BhTransPluCommit *) msg->data;
		int tmp_numplu=pc->num_items;
		BhTransItem *plu_c=pc->items;

#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Inq: TransCommit commiting %d layby PLUs\n", tmp_numplu);
			printf("InQ:: Num Trans: %d this trans items: %d\n", c.NumTrans(), c.NumItems(pc->transID, bhpos::BHITEM_TYPE_ALL));
		}
#endif

		while(tmp_numplu){
		   si.SetPLU(plu_c->PLU);
         si.SetQuantity(plu_c->quantity);
      	si.SetFlags(plu_c->flags);
      	si.SetType(bhpos::BHSI_SITYPE_LAYBY);
			if(c.TransPLU(pc->transID, si)){
#ifdef _DEBUGGERY				
				if(DebugLevel > 2){
					printf("Inq: TransCommit clientinfo refused PLU\n");
				}
#endif
				rmsg.hdr.Ex=BHPOS_REFUSE;
				break;
			}
#ifdef _DEBUGGERY				
			if(DebugLevel > 2){
				printf("Inq: Layby PLU Commit:\nPLU\t%s\nflags\t%d\nqty\t%d\n", plu_c->PLU, plu_c->flags, plu_c->quantity);
			}
#endif
			plu_c++;
			tmp_numplu--;
		}
		// ack this one
		rmsg.hdr.Ex=BHPOS_GRANT;
	}
	
	case BHMSG_POST_TRANSACTION_COMMIT_UPLU: 	
	{
		BhTransUPluCommit *pc=(BhTransUPluCommit *) msg->data;
		int tmp_numuplu=pc->num_items;
		BhTransItemUPLU *plu_c=pc->items;		

#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Inq: TransCommit commiting %d unknown PLUs\n", tmp_numuplu);
		}
#endif


		while(tmp_numuplu){
			if(c.TransPLU(pc->transID, plu_c)){
#ifdef _DEBUGGERY				
		if(DebugLevel){
			printf("Inq: Failed to commit unknown PLU\n");
		}
#endif			   
				rmsg.hdr.Ex=BHPOS_REFUSE;
				break;
			}
#ifdef _DEBUGGERY				
			if(DebugLevel > 2){
				printf("Inq: UPLU Commit:\nPLU\t%s\nflags\t%d\nqty\t%d\n", plu_c->PLU, plu_c->flags, plu_c->quantity);
			}
#endif			
			plu_c++;
			tmp_numuplu--;
		}
		rmsg.hdr.Ex=BHPOS_GRANT;
	}
		break;
   
   case BHMSG_POST_TRANSACTION_COMMIT_PRICEPLU:
	{
		BhTransUPluCommit *pc=(BhTransUPluCommit *) msg->data;
		int tmp_numuplu=pc->num_items;
		BhTransItemUPLU *plu_c=pc->items;		

#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Inq: TransCommit commiting %d price PLUs\n", tmp_numuplu);
		}
#endif


		while(tmp_numuplu){
			if(c.TransPLU(pc->transID, plu_c)){
#ifdef _DEBUGGERY				
		if(DebugLevel){
			printf("Inq: Failed to commit price PLU\n");
		}
#endif			   
				rmsg.hdr.Ex=BHPOS_REFUSE;
				break;
			}
#ifdef _DEBUGGERY				
			if(DebugLevel > 2){
				printf("Inq: PRICEPLU Commit:\nPLU\t%s\nflags\t%d\nqty\t%d\n", plu_c->PLU, plu_c->flags, plu_c->quantity);
			}
#endif			
			plu_c++;
			tmp_numuplu--;
		}
		rmsg.hdr.Ex=BHPOS_GRANT;
	}
		break;

	case BHMSG_POST_TRANSACTION_COMMIT_PT: 
	{
		BhTransPayAuth		    pc;
//		BhTransPayCredit		 pcc;
		BhTransPaymentCommit *pay_commit;
		BhTransPayment       *pay_item;
	
		// cast for auth, or not
		memcpy(&pc, msg->data, sizeof(pc));
		
		pay_commit=(BhTransPaymentCommit *) msg->data;    
      pay_item=pay_commit->payment;

#ifdef _DEBUGGERY				
	   if(DebugLevel > 2){
			printf("InQ: TransCommit: payment: %d, %d, %f\n", pay_item->type,pay_item->flags,pay_item->amount);
		}
#endif

		if(pc.type==bhpos::PTYPE_CREDIT){
			switch(pc.flags){
			case bhpos::PTYPE_FLAG_NONE:
				// add to ccvq for a sale completion...
#ifdef _CCVENGINEMCVE				
				if(CcvQ) CcvQ->Add(msg);
#endif
				return 0;
			
			case bhpos::PTYPE_FLAG_PREAUTH:
				// add to ccvq for a preauth completion...
#ifdef _CCVENGINEMCVE
				if(CcvQ) CcvQ->Add(msg);
#endif				
				return 0;
				
			case bhpos::PTYPE_FLAG_DONE:
				// this recycled from ccv queue indicating all is well
#ifdef _DEBUGGERY				
				if(DebugLevel > 2){
					printf("InQ: TransCommit: credit done\n");
				}
#endif
				// all ok, just add to pile...
				c.TransPay(pay_commit);
				rmsg.hdr.Ex=BHPOS_GRANT;
				break;
				
			case bhpos::PTYPE_FLAG_FAIL:
				rmsg.hdr.Ex=BHPOS_REFUSE;
				break;
			}
		}
		else{
			if(c.TransPay(pay_commit)) rmsg.hdr.Ex=BHPOS_REFUSE;
			else rmsg.hdr.Ex=BHPOS_GRANT;
		}
	}
		break;
	
	case BHMSG_POST_TRANSACTION_COMMIT: 
	{
		BhTransCommit *tc=(BhTransCommit *) msg->data;
		if(c.TransCommit(tc)){
#ifdef _DEBUGGERY				
			if(DebugLevel > 2){
				printf("InQ: Transaction - Commit Refused\n");
			}
#endif			
			rmsg.hdr.Ex=BHPOS_REFUSE;
		}
		else{
			rmsg.hdr.Ex=BHPOS_GRANT;
#ifdef _DEBUGGERY
			if(DebugLevel > 2){
				printf("InQ: Transaction - Commit Granted\n");
			}
#endif
		}
		break;	
	}
	}
	
	p->UpdateClient(c);
	
	rmsg.hdr.size=0;
	if(OutQ) OutQ->Add(&rmsg);
	
	return 0;
}

///////////////////////////////////////////////////////
//
// post list...
//
//
int PostList(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   BhPosMsg     rmsg;
   GlobalPrefix *gp;
   
   memcpy(&rmsg, msg, sizeof(rmsg));
   gp=(GlobalPrefix *) rmsg.data;

	switch(msg->hdr.subcmd){
	case BHMSG_POST_LIST_TAX:
	case BHMSG_POST_LIST_PRICE:
   case BHMSG_POST_LIST_RETURNREASON:
   case BHMSG_POST_LIST_PAYTYPES:
		if(SqlQ) SqlQ->Add(msg);
		break;
      
	case BHMSG_POST_LIST_PREFIX:
#ifdef _DEBUGGERY
		if(DebugLevel > 2){
			printf("InQ: PostList: Prefix\n");
		}
#endif      
	   rmsg.hdr.Ex=BHPOS_GRANT;
	   rmsg.hdr.size=sizeof(GlobalPrefix);
	   p->GetOption(BH_PLUS_OPTION_WEIGHTBCPREFIX, &gp->weight_prefix);
      p->GetOption(BH_PLUS_OPTION_QTYBCPREFIX, &gp->quantity_prefix);
      p->GetOption(BH_PLUS_OPTION_IFBCPREFIX, &gp->interface_prefix);
      p->GetOption(BH_PLUS_OPTION_PRICEBCPREFIX, &gp->price_prefix);
      p->GetOption(BH_PLUS_OPTION_TOKENBCPREFIX, &gp->token_prefix);
      p->GetOption(BH_PLUS_OPTION_VOUCHERBCPREFIX, &gp->voucher_prefix);
      p->GetOption(BH_PLUS_OPTION_INTBCPREFIX, &gp->instore_prefix);
	   if(OutQ) OutQ->Add(msg);
	   break;
	}

	return 0;
}

///////////////////////////////////////////////////////
//
// post request...
//
//
int PostService(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   BhMcPosMsg Mcbc;
//   int        msize;
	
	memset(&Mcbc, 0, sizeof(BhMcPosMsg));
   
   Mcbc.hdr.type=BHMSG_MC_CLIENT;
	Mcbc.hdr.cmd=BHMSG_MC_CLIENT_SET_PRINTERMSG;
	Mcbc.site=MSG_ALL_SITES;
	Mcbc.dept=MSG_ALL_DEPTS;
	Mcbc.group=MSG_ALL_GROUPS;
	Mcbc.post=MSG_ALL_POSTS;
   
   //memset(&rmsg, 0, sizeof(rmsg));
   //memcpy(&rmsg, msg, sizeof(rmsg));
   
	switch(msg->hdr.subcmd){
	case BHMSG_POST_SERVICE_REQUEST_SUPER:
		c.SetRequest(BHMSG_POST_SERVICE_REQUEST_SUPER);
		break;
	case BHMSG_POST_SERVICE_REQUEST_CHANGE:
		c.SetRequest(BHMSG_POST_SERVICE_REQUEST_CHANGE);
		break;
	case BHMSG_POST_SERVICE_REQUEST_AUTH:
		c.SetRequest(BHMSG_POST_SERVICE_REQUEST_AUTH);
		break;
	case BHMSG_POST_SERVICE_REQUEST_PROBLEM:
		c.SetRequest(BHMSG_POST_SERVICE_REQUEST_PROBLEM);
		break;
	case BHMSG_POST_SERVICE_REQUEST_BREAK:
		c.SetRequest(BHMSG_POST_SERVICE_REQUEST_BREAK);
		break;
		
	case BHMSG_POST_SERVICE_SUMMARY_TRANS:
	case BHMSG_POST_SERVICE_SUMMARY_USER:
   case BHMSG_POST_SERVICE_SUMMARY_XTAPE:
   case BHMSG_POST_SERVICE_SUMMARY_ZTAPE:
	   if(SqlQ) SqlQ->Add(msg);
	   return 0;
	   
   case BHMSG_POST_PRINTER_SETUP:
      if(strlen(msg->data)!=0){
   	   p->SetOption(BH_PLUS_OPTION_PRNSETUP, msg->data);
   	   Mcbc.hdr.subcmd=BHMSG_POST_PRINTER_SETUP;
      	Mcbc.hdr.size=strlen(msg->data)+1;
   	   // fire off a multicast...
   	   if(CtrlQ) CtrlQ->Add((BhPosMsg *) &Mcbc);
   	}
	   return 0;
	case BHMSG_POST_PRINTER_INIT:
   	if(strlen(msg->data)!=0){
	      p->SetOption(BH_PLUS_OPTION_PRNINIT, msg->data);
	      Mcbc.hdr.subcmd=BHMSG_POST_PRINTER_INIT;
      	Mcbc.hdr.size=strlen(msg->data)+1;
   	   // fire off a multicast...
   	   if(CtrlQ) CtrlQ->Add((BhPosMsg *) &Mcbc);
	   }
	   return 0;
	case BHMSG_POST_PRINTER_PREPRINT:
	   if(strlen(msg->data)!=0){
	      p->SetOption(BH_PLUS_OPTION_PRNPREPRINT, msg->data);
	      Mcbc.hdr.subcmd=BHMSG_POST_PRINTER_PREPRINT;
      	Mcbc.hdr.size=strlen(msg->data)+1;
   	   // fire off a multicast...
   	   if(CtrlQ) CtrlQ->Add((BhPosMsg *) &Mcbc);
	   }
	   return 0;
	case BHMSG_POST_PRINTER_HDR:
	   if(strlen(msg->data)!=0){
	      p->SetOption(BH_PLUS_OPTION_PRNHDR, msg->data);
	      Mcbc.hdr.subcmd=BHMSG_POST_PRINTER_HDR;
      	Mcbc.hdr.size=strlen(msg->data)+1;
   	   // fire off a multicast...
   	   if(CtrlQ) CtrlQ->Add((BhPosMsg *) &Mcbc);
	   }
	   return 0;
	case BHMSG_POST_PRINTER_ITEM:
	   if(strlen(msg->data)!=0){
	      p->SetOption(BH_PLUS_OPTION_PRNITEM, msg->data);
	      Mcbc.hdr.subcmd=BHMSG_POST_PRINTER_ITEM;
      	Mcbc.hdr.size=strlen(msg->data)+1;
   	   // fire off a multicast...
   	   if(CtrlQ) CtrlQ->Add((BhPosMsg *) &Mcbc);
	   }
	   return 0;
	case BHMSG_POST_PRINTER_TRANSSUM:
	   if(strlen(msg->data)!=0){
	      p->SetOption(BH_PLUS_OPTION_PRNTRANSSUM, msg->data);
	      Mcbc.hdr.subcmd=BHMSG_POST_PRINTER_TRANSSUM;
      	Mcbc.hdr.size=strlen(msg->data)+1;
   	   // fire off a multicast...
   	   if(CtrlQ) CtrlQ->Add((BhPosMsg *) &Mcbc);
	   }
	   return 0;
	case BHMSG_POST_PRINTER_PAYSUM:
	   if(strlen(msg->data)!=0){
	      p->SetOption(BH_PLUS_OPTION_PRNPAYSUM, msg->data);
	      Mcbc.hdr.subcmd=BHMSG_POST_PRINTER_PAYSUM;
      	Mcbc.hdr.size=strlen(msg->data)+1;
   	   // fire off a multicast...
   	   if(CtrlQ) CtrlQ->Add((BhPosMsg *) &Mcbc);
	   }
	   return 0;
	case BHMSG_POST_PRINTER_FTR:
	   if(strlen(msg->data)!=0){
	      p->SetOption(BH_PLUS_OPTION_PRNFOOTER, msg->data);
	      Mcbc.hdr.subcmd=BHMSG_POST_PRINTER_FTR;
      	Mcbc.hdr.size=strlen(msg->data)+1;
   	   // fire off a multicast...
   	   if(CtrlQ) CtrlQ->Add((BhPosMsg *) &Mcbc);
	   }
	   return 0;
	case BHMSG_POST_PRINTER_POSTPRINT:
	   if(strlen(msg->data)!=0){
   	   p->SetOption(BH_PLUS_OPTION_PRNPOSTPRINT, msg->data);
   	   Mcbc.hdr.subcmd=BHMSG_POST_PRINTER_POSTPRINT;
      	Mcbc.hdr.size=strlen(msg->data)+1;
   	   // fire off a multicast...
   	   if(CtrlQ) CtrlQ->Add((BhPosMsg *) &Mcbc);
	   }
	   return 0;
	}
	p->UpdateClient(c);
	return 0;
}

int PostDelivery(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   switch(msg->hdr.subcmd){
   case BHMSG_POST_DELIVERY_ADD:
   case BHMSG_POST_DELIVERY_LIST:
   case BHMSG_POST_DELIVERY_STATUS:
   case BHMSG_POST_DELIVERY_RETURNLIST:
   case BHMSG_POST_DELIVERY_DUE:
   case BHMSG_POST_DELIVERY_PAYMENT:
   case BHMSG_POST_DELIVERY_UNDELIVERED:
      if(SqlQ) SqlQ->Add(msg);
      break;
   }
   return 0;
}

int PostRental(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   switch(msg->hdr.subcmd){
   case BHMSG_POST_RENTAL_ADD:
   case BHMSG_POST_RENTAL_LIST:
   case BHMSG_POST_RENTAL_STATUS:
   case BHMSG_POST_RENTAL_RETURN:
   case BHMSG_POST_RENTAL_DUE:
   case BHMSG_POST_RENTAL_PAYMENT:
      if(SqlQ) SqlQ->Add(msg);
      break;
   }
   return 0;
}

int PostCustomer(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   switch(msg->hdr.subcmd){
   case BHMSG_POST_CUSTOMER_NEW:
   case BHMSG_POST_CUSTOMER_LIST:
   case BHMSG_POST_CUSTOMER_DISCOUNT:
   case BHMSG_POST_CUSTOMER_PRODUCTS:
   case BHMSG_POST_CUSTOMER_CHARGEBAL:
      if(SqlQ) SqlQ->Add(msg);
      break;
   }
   return 0;
}

int PostSignature(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	switch(msg->hdr.subcmd){
	case BHMSG_POST_SIGNATURE_COMMIT:
	case BHMSG_POST_SIGNATURE_DATA:
	case BHMSG_POST_SIGNATURE_END:
	case BHMSG_POST_SIGNATURE_RETRIEVE:
	case BHMSG_POST_SIGNATURE_RETDATA:
	case BHMSG_POST_SIGNATURE_RETEND:
	   if(SqlQ) SqlQ->Add(msg);
		break;
	}
	return 0;
}

///////////////////////////////////////////////////////
//
// Returned from a multicast request...
//
//

int PostQuery(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	switch(msg->hdr.subcmd){
	case BHMSG_POST_QUERY_STATUS:	return QueryStatus(p, c, msg); break;
	case BHMSG_POST_QUERY_MODE: return QueryMode(p, c, msg); break;
	case BHMSG_POST_QUERY_WHO: return QueryWho(p, c, msg); break;
	}
	return 0;
}
int QueryCash(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	int status=*((int *) msg->data);
	c.SetParameter(BHCI_PARAM_STATUS, &status);
	p->UpdateClient(c);
	return status;
}
int AckSetCash(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	int status=*((int *) msg->data);
	c.SetParameter(BHCI_PARAM_STATUS, &status);
	p->UpdateClient(c);
	return status;
}
int QueryStatus(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	int status=*((int *) msg->data);
	c.SetParameter(BHCI_PARAM_STATUS, &status);
	p->UpdateClient(c);
	return status;
}
int QueryMode(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	int mode=*((int *) msg->data);
	c.SetParameter(BHCI_PARAM_MODE, &mode);
	p->UpdateClient(c);
	return 0;
}
int QueryWho(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	int user=*((int *) msg->data);
	c.SetParameter(BHCI_PARAM_USERID, &user);
	p->UpdateClient(c);
	return 0;
}

int PostPrinter(PLUServer *p, BhPosMsg *msg)
{
   BhPosMsg      rmsg;
   int           msgsize;
   
   memset(&rmsg, 0, sizeof(rmsg));
   memcpy(&rmsg, msg, sizeof(rmsg));
   
   BhPrinterMsg *pm=(BhPrinterMsg *) rmsg.data;
   char         *temp=pm->data;
   
   rmsg.hdr.Ex=BHPOS_REFUSE;
   rmsg.hdr.size=0;
   
#ifdef _DEBUGGERY				
   if(DebugLevel > 2){
      printf("InQ: PostPrinter message\n");
   }
#endif   
   
   switch(msg->hdr.subcmd){
	case BHMSG_POST_PRINTER_SETUP:
	   p->GetOption(BH_PLUS_OPTION_PRNSETUP, temp);
	   p->GetOption(BH_PLUS_OPTION_PRNSETUP_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }	   
	   break;
	case BHMSG_POST_PRINTER_INIT:
	   p->GetOption(BH_PLUS_OPTION_PRNINIT, temp);
	   p->GetOption(BH_PLUS_OPTION_PRNINIT_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
	   break;
	case BHMSG_POST_PRINTER_PREPRINT:
	   p->GetOption(BH_PLUS_OPTION_PRNPREPRINT, temp);
	   p->GetOption(BH_PLUS_OPTION_PRNPREPRINT_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
	   break;
	case BHMSG_POST_PRINTER_HDR:
	   p->GetOption(BH_PLUS_OPTION_PRNHDR, temp);
	   p->GetOption(BH_PLUS_OPTION_PRNHDR_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
	   break;
	case BHMSG_POST_PRINTER_ITEM:
	   p->GetOption(BH_PLUS_OPTION_PRNITEM, temp);
	   p->GetOption(BH_PLUS_OPTION_PRNITEM_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }	   
	   break;
	case BHMSG_POST_PRINTER_TRANSSUM:
	   p->GetOption(BH_PLUS_OPTION_PRNTRANSSUM, temp);
	   p->GetOption(BH_PLUS_OPTION_PRNTRANSSUM_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
	   break;
	case BHMSG_POST_PRINTER_PAYSUM:
	   p->GetOption(BH_PLUS_OPTION_PRNPAYSUM, temp);
	   p->GetOption(BH_PLUS_OPTION_PRNPAYSUM_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
	   break;
	case BHMSG_POST_PRINTER_FTR:
	   p->GetOption(BH_PLUS_OPTION_PRNFOOTER, temp);
	   p->GetOption(BH_PLUS_OPTION_PRNFOOTER_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }	   
	   break;
	case BHMSG_POST_PRINTER_POSTPRINT:
	   p->GetOption(BH_PLUS_OPTION_PRNPOSTPRINT, temp);
	   p->GetOption(BH_PLUS_OPTION_PRNPOSTPRINT_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_JRNPREPRINT:
      p->GetOption(BH_PLUS_OPTION_JRNPREPRINT, temp);
	   p->GetOption(BH_PLUS_OPTION_JRNPREPRINT_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_JRNITEM:
      p->GetOption(BH_PLUS_OPTION_JRNITEM, temp);
	   p->GetOption(BH_PLUS_OPTION_JRNITEM_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_JRNPOSTPRINT:
      p->GetOption(BH_PLUS_OPTION_JRNPOSTPRINT, temp);
	   p->GetOption(BH_PLUS_OPTION_JRNPOSTPRINT_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_LBLPREPRINT:
      p->GetOption(BH_PLUS_OPTION_LBLPREPRINT, temp);
	   p->GetOption(BH_PLUS_OPTION_LBLPREPRINT_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;   	
   case BHMSG_POST_PRINTER_LBLITEM:
      p->GetOption(BH_PLUS_OPTION_LBLITEM, temp);
	   p->GetOption(BH_PLUS_OPTION_LBLITEM_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_LBLPOSTPRINT:
      p->GetOption(BH_PLUS_OPTION_LBLPOSTPRINT, temp);
	   p->GetOption(BH_PLUS_OPTION_LBLPOSTPRINT_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_CHKPREPRINT:
      p->GetOption(BH_PLUS_OPTION_CHKPREPRINT, temp);
	   p->GetOption(BH_PLUS_OPTION_CHKPREPRINT_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_CHKITEM:
      p->GetOption(BH_PLUS_OPTION_CHKITEM, temp);
	   p->GetOption(BH_PLUS_OPTION_CHKITEM_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_CHKPOSTPRINT:
      p->GetOption(BH_PLUS_OPTION_CHKPOSTPRINT, temp);
	   p->GetOption(BH_PLUS_OPTION_CHKPOSTPRINT_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_ENDPREPRINT:
      p->GetOption(BH_PLUS_OPTION_ENDPREPRINT, temp);
	   p->GetOption(BH_PLUS_OPTION_ENDPREPRINT_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_ENDITEM:
      p->GetOption(BH_PLUS_OPTION_ENDITEM, temp);
	   p->GetOption(BH_PLUS_OPTION_ENDITEM_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_ENDPOSTPRINT:
      p->GetOption(BH_PLUS_OPTION_ENDPOSTPRINT, temp);
	   p->GetOption(BH_PLUS_OPTION_ENDPOSTPRINT_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
    case BHMSG_POST_PRINTER_SLIPPREPRINT:
      p->GetOption(BH_PLUS_OPTION_SLIPPREPRINT, temp);
	   p->GetOption(BH_PLUS_OPTION_SLIPPREPRINT_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
    case BHMSG_POST_PRINTER_SLIPHEADER:
      p->GetOption(BH_PLUS_OPTION_SLIPHEADER, temp);
	   p->GetOption(BH_PLUS_OPTION_SLIPHEADER_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_SLIPITEM:
      p->GetOption(BH_PLUS_OPTION_SLIPITEM, temp);
	   p->GetOption(BH_PLUS_OPTION_SLIPITEM_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
    case BHMSG_POST_PRINTER_SLIPFOOTER:
      p->GetOption(BH_PLUS_OPTION_SLIPFOOTER, temp);
	   p->GetOption(BH_PLUS_OPTION_SLIPFOOTER_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_SLIPPOSTPRINT:
      p->GetOption(BH_PLUS_OPTION_SLIPPOSTPRINT, temp);
	   p->GetOption(BH_PLUS_OPTION_SLIPPOSTPRINT_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_MICRPREREAD:
      p->GetOption(BH_PLUS_OPTION_MICRPREREAD, temp);
	   p->GetOption(BH_PLUS_OPTION_MICRPREREAD_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_MICRREAD:
      p->GetOption(BH_PLUS_OPTION_MICRREAD, temp);
	   p->GetOption(BH_PLUS_OPTION_MICRREAD_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_MICRSTATUS:
      p->GetOption(BH_PLUS_OPTION_MICRSTATUS, temp);
	   p->GetOption(BH_PLUS_OPTION_MICRSTATUS_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_MICRPOSTREAD:
      p->GetOption(BH_PLUS_OPTION_MICRPOSTREAD, temp);
	   p->GetOption(BH_PLUS_OPTION_MICRPOSTREAD_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_CUSTDISPPRESALE:
      p->GetOption(BH_PLUS_OPTION_CUSTDISPPRESALE, temp);
	   p->GetOption(BH_PLUS_OPTION_CUSTDISPPRESALE_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_CUSTDISPPOSTSALE:
      p->GetOption(BH_PLUS_OPTION_CUSTDISPPOSTSALE, temp);
	   p->GetOption(BH_PLUS_OPTION_CUSTDISPPOSTSALE_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
   case BHMSG_POST_PRINTER_CUSTDISPIDLE:
      p->GetOption(BH_PLUS_OPTION_CUSTDISPIDLE, temp);
	   p->GetOption(BH_PLUS_OPTION_CUSTDISPIDLE_SZ, &msgsize);
	   if(msgsize){
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      pm->size=msgsize;
	      rmsg.hdr.size=sizeof(*pm)+msgsize-1;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 0;
	   }
   	break;
	}
	
	rmsg.hdr.Ex=BHPOS_REFUSE;
   rmsg.hdr.size=0;
   if(OutQ) OutQ->Add(&rmsg);
	
	return 0;
}

int GoodsMessage(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   if(msg->hdr.cmd==BHMSG_GOODS_IN){
      switch(msg->hdr.subcmd){
      case BHMSG_GOODS_IN_POLIST:
      case BHMSG_GOODS_IN_POFETCH:
      case BHMSG_GOODS_IN_POCOMMIT:
	      if(SqlQ) SqlQ->Add(msg);
	      return 0;
      }
   }
   else if(msg->hdr.cmd==BHMSG_GOODS_OUT){
      switch(msg->hdr.subcmd){
      case BHMSG_GOODS_OUT_ORDERLIST:
      case BHMSG_GOODS_OUT_ORDERFETCH:
      case BHMSG_GOODS_OUT_ORDERCOMMIT:
	      if(SqlQ) SqlQ->Add(msg);
	      return 0;
      }
   }
   else if(msg->hdr.cmd==BHMSG_GOODS_TRANSFERIN){
      switch(msg->hdr.subcmd){
      case BHMSG_GOODS_TRANSFERIN_LIST:
      case BHMSG_GOODS_TRANSFERIN_FETCH:
      case BHMSG_GOODS_TRANSFERIN_COMMIT:
	      if(SqlQ) SqlQ->Add(msg);
	      return 0;
      }
   }
   else if(msg->hdr.cmd==BHMSG_GOODS_TRANSFEROUT){
      switch(msg->hdr.subcmd){
      case BHMSG_GOODS_TRANSFEROUT_LIST:
      case BHMSG_GOODS_TRANSFEROUT_FETCH:
      case BHMSG_GOODS_TRANSFEROUT_COMMIT:
	      if(SqlQ) SqlQ->Add(msg);
	      return 0;
      }
   }
   return 1;
}
