/*
main.cpp - a err, main function?
Copyright (C) 2001 Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos, Copyright (C) 2001-2004 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY. 
*/
#include <unistd.h>
#include <grp.h>
#include <pwd.h>
#include <sys/wait.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <syslog.h>

#include <cctype>
#include <cstring>	
#include <cstdio>
#include <cstdlib>
#include <cerrno>	

#include <pos.h>
#include "plusd.h"
#include "cfg.h"

static char *ConfigFile;  // /etc/bhpos/plusd.conf
static bool bDaemon;
static uid_t euid, ruid;

void closeall(int fd);
int steamindaemon(int nochdir, int noclose);
int porkfork();
int ParseCommandLine(PLUServer *, int argc, char *argv[]);
int SetupSig();
void sig_handler(int signal);

/* 
	DebugLevel==0 : no debug info
					1 : minimal info
					2 : verbose mode
					3 : kitchen sink debugging
*/
#ifdef _DEBUGGERY
int	DebugLevel=0;
#endif

bool	bgoingtodie;
char *pidfile;
PLUServer	*plusd=NULL; 

static int check_pidfile()
{
	mode_t mode;
	char pid[32];
	int pid_fd;
	struct flock lock;
	
	if(!pidfile)
		return -1;

	mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;

	if((pid_fd = open(pidfile, O_RDWR | O_CREAT | O_EXCL, mode)) == -1){
		if(errno != EEXIST){
			delete [] pidfile;
			return -1;
		}

		if((pid_fd = open(pidfile, O_RDWR)) == -1){
			delete [] pidfile;
			return -1;
		}
	}

   lock.l_type = F_WRLCK;
   lock.l_whence = SEEK_SET;
   lock.l_start = 0;
   lock.l_len = 0;

   if(fcntl(pid_fd, F_SETLK, &lock) == -1){
      delete [] pidfile;
      return -1;
   }

   close(pid_fd);
   
	return 0;
}

static int create_pidfile()
{
	mode_t mode;
	char pid[32];
	int pid_fd;
	struct flock lock;
	
	if(!pidfile)
		return -1;

	mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;

	if((pid_fd = open(pidfile, O_RDWR | O_EXCL, mode)) == -1){
		if(errno != EEXIST){
			delete [] pidfile;
			return -1;
		}

		if((pid_fd = open(pidfile, O_RDWR)) == -1){
			delete [] pidfile;
			return -1;
		}
	}

	lock.l_type = F_WRLCK;
	lock.l_whence = SEEK_SET;
	lock.l_start = 0;
	lock.l_len = 0;

	if(fcntl(pid_fd, F_SETLK, &lock) == -1){
		delete [] pidfile;
		return -1;
	}

	snprintf(pid, 32, "%d\n", getpid());

	if(write(pid_fd, pid, strlen(pid)) != (int) strlen(pid)){
		return -1;
	}

   close(pid_fd);
   
	return 0;
}

int bhpos_setuid()
{
   int   status;
   
   group *bhpos_group=getgrnam("plusd");
   if(bhpos_group==NULL) return 1;
   status=setegid(bhpos_group->gr_gid);
   if(status < 0){
      return 2;
   }

   passwd *bhpos_user=getpwnam("plusd");
   if(bhpos_user==NULL) return 3;
   status = seteuid(bhpos_user->pw_uid);
   if(status < 0){
      return 4;
   }
   return 0;
}
     
int bhpos_unsetuid()
{
   int status;
     
#ifdef _POSIX_SAVED_IDS
   status = seteuid(ruid);
#else
   status = setreuid(euid, ruid);
#endif
   if(status < 0){
      return 1;
   }
   return 0;
}

int loadconfig()
{
	BhPos::CConfig	cfg;
	char	   value[1024];
	char		x[256];
	double	 d=0.0;
	int		 y;
	bool		b;
	
	if(!ConfigFile) return 1;
	
	if(cfg.Load(ConfigFile)){
#ifdef _DEBUGGERY
		printf("No Config file found.\n");
#endif
		return 1;
	}
	else{

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ default ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("default", "isdaemon", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting daemon: %s\n", value);
#endif
			if(bDaemon){	
				if(*value=='Y' || *value=='y'){
					b=true;
				}
				else{
					bDaemon=false;
					b=false;
				}
				plusd->SetOption(BH_PLUS_OPTION_SETDAEMON, &b);
			}
			else{ // the command line arg for debug overrides
				b=false;
				plusd->SetOption(BH_PLUS_OPTION_SETDAEMON, &b);
			}
		}
	
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("default", "pidfile", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting piddle file: %s\n", value);
#endif
			if(pidfile) delete [] pidfile;
			if(strlen((char *) value)!=0 && strlen((char *) value) < 256){
				pidfile=new char[strlen((char *) value) + 1];
				if(pidfile) strcpy(pidfile, (char *) value);
			}
		}
	
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("default", "protocoltcp", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting Protocol TCP: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_PROTOCOLTCP, &b);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("default", "protocoludp", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting Protocol UDP: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_PROTOCOLUDP, &b);
		}
		
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("default", "host", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting host: %s\n", value);
#endif
			if(strlen(value) > 256)
				value[256]=0;
			strncpy(x, value, 256);
			plusd->SetOption(BH_PLUS_OPTION_HOST, (void *) x);
			
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("default", "port", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting port: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_PORT, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("default", "mchost", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting multicast host: %s\n", value);
#endif
			if(value){ 
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_MCHOST, (void *) value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("default", "mcport", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting multicast port: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_MCPORT, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("default", "loglevel", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting logging level: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_LOGLEVEL, &y);
		}
		

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ system ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/


		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("system", "errorredirect", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting error redirection: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_ERRREDIRECT, &b);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("system", "erroraddr", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting error host: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_ERRHOST, (void *) value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("system", "errorport", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting error redirection port: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_ERRPORT, &y);
		}
		if(!cfg.ParseRecord("system", "errorretries", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting error retries to: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_ERRRETRIES, &y);
		}
		if(!cfg.ParseRecord("system", "errortimeout", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting error timeout to: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_ERRTIMEOUT, &y);
		}		
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("system", "basecurrencyID", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting basecurrencyID: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_BASECURRENCYID, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("system", "baselocaleID", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting baselocaleID: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_BASELOCALEID, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("system", "modulesdir", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting modulesdir: %s\n", value);
#endif
			if(strlen((char *) value)!=0 && strlen((char *) value) < 256){
				strcpy(x, (char *) value);
				plusd->SetOption(BH_PLUS_OPTION_MODULESDIR, x);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("system", "pluandplay", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting PLUandPlay: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_PLUANDPLAY, &b);
		}
      memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("system", "serverside", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting Server side trans: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_SERVERSIDE, &b);
		}
      memset(value, 0, sizeof(value));
      if(!cfg.ParseRecord("system", "servertargets", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting server side targets: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_SERVERTARGETS, &b);
		}
      memset(value, 0, sizeof(value));
      if(!cfg.ParseRecord("system", "servermenu", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting server side products: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_SERVERMENU, &b);
		}
      
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ clients ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/


		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("clients", "keepalive", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting keepalive: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_KEEPALIVE, &b);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("clients", "alivetime", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting alive time: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_ALIVETIME, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("clients", "aliveretries", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting alive retries: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_ALIVERETRIES, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("clients", "addressRange", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting address Range: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_ADDRESSRANGE, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("clients", "filterHosts", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting filter Hosts: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_FILTERHOSTS, &b);
		}		
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("clients", "resettime", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting client reset time to: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_CLIENTRESETTIME, &y);
		}
		
		
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ db ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/		
		
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "cacheplu", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Cache PLUs: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_CACHEPLU, &b);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "cachelimit", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("PLU cache limit: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_CACHELIMIT, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "cachepricebreak", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Cache Price Breaks: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_CACHEPB, &b);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "cachepblimit", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("price break cache limit: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_CACHEPBLIMIT, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "multiplehosts", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("muliple db hosts: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_DBHOSTMULTIPLE, &b);
		}
      memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "datasource", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting ODBC datasource: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_DBSOURCE, x);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "host", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting database host: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_DBHOSTPRIME, x);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "database", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting db: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_DBDATABASE, x);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "user", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting db user: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_DBUSER, x);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "password", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting db password: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_DBPASSWORD, x);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "pctablename", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting pre-cached table name: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_PRECACHETABLE, x);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "sqlfile", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting db sql file: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_DBSQLFILE, x);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "usedefaultcashdrawer", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("use default cash drawer settings: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_USEDEFCASHD, &b);
		}
      memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "currencyroundtype", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("currency round type to: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_CURRRNDTYPE, &y);
		}
		
/*~~~~~~~~~~~~~~~~~~~~~~~~~ barcode stuff ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "wbcprefix", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Internal weight barcode prefix: %s\n", value);
#endif
			if(value){
				y=atoi(value);
				plusd->SetOption(BH_PLUS_OPTION_WEIGHTBCPREFIX, &y);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "qbcprefix", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Internal quantity barcode prefix: %s\n", value);
#endif
			if(value){
				y=atoi(value);
				plusd->SetOption(BH_PLUS_OPTION_QTYBCPREFIX, &y);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "pbcprefix", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Internal price barcode prefix: %s\n", value);
#endif
			if(value){
				y=atoi(value);
				plusd->SetOption(BH_PLUS_OPTION_PRICEBCPREFIX, &y);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "ifbcprefix", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Internal interface barcode prefix: %s\n", value);
#endif
			if(value){
				y=atoi(value);
				plusd->SetOption(BH_PLUS_OPTION_IFBCPREFIX, &y);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "vbcprefix", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Voucher barcode prefix: %s\n", value);
#endif
			if(value){
				y=atoi(value);
				plusd->SetOption(BH_PLUS_OPTION_VOUCHERBCPREFIX, &y);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "tbcprefix", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Token barcode prefix: %s\n", value);
#endif
			if(value){
				y=atoi(value);
				plusd->SetOption(BH_PLUS_OPTION_TOKENBCPREFIX, &y);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "intbcprefix", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("internal barcode prefix: %s\n", value);
#endif
			if(value){
				y=atoi(value);
				plusd->SetOption(BH_PLUS_OPTION_INTBCPREFIX, &y);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("db", "custbcprefix", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("customer barcode prefix: %s\n", value);
#endif
			if(value){
				y=atoi(value);
				plusd->SetOption(BH_PLUS_OPTION_CUSTOMERBCPREFIX, &y);
			}
		}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ ups ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/


		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ups", "haveups", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Have ups to: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_PLUSUPS, &b);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ups", "network", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Network ups to: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_NETWORKPLUSUPS, &b);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ups", "host", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting ups host: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_UPSHOST, (void *) value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ups", "port", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting usp port: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_UPSPORT, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ups", "updateperiod", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("UPS update period: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_UPSPERIOD, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ups", "clientcollect", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting client UPS collection to: %s\n", value);
#endif
			if(*value=='Y' || *value=='y')
				b=true;
			else
				b=false;
			plusd->SetOption(BH_PLUS_OPTION_CLIENTUPS, &b);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ups", "clientperiod", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("client UPS update period: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_CLIENTUPSPERIOD, &y);
		}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ idle ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("idle", "updateperiod", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("setting idle update period: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_IDLEUPDATEPERIOD, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("idle", "threadcheck", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("setting idle thread check update period: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_THREADCHECK, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("idle", "procwriter", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("setting idle procwriter update period: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_PROCWRITER, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("idle", "transrecommit", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("setting idle trans recommit update period: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_RECOMMIT, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("idle", "statsupdate", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("setting idle stats update period: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_STATSUPDATE, &y);
		}

/*~~~~~~~~~~~~~~~~~~~~~~~~~ global printer stuff ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "setup", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer setup: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_PRNSETUP, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "init", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer init: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_PRNINIT, value);
			}
		} 
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "preprint", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer preprint: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_PRNPREPRINT, value);
			}
		} 
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "header", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer header: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_PRNHDR, value);
			}
		} 
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "item", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer item: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_PRNSETUP, value);
			}
		} 
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "transsum", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer transsum: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_PRNTRANSSUM, value);
			}
		} 
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "paysum", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer paysum: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_PRNPAYSUM, value);
			}
		} 
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "footer", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer footer: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_PRNFOOTER, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "postprint", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer postprint: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_PRNPOSTPRINT, value);
			}
		}
		
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "journal_preprint", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer journal preprint: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_JRNPREPRINT, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "journal_item", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer journal preprint: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_JRNITEM, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "journal_postprint", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer journal preprint: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_JRNPOSTPRINT, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "lbl_preprint", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer label preprint: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_LBLPREPRINT, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "lbl_item", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer label item: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_LBLITEM, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "lbl_postprint", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer label postprint: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_LBLPOSTPRINT, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "cheque_preprint", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer cheque preprint: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_CHKPREPRINT, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "cheque_item", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer cheque item: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_CHKITEM, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "cheque_postprint", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer cheque postprint: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_CHKPOSTPRINT, value);
			}
		}
      memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "slip_preprint", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer slip preprint: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_SLIPPREPRINT, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "slip_header", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer slip header: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_SLIPHEADER, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "slip_item", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer slip item: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_SLIPITEM, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "slip_footer", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer slip footer: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_SLIPFOOTER, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "slip_postprint", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer slip postprint: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_SLIPPOSTPRINT, value);
			}
		}
      memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "endorsement_preprint", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer endorsement preprint: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_ENDPREPRINT, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "endorsement_item", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer endorsement item: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_ENDITEM, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "endorsement_postprint", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer endorsement postprint: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_ENDPOSTPRINT, value);
			}
		}

      memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "micr_preread", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer micr preread: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_MICRPREREAD, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "micr_read", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer micr read: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_MICRREAD, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "micr_status", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer micr status: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_MICRSTATUS, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "micr_postread", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer micr postread: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_MICRPOSTREAD, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "custdisp_presale", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer custdisp presale: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_CUSTDISPPRESALE, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "custdisp_postsale", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer custdisp postsale: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_CUSTDISPPOSTSALE, value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("globalprinter", "custdisp_idle", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Global printer custdisp idle: %s\n", value);
#endif
			if(value[0]!=0){
				plusd->SetOption(BH_PLUS_OPTION_CUSTDISPIDLE, value);
			}
		}


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ rps ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("rps", "address", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting rps host: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_RPHOST, (void *) value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("rps", "port", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting rps port: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_RPPORT, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("rps", "retries", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting rps retries to: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_RPRETRIES, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("rps", "timeout", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting rps timeout to: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_RPTIMEOUT, &y);
		}		

/*~~~~~~~~~~~~~~~~~~~~~ global account things ~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

      memset(value, 0, sizeof(value));
      if(!cfg.ParseRecord("account", "defaultsale", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting default sale account to: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_ACCDEFSALE, &y);
		}
      
		memset(value, 0, sizeof(value));
      if(!cfg.ParseRecord("account", "defaulttax", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting default tax account to: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_ACCDEFTAX, &y);
		}
      memset(value, 0, sizeof(value));
      if(!cfg.ParseRecord("account", "returns", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting global returns account to: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_ACCRETURNS, &y);
		}
		memset(value, 0, sizeof(value));
      if(!cfg.ParseRecord("account", "layby", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting global layby account to: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_ACCLAYBY, &y);
		}
		memset(value, 0, sizeof(value));
      if(!cfg.ParseRecord("account", "rental", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting global rentals account to: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_ACCRENTAL, &y);
		}
		memset(value, 0, sizeof(value));
      if(!cfg.ParseRecord("account", "transactionprefix", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting transaction prefix to: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_TRANSPREFIX, &y);
		}
		
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ ccv ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ccv", "host", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting ccv host: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_CCVHOST, (void *) value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ccv", "port", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting ccv port: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_CCVPORT, &y);
		}		
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ccv", "sslhost", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting ccv sslhost: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_CCVSSLHOST, (void *) value);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ccv", "sslport", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting ccv sslport: %d\n", atoi(value));
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_CCVSSLPORT, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ccv", "user", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting ccv user: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_CCVUSER, x);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ccv", "password", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting ccv password: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_CCVPASSWORD, x);
			}			
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ccv", "certificate", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting ccv certificate: %s\n", value);
#endif
			if(value){
				strcpy(x, value);
				plusd->SetOption(BH_PLUS_OPTION_CCVCERTIFICATE, x);
			}
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ccv", "conntypes", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting ccv connection types to: %s\n", value);
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_CCVCONNTYPES, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ccv", "defaultconn", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting ccv default connection to: %s\n", value);
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_CCVDEFAULTCONN, &y);
		}		
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ccv", "authpolicy", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting ccv auth policy to: %s\n", value);
#endif
			y=atoi(value);
			plusd->SetOption(BH_PLUS_OPTION_CCVAUTHPOLICY, &y);
		}
		memset(value, 0, sizeof(value));
		if(!cfg.ParseRecord("ccv", "authamount", value)){
#ifdef _DEBUGGERY
			if(DebugLevel>2)
				printf("Setting ccv auth amount: %s\n", value);
#endif
			d=atof(value);
			plusd->SetOption(BH_PLUS_OPTION_CCVAUTHAMOUNT, &d);
		}
	}
	return 0;
}

// this is only used to place the plusd in the foreground - config file covers rest
int ParseCommandLine(int *dlevel, int argc, char *argv[])
{
	int n;

	while((n=getopt(argc, argv, "?hvd:c:"))!= -1){
		switch(n){
		case 'c':	// config file	
			if(ConfigFile) delete [] ConfigFile;
			ConfigFile= new char[strlen(optarg) + 1];
			if(ConfigFile) strcpy(ConfigFile, optarg);
			break;
			
		case 'd':	// debug mode - expect number
			*dlevel=atoi(optarg);
			break;
			
		case 'v':
			printf("\n%s %s\n", PLUS_STRING, PLUS_VERSION_STRING);
			return 1;
		
		case 'h':
		case '?':
			printf("%s usage:\n\t-h\n\t-?\tthis message\n\t-v\tversion\n\t-d[n]\tdebug level n (1-3)\n", PLUS_STRING);
		default:
			break;
		}
	}
	return 0;
}

int main(int argc, char *argv[])
{
	int		 status=0;
	int		 dl=0;
	
	ConfigFile=NULL;
	pidfile=NULL;
	
	if(ParseCommandLine(&dl, argc, argv)){
		exit(0);
	}
	
   ruid = getuid();
   euid = geteuid();
   
	if(ConfigFile==NULL){
		ConfigFile=new char[22];
		if(ConfigFile) strcpy(ConfigFile, "/etc/plusd.conf");
	}
	
	if(dl!=0){
		bDaemon=false;
#ifdef _DEBUGGERY		
		DebugLevel=dl;
#endif
	}
	else bDaemon=true;
	
#ifdef _DEBUGGERY
	if(DebugLevel)
		printf("\n%s %s\n", PLUS_STRING, PLUS_VERSION_STRING);
#endif
	
	bgoingtodie=false;

	/* the one and only server */
	plusd = new PLUServer();
	
	if(!plusd){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("Can't alloc the server, ciao...\n");
#endif
		exit(1);
	}
	
	if(plusd->GetStatus()<0){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("plusd status failed...\n");
#endif
		exit(1);
	}
#ifdef _DEBUGGERY
	if(DebugLevel > 2){
		printf("\nAllocated PLU Server...\n");
	}

	if(DebugLevel){
		openlog("PLUS", LOG_PID | LOG_PERROR, LOG_USER);
	}
	else{
		openlog("PLUS", LOG_PID, LOG_USER);
	}
#endif

#ifdef _DEBUGGERY
	if(DebugLevel > 0){
	  printf("\nRunning in debug mode...\n");
	}
#endif
	
	if(loadconfig()){
		printf("Can't read config file %s... quitting\n", ConfigFile);
		exit(1);
	}
	if(pidfile==NULL){
		pidfile=new char[21];
		if(pidfile) strcpy(pidfile, "/var/run/plusd/plusd.pid");
	}
	if(check_pidfile() < 0){
		printf("Server already Running...\n");
		exit(1);
	}
	
	if(SetupSig()){
#ifdef _DEBUGGERY
		if(DebugLevel > 2)
			syslog(LOG_ERR, "Can't install sig handlers...");
#endif
		exit(1);
	}
   
	if(bDaemon){
      status=steamindaemon(0, 0);
   }
	else{
      status=plusd->Run();
   }

#ifdef _DEBUGGERY
	if(DebugLevel > 2)
		printf("\n%s going down...\n", PLUS_STRING);
#endif
	
	if(pidfile){
		unlink(pidfile);
		delete [] pidfile;
	}
	return status;
}

/* steamindaemon() - detach process from user and disappear into the background */
int steamindaemon(int nochdir, int noclose)
{
	struct rlimit resourceLimit = { 0 };
	int status = -1; 
	int fd = -1; 
	unsigned int i;

#ifdef _DEBUGGERY	
	if(DebugLevel > 2){
		printf("fork1 - to let parent exit...\n");
	}  
#endif
	switch(fork()){
	case 0:  break;
	case -1: return -1;
	default: _exit(0);          // exit the original process
	}
#ifdef _DEBUGGERY
	if(DebugLevel > 2){
		printf("Process spawned, setting sid...\n");
	}
#endif
	if(setsid() < 0)            // shoudn't fail
		return -1;

#ifdef _DEBUGGERY
	if(DebugLevel > 2){
		printf("sid set...\n");
	}

	if(DebugLevel > 2){
		printf("going for fork2 -  soon to be  a non-session group leader...\n");
	}
#endif	
	resourceLimit.rlim_max = 0; 
	status = getrlimit(RLIMIT_NOFILE, &resourceLimit); 
	if(status==-1){ 	// shouldn't happen 
		perror("resource limit"); 
		exit(1); 
	} 
	if(resourceLimit.rlim_max==0){ 
		printf("Max number of open file descriptors is 0! - work that out\n"); 
		exit(1); 
	} 
	for(i=0; i < resourceLimit.rlim_max; i++){ 
		(void) close(i); 
	} 
   
   switch(fork()){
      case 0:  break;
      case -1: return -1;
      default: _exit(0);
   }
#ifdef _DEBUGGERY
	if(DebugLevel > 2){
		printf("forked -  chdir, and umask\n");
	}
#endif
   if(!nochdir)
     chdir("/");

   umask(0);

   if(!noclose){
      fd=open("/dev/null", 2);
      dup(fd); 
      dup(fd);
   }
   
   if(create_pidfile() < 0){
		printf("no pidfile write...\n");
		exit(1);
	}

	plusd->Run();
	
   // dies here... (should clear up from ~destructor)
   // only this minor bit remains...
	if(pidfile){
		unlink(pidfile);
	}
	
	return 0;
}

int SetupSig()
{
	struct sigaction act;
	
	act.sa_handler = sig_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;

	if(sigaction(SIGALRM, &act, NULL) < 0){
#ifdef _DEBUGGERY
		if(DebugLevel){
			printf("Failed to setup SIGALRM\n");
		}
#endif		
	}
	if(sigaction(SIGTERM, &act, NULL) < 0){
#ifdef _DEBUGGERY
		if(DebugLevel){
			printf("Failed to setup SIGTERM\n");
		}
#endif		
	}
	if(sigaction(SIGINT, &act, NULL) < 0){
#ifdef _DEBUGGERY
		if(DebugLevel){
			printf("Failed to setup SIGINT\n");
		}
#endif		
	}
	if(sigaction(SIGHUP, &act, NULL) < 0){
#ifdef _DEBUGGERY
		if(DebugLevel){
			printf("Failed to setup SIGHIP\n");
		}
#endif		
	}

	return 0;	  
}

void sig_handler(int signal)
{
	switch(signal){
	case SIGINT:
	case SIGTERM:
		
#ifdef _DEBUGGERY
		if(DebugLevel > 2){
			printf("\nSigExtinction\n");
		}
#endif
		plusd->BigCrunch(true);
		break;
		
	case SIGHUP:
      if(loadconfig()){
		   printf("Can't read config file %s... quitting\n", ConfigFile);
		   exit(1);
	   }
      plusd->RestartThreads(true);
		break;
		
	case SIGALRM:
		if(bgoingtodie){
#ifdef _DEBUGGERY
			if(DebugLevel > 2){
				printf("\nHelp! Help! The Universe is closed\n"); 
			}
#endif
			plusd->BigCrunch(true);
		}
		break;
	}
}

// closeall() - close all FDs >= a specified value, a simple make sure function
void closeall(int fd)
{
    int fdlimit = sysconf(_SC_OPEN_MAX);

    while(fd < fdlimit)
      close(fd++);
}

int GetModuleOption(int o, void *d)
{
	if(plusd){
		return plusd->GetOption(o, d);
	}
	return 0;
}
