/*
menu.cpp - menuing things
Copyright (C) 2004  Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos, Copyright (C) 2001-2004 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY.
*/
#include <pthread.h>
#include <cstdio>
#include <cstring>

#include "pos.h"
#include "msgq.h"
#include "plusd.h"
#include "sql.h"

extern CMessageQueue<BhPosMsg>   *InQ;
extern CMessageQueue<BhPosMsg>   *OutQ;
extern CMessageQueue<BhPosMsg>   *SqlQ;
extern CMessageQueue<BhPosMsg>   *CtrlQ;
extern CMessageQueue<BhPosMsg>   *ErrorQ;

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

static int MenuIn(PLUServer *p, bhpos::BhSQL *menudb, bhpos::CClientInfo& c, BhPosMsg *msg);
static int MenuOut(PLUServer *p, bhpos::BhSQL *menudb, bhpos::CClientInfo& c, BhPosMsg *msg);
static int MenuHeading(PLUServer *p, bhpos::BhSQL *menudb, bhpos::CClientInfo& c, BhPosMsg *msg);
static int MenuItem(PLUServer *p, bhpos::BhSQL *menudb, bhpos::CClientInfo& c, BhPosMsg *msg);
static int MenuItemOption(PLUServer *p, bhpos::BhSQL *menudb, bhpos::CClientInfo& c, BhPosMsg *msg);
static int MenuGetImage(PLUServer *p, bhpos::BhSQL *menudb, int id, char *reffile);

int MenuMessage(PLUServer *plu, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   bhpos::BhSQL		*menudb;
   bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow   *sql_row;
   char dbhost[32];
	char dbname[32];
	char dbuser[32];
	char dbpass[32];
   
   plu->GetOption(BH_PLUS_OPTION_DBHOSTPRIME, dbhost);
   plu->GetOption(BH_PLUS_OPTION_DBDATABASE, dbname);
   plu->GetOption(BH_PLUS_OPTION_DBUSER, dbuser);
   plu->GetOption(BH_PLUS_OPTION_DBPASSWORD, dbpass);

   menudb=new bhpos::BhSQL(dbhost, dbname, 0, dbuser, dbpass);
   if(!menudb){
#ifdef _DEBUGGERY
      if(DebugLevel)		
         printf("menu : failed to create SQL object\n");
#endif   
      return 1;
   }

   if(menudb->Connect()){
#ifdef _DEBUGGERY
      if(DebugLevel)		
         printf("menu : failed to connect to SQL database - \n%s\n", menudb->ErrorMsg());
#endif			
      return 2;
   }
   
   switch(msg->hdr.cmd){
   case BHMSG_MENU_IN: return MenuIn(plu, menudb, c, msg);
   case BHMSG_MENU_OUT:  return MenuOut(plu, menudb, c, msg);
   case BHMSG_MENU_HEADING:  return MenuHeading(plu, menudb, c, msg);
   case BHMSG_MENU_ITEM:  return MenuItem(plu, menudb, c, msg);
//   case BHMSG_MENU_ITEMOPTIONS:  return MenuItemOptions(plu, menudb, c, msg);
   }
	return 0;
}

int MenuIn(PLUServer *p, bhpos::BhSQL *menudb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   switch(msg->hdr.subcmd){
   case BHMSG_MENU_IN_LIST:
   case BHMSG_MENU_IN_FETCH:
   case BHMSG_MENU_IN_COMMIT:
      break;
   }
   return 0;
}

int MenuOut(PLUServer *p, bhpos::BhSQL *menudb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   switch(msg->hdr.subcmd){
   case BHMSG_MENU_OUT_LIST:
   case BHMSG_MENU_OUT_FETCH:
   case BHMSG_MENU_OUT_COMMIT:
      break;
   }
   return 0;
}

int MenuHeading(PLUServer *p, bhpos::BhSQL *menudb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   char               qry[512];
   BhPosMsg 	       rmsg;
   bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow   *sql_row;
   BhMenuHeadingList *mhl=(BhMenuHeadingList *) rmsg.data;
   BhMenuHeading     *mh=mhl->items;
	
   memset(&rmsg, 0, sizeof(rmsg));
   memcpy(&rmsg, msg, sizeof(rmsg));

   rmsg.hdr.Ex=BHPOS_REFUSE;
   rmsg.hdr.size=0;

#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
         printf("Menu Heading Message\n");
      }
#endif
      
   switch(msg->hdr.subcmd){
   case BHMSG_MENU_HEADING_GET:
      sql_result.init(menudb, sqlq_qry156, true);
      if(sql_result.NumRows() > 0){
#ifdef _DEBUGGERY
         if(DebugLevel > 2){
            printf("menuheading : menuheading found row...\n");
         }
#endif
         if((sql_row=sql_result.GetRow())){
            sprintf(qry, sqlq_qry157, atoi(sql_row->GetField(2)));
            sql_result.init(menudb, qry, true);
            mhl->num_items=sql_result.NumRows();
            rmsg.hdr.size=sizeof(BhMenuHeadingList) - sizeof(BhMenuHeading);
            while((sql_row=sql_result.GetRow())){
               mh->id=atoi(sql_row->GetField(0));
               strncpy(mh->name, sql_row->GetField(1), 32);
               if(!MenuGetImage(p, menudb, atoi(sql_row->GetField(3)), mh->image)){
               }
               else{
                  strcpy(mh->image, "");
               }
               mh++;
               rmsg.hdr.size+=sizeof(BhMenuHeading);
            }
            rmsg.hdr.Ex=BHPOS_GRANT;
            if(OutQ) OutQ->Add(&rmsg);
         }
         else{
            if(OutQ) OutQ->Add(&rmsg);
   		   return 1;
         }
      }
      else{
   		if(OutQ) OutQ->Add(&rmsg);
   		return 1;
      }
      break;
   }
   return 0;
}

int MenuItem(PLUServer *p, bhpos::BhSQL *menudb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   char               qry[512];
   BhPosMsg 	       rmsg;
   bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow   *sql_row;
   BhMenuItemRequest *req=(BhMenuItemRequest *) msg->data;
   BhMenuItemList    *mil=(BhMenuItemList *) rmsg.data;
   BhMenuItem        *mi=mil->items;
   memset(&rmsg, 0, sizeof(rmsg));
   memcpy(&rmsg, msg, sizeof(rmsg));

   rmsg.hdr.Ex=BHPOS_REFUSE;
   rmsg.hdr.size=0;
   
   switch(msg->hdr.subcmd){
   case BHMSG_MENU_ITEM_GET:
      #ifdef _DEBUGGERY				
      if(DebugLevel > 2){
         printf("Menu item\n");
      }
#endif
      
      sql_result.init(menudb, sqlq_qry158, true);
      if(sql_result.NumRows() > 0){
#ifdef _DEBUGGERY
         if(DebugLevel > 2){
            printf("menu : menu found row...\n");
         }
#endif
         if((sql_row=sql_result.GetRow())){
            sprintf(qry, sqlq_qry158, atoi(sql_row->GetField(2)));
            sql_result.init(menudb, qry, true);
            mil->num_items=sql_result.NumRows();
            rmsg.hdr.size=sizeof(BhMenuItemList) - sizeof(BhMenuItem);
            while((sql_row=sql_result.GetRow())){
               mi->id=atoi(sql_row->GetField(0));
               //strncpy(mi->plu, sql_row->GetField(4), 128);
               strncpy(mi->name, sql_row->GetField(5), 32);
	            strncpy(mi->quantity, sql_row->GetField(7), 32);
               strncpy(mi->rectext, sql_row->GetField(8), 64);
               mi->price=atof(sql_row->GetField(9));
               mi->tax=atof(sql_row->GetField(10));
               
               if(!MenuGetImage(p, menudb, atoi(sql_row->GetField(11)), mi->image)){
               }
               else{
                  strcpy(mi->image, "");
               }
	            mi->printerid==atoi(sql_row->GetField(12));
               mi++;
               rmsg.hdr.size+=sizeof(BhMenuItem);
            }
            rmsg.hdr.Ex=BHPOS_GRANT;
            if(OutQ) OutQ->Add(&rmsg);
         }
         else{
            if(OutQ) OutQ->Add(&rmsg);
   		   return 1;
         }
      }
      else{
   		if(OutQ) OutQ->Add(&rmsg);
   		return 1;
      }
      break;
   }
   return 0;
}

int MenuItemOption(PLUServer *p, bhpos::BhSQL *menudb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   return 0;
}

int MenuGetImage(PLUServer *p, bhpos::BhSQL *menudb, int id, char *reffile)
{
   if(id==0) return 1;
      
   char               qry[512];
   bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow   *sql_row;
   
   sprintf(qry, sqlq_qry159, id);
   sql_result.init(menudb, qry, true);
   if((sql_row=sql_result.GetRow())){
      if(reffile){
         strncpy(reffile, sql_row->GetField(0), 255);
      }
      return 0;
   }
   return 1;
}
/*
int MenuGetDateTime(PLUServer *p, bhpos::BhSQL *menudb, int id)
{
   char               qry[512];
   bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow   *sql_row;
   
   sql_result.init(menudb, sqlq_qry20, true);
   sql_row=sql_result.GetRow();
   BhDateTime dt(BHTSFORMAT_MYSQL_TS14, sql_row->GetField(0));
   
   sprintf(qry, sqlq_qry155, dt.Month(), dt.Day(), dt.Hour(), dt.Minute());
   //sql_result.init(menudb, "select * from menuheading", true);
   if(sql_result.NumRows() > 0){
#ifdef _DEBUGGERY
      if(DebugLevel > 2){
         printf("schedule : found row...\n");
      }
#endif
      while(sql_row=sql_result.GetRow()){
      }
   }
   else{
      return 1;
   }
   
   return 0;
}
*/
