/*
outq.cpp - a queue thing
Copyright (C) 2001  Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos, Copyright (C) 2001-2004 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY.
*/
#include <pthread.h>
#include <cstdio>
#include <cstring>
#include <unistd.h>

#include "pos.h"
#include "msgq.h"
#include "plusd.h"

#include "bhmsgq.cpp"

CMessageQueue<BhPosMsg>   *OutQ;

extern CMessageQueue<BhPosMsg>   *ErrorQ;

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

void *OutQMonitor(void *arg)
{
	PLUServer	  *plu=(PLUServer *) arg;
	BhPosMsg		 msg;
	int			   status;
	unsigned short port;
	char		     host[33];
	sockaddr_in	 me, ret;
	int 			  sock;
	socklen_t   	opt=1;
	
	bhpos::CClientInfo c;
	
	memset(&me, 0, sizeof(me));

	me.sin_family=AF_INET;
	plu->GetOption(BH_PLUS_OPTION_PORT, &port);
	plu->GetOption(BH_PLUS_OPTION_HOST, host);
	me.sin_addr.s_addr=inet_addr(host);
	me.sin_port=htons(port);
	
	OutQ=new CMessageQueue<BhPosMsg>();
	if(!OutQ){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("No OutQ Allocation\n");
#endif			
		return NULL;
	}
	if(OutQ->GetError()){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("OutQ instansiation error\n");
#endif		
	}
/*
	status=::pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	if(status!=0){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("out thread: setcanceltype failed.\n");
#endif			
	}
	*/
	
	sock=socket(AF_INET, SOCK_DGRAM, 0);
	if(sock < 0){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("no outq socket creation");
#endif			
		delete OutQ;
		OutQ=NULL;	
		status=THREAD_DEAD;
		plu->SetThread(1, THREAD_PARAM_STATE, &status);
		return NULL;
	}
	
	if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(int))<0){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("no outq udp socket reuse");
#endif			
	}
	if(bind(sock, (sockaddr *) &me, sizeof(me)) < 0){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("no outq socket bind");
#endif			
		close(sock);
		delete OutQ;
		OutQ=NULL;	
		status=THREAD_DEAD;
		plu->SetThread(OUTQ_THREADID, THREAD_PARAM_STATE, &status);
		return NULL;
	}

   opt=sizeof(ret);

	status=OutQ->Grab();

	if(status!=0){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("OutQ failed to Grab Mutex, %d (%s)\n", status, strerror(status));
#endif		
		close(sock);	
		delete OutQ;
		OutQ=NULL;	
		status=THREAD_DEAD;
		plu->SetThread(OUTQ_THREADID, THREAD_PARAM_STATE, &status);
		return NULL;
	}

	status=THREAD_RUNNING;
	plu->SetThread(OUTQ_THREADID, THREAD_PARAM_STATE, &status);
	
	while(42){
		// wait on condition - q has msg
#ifdef _DEBUGGERY
		if(DebugLevel)				
			printf("OutQ: going for wait on condition\n");
#endif
		status=OutQ->WaitOnMsg();
		if(status != 0){
#ifdef _DEBUGGERY
			if(DebugLevel)		
				printf("OutQ thread wait failed, %d (%s)\n", status, strerror(status));
#endif
			OutQ->Release();
			close(sock);
			delete OutQ;
			OutQ=NULL;	
			status=THREAD_DEAD;
			plu->SetThread(OUTQ_THREADID, THREAD_PARAM_STATE, &status);
			return NULL;
		}
		
		if(!OutQ->HasMsg()){
			printf("OutQ: no outq message!\n");
			OutQ->GrabOnMsg();
			continue;
		}
		OutQ->GetNext(&msg);
		OutQ->DeleteLast();
		status=OutQ->Release();
		if(status != 0) {
#ifdef _DEBUGGERY
			if(DebugLevel)		
				printf("OutQ thread mutex unlock failed, %d (%s)\n", status, strerror(status));
#endif
		}
		
#ifdef _DEBUGGERY
		if(DebugLevel > 2)
			printf("\nOutQ tx packet\nto:%s:%d\ntype:%02d\ncmd:%02d\nsize:%d\ngrant:%d\nmid:%d\n", inet_ntoa((struct in_addr) msg.return_addr.sin_addr), ntohs(msg.return_addr.sin_port), msg.hdr.type, msg.hdr.cmd, msg.hdr.size, msg.hdr.Ex, msg.hdr.msgID);
#endif
		if(!msg.flags){
			if(strcmp(inet_ntoa((struct in_addr) msg.return_addr.sin_addr), "0.0.0.0")==0){
				continue;
			}
			memcpy(&ret, &msg.return_addr, sizeof(msg.return_addr));
			
			if((status=sendto(sock, (void *) &msg.hdr, sizeof(msg.hdr) + msg.hdr.size, 0, (sockaddr *) &ret, sizeof(ret))) < 0){			   
#ifdef _DEBUGGERY
      		if(DebugLevel > 2)		
      			printf("\ntx udp error: %d\n\n", status);
#endif			
			}
		}
		else if(msg.flags==1){
#ifdef _DEBUGGERY
         if(DebugLevel > 2) printf("\nOutQ tx tcp packet\n");
#endif
			if(msg.sd < 1) continue;
			if((status=send(msg.sd, (void *)&msg.hdr, sizeof(msg.hdr)+msg.hdr.size, 0))<0){
#ifdef _DEBUGGERY
      		if(DebugLevel > 2)		
      			printf("\ntx tcp error: %d\n\n", status);
#endif			
			}
		}
		else{
			// a later use else...
		}
		// check cancellation
		plu->GetThread(OUTQ_THREADID, THREAD_PARAM_DEATH, &status);
		if(status){
		   if(!OutQ->HasMsg()){
		      break;
		   }
		}
		
		// lock mutex for the conditional unlock in no msg
		status=OutQ->Grab();
		
		if(status!=0){
#ifdef _DEBUGGERY
   		if(DebugLevel)		
			   printf("OutQ failed to Grab Mutex, %d (%s)\n", status, strerror(status));
#endif			
			close(sock);
			delete OutQ;
			OutQ=NULL;	
			status=THREAD_DEAD;
			plu->SetThread(OUTQ_THREADID, THREAD_PARAM_STATE, &status);
			return NULL;
		}
		else{
#ifdef _DEBUGGERY
			if(DebugLevel)			
				printf("OutQ mutex grabbed\n");
#endif
		}		
	}
	OutQ->Release();
	delete OutQ;
	OutQ=NULL;	
	status=THREAD_DEAD;
	plu->SetThread(OUTQ_THREADID, THREAD_PARAM_STATE, &status);
	close(sock);
	
	return NULL;
}
