/*
plusd.cpp - a plu server class 
Copyright (C) 2001  Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos, Copyright (C) 2001-2004 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY.
    
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <pthread.h>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <cerrno>

#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include "pos.h"
//#include "bhsql.h"
#include "plusd.h"
#include "msgq.h"

#include "bhmsgq.cpp"

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

extern char *pidfile;
extern bool  bgoingtodie;

extern CMessageQueue<BhPosMsg>   *InQ;
extern CMessageQueue<BhPosMsg>   *OutQ;
extern CMessageQueue<BhPosMsg>   *SqlQ;
extern CMessageQueue<BhPosMsg>   *CtrlQ;
extern CMessageQueue<BhPosMsg>   *ErrorQ;
extern CMessageQueue<BhPosMsg>   *TransQ;
extern CMessageQueue<BhPosMsg>   *RPrintQ;

#ifdef _CCVENGINEMCVE
extern CMessageQueue<BhPosMsg>   *CcvQ;
#endif

int GetModuleOption(int o, void *d);
int bhpos_setuid();

PLUServer::PLUServer() 
{
	bPlugandPlay=false;
	
	bModule=true;
	bUDP=true;
	bTCP=true;
	bMulticast=true;
	bDaemon=true;
	
	bErrRedirect=false;
	err_retries=0;
	err_timeout=30;
	
	bPluandPlay=false;
   bServerSide=false;
   bServerTargets=false;
   bServerMenu=false;
	
	bCachedPlu=false;
	plucachelimit=0;
	bPreCached=false;
	precached_tablename=NULL;
	bCachedPb=false;
   pbcachelimit=0;
   currencyrndtype=0;

	bhaveUPS=false;
	bNetworkUPS=false;
	ups_host=NULL;
	ups_port=0;
	ups_period=30;
	bClientUPSCollect=false;
	clientupsperiod=60;
	client_reset_time=0;
	strcpy(int_bcprefix, "");

	price_bcprefix=20;
	quantity_bcprefix=21;
	weight_bcprefix=22;
	interface_bcprefix=24;
	customer_bcprefix=26;
	token_bcprefix=27;
	voucher_bcprefix=28;
	instore_bcprefix=29;
	
	ccvhost=NULL;
	ccvsslhost=NULL;
	ccvport=8333;
	ccvsslport=8444;
	ccvusername=NULL;
	ccvpassword=NULL;
	ccvcertificate=NULL;
	ccvconntypes=0;
	ccvdefaultconn=0;
	ccvauthpolicy=0;
	ccvauthamount=0.0;
	
	def_taxaccount=0;
	def_returnsaccount=0;
	def_transprefix=0;
	def_laybyaccount=0;
	def_rentalsaccount=0;
	
	memset(&globalprint, 0, sizeof(globalprint));

	bMultipleDb=false;;
	dbhost=NULL;
	dbname=NULL;
	dbuser=NULL;
	dbpassword=NULL;
	bUseDefaultCash=false;
	
	bCanShutdown=true;
	//bgoingtodie=false;
	shutdownt=10;

	num_connected=0;
	sessionStatus=0;
	error=0;
	//temp_buf = new char[1024]; // space for return data

	max_load=curr_load=min_time=max_time=total_msg=0;
	msg_elapsed_time=start_time=curr_time=clock();
	msg_count=0;

	mode=0;

	bkeepalive=true;
	bfilterHosts=false;
	alivetime=10;
	aliveretries=3;
	basecurrencyID=1;
	baselocaleID=1;
	modulesdir=NULL;
	addressRange=1;
	loglevel=1;
	idleupdateperiod=1;
	threadupdateperiod=1;
	procwriterperiod=10;
	transrecommitupdate=1;
	statsupdate=5;
	
	strcpy(ctrl_host, "224.0.0.1");
	ctrl_port=6673;

	listener_sock=0;
	 
	strcpy(err_host, "");
	err_port=0;

	bHasModules=false;

	memset(&thread, 0, sizeof(thread));
	
	if(pthread_mutex_init(&mutex, NULL)!=0){
#ifdef _DEBUGGERY
		if(DebugLevel>2) printf("PLUServer Constructor: mutex failed to init: \n");
#endif		
		status=-1;
	}
/*
	if(pthread_mutex_init(&t_mutex, NULL)!=0){
#ifdef _DEBUGGERY
		if(DebugLevel>2) printf("PLUServer Constructor: t_mutex failed to init: \n");
#endif		
		status=-1;
	}*/
	
#ifndef _DEBUGGERY	
	//azfd=fopen("/dev/bhpipe", "wr");
#endif


#ifdef _DEBUGGERY
	if(DebugLevel>2){
		printf("In PLUServer Constructed: \n");
	}
#endif
}

PLUServer::~PLUServer() 
{
	if(listener_sock) close(listener_sock);
	tcpthr.Close();
	if(modules){
		// destroy plug-ins
	}
	clients.clear();
	
	if(baselocale) delete [] baselocale;
	if(modulesdir) delete [] modulesdir;
	if(precached_tablename) delete [] precached_tablename;
	if(dbhost) delete [] dbhost;
	if(dbname) delete [] dbname;
	if(dbuser) delete [] dbuser;
	if(dbpassword) delete [] dbpassword;
	if(sqlfile) delete [] sqlfile;
	if(ups_host) delete [] ups_host;
	if(ccvhost) delete [] ccvhost;
	if(ccvusername) delete [] ccvusername;
	if(ccvpassword) delete [] ccvpassword;
	if(ccvsslhost) delete [] ccvsslhost;
	if(ccvcertificate) delete [] ccvcertificate;

   if(def_transprefix) delete def_transprefix;
   
	if(globalprint.setup) delete globalprint.setup;
	if(globalprint.init) delete globalprint.init;
	if(globalprint.preprint) delete globalprint.preprint;
	if(globalprint.header) delete globalprint.header;
	if(globalprint.item) delete globalprint.item;
	if(globalprint.transsum) delete globalprint.transsum;
	if(globalprint.paysum) delete globalprint.paysum;
	if(globalprint.footer) delete globalprint.footer;
	if(globalprint.postprint) delete globalprint.postprint;
	if(globalprint.slip_preprint) delete globalprint.slip_preprint;
	if(globalprint.slip_header) delete globalprint.slip_header;
   if(globalprint.slip_item) delete globalprint.slip_item;
   if(globalprint.slip_footer) delete globalprint.slip_footer;
   if(globalprint.slip_postprint) delete globalprint.slip_postprint;
	if(globalprint.lbl_preprint) delete globalprint.lbl_preprint;
   if(globalprint.lbl_item) delete globalprint.lbl_item;
   if(globalprint.lbl_postprint) delete globalprint.lbl_postprint;
	if(globalprint.jrn_preprint) delete globalprint.jrn_preprint;
	if(globalprint.jrn_item) delete globalprint.jrn_item;
	if(globalprint.jrn_postprint) delete globalprint.jrn_postprint;
	if(globalprint.end_preprint) delete globalprint.end_preprint;
	if(globalprint.end_item) delete globalprint.end_item;
	if(globalprint.end_postprint) delete globalprint.end_postprint;
	if(globalprint.chk_preprint) delete globalprint.chk_preprint;
	if(globalprint.chk_item) delete globalprint.chk_item;
	if(globalprint.chk_postprint) delete globalprint.chk_postprint;
	if(globalprint.micr_preread) delete globalprint.micr_preread;
	if(globalprint.micr_read) delete globalprint.micr_read;
	if(globalprint.micr_status) delete globalprint.micr_status;
	if(globalprint.micr_postread) delete globalprint.micr_postread;

#ifndef _DEBUGGERY	
/*
	int	azret[2];
	azblock[0]=2;
	azblock[1]=0;
	
	fwrite(&azblock, sizeof(azblock), 1, azfd);
	fread(&azret, sizeof(azret), 1, azfd);
	if(azret[0]==4) fclose(azfd);*/
#endif	
}

int PLUServer::Run()
{
	int count=0;

	Lock();

#ifdef _DEBUGGERY
	if(DebugLevel>2) printf("plusd : Running Server...\n");
#endif

	if(bTCP)
		CreateListener();
			
	if(CreateThread(ALL_THREADS)){
#ifdef _DEBUGGERY
		printf("plusd : Thread Creation Failed\n");
#endif
		return -1;
	}
	
	while(CheckThreads(ALL_THREADS)){
		count++;
		sleep(1);
		if(count==10) return -1;
	}

	//ModuleScan();
	
	Release();
   if(bhpos_setuid()){
      printf("Can't unset userid - exiting\n");
	   exit(0);
   }
	
	return JoinThreads();
}

int PLUServer::ModuleScan()
{	

#ifdef _DEBUGGERY
	if(DebugLevel>2)
		printf("Creating Module Manager\n");
#endif
		
	modules= new bhpos::CModuleCollection();
	if(!modules){
#ifdef _DEBUGGERY
	if(DebugLevel>2)
		printf("Can't alloc the module manager.\n");
#endif
		delete modules;
		modules=NULL;
		status=-4;
		return -1;
	}
#ifdef _DEBUGGERY
	if(DebugLevel>2)
		printf("Directory Scanning\n");
#endif

	if(!modulesdir){
		delete modules;
		modules=NULL;
		return -2;
	}
	
	if(modules->ScanDirectory(this, modulesdir, NULL, (void *) &QueueExists, (void *) &QueueAdd, (void *) &GetModuleOption)){
#ifdef _DEBUGGERY
	   if(DebugLevel>2)
		   printf("No modules found, presuming empty.\n");
#endif
		delete modules;
		modules=NULL;
	}
	else{
		delete [] modulesdir;
      bHasModules=true;	
	}
	
	return 0;
}

int PLUServer::SetOption(int o, void *v)
{
	if(v==NULL) return -1;
	
	Lock();
	
	switch(o){
	case BH_PLUS_OPTION_DEBUGLEVEL:
		mode|=*((int *) v);
		break;
	case BH_PLUS_OPTION_PROTOCOLUDP:
		bUDP = *((bool *) v);
		break;
	case BH_PLUS_OPTION_PROTOCOLTCP:
		 bTCP = *((bool *) v);
		break;
	case BH_PLUS_OPTION_HOST:
		strcpy(host, (char *) v);
		break;
	case BH_PLUS_OPTION_PORT:
		port=*((unsigned short *) v);
		break;
	case BH_PLUS_OPTION_MCHOST:
		strcpy(ctrl_host, (char *) v);
		break;
	case BH_PLUS_OPTION_MCPORT:
		ctrl_port=*((unsigned short *) v);
		break;
	case BH_PLUS_OPTION_LOGLEVEL:
		loglevel=*((int *) v);
		break;
	case BH_PLUS_OPTION_ERRREDIRECT:
		bErrRedirect=*((bool *) v);
		break;
	case BH_PLUS_OPTION_ERRHOST:
		strcpy(err_host, (char *) v);
		break;
	case BH_PLUS_OPTION_ERRPORT:
		err_port=*((unsigned short *) v);
		break;
	case BH_PLUS_OPTION_ERRRETRIES:
		err_retries=*((int *) v);
		break;
	case BH_PLUS_OPTION_ERRTIMEOUT:
		err_timeout=*((int *) v);
		break;
	case BH_PLUS_OPTION_KEEPALIVE:
		bkeepalive=*((bool *) v);
		break;
	case BH_PLUS_OPTION_ALIVETIME:
		alivetime=*((int *) v);
		break;
	case BH_PLUS_OPTION_ALIVERETRIES:
		aliveretries=*((int *) v);
		break;
	case BH_PLUS_OPTION_BASECURRENCYID:
		basecurrencyID=*((int *) v);
		break;
	case BH_PLUS_OPTION_BASELOCALEID:
		baselocaleID=*((int *) v);
		break;
	case BH_PLUS_OPTION_BASELOCALESTR:
		if(baselocale) delete [] baselocale;
		baselocale = new char[strlen((char *) v) + 1];
		if(baselocale) strcpy(baselocale, (char *) v);
		/* check on base local type */
		//locale.SetLocale(baselocale);
		break;
	case BH_PLUS_OPTION_MODULESDIR:
		if(modulesdir) delete [] modulesdir;
		modulesdir = new char[strlen((char *) v) + 1];
		if(modulesdir) strcpy(modulesdir, (char *) v);
		break;
	case BH_PLUS_OPTION_PLUANDPLAY:
		bPluandPlay=*((bool *) v);
		break;
   case BH_PLUS_OPTION_SERVERSIDE:
		bServerSide=*((bool *) v);
		break;
   case BH_PLUS_OPTION_SERVERTARGETS:
      bServerTargets=*((bool *) v);
      break;
   case BH_PLUS_OPTION_SERVERMENU:
      bServerMenu=*((bool *) v);
      break;
	case BH_PLUS_OPTION_ADDRESSRANGE:
		addressRange=*((int *) v);
		break;
	case BH_PLUS_OPTION_FILTERHOSTS:
		bfilterHosts=*((bool *) v);
		break;
	case BH_PLUS_OPTION_LOADMODULES:
		bModule=*((bool *) v);
		break;
	case BH_PLUS_OPTION_SETDAEMON:
		bDaemon=*((bool *) v);
		break;
	case BH_PLUS_OPTION_SETMULTICAST:
		bMulticast=*((bool *) v);
		break;
	case BH_PLUS_OPTION_SETTCP:
		bTCP=*((bool *) v);
		break;
	case BH_PLUS_OPTION_SETUDP:
		bUDP=*((bool *) v);
		break;
	case BH_PLUS_OPTION_CACHEPLU:
		bCachedPlu=*((bool *) v);
		break;
	case BH_PLUS_OPTION_CACHELIMIT:
		plucachelimit=*((int *) v);
		break;
	case BH_PLUS_OPTION_CACHEPB:
		bCachedPb=*((bool *) v);
		break;
	case BH_PLUS_OPTION_CACHEPBLIMIT:
		pbcachelimit=*((int *) v);
		break;
	case BH_PLUS_OPTION_PRECACHED:
		bPreCached=*((bool *) v);
		break;
	case BH_PLUS_OPTION_PRECACHETABLE:
		if(precached_tablename) delete [] precached_tablename;
		precached_tablename = new char[strlen((char *) v) + 1];
		if(precached_tablename) strcpy(precached_tablename, (char *) v);
		break;
	case BH_PLUS_OPTION_DBHOSTMULTIPLE:
		bMultipleDb=*((bool *) v);
		break;
	case BH_PLUS_OPTION_DBHOSTPRIME:
		if(dbhost) delete [] dbhost;
		dbhost = new char[strlen((char *) v) + 1];
		if(dbhost) strcpy(dbhost, (char *) v);
		break;
   case BH_PLUS_OPTION_DBSOURCE:
		if(dbsource) delete [] dbsource;
		dbsource = new char[strlen((char *) v) + 1];
		if(dbsource) strcpy(dbsource, (char *) v);
		break;
	case BH_PLUS_OPTION_DBDATABASE:
		if(dbname) delete [] dbname;
		dbname = new char[strlen((char *) v) + 1];
		if(dbname) strcpy(dbname, (char *) v);
		break;
	case BH_PLUS_OPTION_DBUSER:
		if(dbuser) delete [] dbuser;
		dbuser = new char[strlen((char *) v) + 1];
		if(dbuser) strcpy(dbuser, (char *) v);
		break;
	case BH_PLUS_OPTION_DBPASSWORD:
		if(dbpassword) delete [] dbpassword;
		dbpassword = new char[strlen((char *) v) + 1];
		if(dbpassword) strcpy(dbpassword, (char *) v);
		break;
	case BH_PLUS_OPTION_DBSQLFILE:
		if(sqlfile) delete [] sqlfile;
		sqlfile = new char[strlen((char *) v) + 1];
		if(sqlfile) strcpy(sqlfile, (char *) v);
		break;
	case BH_PLUS_OPTION_USEDEFCASHD:
	   bUseDefaultCash=*((bool *) v);
	   break;
   case BH_PLUS_OPTION_CURRRNDTYPE:
	   currencyrndtype=*((int *) v);
	   break;
	case BH_PLUS_OPTION_PLUSUPS:
		bhaveUPS=*((bool *) v);
		break;
	case BH_PLUS_OPTION_NETWORKPLUSUPS:
		bNetworkUPS=*((bool *) v);
		break;
	case BH_PLUS_OPTION_UPSHOST:
		if(ups_host) delete [] ups_host;
		ups_host = new char[strlen((char *) v) + 1];
		if(ups_host) strcpy(ups_host, (char *) v);
		break;
	case BH_PLUS_OPTION_UPSPORT:
		ups_port=*((unsigned short *) v);
		break;
	case BH_PLUS_OPTION_UPSPERIOD:
		ups_period=*((int *) v);
		break;
	case BH_PLUS_OPTION_CLIENTUPSCOLLECT:
		bClientUPSCollect=*((bool *) v);
		break;
	case BH_PLUS_OPTION_CLIENTUPSPERIOD:
		clientupsperiod=*((int *) v);
		break;

	case BH_PLUS_OPTION_INTBCPREFIX:
		instore_bcprefix=*((int *) v);
		break;
	case BH_PLUS_OPTION_WEIGHTBCPREFIX:
		weight_bcprefix=*((int *) v);
		break;
	case BH_PLUS_OPTION_QTYBCPREFIX:
		quantity_bcprefix=*((int *) v);
		break;
	case BH_PLUS_OPTION_PRICEBCPREFIX:
		price_bcprefix=*((int *) v);
		break;
	case BH_PLUS_OPTION_IFBCPREFIX:
		interface_bcprefix=*((int *) v);
		break;
	case BH_PLUS_OPTION_TOKENBCPREFIX:
		token_bcprefix=*((int *) v);
		break;
	case BH_PLUS_OPTION_VOUCHERBCPREFIX:
		voucher_bcprefix=*((int *) v);
		break;
	case BH_PLUS_OPTION_CUSTOMERBCPREFIX:
		customer_bcprefix=*((int *) v);
		break;

	case BH_PLUS_OPTION_IDLEUPDATEPERIOD:
		idleupdateperiod=*((int *) v);
		break;
	case BH_PLUS_OPTION_THREADCHECK:
		threadupdateperiod=*((int *) v);
		break;
	case BH_PLUS_OPTION_PROCWRITER:
		procwriterperiod=*((int *) v);
		break;
	case BH_PLUS_OPTION_RECOMMIT:
		transrecommitupdate=*((int *) v);
		break;
	case BH_PLUS_OPTION_STATSUPDATE:
		statsupdate=*((int *) v);
		break;		
	case BH_PLUS_OPTION_CCVHOST:
		if(ccvhost) delete [] ccvhost;
		ccvhost = new char[strlen((char *) v) + 1];
		if(ccvhost) strcpy(ccvhost, (char *) v);
		break;
	case BH_PLUS_OPTION_CCVPORT:
		ccvport=*((unsigned short *) v);
		break;
	case BH_PLUS_OPTION_CCVUSER:
		if(ccvusername) delete [] ccvusername;
		ccvusername = new char[strlen((char *) v) + 1];
		if(ccvusername) strcpy(ccvusername, (char *) v);
		break;
	case BH_PLUS_OPTION_CCVPASSWORD:
		if(ccvpassword) delete [] ccvpassword;
		ccvpassword = new char[strlen((char *) v) + 1];
		if(ccvpassword) strcpy(ccvpassword, (char *) v);
		break;	
	case BH_PLUS_OPTION_CCVSSLHOST:
		if(ccvsslhost) delete [] ccvsslhost;
		ccvsslhost = new char[strlen((char *) v) + 1];
		if(ccvsslhost) strcpy(ccvsslhost, (char *) v);
		break;
	case BH_PLUS_OPTION_CCVSSLPORT:
		ccvsslport=*((unsigned short *) v);
		break;
	case BH_PLUS_OPTION_CCVCONNTYPES:
		ccvconntypes=*((int *) v);
		break;
	case BH_PLUS_OPTION_CCVDEFAULTCONN:
		ccvdefaultconn=*((int *) v);
		break;
	case BH_PLUS_OPTION_CCVCERTIFICATE:
		if(ccvcertificate) delete [] ccvcertificate;
		ccvcertificate = new char[strlen((char *) v) + 1];
		if(ccvcertificate) strcpy(ccvcertificate, (char *) v);
		break;
	case BH_PLUS_OPTION_CCVAUTHPOLICY:
		ccvauthpolicy=*((int *) v);
		break;
	case BH_PLUS_OPTION_CCVAUTHAMOUNT:
		ccvauthamount=*((double *) v);
		break;

   case BH_PLUS_OPTION_ACCDEFSALE:
		def_saleaccount=*((int *) v);
	   break;
	case BH_PLUS_OPTION_ACCDEFTAX:
		def_taxaccount=*((int *) v);
	   break;
	case BH_PLUS_OPTION_ACCRETURNS:
	   def_returnsaccount=*((int *) v);
	   break;
	case BH_PLUS_OPTION_ACCLAYBY:
	   def_laybyaccount=*((int *) v);
	   break;
	case BH_PLUS_OPTION_ACCRENTAL:
	   def_rentalsaccount=*((int *) v);
	   break;
	case BH_PLUS_OPTION_TRANSPREFIX:
	   if(def_transprefix) delete def_transprefix;
		def_transprefix = new char[strlen((char *) v) + 1];
		if(def_transprefix) strcpy(def_transprefix, (char *) v);
	   break;

	case BH_PLUS_OPTION_CLIENTRESETTIME:
	   client_reset_time=*((int *) v);
	   break;
		
	case BH_PLUS_OPTION_PRNSETUP:
	   if(globalprint.setup) delete globalprint.setup;
	   globalprint.setup_size=strlen((char *) v) + 1;
		globalprint.setup = new char[globalprint.setup_size];
		if(globalprint.setup) globalprint.setup_size=ParseString((char *) v, globalprint.setup);
		break;
		
	case BH_PLUS_OPTION_PRNINIT:
	   if(globalprint.init) delete globalprint.init;
	   globalprint.init_size=strlen((char *) v) + 1;
		globalprint.init = new char[globalprint.init_size];
		if(globalprint.init) globalprint.init_size=ParseString((char *) v, globalprint.init);
		break;
		
	case BH_PLUS_OPTION_PRNPREPRINT:
	   if(globalprint.preprint) delete globalprint.preprint;
	   globalprint.preprint_size=strlen((char *) v) + 1;
		globalprint.preprint = new char[globalprint.preprint_size];
		if(globalprint.preprint) globalprint.preprint_size=ParseString((char *) v, globalprint.preprint);
		break;
	   
	case BH_PLUS_OPTION_PRNHDR:
	   if(globalprint.header) delete globalprint.header;
	   globalprint.header_size=strlen((char *) v) + 1;
		globalprint.header = new char[globalprint.header_size];
		if(globalprint.header) globalprint.header_size=ParseString((char *) v, globalprint.header);
		break;
	
	case BH_PLUS_OPTION_PRNITEM:
	   if(globalprint.item) delete globalprint.item;
	   globalprint.item_size=strlen((char *) v) + 1;
		globalprint.item = new char[globalprint.item_size];
		if(globalprint.item) globalprint.item_size=ParseString((char *) v, globalprint.item);
		break;
		
	case BH_PLUS_OPTION_PRNTRANSSUM:
	   if(globalprint.transsum) delete globalprint.transsum;
	   globalprint.transsum_size=strlen((char *) v) + 1;
		globalprint.transsum = new char[globalprint.transsum_size];
		if(globalprint.transsum) globalprint.transsum_size=ParseString((char *) v, globalprint.transsum);
		break;

	case BH_PLUS_OPTION_PRNPAYSUM:
	   if(globalprint.paysum) delete globalprint.paysum;
	   globalprint.paysum_size=strlen((char *) v) + 1;
		globalprint.paysum = new char[globalprint.paysum_size];
		if(globalprint.paysum) globalprint.paysum_size=ParseString((char *) v, globalprint.paysum);
		break;

	case BH_PLUS_OPTION_PRNFOOTER:
	   if(globalprint.footer) delete globalprint.footer;
	   globalprint.footer_size=strlen((char *) v) + 1;
		globalprint.footer = new char[globalprint.footer_size];
		if(globalprint.footer) globalprint.footer_size=ParseString((char *) v, globalprint.footer);
		break;

	case BH_PLUS_OPTION_PRNPOSTPRINT:
	   if(globalprint.postprint) delete globalprint.postprint;
	   globalprint.postprint_size=strlen((char *) v) + 1;
		globalprint.postprint = new char[globalprint.postprint_size];
		if(globalprint.postprint) globalprint.postprint_size=ParseString((char *) v, globalprint.postprint);
		break;
		
   /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Journal ~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
   
	case BH_PLUS_OPTION_JRNPREPRINT:
	   if(globalprint.jrn_preprint) delete globalprint.jrn_preprint;
	   globalprint.jrn_preprint_size=strlen((char *) v) + 1;
		globalprint.jrn_preprint = new char[globalprint.jrn_preprint_size];
		if(globalprint.jrn_preprint) globalprint.jrn_preprint_size=ParseString((char *) v, globalprint.jrn_preprint);
		break;
	case BH_PLUS_OPTION_JRNITEM:
	   if(globalprint.jrn_item) delete globalprint.jrn_item;
	   globalprint.jrn_item_size=strlen((char *) v) + 1;
		globalprint.jrn_item = new char[globalprint.jrn_item_size];
		if(globalprint.jrn_item) globalprint.jrn_item_size=ParseString((char *) v, globalprint.jrn_item);
		break;
	case BH_PLUS_OPTION_JRNPOSTPRINT:
	   if(globalprint.jrn_postprint) delete globalprint.jrn_postprint;
	   globalprint.jrn_postprint_size=strlen((char *) v) + 1;
		globalprint.jrn_postprint = new char[globalprint.jrn_postprint_size];
		if(globalprint.jrn_postprint) globalprint.jrn_postprint_size=ParseString((char *) v, globalprint.jrn_postprint);
		break;
		
   /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Label ~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
   
	case BH_PLUS_OPTION_LBLPREPRINT:
	   if(globalprint.lbl_preprint) delete globalprint.lbl_preprint;
	   globalprint.lbl_preprint_size=strlen((char *) v) + 1;
		globalprint.lbl_preprint = new char[globalprint.lbl_preprint_size];
		if(globalprint.lbl_preprint) globalprint.lbl_preprint_size=ParseString((char *) v, globalprint.lbl_preprint);
		break;
	case BH_PLUS_OPTION_LBLITEM:
	   if(globalprint.lbl_item) delete globalprint.lbl_item;
	   globalprint.lbl_item_size=strlen((char *) v) + 1;
		globalprint.lbl_item = new char[globalprint.lbl_item_size];
		if(globalprint.lbl_item) globalprint.lbl_item_size=ParseString((char *) v, globalprint.lbl_item);
		break;
	case BH_PLUS_OPTION_LBLPOSTPRINT:
	   if(globalprint.lbl_postprint) delete globalprint.lbl_postprint;
	   globalprint.lbl_postprint_size=strlen((char *) v) + 1;
		globalprint.lbl_postprint = new char[globalprint.lbl_postprint_size];
		if(globalprint.lbl_postprint) globalprint.lbl_postprint_size=ParseString((char *) v, globalprint.lbl_postprint);
		break;		
		
   /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Endorsement ~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
   
	case BH_PLUS_OPTION_ENDPREPRINT:
	   if(globalprint.end_preprint) delete globalprint.end_preprint;
	   globalprint.end_preprint_size=strlen((char *) v) + 1;
		globalprint.end_preprint = new char[globalprint.end_preprint_size];
		if(globalprint.end_preprint) globalprint.end_preprint_size=ParseString((char *) v, globalprint.end_preprint);
		break;
	case BH_PLUS_OPTION_ENDITEM:
	   if(globalprint.end_item) delete globalprint.end_item;
	   globalprint.end_item_size=strlen((char *) v) + 1;
		globalprint.end_item = new char[globalprint.end_item_size];
		if(globalprint.end_item) globalprint.end_item_size=ParseString((char *) v, globalprint.end_item);
		break;
	case BH_PLUS_OPTION_ENDPOSTPRINT:
	   if(globalprint.end_postprint) delete globalprint.end_postprint;
	   globalprint.end_postprint_size=strlen((char *) v) + 1;
		globalprint.end_postprint = new char[globalprint.end_postprint_size];
		if(globalprint.end_postprint) globalprint.end_postprint_size=ParseString((char *) v, globalprint.end_postprint);
		break;
		
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Cheque ~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
   
	case BH_PLUS_OPTION_CHKPREPRINT:
	   if(globalprint.chk_preprint) delete globalprint.chk_preprint;
	   globalprint.chk_preprint_size=strlen((char *) v) + 1;
		globalprint.chk_preprint = new char[globalprint.chk_preprint_size];
		if(globalprint.chk_preprint) globalprint.chk_preprint_size=ParseString((char *) v, globalprint.chk_preprint);
		break;
	case BH_PLUS_OPTION_CHKITEM:
	   if(globalprint.chk_item) delete globalprint.chk_item;
	   globalprint.chk_item_size=strlen((char *) v) + 1;
		globalprint.chk_item = new char[globalprint.chk_item_size];
		if(globalprint.chk_item) globalprint.chk_item_size=ParseString((char *) v, globalprint.chk_item);
		break;
	case BH_PLUS_OPTION_CHKPOSTPRINT:
	   if(globalprint.chk_postprint) delete globalprint.chk_postprint;
	   globalprint.chk_postprint_size=strlen((char *) v) + 1;
		globalprint.chk_postprint = new char[globalprint.chk_postprint_size];
		if(globalprint.chk_postprint) globalprint.chk_postprint_size=ParseString((char *) v, globalprint.chk_postprint);
		break;
		
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Slip ~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
   
	case BH_PLUS_OPTION_SLIPPREPRINT:
	   if(globalprint.slip_preprint) delete globalprint.slip_preprint;
	   globalprint.slip_preprint_size=strlen((char *) v) + 1;
		globalprint.slip_preprint = new char[globalprint.slip_preprint_size];
		if(globalprint.slip_preprint) globalprint.slip_preprint_size=ParseString((char *) v, globalprint.slip_preprint);
		break;
	case BH_PLUS_OPTION_SLIPHEADER:
	   if(globalprint.slip_header) delete globalprint.slip_header;
	   globalprint.slip_header_size=strlen((char *) v) + 1;
		globalprint.slip_header = new char[globalprint.slip_header_size];
		if(globalprint.slip_header) globalprint.slip_header_size=ParseString((char *) v, globalprint.slip_header);
		break;
	case BH_PLUS_OPTION_SLIPITEM:
	   if(globalprint.slip_item) delete globalprint.slip_item;
	   globalprint.slip_item_size=strlen((char *) v) + 1;
		globalprint.slip_item = new char[globalprint.slip_item_size];
		if(globalprint.slip_item) globalprint.slip_item_size=ParseString((char *) v, globalprint.slip_item);
		break;
	case BH_PLUS_OPTION_SLIPFOOTER:
	   if(globalprint.slip_footer) delete globalprint.slip_footer;
	   globalprint.slip_footer_size=strlen((char *) v) + 1;
		globalprint.slip_footer = new char[globalprint.slip_footer_size];
		if(globalprint.slip_footer) globalprint.slip_footer_size=ParseString((char *) v, globalprint.slip_footer);
		break;
	case BH_PLUS_OPTION_SLIPPOSTPRINT:
	   if(globalprint.slip_postprint) delete globalprint.slip_postprint;
	   globalprint.slip_postprint_size=strlen((char *) v) + 1;
		globalprint.slip_postprint = new char[globalprint.slip_postprint_size];
		if(globalprint.slip_postprint) globalprint.slip_postprint_size=ParseString((char *) v, globalprint.slip_postprint);
		break;
	
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~ MICR ~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
   
	case BH_PLUS_OPTION_MICRPREREAD:
	   if(globalprint.micr_preread) delete globalprint.micr_preread;
	   globalprint.micr_preread_size=strlen((char *) v) + 1;
		globalprint.micr_preread = new char[globalprint.micr_preread_size];
		if(globalprint.micr_preread) globalprint.micr_preread_size=ParseString((char *) v, globalprint.micr_preread);
		break;
	case BH_PLUS_OPTION_MICRREAD:
	   if(globalprint.micr_read) delete globalprint.micr_read;
	   globalprint.micr_read_size=strlen((char *) v) + 1;
		globalprint.micr_read = new char[globalprint.micr_read_size];
		if(globalprint.micr_read) globalprint.micr_read_size=ParseString((char *) v, globalprint.micr_read);
		break;
	case BH_PLUS_OPTION_MICRSTATUS:
	   if(globalprint.micr_status) delete globalprint.micr_status;
	   globalprint.micr_status_size=strlen((char *) v) + 1;
		globalprint.micr_status = new char[globalprint.micr_status_size];
		if(globalprint.micr_status) globalprint.micr_status_size=ParseString((char *) v, globalprint.micr_status);
		break;
	case BH_PLUS_OPTION_MICRPOSTREAD:
	   if(globalprint.micr_postread) delete globalprint.micr_postread;
	   globalprint.micr_postread_size=strlen((char *) v) + 1;
		globalprint.micr_postread = new char[globalprint.micr_postread_size];
		if(globalprint.micr_postread) globalprint.micr_postread_size=ParseString((char *) v, globalprint.micr_postread);
		break;
      
   /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~ customer display ~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
      
	case BH_PLUS_OPTION_CUSTDISPPRESALE:
	   if(globalprint.custdisp_presale) delete globalprint.custdisp_presale;
	   globalprint.custdisp_presale_size=strlen((char *) v) + 1;
		globalprint.custdisp_presale = new char[globalprint.custdisp_presale_size];
		if(globalprint.custdisp_presale) globalprint.custdisp_presale_size=ParseString((char *) v, globalprint.custdisp_presale);
		break;
	case BH_PLUS_OPTION_CUSTDISPPOSTSALE:
	   if(globalprint.custdisp_postsale) delete globalprint.custdisp_postsale;
	   globalprint.custdisp_postsale_size=strlen((char *) v) + 1;
		globalprint.custdisp_postsale = new char[globalprint.custdisp_postsale_size];
		if(globalprint.custdisp_postsale) globalprint.custdisp_postsale_size=ParseString((char *) v, globalprint.custdisp_postsale);
		break;
	case BH_PLUS_OPTION_CUSTDISPIDLE:
	   if(globalprint.custdisp_presale) delete globalprint.custdisp_idle;
	   globalprint.custdisp_idle_size=strlen((char *) v) + 1;
		globalprint.custdisp_idle = new char[globalprint.custdisp_idle_size];
		if(globalprint.custdisp_idle) globalprint.custdisp_idle_size=ParseString((char *) v, globalprint.custdisp_idle);
		break;
	}

	Release();

	return 0;
}

int PLUServer::GetOption(int o, void *v)
{
	if(v==NULL) return -1;
	
	switch(o){
	case BH_PLUS_OPTION_DEBUGLEVEL:
		*((int *) v) = mode;
		break;
	case BH_PLUS_OPTION_PROTOCOLUDP:
		*((bool *) v) = bUDP;
		break;
	case BH_PLUS_OPTION_PROTOCOLTCP:
		*((bool *) v) = bTCP;
		break;
	case BH_PLUS_OPTION_HOST:
		if(host)
			strcpy((char *) v, host);
		break;
	case BH_PLUS_OPTION_PORT:
		*((unsigned short *) v) = port;
		break;
	case BH_PLUS_OPTION_MCHOST:
		if(ctrl_host)
			strcpy((char *) v, ctrl_host);
		break;
	case BH_PLUS_OPTION_MCPORT:
		*((unsigned short *) v) = ctrl_port;
		break;
	case BH_PLUS_OPTION_LOGLEVEL:
		*((int *) v) = loglevel;
		break;
	case BH_PLUS_OPTION_ERRREDIRECT:
		*((bool *) v) = bErrRedirect;
		break;
	case BH_PLUS_OPTION_ERRHOST:
		strcpy((char *) v, err_host);
		break;
	case BH_PLUS_OPTION_ERRPORT:
		*((unsigned short *) v) = err_port;
		break;
		case BH_PLUS_OPTION_ERRRETRIES:
		*((int *) v)=err_retries;
		break;
	case BH_PLUS_OPTION_ERRTIMEOUT:
		*((int *) v)=err_timeout;
		break;
	case BH_PLUS_OPTION_KEEPALIVE:
		*((bool *) v) = bkeepalive;
		break;
	case BH_PLUS_OPTION_ALIVETIME:
		*((int *) v) = alivetime;
		break;
	case BH_PLUS_OPTION_ALIVERETRIES:
		*((int *) v) = aliveretries;
		break;
	case BH_PLUS_OPTION_BASECURRENCYID:
		*((int *) v) = basecurrencyID=*((int *) v);
		break;
	case BH_PLUS_OPTION_BASELOCALEID:
		*((int *) v) = baselocaleID;
		break;
	case BH_PLUS_OPTION_BASELOCALESTR:
		if(baselocale)
			strcpy((char *) v, baselocale);
		break;
	case BH_PLUS_OPTION_MODULESDIR:
		break;
	case BH_PLUS_OPTION_PLUANDPLAY:
		*((bool *) v)=bPluandPlay;
		break;
   case BH_PLUS_OPTION_SERVERSIDE:
		*((bool *) v)=bServerSide;
		break;
   case BH_PLUS_OPTION_SERVERTARGETS:
      *((bool *) v)=bServerTargets;
      break;
   case BH_PLUS_OPTION_SERVERMENU:
      *((bool *) v)=bServerMenu;
      break;
	case BH_PLUS_OPTION_ADDRESSRANGE:
		*((int *) v) = addressRange;
		break;
	case BH_PLUS_OPTION_FILTERHOSTS:
		*((bool *) v) = bfilterHosts;
		break;
	case BH_PLUS_OPTION_LOADMODULES:
		*((bool *) v)=bModule;
		break;
	case BH_PLUS_OPTION_SETDAEMON:
		*((bool *) v)=bDaemon;
		break;
	case BH_PLUS_OPTION_SETMULTICAST:
		*((bool *) v)=bMulticast;
		break;
	case BH_PLUS_OPTION_SETTCP:
		*((bool *) v)=bTCP;
		break;
	case BH_PLUS_OPTION_SETUDP:
		*((bool *) v)=bUDP;
		break;
	case BH_PLUS_OPTION_CACHEPLU:
		*((bool *) v)=bCachedPlu;
		break;
	case BH_PLUS_OPTION_CACHELIMIT:
		*((int *) v)=plucachelimit;
		break;
	case BH_PLUS_OPTION_CACHEPB:
		*((bool *) v)=bCachedPb;
		break;
	case BH_PLUS_OPTION_CACHEPBLIMIT:
		*((int *) v)=pbcachelimit;
		break;
	case BH_PLUS_OPTION_PRECACHED:
		*((bool *) v)=bPreCached;
		break;
	case BH_PLUS_OPTION_PRECACHETABLE:
		if(precached_tablename)
			strcpy((char *) v, precached_tablename);
		break;
	case BH_PLUS_OPTION_DBHOSTMULTIPLE:
		*((bool *) v)=bMultipleDb;
		break;
	case BH_PLUS_OPTION_DBHOSTPRIME:
		if(dbhost)
			strcpy((char *) v, dbhost);
		break;
   case BH_PLUS_OPTION_DBSOURCE:
		if(dbsource) 
         strcpy((char *) v, dbsource);
		break;
	case BH_PLUS_OPTION_DBDATABASE:
		if(dbname)
			strcpy((char *) v, dbname);
		break;
	case BH_PLUS_OPTION_DBUSER:
		if(dbuser)
			strcpy((char *) v,  dbuser);
		break;
	case BH_PLUS_OPTION_DBPASSWORD:
		if(dbpassword)
			strcpy((char *) v, dbpassword);
		break;
	case BH_PLUS_OPTION_DBSQLFILE:
		if(sqlfile)
			strcpy((char *) v, sqlfile);
		break;
	case BH_PLUS_OPTION_USEDEFCASHD:
	   *((bool *) v)=bUseDefaultCash;
	   break;
   case BH_PLUS_OPTION_CURRRNDTYPE:
	   *((int *) v)=currencyrndtype;
	   break;
	case BH_PLUS_OPTION_PLUSUPS:
		*((bool *) v)=bhaveUPS;
		break;
	case BH_PLUS_OPTION_NETWORKPLUSUPS:
		*((bool *) v)=bNetworkUPS;
		break;
	case BH_PLUS_OPTION_UPSHOST:
		strcpy((char *) v, ups_host);
		break;
	case BH_PLUS_OPTION_UPSPORT:
		*((unsigned short *) v)=ups_port;
		break;
	case BH_PLUS_OPTION_UPSPERIOD:
		*((int *) v)=ups_period;
		break;
	case BH_PLUS_OPTION_CLIENTUPSCOLLECT:
		*((bool *) v)=bClientUPSCollect;
		break;
	case BH_PLUS_OPTION_CLIENTUPSPERIOD:
		*((int *) v)=clientupsperiod;
		break;

	case BH_PLUS_OPTION_INTBCPREFIX:
	   *((int *) v)=instore_bcprefix;
		break;
	case BH_PLUS_OPTION_WEIGHTBCPREFIX:
		*((int *) v)=weight_bcprefix;
		break;
	case BH_PLUS_OPTION_QTYBCPREFIX:
		*((int *) v)=quantity_bcprefix;
		break;
	case BH_PLUS_OPTION_PRICEBCPREFIX:
		*((int *) v)=price_bcprefix;
		break;
	case BH_PLUS_OPTION_IFBCPREFIX:
		*((int *) v)=interface_bcprefix;
		break;
	case BH_PLUS_OPTION_TOKENBCPREFIX:
		*((int *) v)=token_bcprefix;
		break;
	case BH_PLUS_OPTION_VOUCHERBCPREFIX:
		*((int *) v)=voucher_bcprefix;
		break;
	case BH_PLUS_OPTION_CUSTOMERBCPREFIX:
		*((int *) v)=customer_bcprefix;
		break;

	case BH_PLUS_OPTION_IDLEUPDATEPERIOD:
		*((int *) v)=idleupdateperiod;
		break;
	case BH_PLUS_OPTION_THREADCHECK:
		*((int *) v)=threadupdateperiod;
		break;
	case BH_PLUS_OPTION_PROCWRITER:
		*((int *) v)=procwriterperiod;
		break;
	case BH_PLUS_OPTION_RECOMMIT:
		*((int *) v)=transrecommitupdate;
		break;
	case BH_PLUS_OPTION_STATSUPDATE:
		*((int *) v)=statsupdate;
		break;		
	case BH_PLUS_OPTION_CCVHOST:
		if(ccvhost)
			strcpy((char *) v, ccvhost);
		break;
	case BH_PLUS_OPTION_CCVPORT:
		*((unsigned short *) v)=ccvport;
		break;
	case BH_PLUS_OPTION_CCVUSER:
		if(ccvusername) strcpy((char *) v, ccvusername);
		break;
	case BH_PLUS_OPTION_CCVPASSWORD:
		if(ccvpassword) strcpy((char *) v, ccvpassword);
		break;	
	case BH_PLUS_OPTION_CCVSSLHOST:
		if(ccvsslhost) strcpy((char *) v, ccvsslhost);
		break;
	case BH_PLUS_OPTION_CCVSSLPORT:
		*((unsigned short *) v)=ccvsslport;
		break;
	case BH_PLUS_OPTION_CCVCONNTYPES:
		*((int *) v)=ccvconntypes;
		break;
	case BH_PLUS_OPTION_CCVDEFAULTCONN:
		*((int *) v)=ccvdefaultconn;
		break;
	case BH_PLUS_OPTION_CCVCERTIFICATE:
		if(ccvcertificate) strcpy((char *) v, ccvcertificate);
		break;	
	case BH_PLUS_OPTION_CCVAUTHPOLICY:
		*((int *) v)=ccvauthpolicy;
		break;
	case BH_PLUS_OPTION_CCVAUTHAMOUNT:
		*((double *) v)=ccvauthamount;
		break;

	case BH_PLUS_OPTION_ACCDEFTAX:
		*((int *) v)=def_taxaccount;
	   break;
	case BH_PLUS_OPTION_ACCRETURNS:
	   *((int *) v)=def_returnsaccount;
	   break;
	case BH_PLUS_OPTION_ACCLAYBY:
	   *((int *) v)=def_laybyaccount;
	   break;
	case BH_PLUS_OPTION_ACCRENTAL:
	   *((int *) v)=def_rentalsaccount;
	   break;
	case BH_PLUS_OPTION_TRANSPREFIX:
	   if(def_transprefix) strcpy((char *) v, def_transprefix);
	   break;

	case BH_PLUS_OPTION_CLIENTRESETTIME:
	   *((int *) v)=client_reset_time;
	   break;

	case BH_PLUS_OPTION_PRNSETUP:
		if(globalprint.setup) memcpy((char *) v, globalprint.setup, globalprint.setup_size);
		break;
	case BH_PLUS_OPTION_PRNINIT:
		if(globalprint.setup) memcpy((char *) v, globalprint.init, globalprint.init_size);
		break;
	case BH_PLUS_OPTION_PRNPREPRINT:
		if(globalprint.preprint) memcpy((char *) v, globalprint.preprint, globalprint.preprint_size);
		break;
	case BH_PLUS_OPTION_PRNHDR:
		if(globalprint.header) memcpy((char *) v, globalprint.header, globalprint.header_size);
		break;
	case BH_PLUS_OPTION_PRNITEM:
		if(globalprint.item) memcpy((char *) v, globalprint.item, globalprint.item_size);
		break;
	case BH_PLUS_OPTION_PRNTRANSSUM:
		if(globalprint.transsum) memcpy((char *) v, globalprint.transsum, globalprint.transsum_size);
		break;
	case BH_PLUS_OPTION_PRNPAYSUM:
		if(globalprint.paysum) memcpy((char *) v, globalprint.paysum, globalprint.paysum_size);
		break;
	case BH_PLUS_OPTION_PRNFOOTER:
		if(globalprint.footer) memcpy((char *) v, globalprint.footer, globalprint.footer_size);
		break;
	case BH_PLUS_OPTION_PRNPOSTPRINT:
		if(globalprint.postprint) memcpy((char *) v, globalprint.postprint, globalprint.postprint_size);
		break;
	case BH_PLUS_OPTION_JRNPREPRINT:
		if(globalprint.jrn_preprint) memcpy((char *) v, globalprint.jrn_preprint, globalprint.jrn_preprint_size);
		break;
	case BH_PLUS_OPTION_JRNITEM:
		if(globalprint.jrn_item) memcpy((char *) v, globalprint.jrn_item, globalprint.jrn_item_size);
		break;
	case BH_PLUS_OPTION_JRNPOSTPRINT:
		if(globalprint.jrn_postprint) memcpy((char *) v, globalprint.jrn_postprint, globalprint.jrn_postprint_size);
		break;
	case BH_PLUS_OPTION_LBLPREPRINT:
		if(globalprint.lbl_preprint) memcpy((char *) v, globalprint.lbl_preprint, globalprint.lbl_preprint_size);
		break;
	case BH_PLUS_OPTION_LBLITEM:
		if(globalprint.lbl_item) memcpy((char *) v, globalprint.lbl_item, globalprint.lbl_item_size);
		break;
	case BH_PLUS_OPTION_LBLPOSTPRINT:
		if(globalprint.lbl_postprint) memcpy((char *) v, globalprint.lbl_postprint, globalprint.lbl_postprint_size);
		break;
	case BH_PLUS_OPTION_ENDPREPRINT:
		if(globalprint.end_preprint) memcpy((char *) v, globalprint.end_preprint, globalprint.end_preprint_size);
		break;
	case BH_PLUS_OPTION_ENDITEM:
		if(globalprint.end_item) memcpy((char *) v, globalprint.end_item, globalprint.end_item_size);
		break;
	case BH_PLUS_OPTION_ENDPOSTPRINT:
		if(globalprint.end_postprint) memcpy((char *) v, globalprint.end_postprint, globalprint.end_postprint_size);
		break;
	case BH_PLUS_OPTION_CHKPREPRINT:
		if(globalprint.chk_preprint) memcpy((char *) v, globalprint.chk_preprint, globalprint.chk_preprint_size);
		break;
	case BH_PLUS_OPTION_CHKITEM:
		if(globalprint.chk_item) memcpy((char *) v, globalprint.chk_item, globalprint.chk_item_size);
		break;
	case BH_PLUS_OPTION_CHKPOSTPRINT:
		if(globalprint.chk_postprint) memcpy((char *) v, globalprint.chk_postprint, globalprint.chk_postprint_size);
		break;
   case BH_PLUS_OPTION_SLIPPREPRINT:
		if(globalprint.chk_preprint) memcpy((char *) v, globalprint.slip_preprint, globalprint.slip_preprint_size);
		break;
   case BH_PLUS_OPTION_SLIPHEADER:
		if(globalprint.slip_header) memcpy((char *) v, globalprint.slip_header, globalprint.slip_header_size);
		break;
	case BH_PLUS_OPTION_SLIPITEM:
		if(globalprint.slip_item) memcpy((char *) v, globalprint.slip_item, globalprint.slip_item_size);
		break;
	case BH_PLUS_OPTION_SLIPFOOTER:
		if(globalprint.slip_footer) memcpy((char *) v, globalprint.slip_footer, globalprint.slip_footer_size);
		break;
	case BH_PLUS_OPTION_SLIPPOSTPRINT:
		if(globalprint.slip_postprint) memcpy((char *) v, globalprint.slip_postprint, globalprint.slip_postprint_size);
		break;
	case BH_PLUS_OPTION_MICRPREREAD:
		if(globalprint.micr_preread) memcpy((char *) v, globalprint.micr_preread, globalprint.micr_preread_size);
		break;
	case BH_PLUS_OPTION_MICRREAD:
		if(globalprint.micr_read) memcpy((char *) v, globalprint.micr_read, globalprint.micr_read_size);
		break;
	case BH_PLUS_OPTION_MICRSTATUS:
		if(globalprint.micr_status) memcpy((char *) v, globalprint.micr_status, globalprint.micr_status_size);
		break;
	case BH_PLUS_OPTION_MICRPOSTREAD:
		if(globalprint.micr_status) memcpy((char *) v, globalprint.micr_postread, globalprint.micr_postread_size);
		break;
	case BH_PLUS_OPTION_CUSTDISPPRESALE:
		if(globalprint.custdisp_presale) memcpy((char *) v, globalprint.custdisp_presale, globalprint.custdisp_presale_size);
		break;
	case BH_PLUS_OPTION_CUSTDISPPOSTSALE:
		if(globalprint.custdisp_postsale) memcpy((char *) v, globalprint.custdisp_postsale, globalprint.custdisp_postsale_size);
		break;
	case BH_PLUS_OPTION_CUSTDISPIDLE:
		if(globalprint.custdisp_idle) memcpy((char *) v, globalprint.custdisp_idle, globalprint.custdisp_idle_size);
		break;
		
	case BH_PLUS_OPTION_PRNSETUP_SZ:
	   *((int *) v)=globalprint.setup_size;
	   break;
	case BH_PLUS_OPTION_PRNINIT_SZ:
	   *((int *) v)=globalprint.init_size;
	   break;
	case BH_PLUS_OPTION_PRNPREPRINT_SZ:
	   *((int *) v)=globalprint.preprint_size;
	   break;
	case BH_PLUS_OPTION_PRNHDR_SZ:
	   *((int *) v)=globalprint.header_size;
	   break;
	case BH_PLUS_OPTION_PRNITEM_SZ:
	   *((int *) v)=globalprint.item_size;
	   break;
	case BH_PLUS_OPTION_PRNTRANSSUM_SZ:
	   *((int *) v)=globalprint.transsum_size;
	   break;
	case BH_PLUS_OPTION_PRNPAYSUM_SZ:
	   *((int *) v)=globalprint.paysum_size;
	   break;
	case BH_PLUS_OPTION_PRNFOOTER_SZ:
	   *((int *) v)=globalprint.footer_size;
	   break;
	case BH_PLUS_OPTION_PRNPOSTPRINT_SZ:
	   *((int *) v)=globalprint.postprint_size;
	   break;
	case BH_PLUS_OPTION_JRNPREPRINT_SZ:
	   *((int *) v)=globalprint.jrn_preprint_size;
	   break;
	case BH_PLUS_OPTION_JRNITEM_SZ:
	   *((int *) v)=globalprint.jrn_item_size;
	   break;
	case BH_PLUS_OPTION_JRNPOSTPRINT_SZ:
	   *((int *) v)=globalprint.jrn_postprint_size;
	   break;	
	case BH_PLUS_OPTION_LBLPREPRINT_SZ:
	   *((int *) v)=globalprint.lbl_preprint_size;
	   break;
	case BH_PLUS_OPTION_LBLITEM_SZ:
	   *((int *) v)=globalprint.lbl_item_size;
	   break;
	case BH_PLUS_OPTION_LBLPOSTPRINT_SZ:
	   *((int *) v)=globalprint.lbl_postprint_size;
	   break; 
	case BH_PLUS_OPTION_ENDPREPRINT_SZ:
	   *((int *) v)=globalprint.end_preprint_size;
	   break;
	case BH_PLUS_OPTION_ENDITEM_SZ:
	   *((int *) v)=globalprint.end_item_size;
	   break;
	case BH_PLUS_OPTION_ENDPOSTPRINT_SZ:
	   *((int *) v)=globalprint.end_postprint_size;
	   break;
	case BH_PLUS_OPTION_CHKPREPRINT_SZ:
	   *((int *) v)=globalprint.chk_preprint_size;
	   break;
	case BH_PLUS_OPTION_CHKITEM_SZ:
	   *((int *) v)=globalprint.chk_item_size;
	   break;
	case BH_PLUS_OPTION_CHKPOSTPRINT_SZ:
	   *((int *) v)=globalprint.chk_postprint_size;
	   break;
	case BH_PLUS_OPTION_SLIPPREPRINT_SZ:
	   *((int *) v)=globalprint.slip_preprint_size;
	   break;
	case BH_PLUS_OPTION_SLIPHEADER_SZ:
	   *((int *) v)=globalprint.slip_header_size;
	   break;
	case BH_PLUS_OPTION_SLIPITEM_SZ:
	   *((int *) v)=globalprint.slip_item_size;
	   break;
	case BH_PLUS_OPTION_SLIPFOOTER_SZ:
	   *((int *) v)=globalprint.slip_footer_size;
	   break;
	case BH_PLUS_OPTION_SLIPPOSTPRINT_SZ:
	   *((int *) v)=globalprint.slip_postprint_size;
	   break;
	case BH_PLUS_OPTION_MICRPREREAD_SZ:
	   *((int *) v)=globalprint.micr_preread_size;
	   break;
	case BH_PLUS_OPTION_MICRREAD_SZ:
	   *((int *) v)=globalprint.micr_read_size;
	   break;
	case BH_PLUS_OPTION_MICRSTATUS_SZ:
	   *((int *) v)=globalprint.micr_status_size;
	   break;
	case BH_PLUS_OPTION_MICRPOSTREAD_SZ:
	   *((int *) v)=globalprint.micr_postread_size;
	   break;
	case BH_PLUS_OPTION_CUSTDISPPOSTSALE_SZ:
   	*((int *) v)=globalprint.custdisp_postsale_size;
	   break;
	case BH_PLUS_OPTION_CUSTDISPPRESALE_SZ:
	   *((int *) v)=globalprint.custdisp_presale_size;
	   break;
	case BH_PLUS_OPTION_CUSTDISPIDLE_SZ:
	   *((int *) v)=globalprint.custdisp_idle_size;
	   break;
	}
	return 0;
}

int PLUServer::GetParameter(int p, void *v)
{
	if(v==NULL) return -1;
	
	Lock();
	switch(p){
	case BH_PLUS_PARM_NUMCLIENTS:
		*((int *) v)=clients.size();
		break;
	case BH_PLUS_PARM_NUMCLIENTACK:
		*((int *) v)=client_collector.NumLeft();
		break;
	case BH_PLUS_PARM_NUMOPENTRANS:
		break;
	case BH_PLUS_PARM_CURRLOAD:
		*((double *) v)=curr_load;
		break;
	case BH_PLUS_PARM_MAXLOAD:
		*((double *) v)=max_load;
		break;
	case BH_PLUS_PARM_CURRTIME:
		*((clock_t *) v)=curr_time;
		break;
	case BH_PLUS_PARM_STARTTIME:
		*((clock_t *) v)=start_time;
		break;
	case BH_PLUS_PARM_UPTIME:
		*((clock_t *) v)=up_time;
		break;
	}
	Release();
	return 0;
}

int PLUServer::SetParameter(int p, void *v)
{
	if(v==NULL) return -1;
	
	Lock();
	switch(p){
	case BH_PLUS_PARM_CURRLOAD:
		curr_load=*((double *) v);
		break;
	case BH_PLUS_PARM_MAXLOAD:
		max_load=*((double *) v);
		break;
	case BH_PLUS_PARM_CURRTIME:
		curr_time=*((clock_t *) v);
		break;
	case BH_PLUS_PARM_STARTTIME:
		start_time=*((clock_t *) v);
		break;
	case BH_PLUS_PARM_UPTIME:
		up_time=*((clock_t *) v);
		break;
	}
	Release();
	return 0;
}

int PLUServer::AdjParameter(int p)
{
	Lock();
	switch(p){
	case BH_PLUS_PARM_MSGCOUNT:
		msg_count++;
		break;	
	case BH_PLUS_PARM_TOTALMSG:
		total_msg++;
		break;
	}
	Release();
	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////////


int PLUServer::GetClientInfo(int cid, unsigned int param, void *data)
{
	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////////

int PLUServer::CreateListener()
{
	socklen_t opt=1;
	sockaddr_in sin;

	memset(&sin, 0, sizeof(sin));
	sin.sin_family=AF_INET;
	sin.sin_addr.s_addr=inet_addr(host);
	sin.sin_port=htons(port);
	
	listener_sock=socket(AF_INET, SOCK_STREAM, 0);
	if(listener_sock < 0){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("no lister socket creation");
#endif
		return 1;
	}
	
	if(setsockopt(listener_sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(int))<0){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("no listener socket reuse");
#endif
	}
	if(bind(listener_sock, (sockaddr *) &sin, sizeof(sin)) < 0){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("no listener socket bind");
#endif
		close(listener_sock);
		return 1;
	}
	if(tcpthr.Create(this, listener_sock)){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("no tcp worker set create");
#endif
		close(listener_sock);
		return 1;
	}

	if(listen(listener_sock, 10) < 0){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("no tcp socket listen");
#endif
		close(listener_sock);
		return 1;
	}
#ifdef _DEBUGGERY
	if(DebugLevel>2) printf("TCP Listening...\n");
#endif

	return 0;
}

int PLUServer::GetListener()
{
	return listener_sock;
}

bool PLUServer::CanShutdown()
{
	return bCanShutdown;
}

int PLUServer::ShutdownIn(int t)
{
	itimerval tv;
	itimerval old_tv;
	int		 ret;

	BhMcPosMsg Mcbc;
	bhpos::CClientInfo f;
	bhpos::service::CClientAck ack;
	std::list<bhpos::CClientInfo>::iterator fi;

	Lock();
	
	if(clients.empty()){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("server shutdown: no clients...");
#endif		
		bgoingtodie=true;
		shutdownt=t;
		tv.it_interval.tv_sec=0;
		tv.it_interval.tv_usec=0;
		tv.it_value.tv_sec=t;
		tv.it_value.tv_usec=0;
	
		Release();
	   
		ret=setitimer(ITIMER_REAL, &tv, NULL);
		if(ret!=0){
#ifdef _DEBUGGERY
			if(DebugLevel)
				printf("server shutdown: set timer failed: %s", strerror(errno));
#endif			
		}
		return 0;
	}
	
	memset(&Mcbc, 0, sizeof(BhMcPosMsg));

	Mcbc.hdr.type=BHMSG_MC_SYSTEM;
	Mcbc.hdr.cmd=BHMSG_MC_SYS_SHUTDOWN;
	Mcbc.hdr.subcmd=0;
	Mcbc.hdr.Ex=0;
	Mcbc.hdr.size=0;
	Mcbc.hdr.flags=0;
	Mcbc.site=MSG_ALL_SITES;
	Mcbc.dept=MSG_ALL_DEPTS;
	Mcbc.group=MSG_ALL_GROUPS;
	Mcbc.post=MSG_ALL_POSTS;

	unsigned int v=POST_STATUS_SHUTDOWNINPROG;

	for(fi=clients.begin(); fi!=clients.end(); fi++){
		fi->SetParameter(BHCI_PARAM_STATUS, (void *) &v);
      ack.Clear();
      client_collector.Add(ack);		
	}
	if(CtrlQ)
		CtrlQ->Add((BhPosMsg *) &Mcbc);
	
	bgoingtodie=true;
	//alarm(0);

	shutdownt=t;
	tv.it_interval.tv_sec=0;
	tv.it_interval.tv_usec=0;
	tv.it_value.tv_sec=t;
	tv.it_value.tv_usec=0;
	
	Release();
	
	return setitimer(ITIMER_REAL, &tv, NULL);
}

void PLUServer::BigCrunch(bool bForce)
{
   CancelThreads(bForce);
   if(modules) modules->Close();
   if(pidfile){
      if(open(pidfile, O_CREAT | O_EXCL) == -1){
	   	if(errno==EEXIST){
		   	unlink(pidfile);
		   }
	   }
   }
	for(unsigned int i=0; i!=clients.size(); i++){
		clients.pop_back();
	}
	
	pthread_mutex_destroy(&mutex);
}

void PLUServer::ClientAck(int cid)
{
	client_collector.Ack(cid);
}

void PLUServer::ResetClients()
{
   std::list<bhpos::CClientInfo>::iterator fi;
   for(fi=clients.begin(); fi!=clients.end(); fi++){
      fi->Reset();
   }
}

void PLUServer::KeepAlive()
{
	BhMcPosMsg mc;
	unsigned int status;
	std::list<bhpos::CClientInfo>::iterator fi;
	
	if(clients.empty()) return;

	Lock();
	memset(&mc, 0, sizeof(mc));

	mc.hdr.type=BHMSG_MC_SYSTEM;
	mc.hdr.cmd=BHMSG_MC_SYS_KEEPALIVE;
	mc.hdr.subcmd=0;
	mc.hdr.Ex=0;
	mc.hdr.size=0;
	mc.hdr.flags=0;
	mc.site=MSG_ALL_SITES;
	mc.dept=MSG_ALL_DEPTS;
	mc.group=MSG_ALL_GROUPS;
	mc.post=MSG_ALL_POSTS;		

	for(fi=clients.begin(); fi!=clients.end(); fi++){
		fi->GetParameter(BHCI_PARAM_STATUS, (void *) &status);
		if(status==POST_STATUS_DISCONNECTED){
#ifdef _DEBUGGERY
			if(DebugLevel)
				printf("keepalive: dead client - removing\n");
#endif			
			//remove client
			clients.erase(fi);
		}
		else{
			fi->FlagKeepalive(true);
			UpdateClient(*fi);
		}
	}
	if(CtrlQ)
		CtrlQ->Add((BhPosMsg *) &mc);
	
	Release();
	
#ifndef _DEBUGGERY	
/*
	int azret[2];
	if(!DebugLevel){
		if(azfd){
			azblock++;
			fwrite(&azblock, sizeof(azblock), 1, azfd);
			fread(&azret, sizeof(azret), 1, azfd);
			if(azret[1]==~azblock[1]){
			}
		}					
	}
*/
#endif
}

int PLUServer::ParseString(char *s, char *r)
{
	char *c=s;
	bool tok1, tok2, tok3;
	char bob[4];
	int  x;
	int  buffpos=0;
	char buff[257];
	
	tok1=tok2=tok3=false;

	while(c){
	   if(*c=='~') break;
	   if(*c=='"' && !tok3){		   
			tok3=true;
			c++;
			continue;
		}
		if(tok3){
		   if(*c=='"'){		   
   			tok3=false;
			   c++;
   			continue;
		   }
		   else{
		      buff[buffpos]=*c;
      		buffpos++;
      		c++;
   			continue;
		   }
		}
		if(*c=='0' && !tok1 && !tok2){	
			tok1=true;
			c++;
			continue;
		}
		if(*c=='x' && tok1){
			tok2=true;
			c++;
			continue;
		}
		if(tok1 && tok2){
			bob[0]=*c;
			c++;
			bob[1]=*c;
			if(*(c+1)!=' ' && *(c+1)!='~'){
			   c++;
			   bob[2]=*c;
			   bob[3]=0;
			}
			else{
   			bob[2]=0;
   	   }
			x=atoi(bob);
			tok1=tok2=false;
			buff[buffpos]=(char) x;
			//printf("strctrl: %s - 0x%03d\n", bob, x);
			buffpos++;
			c++;
			continue;
		}
		c++;
	}
	memcpy(r, buff, buffpos);
	
	return buffpos;
}

///////////////////////////////////////////////////////////////////////////
bool PLUServer::HasModules()
{
	return bHasModules;
}

int PLUServer::QueueExists(bhpos::BhModQ q)
{
   switch(q){
   case bhpos::BHMOD_QUEUE_IN: if(InQ) return true;
   case bhpos::BHMOD_QUEUE_OUT: if(OutQ) return true;
   case bhpos::BHMOD_QUEUE_SQL: if(SqlQ) return true;
   case bhpos::BHMOD_QUEUE_ERROR: if(ErrorQ) return true;
   case bhpos::BHMOD_QUEUE_CTRL: if(CtrlQ) return true;
   case bhpos::BHMOD_QUEUE_TRANS: if(TransQ) return true;
   case bhpos::BHMOD_QUEUE_RPRINT: if(RPrintQ) return true;
#ifdef _CCVENGINEMCVE   
   case bhpos::BHMOD_QUEUE_CCV: if(CcvQ) return true;
#endif   
   }
	return 0;
}

int PLUServer::QueueAdd(bhpos::BhModQ q, BhPosMsg *m)
{
   if(!m) return true;
   
   if(m->hdr.type <= BHPOSMSG_RESERVED) return true;
   
   switch(q){
   case bhpos::BHMOD_QUEUE_IN: if(InQ) InQ->Add(m); break;
   case bhpos::BHMOD_QUEUE_OUT: if(OutQ) OutQ->Add(m); break;
   case bhpos::BHMOD_QUEUE_SQL: if(SqlQ) SqlQ->Add(m); break;
   case bhpos::BHMOD_QUEUE_ERROR: if(ErrorQ) ErrorQ->Add(m); break;
   case bhpos::BHMOD_QUEUE_CTRL: if(CtrlQ) CtrlQ->Add(m); break;
   case bhpos::BHMOD_QUEUE_TRANS: if(TransQ) TransQ->Add(m); break;
   case bhpos::BHMOD_QUEUE_RPRINT: if(RPrintQ) RPrintQ->Add(m); break;
#ifdef _CCVENGINEMCVE   
   case bhpos::BHMOD_QUEUE_CCV: if(CcvQ) CcvQ->Add(m); break;
#endif   
   }
	return 0;
}

bool PLUServer::ModuleHandlesMessages(int m)
{
   //if(m->hdr.type <= BHPOSMSG_RESERVED) return true;
     
	return false;
}
int PLUServer::HandleMsg(BhPosMsg *m)
{
   bhpos::bhm mt;
   
   mt.bmsg[0]=m->hdr.type;
   mt.bmsg[1]=m->hdr.cmd;
   mt.bmsg[2]=m->hdr.subcmd;
   mt.bmsg[3]=m->hdr.Ex=0;
   
   modules->HandleMsg(mt.imsg, m->hdr.size + sizeof(BhHdr), m, 0);
   
   if(m->hdr.type <= BHPOSMSG_RESERVED) return 1;
	return 0;
}

int PLUServer::Lock()
{
	return pthread_mutex_lock(&mutex);
}
int PLUServer::Release()
{
	return pthread_mutex_unlock(&mutex);
}
