/*
rprintq.cpp - an remote printer queue
Copyright (C) 2002  Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos, Copyright (C) 2001-2004 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY.
*/
#include <pthread.h>
#include <cstdio>
#include <cstring>
#include <cerrno>
#include <fcntl.h>
#include <unistd.h>

#include <pos.h>
#include <cprnlist.h>

#include "msgq.h"
#include "plusd.h"
#include "sql.h"

#include "bhmsgq.cpp"

extern CMessageQueue<BhPosMsg>   *SqlQ;
extern CMessageQueue<BhPosMsg>   *InQ;
extern CMessageQueue<BhPosMsg>   *OutQ;
extern CMessageQueue<BhPosMsg>   *CtrlQ;

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

CMessageQueue<BhPosMsg>   *RPrintQ;

static int RpqGetPrinterList(PLUServer *p, bhpos::CPrinterList *pl);

void *RPrintQMonitor(void *arg)
{
	PLUServer	*plu=(PLUServer *) arg;
	BhPosMsg	    msg;
	BhPosMsg	    rmsg;
	int			 status, ret;
	int			 sock;
   int          num_toscan=0;
	sockaddr_in  esin;
	bool		    bRpServ=false;
	bool 		    bConnected=false;
	int			 prnserv_retries=0;
	int			 prnserv_timeout=0;	
	int			 prnserv_retrtcnt=0;
	timespec	    to;
	socklen_t    opt=1;
	unsigned short port;
	char		    host[33];
		
	RPrintQ=new CMessageQueue<BhPosMsg>();
	if(!RPrintQ){
#ifdef _DEBUGGERY
		if(DebugLevel)
			printf("No RPrintQ Allocation\n");
#endif			
		return NULL;
	}
	
	plu->GetOption(BH_PLUS_OPTION_RPHOST, host);
	plu->GetOption(BH_PLUS_OPTION_RPPORT, &port);
	plu->GetOption(BH_PLUS_OPTION_RPRETRIES, &prnserv_retries);
	plu->GetOption(BH_PLUS_OPTION_RPTIMEOUT, &prnserv_timeout);

#ifdef _DEBUGGERY
	if(DebugLevel)
		printf("RPrintQ server - %s:%d\n", host, port);
#endif		
	
	if(bRpServ && prnserv_timeout!=0){
		to.tv_sec=(int) (prnserv_timeout/1000);
		to.tv_nsec=(prnserv_timeout-(to.tv_sec * 10000)) * 1000000;
	}
	
	memset(&esin, 0, sizeof(esin));
	esin.sin_family=AF_INET;
	esin.sin_addr.s_addr=inet_addr(host);
	esin.sin_port=htons(port);
	
	sock=socket(PF_INET, SOCK_STREAM, 0);
	if(sock < 0){
#ifdef _DEBUGGERY
		printf("remote print thread: no socket creation");
#endif	
		ret=THREAD_DEAD;
		plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
		return NULL;
	}
	
	if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(int))<0){
#ifdef _DEBUGGERY
		if(DebugLevel>1)
			printf("remote print thread: no socket reuse");
#endif
	}

	// set non blocking
  ret=fcntl(sock, F_GETFL, 0);
  if(ret == -1){
	  // buggerfish
	  ret=THREAD_DEAD;
	  plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
	  close(sock);
#ifdef _DEBUGGERY
		if(DebugLevel>1)
			printf("remote print thread: can't set non-block - quitting");
#endif
     return NULL;
  }
  if(ret!=0) ret |= O_NONBLOCK;
  else ret &= ~O_NONBLOCK;
  fcntl(sock, F_SETFL, ret);

	
	status=::pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	if(status!=0){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("remote print thread: setcanceltype failed.\n");
#endif			
	}
		
	ret=THREAD_RUNNING;
	plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
	
	while(42){
		if(!bConnected && bRpServ){
			RPrintQ->Release();
			// attempt a connect
#ifdef _DEBUGGERY
			if(DebugLevel)	
				printf("remote print thread: connecting.\n");
#endif			
			if(connect(sock, (sockaddr *) &esin, sizeof(esin)) < 0){
				switch(errno){
				case EBADF:				// The socket socket is not a valid file descriptor.
				case ENOTSOCK:			// File descriptor socket is not a socket.
#ifdef _DEBUGGERY
					if(DebugLevel)	
						printf("remote print thread: connect ENOTSOCK.\n");
#endif				
					// kill thread for restart...
					break;

				case EADDRNOTAVAIL:	// The specified address is not available on the remote machine.
				case EAFNOSUPPORT:	// The namespace of the addr is not supported by this socket
#ifdef _DEBUGGERY
					if(DebugLevel)	
						printf("remote print thread: connect EAFNOSUPPORT.\n");
#endif				
					break;

				case EISCONN:			// The socket socket is already connected. 
#ifdef _DEBUGGERY
					if(DebugLevel)	
						printf("remote print thread: connect EISCONN.\n");
#endif				
					break;

				case ETIMEDOUT:		// The attempt to establish the connection timed out. 
#ifdef _DEBUGGERY
					if(DebugLevel)	
						printf("remote print thread: connect ETIMEDOUT.\n");
#endif				
					break;

				case ECONNREFUSED:	// The server has actively refused to establish the connection. 
				case ENETUNREACH:		// The network of the given addr isn't reachable from this host.
#ifdef _DEBUGGERY
					if(DebugLevel)	
						printf("remote print thread: connect ENETUNREACH.\n");
#endif			
					break;

				case EADDRINUSE:		// The socket address of the given addr is already in use. 
#ifdef _DEBUGGERY
					if(DebugLevel)	
						printf("remote print thread: connect EADDRINUSE.\n");
#endif				
					break;

				case EINPROGRESS:		// The socket socket is non-blocking and the connection could not be established immediately. You can determine when the connection is completely established with select; see section Waiting for Input or Output. Another connect call on the same socket, before the connection is completely established, will fail with EALREADY. 
				case EALREADY:			// The socket socket is non-blocking and already has a pending connection in progress
#ifdef _DEBUGGERY
					if(DebugLevel)	
						printf("remote print thread: connect EINPROGRESS.\n");
#endif				
					break;
				}
				if(prnserv_retries){	// 0 for forever
					prnserv_retrtcnt--;
					if(prnserv_retrtcnt <=0){
						ret=THREAD_DEAD;
						plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
						return NULL;
					}
				}
				else{	// this is await forever, so discard messages to stop queue filling up
					if(RPrintQ->HasMsg()){
						RPrintQ->DeleteLast();
					}
				}
				sleep(5);
				continue;
			}
			else{			
				// select test it
				bConnected=true;
			}
		}
		
		plu->GetThread(ERRQ_THREADID, THREAD_PARAM_DEATH, &status);
		if(status && !RPrintQ->HasMsg()) break;
		
		status=RPrintQ->Grab();
		status=RPrintQ->WaitOnMsg();
		
		if(status != 0){
#ifdef _DEBUGGERY
		if(DebugLevel)		
			printf("RPrintQ thread wait failed, %d (%s)\n", status, strerror(status));
#endif			
			RPrintQ->Release();
			if(bConnected && bRpServ) close(sock);
			ret=THREAD_DEAD;
			plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
			return NULL;
		}
		if(!RPrintQ->HasMsg()){
			continue;
		}
#ifdef _DEBUGGERY
		if(DebugLevel>1)
			printf("remote print thread: popped error msg\n");
#endif		
		RPrintQ->GetNext(&msg);
		
		// redirect some crap
		if(bConnected && bRpServ){
			if((ret=send(sock, &msg, sizeof(msg.hdr)+msg.hdr.size, 0)) < 0){
#ifdef _DEBUGGERY
					if(DebugLevel>1)
						printf("remote print thread: failed to send to error server\n");
#endif				
				if(!prnserv_retries){	// 0 for forever
					prnserv_retrtcnt=prnserv_retries;
				}
				RPrintQ->Release();
				bConnected=false;
			}
			else{
				RPrintQ->DeleteLast();
				RPrintQ->Release();
			}
#ifdef _DEBUGGERY
			if(DebugLevel>1)
				printf("remote print thread: successfully send to error server\n");
#endif
		}

      // function area
      
		//printf("RPrintQ System Message\ntype:%02d\ncmd:%02d", msg.hdr.type, msg.hdr.cmd);
		
		status=RPrintQ->Grab();
		if(status != 0) {
#ifdef _DEBUGGERY
			if(DebugLevel)		
				printf("RPrintQ failed to Grab Mutex, %d (%s)\n", status, strerror(status));
#endif
			ret=THREAD_DEAD;
			plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
			if(!bConnected && bRpServ) close(sock);
			return NULL;
		}
	}
	RPrintQ->Release();
	if(bConnected) close(sock);
	ret=THREAD_DEAD;
	plu->SetThread(ERRQ_THREADID, THREAD_PARAM_STATE, &ret);
	return NULL;
}

int RpqGetPrinterList(PLUServer *plu, bhpos::CPrinterList *pl)
{
   char dbhost[32];
	char dbname[32];
	char dbuser[32];
	char dbpass[32];

	bhpos::BhSQL		 *tempdb;	
	bhpos::BhSQLResult  sql_result;
   bhpos::BhSQLRow    *sql_row;
	
	plu->GetOption(BH_PLUS_OPTION_DBHOSTPRIME, dbhost);
	plu->GetOption(BH_PLUS_OPTION_DBDATABASE, dbname);
	plu->GetOption(BH_PLUS_OPTION_DBUSER, dbuser);
	plu->GetOption(BH_PLUS_OPTION_DBPASSWORD, dbpass);

	tempdb=new bhpos::BhSQL(dbhost, dbname, 0, dbuser, dbpass);
	
	if(!tempdb){
#ifdef _DEBUGGERY
		if(DebugLevel)		
			printf("rpq : failed to create SQL object\n");
#endif
		return 1;
	}

	if(tempdb->Connect()){
#ifdef _DEBUGGERY
		if(DebugLevel)		
			printf("rpq : failed to connect to SQL database - \n%s\n", tempdb->ErrorMsg());
#endif			
		return 2;
	}
   
	sql_result.init(tempdb, "select id, address, port from printer where active='Y'", true);
	if(sql_result.NumRows() > 0){
#ifdef _DEBUGGERY
   	if(DebugLevel > 2){
   		printf("RpQ: printers found row...\n");
	   }
#endif
		while(sql_row=sql_result.GetRow()){
         bhpos::CRPrinter prn(atoi(sql_row->GetField(0)), sql_row->GetField(1), atoi(sql_row->GetField(2)));
         pl->Add(prn);
      }
	}
   return 0;
}
