/*
servmsg.cpp
Copyright (C) 2001  Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Don't write to me at: kingsley@bananahead.com 

   Part of:
   BhPos, Copyright (C) 2001-2004 Al Kingsley
   BhPos comes with ABSOLUTELY NO WARRANTY.
*/
#include <cstdio>
#include <cstring>

#include "pos.h"
#include "msgq.h"
#include "plusd.h"

#include "bhmsgq.cpp"

extern CMessageQueue<BhPosMsg>   *InQ;
extern CMessageQueue<BhPosMsg>   *OutQ;
extern CMessageQueue<BhPosMsg>   *SqlQ;
extern CMessageQueue<BhPosMsg>   *ErrorQ;
extern CMessageQueue<BhPosMsg>   *CtrlQ;

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

static int serviceSystem(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int serviceClient(PLUServer *p, BhPosMsg *msg);
static int serviceSite(PLUServer *p, BhPosMsg *msg);
static int serviceMsg(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int serviceDb(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int serviceCustomer(PLUServer *p, BhPosMsg *msg);
static int serviceFinancial(PLUServer *p, BhPosMsg *msg);
static int servicePnP(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int serviceCcv(PLUServer *p, BhPosMsg *msg);
static int serviceKeymap(PLUServer *p, BhPosMsg *msg);

static int serviceClientGet(PLUServer *p, BhPosMsg *msg);
static int serviceClientSet(PLUServer *p, BhPosMsg *msg);
static int serviceClientReqGet(PLUServer *p, BhPosMsg *msg);
static int serviceClientReqAck(PLUServer *p, BhPosMsg *msg);
static int serviceClientGetCash(PLUServer *p, BhPosMsg *msg);
static int serviceClientSetCash(PLUServer *p, BhPosMsg *msg);

static int serviceSystemNum(PLUServer *p, BhPosMsg *msg);
static int serviceSystemWake(PLUServer *p, BhPosMsg *msg);

int ServiceMessage(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   switch(msg->hdr.cmd){
   case BHMSG_SERVICE_SYSTEM:    return serviceSystem(p, c, msg);
   case BHMSG_SERVICE_CLIENT:    return serviceClient(p, msg); 
   case BHMSG_SERVICE_SITE:      return serviceSite(p, msg); 
   case BHMSG_SERVICE_MESSAGE:   return serviceMsg(p, c, msg);
   case BHMSG_SERVICE_DATABASE:  return serviceDb(p, c, msg);
   case BHMSG_SERVICE_CUSTOMER:  return serviceCustomer(p, msg);
   case BHMSG_SERVICE_FINANCIAL: return serviceFinancial(p, msg);
   case BHMSG_SERVICE_PNP:       return servicePnP(p, c, msg);
   case BHMSG_SERVICE_CCV:       return serviceCcv(p, msg);
   case BHMSG_SERVICE_KEYMAP:    return serviceKeymap(p, msg);
   
   default: break;
   }
   return 0;
}
////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////

int serviceSystem(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   int      t;
   double   d;
   BhPosMsg rmsg;
//   BhMcPosMsg Mcrmsg;
   BhVersion  ver;
   
   memcpy(&rmsg, msg, sizeof(rmsg));
   
   switch(msg->hdr.subcmd){
   case BHMSG_SERVICE_SYSTEM_STATUS:
#ifdef _DEBUGGERY
      if(DebugLevel > 2)
         printf("servmsg: server status\n");
#endif   
   	if(OutQ) OutQ->Add(&rmsg);
      break;
   
   case BHMSG_SERVICE_SYSTEM_VERSION_SERVER:
#ifdef _DEBUGGERY
      if(DebugLevel > 2)
         printf("servmsg: server version\n");
#endif         
      rmsg.hdr.Ex=BHPOS_GRANT;
      ver.MajorVersion=BH_PLUS_SERVER_MAJVERSION;
      ver.MinorVersion=BH_PLUS_SERVER_MINVERSION;
      ver.BuildVersion=BH_PLUS_SERVER_BUILDVERSION;
      ver.packermack=0;
      memcpy(rmsg.data, &ver, sizeof(BhVersion));
		rmsg.hdr.size=sizeof(BhVersion);
      if(OutQ) OutQ->Add(&rmsg);
      break;
   
   case BHMSG_SERVICE_SYSTEM_VERSION_CLIENT:
      //strcpy(rmsg.data, CLIENT_VERSION);
      break;

   case BHMSG_SERVICE_SYSTEM_VERSION_PROTOCOL:
      rmsg.hdr.Ex=BHPOS_GRANT;
      ver.MajorVersion=BH_PLUS_PROTOCOL_MAJVERSION;
      ver.MinorVersion=BH_PLUS_PROTOCOL_MINVERSION;
      ver.BuildVersion=0;
      ver.packermack=0;
      memcpy(rmsg.data, &ver, sizeof(BhVersion));
		rmsg.hdr.size=sizeof(BhVersion);
      if(OutQ) OutQ->Add(&rmsg);
      break;

	case BHMSG_SERVICE_SYSTEM_PING:
		rmsg.hdr.Ex=BHPOS_GRANT;
	rmsg.hdr.size=0;
	if(OutQ) OutQ->Add(&rmsg);
		break;

   case BHMSG_SERVICE_SYSTEM_SHUTDOWN1:   // are you sure?
#ifdef _DEBUGGERY
      if(DebugLevel > 2)
         printf("servmsg: shutdown 1\n");
#endif   
      rmsg.hdr.Ex=BHPOS_GRANT;
		rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      break;
      
   case BHMSG_SERVICE_SYSTEM_SHUTDOWN2:   // yes? maybe not.
      if(p->CanShutdown()){
         rmsg.hdr.Ex=BHPOS_GRANT;
      }
      else{
         rmsg.hdr.Ex=BHPOS_REFUSE;
      }
		rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      break;
      
   case BHMSG_SERVICE_SYSTEM_SHUTDOWN3:   // ok, when?
      memcpy(&t, msg->data, sizeof(t));
#ifdef _DEBUGGERY
      if(DebugLevel > 2)
         printf("servmsg: shutdown 3 - down in %d seconds\n", t);
#endif       
      p->ShutdownIn(t);
      rmsg.hdr.Ex=BHPOS_GRANT;
		rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      break;

   case BHMSG_SERVICE_SYSTEM_NUMCLIENTS:
      p->GetParameter(BH_PLUS_PARM_NUMCLIENTS, &t);
      memcpy(rmsg.data, &t, sizeof(t));
		rmsg.hdr.size=sizeof(int);
#ifdef _DEBUGGERY
      if(DebugLevel > 2)
         printf("servmsg: num clients - %d\n", t);
#endif 		
      if(OutQ) OutQ->Add(&rmsg);
      break;
      
   case BHMSG_SERVICE_SYSTEM_CLIENTACK:
   	p->GetParameter(BH_PLUS_PARM_NUMCLIENTACK, &t);
   	memcpy(rmsg.data, &t, sizeof(t));
		rmsg.hdr.size=sizeof(int);
		if(OutQ) OutQ->Add(&rmsg);
   	break;

   case BHMSG_SERVICE_SYSTEM_NUMTRANSOPEN:
      p->GetParameter(BH_PLUS_PARM_NUMOPENTRANS, &t);
      memcpy(rmsg.data, &t, sizeof(t));
		rmsg.hdr.size=sizeof(int);
      if(OutQ) OutQ->Add(&rmsg);
      break;

   case BHMSG_SERVICE_SYSTEM_LOAD:
      p->GetParameter(BH_PLUS_PARM_CURRLOAD, &d);
      memcpy(rmsg.data, &d, sizeof(d));
		rmsg.hdr.size=sizeof(double);
		if(OutQ) OutQ->Add(&rmsg);
      break;

   case BHMSG_SERVICE_SYSTEM_MAXLOAD:
      p->GetParameter(BH_PLUS_PARM_MAXLOAD, &d);
      memcpy(rmsg.data, &d, sizeof(d));
		rmsg.hdr.size=sizeof(double);
      if(OutQ) OutQ->Add(&rmsg);
      break;
      
   case BHMSG_SERVICE_SYSTEM_WAKE_POST:
   case BHMSG_SERVICE_SYSTEM_WAKE_ALL:
      serviceSystemWake(p, msg);
      break;
   }
   return 0;
}

////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////

int serviceSite(PLUServer *p, BhPosMsg *msg)
{
   if(SqlQ) SqlQ->Add(msg);
   return 0;
}

////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////

int serviceClient(PLUServer *p, BhPosMsg *msg)
{
   switch(msg->hdr.subcmd){
   case BHMSG_SERVICE_CLIENT_GET: return serviceClientGet(p, msg); break;
   case BHMSG_SERVICE_CLIENT_SET: return serviceClientSet(p, msg); break;
   case BHMSG_SERVICE_CLIENT_GETREQ: return serviceClientReqGet(p, msg); break;
   case BHMSG_SERVICE_CLIENT_ACKREQ: return serviceClientReqAck(p, msg); break;
   case BHMSG_SERVICE_CLIENT_GETCASH: return serviceClientGetCash(p, msg); break;
   case BHMSG_SERVICE_CLIENT_SETCASH: return serviceClientSetCash(p, msg); break;
   }
   return 0;
}

int serviceClientGet(PLUServer *p, BhPosMsg *msg)
{
   /*
   BhPosMsg       rmsg;
   bhpos::CClientInfo      cl;
   bhClientInfo   bhcl;
   int who=(int) *((int *) msg->data);

   memcpy(&rmsg, msg, sizeof(rmsg));
   p->GetClient(who, cl);
   memcpy(&bhcl, &cl, sizeof(bhcl));
   memcpy(rmsg.data, &bhcl, sizeof(bhcl));
   
   if(OutQ) OutQ->Add(&rmsg);*/

   return 0;
}
int serviceClientSet(PLUServer *p, BhPosMsg *msg)
{
   if(CtrlQ) CtrlQ->Add(msg);
   return 0;
}

int serviceClientReqGet(PLUServer *p, BhPosMsg *msg)
{
   int          req;
   time_t        req_when;
   BhPosMsg      rmsg;
   bhpos::CClientInfo   cl;
   int who=(int) *((int *) msg->data);

   memcpy(&rmsg, msg, sizeof(rmsg));
   p->GetClient(who, cl);
   
   req=cl.GetRequest(&req_when);
   
   if(req!=0){
   }
   
   return 0;   
}

int serviceClientReqAck(PLUServer *p, BhPosMsg *msg)
{
   /*
   BhMcPosMsg     McMsg;
   BhMcCash       cash;
   BhServiceClient  *serc = (BhServiceClient  *) msg->data;
   BhServiceCash   *sc=(BhServiceCash *) (msg->data + sizeof(SericeClientMsg);
   bhpos::CClientInfo      cl;
   
   memset(McMsg, 0, sizeof(McMsg));
   memcy(&McMsg.hdr, msg->hdr, sizeof(McMsg.hdr));
   
   p->GetClient(serc->clientID, cl);

   McMsg.site=serc->site;
   McMsg.department=serc->department;
   McMsg.group=serc->group;
   McMsg.post=serc->station;

   cash.targetUnit=targetID;
   cash.Amount=sc.amount;
   McMsg.hdr.size=sizeof(BhMcCash);
   memcpy(McMsg.data, &cash, sizeof(cash));
   
   if(CtrlQ) CtrlQ->Add((BhPosMsg *) &McMsg);
   */
   return 0;
}

int serviceClientGetCash(PLUServer *p, BhPosMsg *msg)
{
   return 0;
}

int serviceClientSetCash(PLUServer *p, BhPosMsg *msg)
{
   /*
   BhServiceClient   *serc = (BhServiceClient  *) msg->data;
   BhServiceCash     *sc=(BhServiceCash *) (msg->data + sizeof(BhServiceClient));
   BhPosMonetaryUnit *mu=sc->unit;
   bhpos::CMoney      m;
   //bhpos::CClientInfo cl;
   
   //p->GetClient(serc->clientID, cl);
   
   //cl.SetCash(sc->targetUnit, sc->targetValue, sc->amount);

   //rmsg.hdr.size=sizeof(BhPosMonetaryUnit));		
			
	int num_units=(msg->hdr.size-sizeof(BhServiceClient)-sizeof(BhServiceCash)+sizeof(BhPosMonetaryUnit))/sizeof(BhPosMonetaryUnit);
	
#ifdef _DEBUGGERY				
   if(DebugLevel > 2){
		printf("SqlQ: Connect: Qry:MonetaryUnits returned %d rows\n",sql_result.NumRows());
   }
#endif				
		
			
	BhPosMonetaryUnit *mu=rconn.unit;
	while(num_units){
		m.Set(mu->unit, mu->value, 10, mu->amount);
		c.AddMonetaryUnit(m);				
		mu++;
		num_units--;
		rmsg.hdr.size+=sizeof(BhPosMonetaryUnit);
	}
	
	// set the client
	
	// multicast it out...
*/
   return 0;
}

////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////

int serviceCustomer(PLUServer *p, BhPosMsg *msg)
{
   return 0;
}

////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////

int serviceFinancial(PLUServer *p, BhPosMsg *msg)
{
   return 0;
}

////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////

int serviceCcv(PLUServer *p, BhPosMsg *msg)
{
   return 0;
}

////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////
int serviceDb(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   switch(msg->hdr.subcmd){
   case BHMSG_SERVICE_DATABASE_PLUGET:
   case BHMSG_SERVICE_DATABASE_PLUSET:
      break;
   case BHMSG_SERVICE_DATABASE_PLUEADD:
   case BHMSG_SERVICE_DATABASE_PLUEDELETE:
      if(SqlQ) SqlQ->Add(msg);
      break;
   }
   return 0;
}

////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////
int serviceKeymap(PLUServer *p, BhPosMsg *msg)
{
   // moduled up - in theory
   // let's see if the mod is loaded...
   return 0;
}

////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////

int serviceSystemWake(PLUServer *p, BhPosMsg *msg)
{
   int      s, i;
   int      clientid;
   int      one = 1;
   u_char   pkt[1024];
   u_char   mac_addr[6];
   struct sockaddr saddr;
   char *ifname = "eth0";

   clientid=(long) msg->data;

   // hdr info
   // ether   IP   TCP/UCP
   // 14      20   20

   // ethernet  hdr
   //

   // IP hdr
   //

   // TCP hdr
   //

   
   if((s=socket(AF_INET, SOCK_PACKET, SOCK_PACKET)) < 0){
      return -1;
   }

   switch(msg->hdr.subcmd){
   case BHMSG_SERVICE_SYSTEM_WAKE_POST:
      // get mac from db...
      p->GetClientInfo(clientid, CLIENTINFO_MACADDRESS, &mac_addr);
      memcpy(pkt, mac_addr, 6);
      break;
      
   case BHMSG_SERVICE_SYSTEM_WAKE_ALL:
      memset(pkt, 0xff, 6);
      if(setsockopt(s, SOL_SOCKET, SO_BROADCAST, (char *)&one, sizeof(one)) < 0)
      return -2;
      break;
   }


   memcpy(pkt+6, mac_addr, 6);
   pkt[12] = 0x08;      /* Or 0x0806 for ARP, 0x8035 for RARP */
   pkt[13] = 0x42;

   // ip hdr
   int offset = 14;

   memset(pkt+offset, 0xff, 6);
   offset+=6;

   for(i=0; i < 16; i++){
      memcpy(pkt+offset, mac_addr, 6);
      offset += 6;
   }
   offset+=6;
   
   saddr.sa_family = 0;
   strcpy(saddr.sa_data, ifname);

   if((i=sendto(s, pkt, offset, 0, &saddr, sizeof(saddr))) < 0)
      return -3;

   if(bind(s, &saddr, sizeof(saddr)) < 0)
      return -4;
   
   if(send(s, pkt, 100, 0) < 0)
      return -5;

   close(s);

   return 0;
}

////////////////////////////////////////////////////////////////////////////////////

int serviceMsg(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
//   BhPosMsg rmsg;
   
   if(CtrlQ) CtrlQ->Add(msg);
   
   return 0;
}

////////////////////////////////////////////////////////////////////////////////////

int servicePnP(PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   switch(msg->hdr.subcmd){
   case BHMSG_SERVICE_PNP_START:
      p->SetPlugAndPlay(true);
      break;
      
   case BHMSG_SERVICE_PNP_END:
      p->SetPlugAndPlay(false);
      break;
   }

   return 0;
}
