/*
servside.cpp - serverside transactions
Copyright (C) 2003  Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos, Copyright (C) 2001-2005 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY.
*/
#include <pthread.h>
#include <cstdio>

#include "pos.h"
#include "sql.h"
#include "msgq.h"
#include "plusd.h"
#include "sql.h"
#include "plucache.h"

#include "bhmsgq.cpp"

extern CMessageQueue<BhPosMsg> *TransQ;
extern CMessageQueue<BhPosMsg> *InQ;
extern CMessageQueue<BhPosMsg> *OutQ;
extern CMessageQueue<BhPosMsg> *CtrlQ;
extern CMessageQueue<BhPosMsg> *ErrorQ;

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

static int TransServerSideTrans(PLUServer *p, bhpos::BhSQL *db, BhPosMsg *msg);

int TransServerSide(PLUServer *p, bhpos::BhSQL *db, BhPosMsg *msg)
{
   switch(msg->hdr.cmd){
   case BHMSG_SERVERSIDE_TRANS: return TransServerSideTrans(p, db, msg);
   case BHMSG_SERVERSIDE_PLU:
   case BHMSG_SERVERSIDE_ADD:
   case BHMSG_SERVERSIDE_EDIT:
   case BHMSG_SERVERSIDE_DELETE:
      break;
   }
   
   return 0;
}

int TransServerSideTrans(PLUServer *p, bhpos::BhSQL *db, BhPosMsg *msg)
{
   bhpos::BhSQLResult sql_result;
   bhpos::BhSQLRow	*sql_row;
	BhPosMsg			    rmsg;
   BhMcPosMsg         mcmsg;
	int					 status, transid; //, mode;
   bool               bServSide;
	char				    buff[256];
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
   memset(&mcmsg, 0, sizeof(mcmsg));
   memcpy(&mcmsg.hdr, &rmsg.hdr, sizeof(rmsg.hdr));
	
	rmsg.hdr.Ex=BHPOS_GRANT;
   rmsg.hdr.size=0;
   
   p->GetOption(BH_PLUS_OPTION_SERVERSIDE, &bServSide);
   
   switch(msg->hdr.subcmd){
   case BHMSG_SERVERSIDE_TRANS_NEW:
   {
      int  id;
      int  trans_prefix=0;
      char tp[8], *ch;
      
      BhTransRovernew *nt=(BhTransRovernew *) msg->data;
      BhMcTransNew    *mt=(BhMcTransNew *) mcmsg.data;
      
      p->GetOption(BH_PLUS_OPTION_TRANSPREFIX, &tp);
      
      status=db->SimpleQuery(sqlq_qry70);
	   if(status!=0){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to lock trans table for new insert: %s - %d\n", db->ErrorMsg(), status);
		   }
#endif
		   rmsg.hdr.Ex=BHPOS_REFUSE;
	   	rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
		   return 1;
	   }
      
      status=db->SimpleQuery(sqlq_qry73);
	   if(status!=0){
#ifdef _DEBUGGERY			 
		   if(DebugLevel > 2){
			   printf("Failed to insert new transaction: %s\n", db->ErrorMsg());
		   }
#endif
         db->SimpleQuery(sqlq_qry146);
		   rmsg.hdr.Ex=BHPOS_REFUSE;
		   rmsg.hdr.size=0;
		   if(OutQ) OutQ->Add(&rmsg);
		   return 1;
	   }
      
      // last insert id
      if(sql_result.init(db, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
         if(DebugLevel > 2){
			   printf("Failed to query 145: %s\n", db->ErrorMsg());
         }
#endif
         db->SimpleQuery(sqlq_qry146);
         rmsg.hdr.Ex=BHPOS_REFUSE;
         rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
         return 2;
      }
      if(sql_result.NumRows()!=0){
         sql_row=sql_result.GetRow();
         nt->ID=atoi(sql_row->GetField(0));
      }
   
      if(trans_prefix!=0){
	   	sprintf(buff, "%d", trans_prefix);
		   transid=10-strlen(tp)-status;
   		strcpy(buff, tp);
	   	ch=&buff[transid];
		   for(unsigned int i=0; i<=transid; i++){
			   *ch='0';
			   ch++;
		   }
		   nt->ID=atoi(buff);
		   //nt.transID+=transid;
	   }
	   sql_result.free();

      sprintf(buff, sqlq_qry74, nt->ID, nt->ID);
	   status=db->SimpleQuery(buff);
	   if(status!=0){
#ifdef _DEBUGGERY			 
		   if(DebugLevel > 2){
			   printf("Failed To update new transaction: %s\n", db->ErrorMsg());
		   }
#endif
      }
      
      status=db->SimpleQuery(sqlq_qry77);
	   if(status!=0){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to unlock trans table for new insert: %s\n", db->ErrorMsg());
		   }
#endif
	   }
      
      memcpy(rmsg.data, nt, sizeof(BhTransRovernew));
      rmsg.hdr.size=sizeof(BhTransRovernew);
      
      p->ssts.Add(nt->ID, nt->postID, nt->userID);
		
		if(OutQ) OutQ->Add(&rmsg);

      mcmsg.hdr.type=BHMSG_MC_CLIENT;
      mcmsg.hdr.cmd=BHMSG_MC_CLIENT_TRANS;
      mcmsg.hdr.subcmd=BHMSG_MC_CLIENT_TRANS_NEW;
      mcmsg.site=0;
      mcmsg.dept=0;
      mcmsg.group=0;
      mcmsg.post=0;
      mt->transid=nt->ID;
      mt->stationid=nt->postID;
      mt->userid=nt->userID;
      mt->status=BHMC_CLIENT_TRANSSTATUS_NEW;
      mcmsg.hdr.size=sizeof(BhMcTransNew);
      CtrlQ->Add(&rmsg);
   }
      break;
   
   case BHMSG_SERVERSIDE_TRANS_LIST:
   {
      BhTransRoverListRequest *lreq=(BhTransRoverListRequest *) msg->data;
      BhTransRoverList        *llist=(BhTransRoverList *) rmsg.data;
      
      status=p->ssts.List((bhpos::BhSSTSNumber) lreq->flags, lreq->userID, llist);
      rmsg.hdr.size==sizeof(BhTransRoverList) + (sizeof(BhTransRoverListItem) * (status-1));
      if(OutQ) OutQ->Add(&rmsg);
   }
      break;
   
   case BHMSG_SERVERSIDE_TRANS_STATUS:
   case BHMSG_SERVERSIDE_TRANS_RETRIEVE:
   case BHMSG_SERVERSIDE_TRANS_ADD:
      break;
   
   case BHMSG_SERVERSIDE_TRANS_TARGETLIST:
   {
      BhTransRoverTargetListRequest *lreq=(BhTransRoverTargetListRequest *) msg->data;
      BhTransRoverTargetList        *llist=(BhTransRoverTargetList *) rmsg.data;
      BhTransRoverTargetListItem    *li=llist->item;
      
      if(sql_result.init(db, "select * from sstarget where available='Y'", true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
      	if(DebugLevel > 2){
		      printf("Failed To select target: %s\n", db->ErrorMsg());
   	   }
#endif
   	   return 2;
	   }

      if(sql_result.NumRows()==0){
         rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 3;
      }
      llist->num_items=sql_result.NumRows();
      while((sql_row = sql_result.GetRow())){
         li->ID=atoi(sql_row->GetField(0));
         li->status=0;
         strncpy(li->target, sql_row->GetField(1), 32);
         li++;
      }
      if(OutQ) OutQ->Add(&rmsg);
   }
      break;
   
   case BHMSG_SERVERSIDE_TRANS_PICKUP:
      // get current trans status
   
      // lock down trans
   
      // return the trans items
   
      //unlock it
   
      // send multicast trans status msg
      break;
   
   case BHMSG_SERVERSIDE_TRANS_DROP:
      // get current trans status
   
      // lock down trans
   
      // update current ss trans
   
      //unlock it
   
      // send multicast trans status msg
      break;
   
   case BHMSG_SERVERSIDE_TRANS_COMMIT:
      break;
/*
   case BHMSG_SERVERSIDE_TRANS_AUTH: 
   case BHMSG_SERVERSIDE_TRANS_SELECT:
   case BHMSG_SERVERSIDE_TRANS_COMMIT:
   case BHMSG_SERVERSIDE_TRANS_DELETE:
   case BHMSG_SERVERSIDE_TRANS_STORE:
   case BHMSG_SERVERSIDE_TRANS_RET:
   case BHMSG_SERVERSIDE_TRANS_END:
   case BHMSG_SERVERSIDE_TRANS_CANCEL:
   case BHMSG_SERVERSIDE_TRANS_PAYAUTH:
   case BHMSG_SERVERSIDE_TRANS_DEFER:
   case BHMSG_SERVERSIDE_TRANS_INFOREDIR:
   case BHMSG_SERVERSIDE_TRANS_ADDTO:
   case BHMSG_SERVERSIDE_TRANS_NEXT:
   case BHMSG_SERVERSIDE_TRANS_PREV:
      break;*/
   }
   
   return 0;
}
