/*
threads.cpp - implements threading gear 
Copyright (C) 2001  Al Kingsley


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Part of:
	BhPos, Copyright (C) 2001-2004 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY.
    
The shape of things to come:
*/
#include <pthread.h>
#include <cstdio>
#include <cerrno>
#include "pos.h"

#include "plusd.h"
#include "msgq.h"
#include "cfg.h"

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif


void *InQMonitor(void *arg);
void *CtrlQMonitor(void *arg);
void *OutQMonitor(void *arg);
void *SqlQMonitor(void *arg);
void *ErrorQMonitor(void *arg);
void *IdleThread(void *arg);
void *UDPThread(void *arg);
void *TransQMonitor(void *arg);
void *RPrintQMonitor(void *arg);

#ifdef _CCVENGINEMCVE
void *CcvQMonitor(void *arg);
#endif

extern PLUServer	*plusd; 

pthread_mutex_t  t_mutex=PTHREAD_MUTEX_INITIALIZER;

int PLUServer::CreateThread(unsigned int threadid)
{
	int				status;
	pthread_attr_t attrib;
	sched_param	   sp;

	sp.sched_priority=0;

	if(threadid & ERRQ_THREAD){
#ifdef _DEBUGGERY	
		if(DebugLevel>2) printf("plusd::CreateThread - Error Thread\n");
#endif		
		status=::pthread_attr_init(&attrib);
		if(status!=0){
#ifdef _DEBUGGERY
			if(DebugLevel>2) printf("No Thread attrib init\n");
#endif			
			return status;
		}
		pthread_attr_setschedpolicy(&attrib, SCHED_RR);
		pthread_attr_setschedparam(&attrib, &sp);
	
		status=::pthread_create(&thread[ERRQ_THREADID].thread, &attrib, ErrorQMonitor, this);
		if(status!=0){
#ifdef _DEBUGGERY
			if(DebugLevel>2) printf("ErrorQ : No Thread Create\n");
#endif			
			return status;
		}
		thread[ERRQ_THREADID].die=0;
	}

	if(threadid & INQ_THREAD){
		if(threadid==ALL_THREADS){
			while(thread[ERRQ_THREADID].status!=THREAD_RUNNING){
				usleep(10000);
			}
		}
#ifdef _DEBUGGERY	
		if(DebugLevel>2) printf("plusd::CreateThread - InQ Thread\n");
#endif		

		sp.sched_priority=-20;
		status=::pthread_attr_init(&attrib);
		if(status!=0){
#ifdef _DEBUGGERY
			if(DebugLevel>2) printf("No InQ Thread attrib init\n");
#endif			
			return status;
		}
		pthread_attr_setschedpolicy(&attrib, SCHED_RR);
		pthread_attr_setschedparam(&attrib, &sp);
		status=::pthread_create(&thread[INQ_THREADID].thread, &attrib, InQMonitor, this);
		if(status!=0){
#ifdef _DEBUGGERY		
			if(DebugLevel>2) printf("InQ : No Thread Create\n");
#endif
			return status;
		}
		thread[INQ_THREADID].die=0;
	}

	if(threadid & OUTQ_THREAD){
		if(threadid==ALL_THREADS){
			while(thread[INQ_THREADID].status!=THREAD_RUNNING){
				usleep(10000);
			}
		}
#ifdef _DEBUGGERY	
		if(DebugLevel>2) printf("plusd::CreateThread - Out Thread\n");
#endif		
		
		status=::pthread_attr_init(&attrib);
		if(status!=0){
#ifdef _DEBUGGERY	
			if(DebugLevel>2) printf("No Thread attrib init\n");
#endif		
			return status;
		}
		pthread_attr_setschedpolicy(&attrib, SCHED_RR);
		pthread_attr_setschedparam(&attrib, &sp);
		status=::pthread_create(&thread[OUTQ_THREADID].thread, &attrib, OutQMonitor, this);
		if(status!=0){
#ifdef _DEBUGGERY	
			if(DebugLevel>2) printf("OutQ : No Thread Create\n");
#endif		
			return status;
		}
		thread[OUTQ_THREADID].die=0;
	}
	
	if(threadid & SQLQ_THREAD){
		if(threadid==ALL_THREADS){
			while(thread[OUTQ_THREADID].status!=THREAD_RUNNING){
				usleep(10000);
			}
		}
#ifdef _DEBUGGERY	
		if(DebugLevel>2) printf("plusd::CreateThread - SqlQ Thread\n");
#endif		

		status=::pthread_attr_init(&attrib);
		if(status!=0){
#ifdef _DEBUGGERY	
			if(DebugLevel>2) printf("No Thread attrib init\n");
#endif		
			return status;
		}
		pthread_attr_setschedpolicy(&attrib, SCHED_RR);
		pthread_attr_setschedparam(&attrib, &sp);
		status=::pthread_create(&thread[SQLQ_THREADID].thread, &attrib, SqlQMonitor, this);
		if(status!=0){
#ifdef _DEBUGGERY	
			if(DebugLevel>2) printf("SqlQ : No Thread Create\n");
#endif		
			return status;
		}
		thread[SQLQ_THREADID].die=0;
	}

	if(threadid & CTRLQ_THREAD){
		if(threadid==ALL_THREADS){
			while(thread[SQLQ_THREADID].status!=THREAD_RUNNING){
				usleep(10000);
			}
		}
#ifdef _DEBUGGERY	
		if(DebugLevel>2) printf("plusd::CreateThread - Ctrl Thread\n");
#endif		
		
		sp.sched_priority=0;
		status=::pthread_attr_init(&attrib);
		if(status!=0){
#ifdef _DEBUGGERY	
			if(DebugLevel>2) printf("No Ctrl Thread attrib init\n");
#endif		
			return status;
		}
		pthread_attr_setschedpolicy(&attrib, SCHED_RR);
		pthread_attr_setschedparam(&attrib, &sp);
		status=::pthread_create(&thread[CTRLQ_THREADID].thread, &attrib, CtrlQMonitor, this);
		if(status!=0){
#ifdef _DEBUGGERY	
			if(DebugLevel>2) printf("CtrlQ : No Thread Create\n");
#endif
			return status;
		}
		thread[CTRLQ_THREADID].die=0;
	}

	if(threadid & UDP_THREAD){
		if(threadid==ALL_THREADS){
			while(thread[CTRLQ_THREADID].status!=THREAD_RUNNING){
				usleep(10000);
			}
		}
#ifdef _DEBUGGERY	
		if(DebugLevel>2) printf("plusd::CreateThread - UDP Thread\n");
#endif		
		
		sp.sched_priority=-20;
		status=::pthread_attr_init(&attrib);
		if(status!=0){
#ifdef _DEBUGGERY	
			if(DebugLevel>2) printf("No UDP Thread attrib init\n");
#endif		
			return status;
		}
		pthread_attr_setschedpolicy(&attrib, SCHED_RR);
		pthread_attr_setschedparam(&attrib, &sp);
		status=::pthread_create(&thread[UDP_THREADID].thread, &attrib, UDPThread, this);
		if(status!=0){
#ifdef _DEBUGGERY	
			if(DebugLevel>2) printf("UDP : No Thread Create\n");
#endif		
			return status;
		}
		thread[UDP_THREADID].die=0;
	}

	if(threadid & TRANSQ_THREAD){
		if(threadid==ALL_THREADS){
			while(thread[UDP_THREADID].status!=THREAD_RUNNING){
				usleep(10000);
			}
		}
#ifdef _DEBUGGERY	
		if(DebugLevel>2) printf("plusd::CreateThread - TransQ Thread\n");
#endif
		sp.sched_priority=-10;
		status=::pthread_attr_init(&attrib);
		if(status!=0){
#ifdef _DEBUGGERY	
			if(DebugLevel>2) printf("plusd : No TransQ attrib init\n");
#endif		
		}
		pthread_attr_setschedpolicy(&attrib, SCHED_RR);
		pthread_attr_setschedparam(&attrib, &sp);

		status=::pthread_create(&thread[TRANSQ_THREADID].thread, &attrib, TransQMonitor, this);
		if(status!=0){
#ifdef _DEBUGGERY
			if(DebugLevel>2){		
				printf("plusd : No TransQ Thread Create - error: %d - %s\n", status, strerror(status));
			}
#endif
         return status;
		}
		thread[TRANSQ_THREADID].die=0;
	}
	
	if(threadid & IDLE_THREAD){
		if(threadid==ALL_THREADS){
			while(thread[TRANSQ_THREADID].status!=THREAD_RUNNING){
				usleep(10000);
			}
		}
#ifdef _DEBUGGERY	
		if(DebugLevel>2) printf("plusd::CreateThread - Idle Thread\n");
#endif
		sp.sched_priority=0;
		status=::pthread_attr_init(&attrib);
		if(status!=0){
#ifdef _DEBUGGERY	
			if(DebugLevel>2) printf("plusd : No Idle Thread attrib init\n");
#endif		
		}
		pthread_attr_setschedpolicy(&attrib, SCHED_RR);
		pthread_attr_setschedparam(&attrib, &sp);

		status=::pthread_create(&thread[IDLE_THREADID].thread, &attrib, IdleThread, this);
		if(status!=0){
#ifdef _DEBUGGERY
			if(DebugLevel>2){		
				printf("plusd : No Idle Thread Create, who cares? - error: %d - %s\n", status, strerror(status));
			}
#endif
		}
		thread[IDLE_THREADID].die=0;
	}
	
	if(threadid & RPRINTQ_THREAD){
		/*
		if(threadid==ALL_THREADS){
			while(&thread[IDLE_THREADID].status!=THREAD_RUNNING){
				usleep(100000);
			}
		}*/
#ifdef _DEBUGGERY	
		if(DebugLevel>2) printf("plusd::CreateThread - Remote Print Thread\n");
#endif
		sp.sched_priority=0;
		status=::pthread_attr_init(&attrib);
		if(status!=0){
#ifdef _DEBUGGERY	
			if(DebugLevel>2) printf("plusd : No Remote Print Thread attrib init\n");
#endif		
		}
		pthread_attr_setschedpolicy(&attrib, SCHED_RR);
		pthread_attr_setschedparam(&attrib, &sp);

		status=::pthread_create(&thread[RPRINTQ_THREADID].thread, &attrib, RPrintQMonitor, this);
		if(status!=0){
#ifdef _DEBUGGERY
			if(DebugLevel>2){		
				printf("plusd : No Remote Print Thread Create - error: %d - %s\n", status, strerror(status));
			}
#endif
		}
		thread[RPRINTQ_THREADID].die=0;
	}

#ifdef _CCVENGINEMCVE
	if(threadid & CCVQ_THREAD){
		if(threadid==ALL_THREADS){
			while(&thread[RPRINTQ_THREADID].status!=THREAD_RUNNING){
				usleep(10000);
			}
		}
#ifdef _DEBUGGERY	
		if(DebugLevel>2) printf("plusd::CreateThread - CCV Thread\n");
#endif
		sp.sched_priority=0;
		status=::pthread_attr_init(&attrib);
		if(status!=0){
#ifdef _DEBUGGERY	
			if(DebugLevel>2) printf("plusd : No Idle CCV attrib init\n");
#endif		
		}
		pthread_attr_setschedpolicy(&attrib, SCHED_RR);
		pthread_attr_setschedparam(&attrib, &sp);

		status=::pthread_create(&thread[CCVQ_THREADID].thread, &attrib, CcvQMonitor, this);
		if(status!=0){
#ifdef _DEBUGGERY
			if(DebugLevel>2){		
				printf("plusd : No CCV Thread Create - error: %d - %s\n", status, strerror(status));
			}
#endif			
		}
		thread[CCVQ_THREADID].die=0;
	}	
#endif

	return 0;	
}

int PLUServer::JoinThreads()
{
	int ret;
	
	if((ret=pthread_join(thread[INQ_THREADID].thread, NULL))){
#ifdef _DEBUGGERY		
		if(ret==EDEADLK){
			printf("plusd: join: InQThread - deadlock\n");
		}
		if(ret==EINVAL){
			printf("plusd: join: InQThread - no such joinable thread\n");
		}
		if(ret==ESRCH){
			printf("plusd: join: InQThread - no such thread\n");
		}
		if(ret==EFAULT){
			printf("plusd: join: InQThread - invalid status pointer\n");
		}
#endif		
	}
#ifdef _DEBUGGERY	
	printf("plusd: InQThread joined\n");
#endif

	if((ret=pthread_join(thread[OUTQ_THREADID].thread, NULL))){
#ifdef _DEBUGGERY			
		if(ret==EDEADLK){
			printf("plusd: join: OutQThread - deadlock\n");
		}
		if(ret==EINVAL){
			printf("plusd: join: OutQThread - no such joinable thread\n");
		}
		if(ret==ESRCH){
			printf("plusd: join: OutQThread - no such thread\n");
		}
		if(ret==EFAULT){
			printf("plusd: join: OutQThread - invalid status pointer\n");
		}
#endif		
	}
#ifdef _DEBUGGERY		
	printf("plusd: OutQThread joined\n");
#endif
	if((ret=pthread_join(thread[SQLQ_THREADID].thread, NULL))){
#ifdef _DEBUGGERY			
		if(ret==EDEADLK){
			printf("plusd: join: SqlQThread - deadlock\n");
		}
		if(ret==EINVAL){
			printf("plusd: join: SqlQThread - no such joinable thread\n");
		}
		if(ret==ESRCH){
			printf("plusd: join: SqlQThread - no such thread\n");
		}
		if(ret==EFAULT){
			printf("plusd: join: SqlQThread - invalid status pointer\n");
		}
#endif		
	}
#ifdef _DEBUGGERY		
	printf("plusd: SqlQThread joined\n");
#endif

	if((ret=pthread_join(thread[CTRLQ_THREADID].thread, NULL))){
#ifdef _DEBUGGERY			
		if(ret==EDEADLK){
			printf("plusd: join: CtrlQThread - deadlock\n");
		}
		if(ret==EINVAL){
			printf("plusd: join: CtrlQThread - no such joinable thread\n");
		}
		if(ret==ESRCH){
			printf("plusd: join: CtrlQThread - no such thread\n");
		}
		if(ret==EFAULT){
			printf("plusd: join: CtrlQThread - invalid status pointer\n");
		}
#endif		
	}
#ifdef _DEBUGGERY	
		printf("plusd: CtrlQThread joined\n");
#endif		

	if((ret=pthread_join(thread[ERRQ_THREADID].thread, NULL))){
#ifdef _DEBUGGERY			
		if(ret==EDEADLK){
			printf("plusd: join: ErrorQThread - deadlock\n");
		}
		if(ret==EINVAL){
			printf("plusd: join: ErrorQThread - no such joinable thread\n");
		}
		if(ret==ESRCH){
			printf("plusd: join: ErrorQThread - no such thread\n");
		}
		if(ret==EFAULT){
			printf("plusd: join: ErrorQThread - invalid status pointer\n");
		}
#endif		
	}
#ifdef _DEBUGGERY		
	printf("plusd: ErrorQThread joined\n");
#endif

	if((ret=pthread_join(thread[UDP_THREADID].thread, NULL))){
#ifdef _DEBUGGERY			
		if(ret==EDEADLK){
			printf("plusd: join: UDPQThread - deadlock\n");
		}
		if(ret==EINVAL){
			printf("plusd: join: UDPQThread - no such joinable thread\n");
		}
		if(ret==ESRCH){
			printf("plusd: join: UDPQThread - no such thread\n");
		}
		if(ret==EFAULT){
			printf("plusd: join: UDPQThread - invalid status pointer\n");
		}
#endif		
	}
#ifdef _DEBUGGERY		
	printf("plusd: UDPQThread joined\n");
#endif

	if((ret=pthread_join(thread[IDLE_THREADID].thread, NULL))){
#ifdef _DEBUGGERY			
		if(ret==EDEADLK){
			printf("plusd: join: IdleThread - deadlock\n");
		}
		if(ret==EINVAL){
			printf("plusd: join: IdleThread - no such joinable thread\n");
		}
		if(ret==ESRCH){
			printf("plusd: join: IdleThread - no such thread\n");
		}
		if(ret==EFAULT){
			printf("plusd: join: IdleThread - invalid status pointer\n");
		}
#endif		
	}
#ifdef _DEBUGGERY		
	printf("plusd: Idle Thread joined\n");
#endif
	
		if((ret=pthread_join(thread[TRANSQ_THREADID].thread, NULL))){
#ifdef _DEBUGGERY			
		if(ret==EDEADLK){
			printf("plusd: join: TransThread - deadlock\n");
		}
		if(ret==EINVAL){
			printf("plusd: join: TransThread - no such joinable thread\n");
		}
		if(ret==ESRCH){
			printf("plusd: join: TransThread - no such thread\n");
		}
		if(ret==EFAULT){
			printf("plusd: join: TransThread - invalid status pointer\n");
		}
#endif		
	}
#ifdef _DEBUGGERY		
	printf("plusd: TransThread joined\n");
#endif

   if((ret=pthread_join(thread[RPRINTQ_THREADID].thread, NULL))){
#ifdef _DEBUGGERY			
		if(ret==EDEADLK){
			printf("plusd: join: RemotePrintThread - deadlock\n");
		}
		if(ret==EINVAL){
			printf("plusd: join: RemotePrintThread - no such joinable thread\n");
		}
		if(ret==ESRCH){
			printf("plusd: join: RemotePrintThread - no such thread\n");
		}
		if(ret==EFAULT){
			printf("plusd: join: RemotePrintThread - invalid status pointer\n");
		}
#endif		
	}
#ifdef _DEBUGGERY		
	printf("plusd: RemotePrint joined\n");
#endif

#ifdef _CCVENGINEMCVE
		if((ret=pthread_join(thread[CCVQ_THREADID].thread, NULL))){
#ifdef _DEBUGGERY			
		if(ret==EDEADLK){
			printf("plusd: join: CcvThread - deadlock\n");
		}
		if(ret==EINVAL){
			printf("plusd: join: CcvThread - no such joinable thread\n");
		}
		if(ret==ESRCH){
			printf("plusd: join: CcvThread - no such thread\n");
		}
		if(ret==EFAULT){
			printf("plusd: join: CcvThread - invalid status pointer\n");
		}
#endif		
	}
#ifdef _DEBUGGERY		
	printf("plusd: CcvThread joined\n");
#endif
#endif
	return ret;
}

void PLUServer::CancelThreads(bool bForce)
{
	int biggles=1;
	
	if(bForce){
		if(thread[INQ_THREADID].status==THREAD_RUNNING)
			pthread_cancel(thread[INQ_THREADID].thread);
		if(thread[OUTQ_THREADID].status==THREAD_RUNNING)
			pthread_cancel(thread[OUTQ_THREADID].thread);
		if(thread[SQLQ_THREADID].status==THREAD_RUNNING)
			pthread_cancel(thread[SQLQ_THREADID].thread);
		if(thread[CTRLQ_THREADID].status==THREAD_RUNNING)
			pthread_cancel(thread[CTRLQ_THREADID].thread);
		if(thread[ERRQ_THREADID].status==THREAD_RUNNING)
			pthread_cancel(thread[ERRQ_THREADID].thread);
		if(thread[UDP_THREADID].status==THREAD_RUNNING)
			pthread_cancel(thread[UDP_THREADID].thread);
		if(thread[IDLE_THREADID].status==THREAD_RUNNING)
			pthread_cancel(thread[IDLE_THREADID].thread);
		if(thread[TRANSQ_THREADID].status==THREAD_RUNNING)
			pthread_cancel(thread[TRANSQ_THREADID].thread);
		if(thread[RPRINTQ_THREADID].status==THREAD_RUNNING)
			pthread_cancel(thread[RPRINTQ_THREADID].thread);
#ifdef _CCVENGINEMCVE			
		if(thread[CCVQ_THREADID].status==THREAD_RUNNING)
			pthread_cancel(thread[CCVQ_THREADID].thread);
#endif			
	}
	else{
		SetThread(INQ_THREADID, THREAD_PARAM_DEATH, &biggles);
		SetThread(OUTQ_THREADID, THREAD_PARAM_DEATH, &biggles);
		SetThread(SQLQ_THREADID, THREAD_PARAM_DEATH, &biggles);
		SetThread(CTRLQ_THREADID, THREAD_PARAM_DEATH, &biggles);
		SetThread(ERRQ_THREADID, THREAD_PARAM_DEATH, &biggles);
		SetThread(UDP_THREADID, THREAD_PARAM_DEATH, &biggles);
		SetThread(IDLE_THREADID, THREAD_PARAM_DEATH, &biggles);		
		SetThread(TRANSQ_THREADID, THREAD_PARAM_DEATH, &biggles);
		SetThread(RPRINTQ_THREADID, THREAD_PARAM_DEATH, &biggles);
#ifdef _CCVENGINEMCVE		
		SetThread(CCVQ_THREADID, THREAD_PARAM_DEATH, &biggles);
#endif		
	}
	pthread_mutex_destroy(&t_mutex);
}

int PLUServer::RestartThreads(bool bForce)
{
   pthread_mutex_lock(&t_mutex);
   CancelThreads(bForce);
   CreateThread(ALL_THREADS);
   pthread_mutex_unlock(&t_mutex);
	return 0;
}

int PLUServer::SetThread(unsigned int id, int param, void *value)
{
	if(value==NULL) return -1;
	
	pthread_mutex_lock(&t_mutex);
	
	// no more params on death notice...
	if(thread[id].die){
		pthread_mutex_unlock(&t_mutex);
		return 1;
	}
	switch(param){
	case THREAD_PARAM_STATE:
		thread[id].status=*((unsigned char *) value);
		break;
	case THREAD_PARAM_RSCOUNT:
		thread[id].restart_count=*((unsigned char *) value);
		break;
	case THREAD_PARAM_NORS:
		thread[id].norestart=*((unsigned char *) value);
		break;
	case THREAD_PARAM_DEATH:
		thread[id].die=*((unsigned char *) value);
		break;
	}
	pthread_mutex_unlock(&t_mutex);
	return 0;
}

int PLUServer::GetThread(unsigned int id, int param, void * value)
{
	if(value==NULL) return -1;
	
	pthread_mutex_lock(&t_mutex);
	switch(param){
	case THREAD_PARAM_STATE:
		*((int *) value)=thread[id].status;
		break;
	case THREAD_PARAM_RSCOUNT:
		*((int *) value)=thread[id].restart_count;
		break;
	case THREAD_PARAM_NORS:
		*((int *) value)=thread[id].norestart;
		break;
	case THREAD_PARAM_DEATH:
		*((int *) value)=thread[id].die;
		break;
	}
	pthread_mutex_unlock(&t_mutex);
	return 0;
}

int PLUServer::CheckQueues(int queue)
{
	return 0;
}

int PLUServer::CheckThreads(unsigned int threadid)
{
	pthread_mutex_lock(&t_mutex);
	
	if(threadid & INQ_THREAD){
		if(thread[INQ_THREADID].status==THREAD_DEAD){
			if(thread[INQ_THREADID].restart_count>=10){
				thread[INQ_THREADID].norestart=1;
			}
			else{
				thread[INQ_THREADID].restart_count++;
#ifdef _DEBUGGERY	
				if(DebugLevel>2) printf("InQ Thread dead, restarting: %d\n", thread[INQ_THREADID].restart_count);
#endif				
			}
			if(!thread[INQ_THREADID].norestart){
				CreateThread(INQ_THREAD);
			}
		}
	}

	if(threadid & OUTQ_THREAD){
		if(thread[OUTQ_THREADID].status==THREAD_DEAD){
			if(thread[OUTQ_THREADID].restart_count>=10){
				thread[OUTQ_THREADID].norestart=1;
			}
			else{
				thread[OUTQ_THREADID].restart_count++;
#ifdef _DEBUGGERY	
				if(DebugLevel>2) printf("OutQ Thread dead, restarting: %d\n", thread[OUTQ_THREADID].restart_count);
#endif				
			}
			if(!thread[OUTQ_THREADID].norestart){
				CreateThread(OUTQ_THREAD);
			}
		}
	}

	if(threadid & SQLQ_THREAD){
		if(thread[SQLQ_THREADID].status==THREAD_DEAD){
			if(thread[SQLQ_THREADID].restart_count>=10){
				thread[SQLQ_THREADID].norestart=1;
			}
			else{
				thread[2].restart_count++;
#ifdef _DEBUGGERY	
				if(DebugLevel>2) printf("SqlQ Thread dead, restarting: %d\n", thread[SQLQ_THREADID].restart_count);
#endif				
			}
			if(!thread[SQLQ_THREADID].norestart){
				CreateThread(SQLQ_THREAD);
			}
		}
	}

	if(threadid & CTRLQ_THREAD){
		if(thread[CTRLQ_THREADID].status==THREAD_DEAD){
			if(thread[CTRLQ_THREADID].restart_count>=10){
				thread[CTRLQ_THREADID].norestart=1;
			}
			else{
				thread[CTRLQ_THREADID].restart_count++;
#ifdef _DEBUGGERY	
				if(DebugLevel>2) printf("CtrlQ Thread dead, restarting: %d\n", thread[CTRLQ_THREADID].restart_count);
#endif				
			}
			if(!thread[CTRLQ_THREADID].norestart){
				CreateThread(CTRLQ_THREAD);
			}
		}
	}

	if(threadid & ERRQ_THREAD){
		if(thread[ERRQ_THREADID].status==THREAD_DEAD){
			if(thread[ERRQ_THREADID].restart_count>=10){
				thread[ERRQ_THREADID].norestart=1;
			}
			else{
				thread[ERRQ_THREADID].restart_count++;
#ifdef _DEBUGGERY	
				if(DebugLevel>2) printf("ErrorQ Thread dead, restarting: %d\n", thread[ERRQ_THREADID].restart_count);
#endif				
				if(!thread[ERRQ_THREADID].norestart){
					CreateThread(ERRQ_THREAD);
				}
			}
		}
	}

	if(threadid & UDP_THREAD){
		if(thread[UDP_THREADID].status==THREAD_DEAD){
			if(thread[UDP_THREADID].restart_count>=10){
				thread[UDP_THREADID].norestart=1;
			}
			else{
				thread[UDP_THREADID].restart_count++;
#ifdef _DEBUGGERY	
				if(DebugLevel>2) printf("UDP Thread dead, restarting: %d\n", thread[UDP_THREADID].restart_count);
#endif			
			}	
			if(!thread[UDP_THREADID].norestart){
				CreateThread(UDP_THREAD);
			}
		}
	}

	if(threadid & IDLE_THREAD){
		if(thread[IDLE_THREADID].status==THREAD_DEAD){
			if(thread[IDLE_THREADID].restart_count>=10){
				thread[IDLE_THREADID].norestart=1;
			}
			else{
				thread[IDLE_THREADID].restart_count++;
#ifdef _DEBUGGERY	
				if(DebugLevel>2) printf("Idle Thread dead, restarting: %d\n", thread[IDLE_THREADID].restart_count);
#endif				
			}
			if(!thread[IDLE_THREADID].norestart){
				CreateThread(IDLE_THREAD);
			}
		}
	}
	
	if(threadid & TRANSQ_THREAD){
		if(thread[TRANSQ_THREADID].status==THREAD_DEAD){
			if(thread[TRANSQ_THREADID].restart_count>=10){
				thread[TRANSQ_THREADID].norestart=1;
			}
			else{
				thread[TRANSQ_THREADID].restart_count++;
#ifdef _DEBUGGERY	
				if(DebugLevel>2) printf("Trans Thread dead, restarting: %d\n", thread[TRANSQ_THREADID].restart_count);
#endif				
			}
			if(!thread[TRANSQ_THREADID].norestart){
				CreateThread(TRANSQ_THREADID);
			}
		}
	}
	
	if(threadid & RPRINTQ_THREAD){
		if(thread[RPRINTQ_THREADID].status==THREAD_DEAD){
			if(thread[RPRINTQ_THREADID].restart_count>=10){
				thread[RPRINTQ_THREADID].norestart=1;
			}
			else{
				thread[RPRINTQ_THREADID].restart_count++;
#ifdef _DEBUGGERY	
				if(DebugLevel>2) printf("Remote Print Thread dead, restarting: %d\n", thread[RPRINTQ_THREADID].restart_count);
#endif				
			}
			if(!thread[RPRINTQ_THREADID].norestart){
				CreateThread(RPRINTQ_THREADID);
			}
		}
	}
	
#ifdef _CCVENGINEMCVE	
	if(threadid & CCVQ_THREAD){
		if(thread[CCVQ_THREADID].status==THREAD_DEAD){
			if(thread[CCVQ_THREADID].restart_count>=10){
				thread[CCVQ_THREADID].norestart=1;
			}
			else{
				thread[CCVQ_THREADID].restart_count++;
#ifdef _DEBUGGERY	
				if(DebugLevel>2) printf("CCV Thread dead, restarting: %d\n", thread[CCVQ_THREADID].restart_count);
#endif				
			}
			if(!thread[CCVQ_THREADID].norestart){
				CreateThread(CCVQ_THREAD);
			}
		}
	}	
#endif
	
	pthread_mutex_unlock(&t_mutex);
	return 0;
}
