/*
TransQ.cpp - transaction commit/retrieve/update/store queue
Copyright (C) 2002  Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos, Copyright (C) 2001-2005 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY.
*/
#include <pthread.h>
#include <cstdio>

#include "pos.h"
#include "sql.h"
#include "msgq.h"
#include "plusd.h"
#include "sql.h"
#include "plucache.h"

#include "bhmsgq.cpp"

CMessageQueue<BhPosMsg>          *TransQ;

extern CMessageQueue<BhPosMsg>   *InQ;
extern CMessageQueue<BhPosMsg>   *OutQ;
extern CMessageQueue<BhPosMsg>   *ErrorQ;
#ifdef _CCVENGINEMCVE
extern CMessageQueue<BhPosMsg>   *CcvQ;
#endif
static int TransEnd(PLUServer *p, bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg);
static int TransStore(PLUServer *p, bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg);
static int TransCancel(PLUServer *p, bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg);
static int TransUpdate(PLUServer *p, bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg);
static int TransRetrieve(PLUServer *p, bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg);
static int TransRetItems(PLUServer *p, bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg);
static int TransPluAdd(bhpos::BhSQL *, bhpos::CSaleItem *);
static int ReturnsEnd(PLUServer *p, bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg);
static int GoodsInEnd(PLUServer *p, bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg);
static int LaybySummary(PLUServer *p, bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg);
static int PluCommit(bhpos::CTransItems *ct, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg);
static int UPluCommit(bhpos::CTransItems *ct, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg);
static int PricePluCommit(bhpos::CTransItems *ct, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg);
static int ReturnPluCommit(PLUServer *p, bhpos::CTransItems *ct, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg);
static int RentalPluCommit(PLUServer *p, bhpos::CTransItems *ct, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg);
static int LaybyPluCommit(PLUServer *p, bhpos::CTransItems *ct, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg);
static int ReturnReasonCommit(PLUServer *p, bhpos::BhSQL *transdb, int id, int reason, int type);
static int ChequeCommit(bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg);
static int ChargeCommit(bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg);
static int CreditCommit(bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg);
static int CreditDone(bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg);

static int FillTrans(int transID, PLUServer *p, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, unsigned int, BhPosMsg *);
static int PluRetrieve(bhpos::CTransItems *ct, BhPosMsg *msg);
static int PayRetrieve(bhpos::CTransItems *ct, BhPosMsg *msg);
static int TransqPluLookup(bhpos::BhSQL *transdb, bhpos::CSaleItem *si);

int TransServerSide(PLUServer *p, bhpos::BhSQL *, BhPosMsg *msg);

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

void *TransQMonitor(void *arg)
{
	PLUServer	      *plu=(PLUServer *) arg;
	BhPosMsg		       msg;
	int			       status, old_state;
	bhpos::CClientInfo c;

   bhpos::CServSideTransStack  transstack;
//   bhpos::CServSideTrans      *trans;

	bool bPluandPlay=false;

	char dbhost[32];
	char dbname[32];
	char dbuser[32];
	char dbpass[32];

	bhpos::BhSQL			*transdb;
	bhpos::BhSQLResult	sql_result;
	
	TransQ=new CMessageQueue<BhPosMsg>();
	if(!TransQ){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("No TransQ Allocation\n");
#endif			
		return NULL;
	}
	
	if(TransQ->GetError()){
#ifdef _DEBUGGERY
		if(DebugLevel)	
			printf("TransQ instansiation error\n");
#endif		
	}
	
	if(TransQ->GetError()!=0){
#ifdef _DEBUGGERY
		if(DebugLevel)		
			printf("TransQ Msg Queue Error: %d\n", TransQ->GetError());
#endif
		
		TransQ=NULL;	
		status=THREAD_DEAD;
		plu->SetThread(TRANSQ_THREADID, THREAD_PARAM_STATE, &status);			
		return NULL;
	}	
	
	plu->GetOption(BH_PLUS_OPTION_DBHOSTPRIME, dbhost);
	plu->GetOption(BH_PLUS_OPTION_DBDATABASE, dbname);
	plu->GetOption(BH_PLUS_OPTION_DBUSER, dbuser);
	plu->GetOption(BH_PLUS_OPTION_DBPASSWORD, dbpass);
	plu->GetOption(BH_PLUS_OPTION_PLUANDPLAY, &bPluandPlay);

	transdb=new bhpos::BhSQL(dbhost, dbname, 0, dbuser, dbpass);
	
	if(!transdb){
#ifdef _DEBUGGERY
		if(DebugLevel)		
			printf("PLUServer - Sql Thread : failed to create SQL object\n");
#endif
		delete TransQ;
		TransQ=NULL;	
		status=THREAD_DEAD;
		plu->SetThread(TRANSQ_THREADID, THREAD_PARAM_STATE, &status);			
		return NULL;
	}

	if(transdb->Connect()){
#ifdef _DEBUGGERY
		if(DebugLevel)		
			printf("PLUServer - Sql Thread : failed to connect to SQL database - \n%s\n", transdb->ErrorMsg());
#endif			
		delete TransQ;
		TransQ=NULL;	
		status=THREAD_DEAD;
		plu->SetThread(TRANSQ_THREADID, THREAD_PARAM_STATE, &status);
		return NULL;
	}
	
	status=TransQ->Grab();
	if(status != 0) {
#ifdef _DEBUGGERY
		if(DebugLevel)		
			printf("TransQ failed to Grab Mutex, %d (%s)\n", status, strerror(status));
#endif
		delete TransQ;
		TransQ=NULL;	
		status=THREAD_DEAD;
		plu->SetThread(TRANSQ_THREADID, THREAD_PARAM_STATE, &status);	
		return NULL;
	}
	
	status=THREAD_RUNNING;
	plu->SetThread(TRANSQ_THREADID, THREAD_PARAM_STATE, &status);
	
	while(1){
		status=TransQ->WaitOnMsg();
		if(status != 0) {
#ifdef _DEBUGGERY
			if(DebugLevel)			
				printf("TransQ thread wait failed, %d (%s)\n", status, strerror(status));
#endif			
			TransQ->Release();
			delete TransQ;
			TransQ=NULL;	
			status=THREAD_DEAD;
			plu->SetThread(TRANSQ_THREADID, THREAD_PARAM_STATE, &status);
			return NULL;
		}
		if(!TransQ->HasMsg()) continue;

		TransQ->GetNext(&msg);
		TransQ->DeleteLast();
		TransQ->Release();
		
#ifdef _DEBUGGERY
		if(DebugLevel > 2) 	
		printf("TransQ popped type:%02d\ncmd:%02d\nsubcmd:%02d\n", msg.hdr.type, msg.hdr.cmd, msg.hdr.subcmd);
#endif

		if(plu->ClientExists(c, &msg)){
			// check client status - i.e. ready for transcommit
		}
		else{
			// this should never happen...
#ifdef _DEBUGGERY
			if(DebugLevel > 2) 	
				printf("TransQ client doesn't exist\n");
#endif
			msg.hdr.Ex=BHPOS_REFUSE;
			if(OutQ)
				OutQ->Add(&msg);
			status=TransQ->Grab();
			continue;
		}

		if(msg.hdr.type==BHPOS_TYPE_POST && msg.hdr.cmd==BHMSG_POST_TRANSACTION){
			// disable thread cancelling on the following
			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &old_state);
			switch(msg.hdr.subcmd){
			case BHMSG_POST_TRANSACTION_STORE:
				TransStore(plu, transdb, c, &msg);
				break;
			case BHMSG_POST_TRANSACTION_CANCEL:
			   TransCancel(plu, transdb, c, &msg);
			   break;
			case BHMSG_POST_TRANSACTION_RETRIEVE:
			case BHMSG_POST_TRANSACTION_RECALL:
         case BHMSG_POST_TRANSACTION_RESTORE:
            TransRetrieve(plu, transdb, c, &msg);
				break;
			case BHMSG_POST_TRANSACTION_RET_PLU:
         case BHMSG_POST_TRANSACTION_RET_PT:         
				TransRetItems(plu, transdb, c, &msg);
				break;
			case BHMSG_POST_TRANSACTION_END:
				TransEnd(plu, transdb, c, &msg);
				break;
			case BHMSG_POST_TRANSACTION_RETURNS_END:
			   ReturnsEnd(plu, transdb, c, &msg);
			   break;
      	case BHMSG_POST_TRANSACTION_GOODSIN_END:
      	   GoodsInEnd(plu, transdb, c, &msg);
			   break;
			case BHMSG_POST_TRANSACTION_CUSTSUMMARY:
			   LaybySummary(plu, transdb, c, &msg);
			   break;
			case BHMSG_POST_TRANSACTION_ADDITEM:
      	case BHMSG_POST_TRANSACTION_ADDPAYITEM:
      	   break;
			case BHMSG_POST_TRANSACTION_MICRCOMMIT:
				ChequeCommit(transdb, c, &msg);
				break;
         /*
         case BHMSG_POST_TRANSACTION_CHARGECOMMIT:
				ChargeCommit(transdb, c, &msg);
            break;*/
			case BHMSG_POST_TRANSACTION_CREDITCOMMIT:
				CreditCommit(transdb, c, &msg);
            break;
			case BHMSG_POST_TRANSACTION_CREDITDONE:
				CreditDone(transdb, c, &msg);
				break;
			}
			pthread_setcancelstate(old_state, NULL);
		}
		if(msg.hdr.type==BHPOS_TYPE_SERVICE){
		}
		if(msg.hdr.type==BHPOS_TYPE_SERVERSIDE){
		   TransServerSide(plu, transdb, &msg);
		}
		plu->UpdateClient(c);
		// check cancellation
		plu->GetThread(TRANSQ_THREADID, THREAD_PARAM_DEATH, &status);
		if(status && !TransQ->HasMsg()) break;
		
		status=TransQ->Grab();
		if(status != 0) {
#ifdef _DEBUGGERY
			if(DebugLevel)			
				printf("TransQ failed to Grab Mutex, %d (%s)\n", status, strerror(status));
#endif				
			delete TransQ;
			TransQ=NULL;	
			status=THREAD_DEAD;
			plu->SetThread(TRANSQ_THREADID, THREAD_PARAM_STATE, &status);
			return NULL;
		}
		printf("mutex grabbed");
	}
	TransQ->Release();
	delete TransQ;
	TransQ=NULL;	
	status=THREAD_DEAD;
	plu->SetThread(TRANSQ_THREADID, THREAD_PARAM_STATE, &status);
	return NULL;
}


/*****************************************************************************************************/

int TransEnd(PLUServer *p, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	bhpos::CTransItems ct;
	BhTransEnd        *te=(BhTransEnd  *) msg->data;
	BhPosMsg 	       rmsg;
	int			       pay_id=0, trans_id=0, status=0, postID=0, userID=0, accountid;
	double		       amount=0.0;
	bhpos::BhSQLResult sql_result;
   bhpos::BhSQLRow   *sql_row;
	bhpos::CSaleItem	 si;
	bhpos::CPayItem	 pi;
	char				    buff[256], eaid[21];
	
	trans_id=te->transID;
	
#ifdef _DEBUGGERY				
	if(DebugLevel > 2){
		printf("Trans commit for trans ID: %d\n", trans_id);
	}
#endif	
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	c.GetTransaction(trans_id, ct);
	
	c.GetParameter(BHCI_PARAM_STATIONID, &postID);
	c.GetParameter(BHCI_PARAM_USERID, &userID);
   p->GetOption(BH_PLUS_OPTION_ACCDEFSALE, &accountid);
	memset(eaid, 0, sizeof(eaid));
	c.GetEnteredAt(eaid);
#ifdef _DEBUGGERY				
	if(DebugLevel > 2){
		printf("Trans commit: EnteredAt: %s\n", eaid);
	}
#endif
	
	// is a delivery?
	if(c.IsDelivery()){
	   int cid;
	   c.GetParameter(BHCI_PARAM_CUSTID, &cid);
      
      status=transdb->SimpleQuery(sqlq_qry70);
   	if(status!=0){
#ifdef _DEBUGGERY				
   		if(DebugLevel > 2){
			   printf("Failed to lock createmod: %s\n", transdb->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 1;
   	}
      
	   sprintf(buff, sqlq_qry90, userID, userID);
	   status=transdb->SimpleQuery(buff);
   	if(status!=0){
#ifdef _DEBUGGERY				
   		if(DebugLevel > 2){
			   printf("Failed to insert createmod: %s\n", transdb->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 1;
   	}
      
      status=transdb->SimpleQuery(sqlq_qry77);
   	if(status!=0){
#ifdef _DEBUGGERY				
   		if(DebugLevel > 2){
			   printf("Failed to unlock createmod: %s\n", transdb->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 1;
   	}
      
      if(sql_result.init(transdb, transq_qry31, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
         if(DebugLevel > 2){
	         printf("Failed to query transq_qry31: %s\n", transdb->ErrorMsg());
         }
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
         rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
         return 2;
      }
      sql_row=sql_result.GetRow();
      pay_id=atoi(sql_row->GetField(0));
      
	   //pay_id=sql_result.GetLastID(transdb);
      
	   status=transdb->SimpleQuery(sqlq_qry70);
   	if(status!=0){
#ifdef _DEBUGGERY				
   		if(DebugLevel > 2){
			   printf("Failed to lock delivery: %s\n", transdb->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 1;
   	}
	   sprintf(buff, transq_qry1, pay_id, cid, ct.GetID(), SQL_DELIVERY_READY, ct.GetNumber(bhpos::BHNUM_TOTAL));
	
   	status=transdb->SimpleQuery(buff);
   	if(status!=0){
#ifdef _DEBUGGERY				
   		if(DebugLevel > 2){
			   printf("Failed to insert delivery: %s\n", transdb->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 1;
   	}
	
      status=transdb->SimpleQuery(sqlq_qry77);
   	if(status!=0){
#ifdef _DEBUGGERY				
   	   if(DebugLevel > 2){
   	      printf("Failed to unlock delivery: %s\n", transdb->ErrorMsg());
      	}
#endif
   	   rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
      	if(OutQ) OutQ->Add(&rmsg);
      	return 1;
   	}
   }
	
	// let's clear linked trans items first...
	if(ct.NumReturnItems()!=0){
#ifdef _DEBUGGERY				
   	if(DebugLevel > 2){
		   printf("Trans has return Items: %d\n", ct.NumReturnItems());
   	}
#endif      
	   if(ReturnPluCommit(p, &ct, transdb, c, msg)){
         //returns only trans
         memcpy(&rmsg, msg, sizeof(rmsg));	
	      rmsg.hdr.Ex=BHPOS_GRANT;
	      rmsg.hdr.size=0;
	      if(OutQ)
		      OutQ->Add(&rmsg);
	      return 0;
      }
	}
	if(ct.NumLaybyItems()!=0){
#ifdef _DEBUGGERY				
   	if(DebugLevel > 2){
		   printf("Trans has layby Items: %d\n", ct.NumLaybyItems());
   	}
#endif
	   if(LaybyPluCommit(p, &ct, transdb, c, msg)) return 1; // failure
	}
	if(ct.NumRentalItems()!=0){
	   if(RentalPluCommit(p, &ct, transdb, c, msg)) return 1; // failure
	}
	
	// let's try a commit 	
#ifdef _DEBUGGERY				
	if(DebugLevel > 2){
		printf("Trans items ID: %d\nNum Items: %d\n", ct.GetID(), ct.NumItems());
	}
#endif

   if(ct.NumItems()!=0){
      if(PluCommit(&ct, transdb, c, &rmsg)){
#ifdef _DEBUGGERY				
			if(DebugLevel > 2){
				printf("TransEnd: PluCommit Failed!\n");
			}
#endif		
	      return 1;
	   }
	}
   if(ct.NumUnknownItems()!=0){
		if(UPluCommit(&ct, transdb, c, &rmsg)){
	      return 1;
	   }
	}
   
   if(ct.NumPriceItems()!=0){
		if(PricePluCommit(&ct, transdb, c, &rmsg)){
	      return 1;
	   }
	}
   
   if(!c.IsDelivery()){
      // insert, get last ID, group for split payment
      while(ct.PopPayment(pi)){
         //amount=ct.GetNumber(bhpos::BHNUM_TOTAL);
         if(pi.GetID()==bhpos::PTYPE_CASH){
            if(ct.GetNumber(bhpos::BHNUM_CHANGE) > 0.0){
#ifdef _DEBUGGERY
         		if(DebugLevel > 2){
			         printf("payment cash with change: %f - %f\n", pi.GetAmount(), ct.GetNumber(bhpos::BHNUM_CHANGE));
         		}
#endif
               status=transdb->SimpleQuery(sqlq_qry70);
   	         if(status!=0){
#ifdef _DEBUGGERY				
            		if(DebugLevel > 2){
			            printf("Failed to lock adjustment: %s\n", transdb->ErrorMsg());
   		         }
#endif
            		rmsg.hdr.Ex=BHPOS_REFUSE;
   		         rmsg.hdr.size=0;
            		if(OutQ) OutQ->Add(&rmsg);
   		         return 1;
   	         }
               
               sprintf(buff, sqlq_qry120, eaid, userID, SQL_A_CHANGE, ct.GetNumber(bhpos::BHNUM_CHANGE));
               
               status=transdb->SimpleQuery(buff);
          	   if(status!=0){
#ifdef _DEBUGGERY				
            		if(DebugLevel > 2){
      			      printf("Failed To insert change adjustment item: %s\n", transdb->ErrorMsg());
   		         }
#endif
            		rmsg.hdr.Ex=BHPOS_REFUSE;
            		rmsg.hdr.size=0;
            		if(OutQ) OutQ->Add(&rmsg);
            		return 1;
      	      }
               
               status=transdb->SimpleQuery(sqlq_qry77);
   	         if(status!=0){
#ifdef _DEBUGGERY				
            		if(DebugLevel > 2){
			            printf("Failed to unlock adjustment: %s\n", transdb->ErrorMsg());
   		         }
#endif
            		rmsg.hdr.Ex=BHPOS_REFUSE;
   		         rmsg.hdr.size=0;
            		if(OutQ) OutQ->Add(&rmsg);
   		         return 1;
   	         }
               
               amount=pi.GetAmount() - ct.GetNumber(bhpos::BHNUM_CHANGE);
               sprintf(buff, transq_qry2, ct.GetID(), pi.GetID(), amount, userID, eaid);
               
            }
            else{
               sprintf(buff, transq_qry2,	ct.GetID(), pi.GetID(), pi.GetAmount(), userID, eaid);
            }
         }
         else if(pi.GetID()==bhpos::PTYPE_CHEQUE){
            if(pi.GetFlags()!=bhpos::PTYPE_FLAG_DONE){ // pre-committed
               sprintf(buff, transq_qry2, ct.GetID(), pi.GetID(), pi.GetAmount(), userID, eaid);
            }
         }
         else if(pi.GetID()==bhpos::PTYPE_INHOUSECHARGE){ // in-house charge
            int cid, chargeaccountid;
	         c.GetParameter(BHCI_PARAM_CUSTID, &cid);
            
            // get the chargeaccountid
            sprintf(buff, transq_qry34, cid);
            if(sql_result.init(transdb, buff, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
               if(DebugLevel > 2){
   	            printf("Failed to query transq_qry34: %s\n", transdb->ErrorMsg());
               }
#endif
               rmsg.hdr.Ex=BHPOS_REFUSE;
               rmsg.hdr.size=0;
               if(OutQ) OutQ->Add(&rmsg);
               return 2;
            }
            if(sql_result.NumRows()!=0){
               rmsg.hdr.Ex=BHPOS_REFUSE;
               rmsg.hdr.size=0;
               if(OutQ) OutQ->Add(&rmsg);
               return 3;
            }
            sql_row=sql_result.GetRow();
            chargeaccountid=atoi(sql_row->GetField(0));
            
            // insert chargeitem
            sprintf(buff, transq_qry33, chargeaccountid, 0, ct.GetID(), pi.GetAmount());
            if(sql_result.init(transdb, buff, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
               if(DebugLevel > 2){
   	            printf("Failed to query transq_qry34: %s\n", transdb->ErrorMsg());
               }
#endif
               rmsg.hdr.Ex=BHPOS_REFUSE;
               rmsg.hdr.size=0;
               if(OutQ) OutQ->Add(&rmsg);
               return 4;
            }
            
            // standard pay item
            sprintf(buff, transq_qry2, ct.GetID(), pi.GetID(), pi.GetAmount(), userID, eaid);
         }
         else{
            sprintf(buff, transq_qry2, ct.GetID(), pi.GetID(), pi.GetAmount(), userID, eaid);
         }
#ifdef _DEBUGGERY				
   		if(DebugLevel > 2){
			   printf("pay insert: %s\n", buff);
   		}
#endif
         status=transdb->SimpleQuery(sqlq_qry70);
   	   if(status!=0){
#ifdef _DEBUGGERY				
            if(DebugLevel > 2){
			      printf("Failed to lock paymentitem: %s\n", transdb->ErrorMsg());
   		   }
#endif
            rmsg.hdr.Ex=BHPOS_REFUSE;
   		   rmsg.hdr.size=0;
            if(OutQ) OutQ->Add(&rmsg);
   		       return 1;
   	   }
               
      	status=transdb->SimpleQuery(buff);
   	   if(status!=0){
#ifdef _DEBUGGERY				
      		if(DebugLevel > 2){
			      printf("Failed To insert pay item: %s\n", transdb->ErrorMsg());
   		   }
#endif
      		rmsg.hdr.Ex=BHPOS_REFUSE;
      		rmsg.hdr.size=0;
      		if(OutQ) OutQ->Add(&rmsg);
      		return 1;
      	}
         status=transdb->SimpleQuery(sqlq_qry77);
   	   if(status!=0){
#ifdef _DEBUGGERY				
            if(DebugLevel > 2){
			      printf("Failed to unlock paymentitem: %s\n", transdb->ErrorMsg());
   		   }
#endif
            rmsg.hdr.Ex=BHPOS_REFUSE;
   		   rmsg.hdr.size=0;
            if(OutQ) OutQ->Add(&rmsg);
   		       return 1;
   	   }
         
         if(sql_result.init(transdb, transq_qry30, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
            if(DebugLevel > 2){
	            printf("Failed to query transq_qry30: %s\n", transdb->ErrorMsg());
            }
#endif
            //db->SimpleQuery(sqlq_qry146);
            rmsg.hdr.Ex=BHPOS_REFUSE;
            rmsg.hdr.size=0;
            if(OutQ) OutQ->Add(&rmsg);
            return 2;
         }
         sql_row=sql_result.GetRow();
         if(sql_row==NULL){
         }
         pay_id=atoi(sql_row->GetField(0));
	      sprintf(buff, transq_qry3, pay_id, ct.GetID());
   	   status=transdb->SimpleQuery(buff);
      	if(status!=0){
#ifdef _DEBUGGERY				
      		if(DebugLevel > 2){
			      printf("Failed To update trans pay ID: %s\n", transdb->ErrorMsg());
      		}
#endif
      		rmsg.hdr.Ex=BHPOS_REFUSE;
      		rmsg.hdr.size=0;
      		if(OutQ) OutQ->Add(&rmsg);
      		return 1;
      	}
         
      	if(pi.GetID()==bhpos::PTYPE_CASH){
      		amount=pi.GetAmount();
   	   	c.SetParameter(BHCI_PARAM_CASHADD, &amount);
   	   	amount=ct.GetNumber(bhpos::BHNUM_CHANGE);
   	   	c.SetParameter(BHCI_PARAM_CASHSUBTRACT, &amount);
   	   	c.GetParameter(BHCI_PARAM_TIDEMARK, &amount);
            if(amount!=0.0){
#ifdef _DEBUGGERY				
      		   if(DebugLevel > 2){
			         printf("Cash Drawer Tidemark amount : %0.2f\n", amount);
      		   }
#endif
               // tidemark warning...
               BhPosMsg emsg;
					BhWarningTidemark *tm=(BhWarningTidemark *) emsg.data;

               emsg.hdr.type=BHPOS_TYPE_ERROR;
               emsg.hdr.cmd=BHMSG_ERROR_WARNING;
               if(amount > 0.0){
                  emsg.hdr.subcmd=BHMSG_ERROR_WARNING_HIGHTIDEMARK;
               }
               else{
                  emsg.hdr.subcmd=BHMSG_ERROR_WARNING_LOWTIDEMARK;
               }
					tm->tidemark=amount;
					c.GetNumber(BHCI_NUM_CASHIND, NULL, &amount);
					tm->amount=amount;
               emsg.hdr.size=sizeof(BhWarningTidemark);
               
               if(ErrorQ) ErrorQ->Add(&emsg);
            }
   	   }
	   }
		// if not delivery, close
		
	   status=0;
   	c.GetParameter(BHCI_PARAM_TRANSOPEN, &status);
		
   	sprintf(buff, transq_qry4,	userID, eaid, 15, pay_id, ct.GetID());
   }
   else{
      status=0;
   	c.GetParameter(BHCI_PARAM_TRANSOPEN, &status);
      sprintf(buff, transq_qry5, userID, eaid, ct.GetID());
   }
	status=transdb->SimpleQuery(buff);
	if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Failed to update trans: %s\n", transdb->ErrorMsg());
		}
#endif
		rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
   
   // if not standard price policy
   int nid=ct.GetNameID();
   if(nid!=1){
      sprintf(buff, transq_qry6, ct.GetID(), nid);
	   status=transdb->SimpleQuery(buff);
	   if(status > 0){
#ifdef _DEBUGGERY				
		   if(DebugLevel > 2){
			   printf("Failed to insert transtype: %s\n", transdb->ErrorMsg());
   		}
#endif
	   	rmsg.hdr.Ex=BHPOS_REFUSE;
		   rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
	   	return 1;
	   }
   }

   // add a ledger entry...
   status=transdb->SimpleQuery(sqlq_qry163);
   if(status!=0){
#ifdef _DEBUGGERY				
   	if(DebugLevel > 2){
		   printf("Failed to lock ledgerentry: %s\n", transdb->ErrorMsg());
   	}
#endif
   	rmsg.hdr.Ex=BHPOS_REFUSE;
   	rmsg.hdr.size=0;
   	if(OutQ) OutQ->Add(&rmsg);
   	return 1;
   }

   amount=ct.GetNumber(bhpos::BHNUM_TOTAL);
   sprintf(buff, transq_qry37, 1,1,accountid, "sale", "N",amount,"N",trans_id);
	status=transdb->SimpleQuery(buff);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
         printf("Failed to insert ledgerentry: %s\n", transdb->ErrorMsg());
      }
#endif
      rmsg.hdr.Ex=BHPOS_REFUSE;
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
   	return 1;
   }
      
   status=transdb->SimpleQuery(sqlq_qry77);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
         printf("Failed to unlock ledgerentry: %s\n", transdb->ErrorMsg());
      }
#endif
      rmsg.hdr.Ex=BHPOS_REFUSE;
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
   	return 1;
   }
   
   status=POST_MODE_TRANSEND;
	c.SetParameter(BHCI_PARAM_MODE, &status);
	status=POST_STATUS_IDLE;
	c.SetParameter(BHCI_PARAM_STATUS, &status);
	c.ClearTrans(trans_id);
   
	memset(&rmsg, 0, sizeof(rmsg));
	
#ifdef _DEBUGGERY				
	if(DebugLevel > 2){
		printf("TransQ: TransEnd completed\n");
	}
#endif
	memcpy(&rmsg, msg, sizeof(rmsg));	
	rmsg.hdr.Ex=BHPOS_GRANT;
	rmsg.hdr.size=0;
	if(OutQ)
		OutQ->Add(&rmsg);
	return 0;
}

static int TransCancel(PLUServer *p, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	bhpos::BhSQLResult   sql_result;
	bhpos::BhSQLRow     *sql_row;
	BhPosMsg			      rmsg;
	char				      buff[128];
	int                  status, postid, userid;
   BhTransactionCancel *tc=(BhTransactionCancel *) msg->data;
	//unsigned int		  *tid=(unsigned int *) msg->data;
	
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	rmsg.hdr.size=0;
	rmsg.hdr.Ex=BHPOS_GRANT;	// an ack
	
	c.GetParameter(BHCI_PARAM_STATIONID, &postid);
   c.GetParameter(BHCI_PARAM_USERID, &userid);
	
   if(tc->type==TRANSACTIONCANCEL_TYPE_VOID){
      c.SetParameter(BHCI_PARAM_TRANSVOID, NULL);
   
	   sprintf(buff, sqlq_qry18, postid, SQL_P_VOIDSALE, userid);
      status=transdb->SimpleQuery(buff);
      if(status!=0){
#ifdef _DEBUGGERY          
         if(DebugLevel > 2){
            printf("Failed log connect message for post %d\n", postid);
         }
#endif
      }
      transdb->SimpleQuery(sqlq_qry70);
      sprintf(buff, sqlq_qry19, userid, SQL_U_VOIDSALE);
      status=transdb->SimpleQuery(buff);
      transdb->SimpleQuery(sqlq_qry77);
      if(status!=0){
#ifdef _DEBUGGERY          
         if(DebugLevel > 2){
            printf("Failed log connect message for user %d\n", postid);
         }
#endif
      }
   }
   	
#ifdef _DEBUGGERY				
   if(DebugLevel > 2){
   	printf("TransQ: TransCancel for %d\n", tc->transID);
	}
#endif
	transdb->SimpleQuery(sqlq_qry70);
	sprintf(buff, transq_qry7, tc->transID);
	sql_result.init(transdb, buff, true);
	if(sql_result.NumRows()==1){
#ifdef _DEBUGGERY				
   	if(DebugLevel > 2){
   		printf("TransQ: TransCancel found row...\n");
	   }
#endif
		sql_row=sql_result.GetRow();
      if(tc->type==TRANSACTIONCANCEL_TYPE_VOID){
		   sprintf(buff, transq_qry8, (unsigned int) atoi(sql_row->GetField(0)));
      }
      else{
         sprintf(buff, transq_qry35, (unsigned int) atoi(sql_row->GetField(0)));
      }
		sql_result.init(transdb, buff, true);
      transdb->SimpleQuery(sqlq_qry77);
		status=POST_STATUS_IDLE;
		c.SetParameter(BHCI_PARAM_STATUS, &status);
	}
	if(OutQ)
		OutQ->Add(&rmsg);

	return 0;
}

int TransStore(PLUServer *p, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	//c.SetTransOpen(true);
	return 0;
}

///////////////////////////////////////////////////////
//
// This fills a Ctransitems class with the transaction
// from the dB, then empty it to the client

int TransRetrieve(PLUServer *p, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	bhpos::BhSQLResult 	 sql_result;
	bhpos::BhSQLRow		*sql_row;
	BhPosMsg			  		 rmsg;
	char				   	 buff[128];
	int                   status;
	BhTransRetrieve      *tr;
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	tr=(BhTransRetrieve *) msg->data;
	
	rmsg.hdr.size=sizeof(BhTransRetrieveReply);
	rmsg.hdr.Ex=BHPOS_GRANT;
	
#ifdef _DEBUGGERY				
   	if(DebugLevel > 2){
   		printf("TransQ: TransRetrieve\n");
	   }
#endif

	if(!c.NewTrans(tr->transID, 1)){
#ifdef _DEBUGGERY
     	if(DebugLevel > 2){
        	printf("Fill Trans failed to create new trans...\n");
     	}
#endif
		return 1;
	}
	
	switch(msg->hdr.subcmd){
	case BHMSG_POST_TRANSACTION_RECALL:
#ifdef _DEBUGGERY				
   	if(DebugLevel > 2){
   		printf("TransQ: TransRetrieve : recall\n");
	   }
#endif	
	   sprintf(buff, transq_qry9, (unsigned int) tr->transID);
   	sql_result.init(transdb, buff, true);
   	if(sql_result.NumRows()==1){
   		sql_row=sql_result.GetRow();
   		if(strcmp(sql_row->GetField(0), "N")==0){   // trans is closed
			   rmsg.hdr.Ex=BHPOS_REFUSE;
			   rmsg.hdr.size=0;
			   if(OutQ) OutQ->Add(&rmsg);
			   return 1;
   		}
   		else{   // build a transaction in client info
         	if(FillTrans(tr->transID, p, transdb, c, tr->flags, &rmsg)){
	            rmsg.hdr.Ex=BHPOS_REFUSE;
			      rmsg.hdr.size=0;
			      if(OutQ) OutQ->Add(&rmsg);
			      return 2;
	         } 
         }
      }
	   break;
	   
   case BHMSG_POST_TRANSACTION_RESTORE:
#ifdef _DEBUGGERY				
   	if(DebugLevel > 2){
   		printf("TransQ: TransRetrieve : restore\n");
	   }
#endif	
      sprintf(buff, transq_qry10, (unsigned int) tr->transID);
   	sql_result.init(transdb, buff, true);
   	if(sql_result.NumRows()!=0){
   	   if(FillTrans(tr->transID, p, transdb, c, tr->flags, &rmsg)){
#ifdef _DEBUGGERY				
         	if(DebugLevel > 2){
         		printf("TransQ: TransRestore failed to fill\n");
      	   }
#endif   	      
	         rmsg.hdr.Ex=BHPOS_REFUSE;
		      rmsg.hdr.size=0;
		      if(OutQ) OutQ->Add(&rmsg);
		      return 3;
         } 
   	}
   	else{
#ifdef _DEBUGGERY				
      	if(DebugLevel > 2){
      		printf("TransQ: TransRestore no rows\n");
         }
#endif   	   
         rmsg.hdr.Ex=BHPOS_REFUSE;
	      rmsg.hdr.size=0;
	      if(OutQ) OutQ->Add(&rmsg);
	      return 4;
      }
      break;
      
	case BHMSG_POST_TRANSACTION_RETRIEVE:
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
         printf("TransQ: TransRetreive: id=%d, flags=%d\n", tr->transID, tr->flags);
      }
#endif
	   sprintf(buff, transq_qry9, (unsigned int) tr->transID);
   	if(sql_result.init(transdb, buff, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
         if(DebugLevel > 2){
            printf("TransRetreive: trans query failed: %s\n", transdb->ErrorMsg());
         }
			rmsg.hdr.Ex=BHPOS_REFUSE;
   	   rmsg.hdr.size=0;
			if(OutQ) OutQ->Add(&rmsg);
			return 5;
#endif
		}
   	if(sql_result.NumRows()==0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
            printf("TransQ: TransRetreive: 0 rows\n");
         }
#endif          
         rmsg.hdr.Ex=BHPOS_REFUSE;
   	   rmsg.hdr.size=0;
			if(OutQ) OutQ->Add(&rmsg);
			return 6;
   	}
   	else{
   		sql_row=sql_result.GetRow();
   		if(strcmp(sql_row->GetField(0), "N")==0){   // trans is closed
   		   // reopen 
   		   sprintf(buff, transq_qry11, (unsigned int) tr->transID);
         	status=transdb->SimpleQuery(buff);
            if(status!=0){
#ifdef _DEBUGGERY          
               if(DebugLevel > 2){
                  printf("Failed update trans: %s\n", transdb->ErrorMsg());
               }
#endif
					rmsg.hdr.Ex=BHPOS_REFUSE;
		   	   rmsg.hdr.size=0;
					if(OutQ) OutQ->Add(&rmsg);
               return 7;
            }
         }
      	if(FillTrans(tr->transID, p, transdb, c, tr->flags, &rmsg)){
#ifdef _DEBUGGERY				
         	if(DebugLevel > 2){
         		printf("TransQ: TransRecall failed to fill\n");
      	   }
#endif
            rmsg.hdr.Ex=BHPOS_REFUSE;
			   rmsg.hdr.size=0;
			   if(OutQ) OutQ->Add(&rmsg);
			   return 1;
         }
      }
		// printf("trans num items: %d\n", c.NumItems(tr->transID, bhpos::BHITEM_TYPE_ALL));
	   break;
	}
	if(OutQ) OutQ->Add(&rmsg);
	return 0;
}

/////////////////////////////////////////////////

int TransRetItems(PLUServer *p, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   bhpos::CTransItems  ct;
	BhTransRetrieve    *tr=(BhTransRetrieve *) msg->data;
	int                 trans_id=tr->transID;
	BhPosMsg            rmsg;

	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	rmsg.hdr.size=0;
	rmsg.hdr.Ex=BHPOS_GRANT;
	
	c.GetTransaction(trans_id, ct);
	printf("TransQ: TransRetreive: id=%d : num_item:%d\n", trans_id, ct.NumItems());
	
   
	switch(msg->hdr.subcmd){
   case BHMSG_POST_TRANSACTION_RET_PLU:
      PluRetrieve(&ct, msg);
      break;
   case BHMSG_POST_TRANSACTION_RET_PT:
      PayRetrieve(&ct, msg);
      break;
   case BHMSG_POST_TRANSACTION_RET_END:
		c.ClearTrans(trans_id);
		if(OutQ) OutQ->Add(&rmsg);
      break;
   }
	c.UpdateTrans(trans_id, ct);
	return 0;
}

//////////////////////////////////////////////////////////////////

int TransPluAdd(bhpos::BhSQL *transdb, bhpos::CSaleItem *si)
{
   int                 status;
	char                buff[255];
	bhpos::BhSQLResult  sql_result;
	
   transdb->SimpleQuery(sqlq_qry70);
	sprintf(buff, transq_qry12, si->GetPLU().data(), si->GetDispText().data(), si->GetPrice(), si->GetTax());
   status=transdb->SimpleQuery(buff);
   transdb->SimpleQuery(sqlq_qry77);
#ifdef _DEBUGGERY
	if(status!=0){
   	if(DebugLevel > 2){
			printf("Failed To insert unknown plu: %s\n", transdb->ErrorMsg());
   	}
   }
#endif
	   
	return 0;
}

///////////////////////////////////////////////////////////
//
// This is a dedicated active returns end
//   The original trans is linked to a new trans
//
//
int ReturnsEnd(PLUServer *p, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   bhpos::CTransItems ct;
	BhTransEnd        *te=(BhTransEnd *) msg->data;
	BhPosMsg 	       rmsg;
	int			       pay_id=0, trans_id=0, status=0, postID=0, userID=0, ret_id;
	double		       amount=0.0;
	bhpos::BhSQLResult sql_result;
   bhpos::BhSQLRow   *sql_row;
	bhpos::CSaleItem	 si;
	bhpos::CPayItem	 pi;
	char				    buff[256], eaid[21];
	
	trans_id=te->transID;
	
#ifdef _DEBUGGERY				
	if(DebugLevel > 2){
		printf("Returns commit\n");
	}
#endif	
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	c.GetTransaction(trans_id, ct);
	c.GetParameter(BHCI_PARAM_STATIONID, &postID);
	c.GetParameter(BHCI_PARAM_USERID, &userID);
	memset(eaid, 0, sizeof(eaid));
	c.GetEnteredAt(eaid);

	// let's try a commit
#ifdef _DEBUGGERY				
	if(DebugLevel > 2){
		printf("Returns: Items: %d\n", ct.NumItems());
	}
#endif	

   status=transdb->SimpleQuery(sqlq_qry70);
   if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
   		printf("Failed to lock trans table for returns insert: %s\n", transdb->ErrorMsg());
      }
#endif
		rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
	/* the transID is the original sale ID, so let's link up...
	sprintf(buff, "INSERT into transaction VALUES ('','','%d','%d','%d','22','0','','Item Returns','N','N','','','%d')",
   	trans_id, postID, userID, eaid);
	status=transdb->SimpleQuery(buff);
   */
	if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Failed To insert new transaction: %s\n", transdb->ErrorMsg());
		}
#endif
       status=transdb->SimpleQuery(sqlq_qry77);
      if(status!=0){
#ifdef _DEBUGGERY				
		   if(DebugLevel > 2){
            printf("Failed to unlock trans table for new insert: %s\n", transdb->ErrorMsg());
		   }
#endif
      }
		rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
   
   if(sql_result.init(transdb, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
     if(DebugLevel > 2){
	     printf("Failed to query 145: %s\n", transdb->ErrorMsg());
     }
#endif
      //db->SimpleQuery(sqlq_qry146);
      rmsg.hdr.Ex=BHPOS_REFUSE;
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      return 2;
   }
   sql_row=sql_result.GetRow();
   ct.SetID(atoi(sql_row->GetField(0)));
	
	//ct.SetID(sql_result.GetLastID(transdb));
   
	sprintf(buff, sqlq_qry74, ct.GetID(), ct.GetID());
	status=transdb->SimpleQuery(buff);
	if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Failed To update returns transaction: %s\n", transdb->ErrorMsg());
		}
#endif
   }
   status=transdb->SimpleQuery(sqlq_qry77);
   if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
         printf("Failed to unlock trans table for new insert: %s\n", transdb->ErrorMsg());
		}
#endif
   }

	if(PluCommit(&ct, transdb, c, &rmsg)){
	   return 1;
	}
	
	// pay item commit
	if(ct.PopPayment(pi)){
   	sprintf(buff, transq_qry2, ct.GetID(), pi.GetID(), pi.GetAmount(), userID, eaid);

      if(sql_result.init(transdb, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
         if(DebugLevel > 2){
	         printf("Failed to query 145: %s\n", transdb->ErrorMsg());
         }
#endif
         //db->SimpleQuery(sqlq_qry146);
         rmsg.hdr.Ex=BHPOS_REFUSE;
         rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
         return 2;
      }
      sql_row=sql_result.GetRow();
      pay_id=atoi(sql_row->GetField(0));
      //pay_id=sql_result.GetLastID(transdb);
	
	   sprintf(buff, transq_qry3, pay_id, ct.GetID());
   	status=transdb->SimpleQuery(buff);
      if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
			   printf("Failed To update return pay ID: %s\n", transdb->ErrorMsg());
      	}
#endif
      	rmsg.hdr.Ex=BHPOS_REFUSE;
      	rmsg.hdr.size=0;
      	if(OutQ) OutQ->Add(&rmsg);
      	return 1;
      }
   }
	
	rmsg.hdr.Ex=BHPOS_GRANT;
   rmsg.hdr.size=0;
	if(OutQ) OutQ->Add(&rmsg);

	return 0;
}

int GoodsInEnd(PLUServer *p, bhpos::BhSQL *, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   return 0;
}

int LaybySummary(PLUServer *p, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   bhpos::CTransItems  ct;
   
   BhPosMsg            rmsg;
   
	BhCustomerLookup   *cl=(BhCustomerLookup *) msg->data;
	BhLaybyItem        *li;
	BhLaybyList        *ll;
   bhpos::BhSQLResult  sql_result, sql_res2;
   bhpos::BhSQLRow    *sql_row; //, *sql_row2;
	int	              status, cust_id;
	//double		                 amount;
	char		          buff[256], guff[16], cuff[8];
	
	cust_id=cl->customerID;
	
#ifdef _DEBUGGERY				
	if(DebugLevel > 2){
		printf("Trans customer summary for cust ID: %d\n", cust_id);
	}
#endif	
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	rmsg.hdr.Ex=BHPOS_GRANT;
	
	ll=(BhLaybyList *) &rmsg.data;
	li=ll->item;
	
	rmsg.hdr.size=sizeof(BhLaybyList) - sizeof(BhLaybyItem);
	
	// let's try a query
	sprintf(buff, transq_qry13, cust_id);
	if(sql_result.init(transdb, buff, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
      if(DebugLevel > 2){
         printf("transq: customer summary query failed: %s\n", transdb->ErrorMsg());
      }
#endif         
      return 1;
   }
   if(sql_result.NumRows()==0){
      rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
   }
   else{
      ll->customerID=cust_id;
      ll->numtrans=sql_result.NumRows();
      
      while((sql_row=sql_result.GetRow())){
         li->transID=atoi(sql_row->GetField(0));
         
         sprintf(buff, transq_qry14, li->transID);
         if(sql_res2.init(transdb, buff, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
            if(DebugLevel > 2){
               printf("transq: custsumm items query failed: %s\n", transdb->ErrorMsg());
            }
#endif         
            return 1;
         }
         li->num_items=sql_res2.NumRows();
         sprintf(buff, transq_qry15, li->transID);
         if(sql_res2.init(transdb, buff, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
            if(DebugLevel > 2){
               printf("transq: custsumm pitems query failed: %s\n", transdb->ErrorMsg());
            }
#endif         
            return 1;
         }
         li->num_payitems==sql_res2.NumRows();
         strcpy(guff, sql_row->GetField(1));
         memcpy(cuff, guff, 4);
         cuff[4]=0;
         li->when.flags=8;
         li->when.year=atoi(cuff);
         memcpy(cuff, &guff[4], 2);
         cuff[2]=0;
      	li->when.month=atoi(cuff);
      	memcpy(cuff, &guff[6], 2);
         cuff[2]=0;
         li->when.date=atoi(cuff);
      	li->when.hour=0;
         li->when.minute=0;
         li->when.second=0;
         
         rmsg.hdr.size+=sizeof(BhLaybyItem);
      }
   }

	status=POST_MODE_LAYBYSUMMARY;
	c.SetParameter(BHCI_PARAM_MODE, &status);
	status=POST_STATUS_IDLE;
	c.SetParameter(BHCI_PARAM_STATUS, &status);

	if(OutQ)
		OutQ->Add(&rmsg);

	return 0;   
}

///////////////////////////////////////////////////////////////////////////////////
//
//   Commit things
//
//

int PluCommit(bhpos::CTransItems *ct, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *rmsg)
{
   bhpos::BhSQLResult sql_result;
	bhpos::CSaleItem	 si;
	int                status, userID, nameID;
	char				    buff[256], eaid[21];
	
#ifdef _DEBUGGERY				
   if(DebugLevel > 2){
   	printf("Inserting PLUs: %d\n", ct->NumItems());
   }
#endif
  
   c.GetParameter(BHCI_PARAM_USERID, &userID);
	memset(eaid, 0, sizeof(eaid));
	c.GetEnteredAt(eaid);
	nameID=ct->GetNameID();
   
   if(ct->NumItems()!=0){
      status=transdb->SimpleQuery(sqlq_qry70);
      if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
	         printf("Failed to lock transitems: %s\n", transdb->ErrorMsg());
      	}
#endif
         rmsg->hdr.Ex=BHPOS_REFUSE;
   	   rmsg->hdr.size=0;
         if(OutQ) OutQ->Add(rmsg);
   	      return 1;
      }
   
      while(ct->Pop(si)){
		   sprintf(buff, transq_qry16, ct->GetID(), userID, eaid, SQL_ITEM_ITEM, si.GetPLU().data(), si.GetQuantity());
		   status=transdb->SimpleQuery(buff);
			if(status!=0){
#ifdef _DEBUGGERY				
		   	if(DebugLevel > 2){
					printf("Failed To insert trans items: %s\n", transdb->ErrorMsg());
		   	}
#endif
            transdb->SimpleQuery(sqlq_qry146);
				rmsg->hdr.Ex=BHPOS_REFUSE;
		   	rmsg->hdr.size=0;
				if(OutQ) OutQ->Add(rmsg);
		   	return 1;
			}
#ifdef _DEBUGGERY				
			if(DebugLevel > 2){
   			printf("trans insert: %s\n", buff);
			}
#endif   		
	   }
      
      status=transdb->SimpleQuery(sqlq_qry77);
      if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
	         printf("Failed to lock transitems: %s\n", transdb->ErrorMsg());
      	}
#endif
         rmsg->hdr.Ex=BHPOS_REFUSE;
   	   rmsg->hdr.size=0;
         if(OutQ) OutQ->Add(rmsg);
            return 1;
      }
	}
   
   return 0;
}

int UPluCommit(bhpos::CTransItems *ct, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *rmsg)
{
   bhpos::BhSQLResult sql_result;
	bhpos::CSaleItem	 si;
	int                status, userID;
	char				    buff[256], eaid[21];

#ifdef _DEBUGGERY				
   if(DebugLevel > 2){
   	printf("Inserting unknown PLUs: %d\n", ct->NumUnknownItems());
   }
#endif

   c.GetParameter(BHCI_PARAM_USERID, &userID);
	memset(eaid, 0, sizeof(eaid));
	c.GetEnteredAt(eaid);
	
   if(ct->NumUnknownItems()!=0){
      printf("\nUPLU???????\n");
      return 1;
      status=transdb->SimpleQuery(sqlq_qry70);
      if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
	         printf("Failed to lock transitems: %s\n", transdb->ErrorMsg());
      	}
#endif
         rmsg->hdr.Ex=BHPOS_REFUSE;
   	   rmsg->hdr.size=0;
         if(OutQ) OutQ->Add(rmsg);
   	      return 1;
      }
      
		for(int i=0; i <= ct->NumUnknownItems(); i++){
			ct->PopUnknown(si);
			sprintf(buff, transq_qry16,ct->GetID(), userID, eaid, SQL_ITEM_UNKNOWN, si.GetPLU().data(), si.GetQuantity());
			TransPluAdd(transdb, &si);
			status=transdb->SimpleQuery(buff);
#ifdef _DEBUGGERY				
			if(DebugLevel > 2){
				printf("Insert unknown PLU item: %s:%s:%0.2f\n", si.GetPLU().data(), si.GetDispText().data(), si.GetPrice());
			}
#endif         
			if(status!=0){
#ifdef _DEBUGGERY				
				if(DebugLevel > 2){
					printf("Failed To insert unknown uplu items: %s\n", transdb->ErrorMsg());
				}
#endif
				rmsg->hdr.Ex=BHPOS_REFUSE;
				rmsg->hdr.size=0;
				if(OutQ) OutQ->Add(rmsg);
				return 1;
			}
		}
      
      status=transdb->SimpleQuery(sqlq_qry77);
      if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
	         printf("Failed to lock transitems: %s\n", transdb->ErrorMsg());
      	}
#endif
         rmsg->hdr.Ex=BHPOS_REFUSE;
   	   rmsg->hdr.size=0;
         if(OutQ) OutQ->Add(rmsg);
   	      return 1;
      }
	}
   return 0;
}

int PricePluCommit(bhpos::CTransItems *ct, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *rmsg)
{
   bhpos::BhSQLResult sql_result;
	bhpos::CSaleItem	 si;
	int                status, userID;
	char				    buff[256], eaid[21];

#ifdef _DEBUGGERY				
   if(DebugLevel > 2){
   	printf("Inserting price PLUs: %d\n", ct->NumPriceItems());
   }
#endif

   c.GetParameter(BHCI_PARAM_USERID, &userID);
	memset(eaid, 0, sizeof(eaid));
	c.GetEnteredAt(eaid);
	
   if(ct->NumPriceItems()!=0){
      status=transdb->SimpleQuery(sqlq_qry70);
      if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
	         printf("Failed to lock transitems: %s\n", transdb->ErrorMsg());
      	}
#endif
         rmsg->hdr.Ex=BHPOS_REFUSE;
   	   rmsg->hdr.size=0;
         if(OutQ) OutQ->Add(rmsg);
   	      return 1;
      }
		for(int i=0; i <= ct->NumPriceItems(); i++){
			ct->PopPrice(si);
			sprintf(buff, transq_qry16, ct->GetID(), userID, eaid, SQL_ITEM_PRICE, si.GetPLU().data(), si.GetQuantity());
			TransPluAdd(transdb, &si);
			status=transdb->SimpleQuery(buff);
			if(status!=0){
#ifdef _DEBUGGERY				
				if(DebugLevel > 2){
					printf("Failed To insert unknown trans items: %s\n", transdb->ErrorMsg());
				}
#endif
				rmsg->hdr.Ex=BHPOS_REFUSE;
				rmsg->hdr.size=0;
				if(OutQ) OutQ->Add(rmsg);
				return 1;
			}
		}
      status=transdb->SimpleQuery(sqlq_qry77);
      if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
	         printf("Failed to lock transitems: %s\n", transdb->ErrorMsg());
      	}
#endif
         rmsg->hdr.Ex=BHPOS_REFUSE;
   	   rmsg->hdr.size=0;
         if(OutQ) OutQ->Add(rmsg);
   	      return 1;
      }
	}
   return 0;
}

int RentalPluCommit(PLUServer *p, bhpos::CTransItems *ct, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *rmsg)
{
   //BhPosMsg           rmsg;
   bhpos::BhSQLResult sql_result;
   bhpos::BhSQLRow   *sql_row;
	bhpos::CSaleItem	 si;
	int                status, userID, transID, customerID, duration, retid;
	char				    buff[256], eaid[21];
	
	//memcpy(&rmsg, msg, sizeof(rmsg));
	
	status=transdb->SimpleQuery(sqlq_qry70);
   if(status!=0){
#ifdef _DEBUGGERY          
      if(DebugLevel > 2){
         printf("Failed to lock trans table for new insert: %s\n", transdb->ErrorMsg());
      }
#endif
      rmsg->hdr.Ex=BHPOS_REFUSE;
      rmsg->hdr.size=0;
      if(OutQ) OutQ->Add(rmsg);
      return 1;
   }
   
   p->GetOption(BH_PLUS_OPTION_ACCRENTAL, &retid);
   c.GetParameter(BHCI_PARAM_USERID, &userID);
	memset(eaid, 0, sizeof(eaid));
	c.GetEnteredAt(eaid);
   
   status=transdb->SimpleQuery(sqlq_qry70);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
	      printf("Failed to lock transitems: %s\n", transdb->ErrorMsg());
     	}
#endif
      rmsg->hdr.Ex=BHPOS_REFUSE;
      rmsg->hdr.size=0;
      if(OutQ) OutQ->Add(rmsg);
         return 1;
   }

   sprintf(buff, transq_qry19, retid, ct->GetID());
	status=transdb->SimpleQuery(buff);

   if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Failed To insert new rental transaction: %s\n", transdb->ErrorMsg());
		}
#endif
      transdb->SimpleQuery(sqlq_qry146);
      rmsg->hdr.Ex=BHPOS_REFUSE;
		rmsg->hdr.size=0;
		if(OutQ) OutQ->Add(rmsg);
		return 1;
   }
   
   status=transdb->SimpleQuery(sqlq_qry77);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
	      printf("Failed to unlock transitems: %s\n", transdb->ErrorMsg());
     	}
#endif
      rmsg->hdr.Ex=BHPOS_REFUSE;
      rmsg->hdr.size=0;
      if(OutQ) OutQ->Add(rmsg);
         return 1;
   }
	
   if(sql_result.init(transdb, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
      if(DebugLevel > 2){
         printf("Failed to query 145: %s\n", transdb->ErrorMsg());
      }
#endif
      rmsg->hdr.Ex=BHPOS_REFUSE;
      rmsg->hdr.size=0;
      if(OutQ) OutQ->Add(rmsg);
      return 2;
   }
   sql_row=sql_result.GetRow();
   transID=atoi(sql_row->GetField(0));
      
	//transID=sql_result.GetLastID(transdb);
   status=transdb->SimpleQuery(sqlq_qry70);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
	      printf("Failed to lock transitems: %s\n", transdb->ErrorMsg());
     	}
#endif
      rmsg->hdr.Ex=BHPOS_REFUSE;
      rmsg->hdr.size=0;
      if(OutQ) OutQ->Add(rmsg);
         return 1;
   }
	sprintf(buff, transq_qry20, transID, userID, eaid, transID);
	
	status=transdb->SimpleQuery(buff);
	if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Failed to update trans: %s\n", transdb->ErrorMsg());
		}
#endif
      transdb->SimpleQuery(sqlq_qry146);
		rmsg->hdr.Ex=BHPOS_REFUSE;
		rmsg->hdr.size=0;
		if(OutQ) OutQ->Add(rmsg);
		return 1;
	}
   
   status=transdb->SimpleQuery(sqlq_qry77);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
	      printf("Failed to lock transitems: %s\n", transdb->ErrorMsg());
     	}
#endif
      rmsg->hdr.Ex=BHPOS_REFUSE;
      rmsg->hdr.size=0;
      if(OutQ) OutQ->Add(rmsg);
         return 1;
   }
	
   c.GetParameter(BHCI_PARAM_USERID, &userID);
	memset(eaid, 0, sizeof(eaid));
	c.GetEnteredAt(eaid);
   
   if(ct->NumRentalItems() > 0){
      status=transdb->SimpleQuery(sqlq_qry70);
      if(status!=0){
#ifdef _DEBUGGERY				
	   	if(DebugLevel > 2){
            printf("Failed to unlock trans table for new insert: %s\n", transdb->ErrorMsg());
   		}
#endif
	   	rmsg->hdr.Ex=BHPOS_REFUSE;
		   rmsg->hdr.size=0;
   		if(OutQ) OutQ->Add(rmsg);
	   	return 1;
	   }
	
      for(int i=0; i <= ct->NumRentalItems(); i++){
		   if(ct->PopReturn(si)){
		      sprintf(buff, transq_qry16, transID, userID, eaid, SQL_ITEM_RENTAL, si.GetPLU().data(), si.GetQuantity());
   		   status=transdb->SimpleQuery(buff);
	   		if(status!=0){
#ifdef _DEBUGGERY				
		      	if(DebugLevel > 2){
			   		printf("Failed To insert trans return items: %s\n", transdb->ErrorMsg());
		   	   }
#endif
               transdb->SimpleQuery(sqlq_qry146);
      			rmsg->hdr.Ex=BHPOS_REFUSE;
		   	   rmsg->hdr.size=0;
   		   	if(OutQ) OutQ->Add(rmsg);
			         return 1;
   		   }
   	   }
   	   else{
#ifdef _DEBUGGERY				
			   if(DebugLevel > 2){
      		   printf("Failed To Pop trans return item: %s\n", transdb->ErrorMsg());
			   }
#endif
	      }
      }
      status=transdb->SimpleQuery(sqlq_qry77);
      if(status!=0){
#ifdef _DEBUGGERY				
	   	if(DebugLevel > 2){
            printf("Failed to unlock trans table for new insert: %s\n", transdb->ErrorMsg());
   		}
#endif
	   	rmsg->hdr.Ex=BHPOS_REFUSE;
		   rmsg->hdr.size=0;
   		if(OutQ) OutQ->Add(rmsg);
	   	return 1;
	   }
	}
	
	c.GetParameter(BHCI_PARAM_CUSTID, &customerID);
	c.GetParameter(BHCI_PARAM_DURATION, &duration);
	c.GetParameter(BHCI_PARAM_SITEID , &status);
   
   status=transdb->SimpleQuery(sqlq_qry70);
   if(status!=0){
#ifdef _DEBUGGERY				
	  	if(DebugLevel > 2){
         printf("Failed to lock rental table: %s\n", transdb->ErrorMsg());
   	}
#endif
	  	rmsg->hdr.Ex=BHPOS_REFUSE;
	   rmsg->hdr.size=0;
   	if(OutQ) OutQ->Add(rmsg);
	   	return 1;
	}
   
	sprintf(buff, transq_qry22, transID, customerID, status, duration);
	status=transdb->SimpleQuery(buff);
	if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Failed To insert rental: %s\n", transdb->ErrorMsg());
		}
#endif
      transdb->SimpleQuery(sqlq_qry146);
   	rmsg->hdr.Ex=BHPOS_REFUSE;
		rmsg->hdr.size=0;
   	if(OutQ) OutQ->Add(rmsg);
		return 1;
   }
   
   status=transdb->SimpleQuery(sqlq_qry77);
   if(status!=0){
#ifdef _DEBUGGERY				
	  	if(DebugLevel > 2){
         printf("Failed to unlock rental table: %s\n", transdb->ErrorMsg());
   	}
#endif
	  	rmsg->hdr.Ex=BHPOS_REFUSE;
	   rmsg->hdr.size=0;
   	if(OutQ) OutQ->Add(rmsg);
	   	return 1;
	}
	
	return 0;
}

int ReturnPluCommit(PLUServer *p, bhpos::CTransItems *ct, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   BhPosMsg           rmsg;
   bhpos::BhSQLResult sql_result;
   bhpos::BhSQLRow   *sql_row;
	bhpos::CSaleItem	 si;
   bool               bRetOnly=false;
	int                status, userID, transID=0, retid, numr=ct->NumReturnReasons();
	char				    buff[256], eaid[21];
	
	memcpy(&rmsg, msg, sizeof(rmsg));
   
   c.GetParameter(BHCI_PARAM_USERID, &userID);
	memset(eaid, 0, sizeof(eaid));
	c.GetEnteredAt(eaid);
   
   // check if only return items...
   if(ct->NumReturnItems()==ct->NumAllItems()){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
	      printf("Return Transaction All items return\n");
  	   }
#endif      
      bRetOnly=true;
   }

   // firstly, either insert linked transaction for returns in sale, or
   // close current transaction and mark as returns transaction.
   //
   // secondly, it number of return reasons > 1, then insert a return for
   // each trans item, and mark entry as trans item, else
   // interst single return entry and mark as transaction.
   if(!bRetOnly){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
	      printf("Return Transaction: linked\n");
  	   }
#endif
      status=transdb->SimpleQuery(sqlq_qry70);
   	if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
		      printf("Failed to lock trans: %s\n", transdb->ErrorMsg());
   		}
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
   		    return 1;
   	}
         
      sprintf(buff, transq_qry23, ct->GetID());
      status=transdb->SimpleQuery(buff);

	   if(status!=0){
#ifdef _DEBUGGERY				
	      if(DebugLevel > 2){
		      printf("Failed To insert new return transaction: %s\n", transdb->ErrorMsg());
   	   }
#endif
         transdb->SimpleQuery(sqlq_qry146);
         rmsg.hdr.Ex=BHPOS_REFUSE;
	     	rmsg.hdr.size=0;
	      if(OutQ) OutQ->Add(&rmsg);
     		return 1;
      }
      
      status=transdb->SimpleQuery(sqlq_qry77);
   	if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
		      printf("Failed to unlock tran: %s\n", transdb->ErrorMsg());
   		}
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
   		    return 1;
   	}
      
      
      if(sql_result.init(transdb, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
         if(DebugLevel > 2){
            printf("Failed to query 145: %s\n", transdb->ErrorMsg());
         }
#endif
         //db->SimpleQuery(sqlq_qry146);
         rmsg.hdr.Ex=BHPOS_REFUSE;
         rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
         return 2;
      }
      sql_row=sql_result.GetRow();
      transID=atoi(sql_row->GetField(0));
      
      //transID=sql_result.GetLastID(transdb);
      status=transdb->SimpleQuery(sqlq_qry70);
   	if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
		      printf("Failed to lock trans: %s\n", transdb->ErrorMsg());
   		}
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
   		    return 1;
   	}
      
      sprintf(buff, transq_qry24, transID, userID, eaid, transID);
	
  	   status=transdb->SimpleQuery(buff);
      if(status!=0){
#ifdef _DEBUGGERY				
	      if(DebugLevel > 2){
  		   	printf("Failed to update trans: %s\n", transdb->ErrorMsg());
  	   	}
#endif
         transdb->SimpleQuery(sqlq_qry146);
   	   rmsg.hdr.Ex=BHPOS_REFUSE;
  	   	rmsg.hdr.size=0;
   	   if(OutQ) OutQ->Add(&rmsg);
  		   return 1;
     	}
      
      status=transdb->SimpleQuery(sqlq_qry77);
   	if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
		      printf("Failed to unlock trans: %s\n", transdb->ErrorMsg());
   		}
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
   		    return 1;
   	}
   }
   else{
      status=0;
  	   c.GetParameter(BHCI_PARAM_TRANSOPEN, &status);
      status=transdb->SimpleQuery(sqlq_qry70);
   	if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
		      printf("Failed to lock trans: %s\n", transdb->ErrorMsg());
   		}
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
   		    return 1;
   	}
      sprintf(buff, transq_qry25, userID, eaid, ct->GetID());
      
      status=transdb->SimpleQuery(buff);
      if(status!=0){
#ifdef _DEBUGGERY				
     		if(DebugLevel > 2){
	      	printf("Failed to update trans: %s\n", transdb->ErrorMsg());
     		}
#endif
         transdb->SimpleQuery(sqlq_qry146);
	      rmsg.hdr.Ex=BHPOS_REFUSE;
     		rmsg.hdr.size=0;
	      if(OutQ) OutQ->Add(&rmsg);
     		return 1;
      }
      
      status=transdb->SimpleQuery(sqlq_qry77);
   	if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
		      printf("Failed to lock trans: %s\n", transdb->ErrorMsg());
   		}
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
   		    return 1;
   	}
   }
   
   // transaction items
   //ct->GetItem(-1, &si);
   if(numr < 2){  // returns transaction, 1 reason
      
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
	      printf("Return single return reason\n");
  	   }
#endif         
      if(!bRetOnly){ // linked trans
         if(ReturnReasonCommit(p, transdb, transID, ct->GetReturnReason(), 0)){
#ifdef _DEBUGGERY
       		if(DebugLevel > 2){
	         	printf("Failed to insert return reason\n");
     	   	}
#endif
         }
      }
      else{ // dedicated returns trans
         if(ReturnReasonCommit(p, transdb, ct->GetID(), ct->GetReturnReason(), 0)){
#ifdef _DEBUGGERY
       		if(DebugLevel > 2){
	         	printf("Failed to insert return reason\n");
     	   	}
#endif
         }
      }
      
      status=transdb->SimpleQuery(sqlq_qry70);
   	if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
		      printf("Failed to lock transitems: %s\n", transdb->ErrorMsg());
   		}
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
   		    return 1;
   	}
      
      while(ct->PopReturn(si)){
		   sprintf(buff, transq_qry16, ct->GetID(), userID, eaid, SQL_ITEM_RETURN, si.GetPLU().data(), si.GetQuantity());
		   status=transdb->SimpleQuery(buff);
			if(status!=0){
#ifdef _DEBUGGERY				
		   	if(DebugLevel > 2){
					printf("Failed To insert trans items: %s\n", transdb->ErrorMsg());
		   	}
#endif
            transdb->SimpleQuery(sqlq_qry146);
				rmsg.hdr.Ex=BHPOS_REFUSE;
		   	rmsg.hdr.size=0;
				if(OutQ) OutQ->Add(&rmsg);
		   	return 1;
			}
#ifdef _DEBUGGERY				
			if(DebugLevel > 2){
   			printf("return plu trans insert: %s\n", buff);
			}
#endif   		
	   }
      printf("1\n");
      status=transdb->SimpleQuery(sqlq_qry77);
   	if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
		      printf("Failed to unlock transitems: %s\n", transdb->ErrorMsg());
   		}
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
   		    return 1;
   	}
      printf("2\n");
      status=POST_MODE_TRANSEND;
     	c.SetParameter(BHCI_PARAM_MODE, &status);
     	status=POST_STATUS_IDLE;
      c.SetParameter(BHCI_PARAM_STATUS, &status);
      c.ClearTrans(transID);
   }
   else{    // return trans items with multiple reasons
      status=transdb->SimpleQuery(sqlq_qry70);
   	if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
		      printf("Failed to unlock transitems: %s\n", transdb->ErrorMsg());
   		}
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
   		    return 1;
   	}
     	while(ct->PopReturn(si)){
         if(!bRetOnly){
	         sprintf(buff, transq_qry16, transID, userID, eaid, SQL_ITEM_RETURN, si.GetPLU().data(), si.GetQuantity());
         }
         else{
            sprintf(buff, transq_qry16, ct->GetID(), userID, eaid, SQL_ITEM_RETURN, si.GetPLU().data(), si.GetQuantity());
         }
  		   status=transdb->SimpleQuery(buff);
   		if(status!=0){
#ifdef _DEBUGGERY				
	      	if(DebugLevel > 2){
		   		printf("Failed To insert trans return items: %s\n", transdb->ErrorMsg());
	   	   }
#endif
            transdb->SimpleQuery(sqlq_qry146);
     			rmsg.hdr.Ex=BHPOS_REFUSE;
   		   rmsg.hdr.size=0;
  	   		if(OutQ) OutQ->Add(&rmsg);
		      return 1;
     		}
     	}
      
      status=transdb->SimpleQuery(sqlq_qry77);
   	if(status!=0){
#ifdef _DEBUGGERY				
         if(DebugLevel > 2){
		      printf("Failed to unlock transitems: %s\n", transdb->ErrorMsg());
   		}
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
   		    return 1;
   	}
      /* should be placed in loop
      if(sql_result.init(transdb, transq_qry32, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
        if(DebugLevel > 2){
            printf("Failed to query transq_qry32: %s\n", transdb->ErrorMsg());
         }
#endif
         //db->SimpleQuery(sqlq_qry146);
         rmsg.hdr.Ex=BHPOS_REFUSE;
         rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
         return 2;
      }
      sql_row=sql_result.GetRow();
      retid=atoi(sql_row->GetField(0));
      
      //retid=sql_result.GetLastID(transdb);
      ReturnReasonCommit(p, transdb, retid, si.GetReason(), 0);
      */
   }
   
   if(bRetOnly) return 1;
      
	return 0;
}

int ReturnReasonCommit(PLUServer *p, bhpos::BhSQL *transdb, int id, int reason, int type)
{
   int  status;
   char buff[256];
   
   status=transdb->SimpleQuery(sqlq_qry70);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
	      printf("Failed to lock retreason: %s\n", transdb->ErrorMsg());
   	}
#endif
      return 1;
   }
   
   if(type==0){ //transaction
      sprintf(buff, transq_qry27, id, SQL_RETURN_TRANSACTION, SQL_RETURN_STOCK, reason);
   	status=transdb->SimpleQuery(buff);
      if(status!=0){
#ifdef _DEBUGGERY				
	  	   if(DebugLevel > 2){
	   	   printf("Failed to update returns: %s\n", transdb->ErrorMsg());
      	}
#endif
         transdb->SimpleQuery(sqlq_qry146);
	     	return 1;
	   }
   }
   else if(type==1){ //transactionitem
      sprintf(buff, transq_qry27, id, SQL_RETURN_TRANSITEM, SQL_RETURN_STOCK, reason);
   	status=transdb->SimpleQuery(buff);
      if(status!=0){
#ifdef _DEBUGGERY				
	  	   if(DebugLevel > 2){
	   	   printf("Failed to update returns: %s\n", transdb->ErrorMsg());
      	}
#endif
         transdb->SimpleQuery(sqlq_qry146);
	     	return 1;
	   }
   }
   
   status=transdb->SimpleQuery(sqlq_qry77);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
	      printf("Failed to unlock retreason: %s\n", transdb->ErrorMsg());
   	}
#endif
      return 1;
   }
   
   return 0;
}

int LaybyPluCommit(PLUServer *p, bhpos::CTransItems *ct, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   BhPosMsg           rmsg;
   bhpos::BhSQLResult sql_result;
   bhpos::BhSQLRow   *sql_row;
	bhpos::CSaleItem	 si;
	int                status, userID, transID, customerID, duration, retid;
	char				    buff[256], eaid[21];
	
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	status=transdb->SimpleQuery(sqlq_qry70);
   if(status!=0){
#ifdef _DEBUGGERY          
      if(DebugLevel > 2){
         printf("Failed to lock trans table for new insert: %s\n", transdb->ErrorMsg());
      }
#endif
      rmsg.hdr.Ex=BHPOS_REFUSE;
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      return 1;
   }
   
   p->GetOption(BH_PLUS_OPTION_ACCLAYBY, &retid);
   c.GetParameter(BHCI_PARAM_USERID, &userID);
	memset(eaid, 0, sizeof(eaid));
	c.GetEnteredAt(eaid);

   sprintf(buff, transq_qry28, retid, ct->GetID());
	status=transdb->SimpleQuery(buff);

	if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Failed To insert new layby transaction: %s\n", transdb->ErrorMsg());
		}
#endif
      transdb->SimpleQuery(sqlq_qry146);
      rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
   }
   
   status=transdb->SimpleQuery(sqlq_qry77);
   if(status!=0){
#ifdef _DEBUGGERY          
      if(DebugLevel > 2){
         printf("Failed to lock trans table for new insert: %s\n", transdb->ErrorMsg());
      }
#endif
      rmsg.hdr.Ex=BHPOS_REFUSE;
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      return 1;
   }
   
   if(sql_result.init(transdb, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
      if(DebugLevel > 2){
         printf("Failed to query 145: %s\n", transdb->ErrorMsg());
      }
#endif
      rmsg.hdr.Ex=BHPOS_REFUSE;
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      return 2;
   }
   sql_row=sql_result.GetRow();
   transID=atoi(sql_row->GetField(0));
   
	//transID=sql_result.GetLastID(transdb);
   status=transdb->SimpleQuery(sqlq_qry70);
   if(status!=0){
#ifdef _DEBUGGERY          
      if(DebugLevel > 2){
         printf("Failed to lock trans table for new insert: %s\n", transdb->ErrorMsg());
      }
#endif
      rmsg.hdr.Ex=BHPOS_REFUSE;
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      return 1;
   }
   
	sprintf(buff, transq_qry29, transID, userID, eaid, transID);
	
	status=transdb->SimpleQuery(buff);
	if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Failed to update trans: %s\n", transdb->ErrorMsg());
		}
#endif
      transdb->SimpleQuery(sqlq_qry146);
		rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
	
	status=transdb->SimpleQuery(sqlq_qry77);
   if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
         printf("Failed to unlock trans table for new insert: %s\n", transdb->ErrorMsg());
		}
#endif
		rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
   
   c.GetParameter(BHCI_PARAM_USERID, &userID);
	memset(eaid, 0, sizeof(eaid));
	c.GetEnteredAt(eaid);
	
   if(ct->NumLaybyItems() > 0){
      status=transdb->SimpleQuery(sqlq_qry70);
      if(status!=0){
#ifdef _DEBUGGERY				
	   	if(DebugLevel > 2){
            printf("Failed to unlock trans table for new insert: %s\n", transdb->ErrorMsg());
		   }
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
	   	rmsg.hdr.size=0;
		   if(OutQ) OutQ->Add(&rmsg);
   		return 1;
	   }
      for(int i=0; i <= ct->NumLaybyItems(); i++){
		   if(ct->PopReturn(si)){
		      sprintf(buff, transq_qry16, transID, userID, eaid, SQL_ITEM_LAYBY, si.GetPLU().data(), si.GetQuantity());
   		   status=transdb->SimpleQuery(buff);
	   		if(status!=0){
#ifdef _DEBUGGERY				
		      	if(DebugLevel > 2){
			   		printf("Failed To insert trans layby items: %s\n", transdb->ErrorMsg());
		   	   }
#endif
               transdb->SimpleQuery(sqlq_qry146);
      			rmsg.hdr.Ex=BHPOS_REFUSE;
	   		   rmsg.hdr.size=0;
   	   		if(OutQ) OutQ->Add(&rmsg);
			      return 1;
   		   }
   	   }
   	   else{
#ifdef _DEBUGGERY				
			   if(DebugLevel > 2){
      		   printf("Failed To Pop trans layby item: %s\n", transdb->ErrorMsg());
			   }
#endif
	      }
      }
      status=transdb->SimpleQuery(sqlq_qry77);
      if(status!=0){
#ifdef _DEBUGGERY				
	   	if(DebugLevel > 2){
            printf("Failed to unlock trans table for new insert: %s\n", transdb->ErrorMsg());
		   }
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
	   	rmsg.hdr.size=0;
		   if(OutQ) OutQ->Add(&rmsg);
   		return 1;
	   }
	}
	
	c.GetParameter(BHCI_PARAM_CUSTID, &customerID);
	c.GetParameter(BHCI_PARAM_DURATION, &duration);
	
	sprintf(buff, transq_qry17,customerID, transID, ct->GetNumber(bhpos::BHNUM_LAYBYTOTAL), ct->GetNumber(bhpos::BHNUM_LAYBYDEPOSIT));
	
   status=transdb->SimpleQuery(sqlq_qry70);
   if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
         printf("Failed to unlock trans table for new insert: %s\n", transdb->ErrorMsg());
		}
#endif
		rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
   
	status=transdb->SimpleQuery(buff);
	if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
			printf("Failed To insert layby: %s\n", transdb->ErrorMsg());
		}
#endif
      transdb->SimpleQuery(sqlq_qry146);
   	rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
   	if(OutQ) OutQ->Add(&rmsg);
			return 1;
   }
   
   status=transdb->SimpleQuery(sqlq_qry77);
   if(status!=0){
#ifdef _DEBUGGERY				
		if(DebugLevel > 2){
         printf("Failed to unlock trans table for new insert: %s\n", transdb->ErrorMsg());
		}
#endif
		rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
	
	return 0;
}

/////////////////////////////////////////////////////////////////////////
//
//	Retrieve functions
//
//
int FillTrans(int transID, PLUServer *p, bhpos::BhSQL *transdb, bhpos::CClientInfo& c, unsigned int flags, BhPosMsg *rmsg)
{
   bhpos::BhSQLResult sql_result;
   bhpos::BhSQLResult trans_result;
	bhpos::BhSQLRow   *sql_row, *trans_row;
	bhpos::CSaleItem   si;
	bhpos::CPayItem    pi;
	char               buff[128];
	
	BhTransRetrieveReply  *trr=(BhTransRetrieveReply *) rmsg->data; 
	
	trr->transID=transID;
	
#ifdef _DEBUGGERY				
   if(DebugLevel > 2){
      printf("TransQ: Trans Fill\n");
   }
#endif	
	
	if((flags & TRANSRETRIEVE_FLAGS_RETURNS) || (flags & TRANSRETRIEVE_FLAGS_LAYBYS) || (flags & TRANSRETRIEVE_FLAGS_RENTALS)){
      sprintf(buff, transq_qry18, (unsigned int) transID);
      sql_result.init(transdb, buff, true);
      if(sql_result.NumRows()!=0){   // we have multi-trans
      	while((sql_row=sql_result.GetRow())){
      	   if(flags & BHRET_TYPE_LAYBYS){
      	      if(atoi(sql_row->GetField(1))==21){   // has laybys
         	      sprintf(buff, transq_qry21, (unsigned int) atoi(sql_row->GetField(0)));
                  trans_result.init(transdb, buff, true);
                  trr->total_layby=sql_result.NumRows();
                  if(trans_result.NumRows()!=0){
                     while((trans_row=trans_result.GetRow())){
                        BhDateTime dt(BHTSFORMAT_MYSQL_TS14, trans_row->GetField(2));
                        si.SetTimestamp(dt); // EnteredAt
                        si.SetUser(atoi(trans_row->GetField(3)));     // EnteredBy
                        si.SetWhere(trans_row->GetField(4));          //EnteredStation 
                        //si.SetType(atoi(trans_row->GetField(5)));     // type
                        si.SetPLU(trans_row->GetField(6));
                        si.SetQuantity(atoi(trans_row->GetField(7)));
                        si.SetType(bhpos::BHSI_SITYPE_LAYBY);
                        TransqPluLookup(transdb, &si);
                        c.TransPLU(transID, si);
                     }
                  }
   	         }
   	      }
   	      if(flags & BHRET_TYPE_RETURNS){
      	      if(atoi(sql_row->GetField(1))==22){   // has returns
         	      sprintf(buff, transq_qry21, (unsigned int) atoi(sql_row->GetField(0)));
                  trans_result.init(transdb, buff, true);
                  trr->total_returns=sql_result.NumRows();
                  if(trans_result.NumRows()!=0){
                     while((trans_row=trans_result.GetRow())){
                        BhDateTime dt(BHTSFORMAT_MYSQL_TS14, trans_row->GetField(2));
                        si.SetTimestamp(dt); // EnteredAt
                        si.SetUser(atoi(trans_row->GetField(3))); // EnteredBy
                        si.SetWhere(trans_row->GetField(4));   //EnteredStation 
                        //si.SetType(atoi(trans_row->GetField(5)));     //type
                        si.SetPLU(trans_row->GetField(6));
                        si.SetQuantity(atoi(trans_row->GetField(7)));
                        si.SetType(bhpos::BHSI_SITYPE_RETURN);
                        TransqPluLookup(transdb, &si);
                        c.TransPLU(transID, si);
                     }
                  }
   	         }
   	      }
   	      if(flags & BHRET_TYPE_RENTALS){
         	   if(atoi(sql_row->GetField(1))==20){   // has rental
         	      sprintf(buff, transq_qry21, (unsigned int) atoi(sql_row->GetField(0)));
                  trans_result.init(transdb, buff, true);
                  trr->total_rental=sql_result.NumRows();
                  if(trans_result.NumRows()!=0){
                     while((trans_row=trans_result.GetRow())){
                        BhDateTime dt(BHTSFORMAT_MYSQL_TS14, trans_row->GetField(2));
                        si.SetTimestamp(dt); // EnteredAt
                        si.SetUser(atoi(trans_row->GetField(3))); // EnteredBy
                        si.SetWhere(trans_row->GetField(4));   //EnteredStation 
                        //si.SetType(atoi(trans_row->GetField(5)));     //type
                        si.SetPLU(trans_row->GetField(6));
                        si.SetQuantity(atoi(trans_row->GetField(7)));
                        si.SetType(bhpos::BHSI_SITYPE_RENTAL);
                        TransqPluLookup(transdb, &si);
                        c.TransPLU(transID, si);
                     }
                  }
   	         }
   	      }
   	   }
   	}
   }
   if(flags & TRANSRETRIEVE_FLAGS_SALE){
#ifdef _DEBUGGERY
      if(DebugLevel > 2){
         printf("Fill Trans sale items flag\n");
      }
#endif
		      
      // standard items
      sprintf(buff, transq_qry21, (unsigned int) transID);
      if(sql_result.init(transdb, buff, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
         if(DebugLevel > 2){
            printf("TransFill: query failed: %s\n", transdb->ErrorMsg());
         }
#endif
         return 1;
      }
      if(sql_result.NumRows()!=0){
         trr->total_items=sql_result.NumRows();
#ifdef _DEBUGGERY
         if(DebugLevel > 2){
            printf("Fill Trans num items: %d\n", trr->total_items);
		   }
#endif
         while((sql_row=sql_result.GetRow())){
            BhDateTime dt(BHTSFORMAT_MYSQL_TS14, sql_row->GetField(2));
            si.SetTimestamp(dt);                      // EnteredAt
            si.SetUser(atoi(sql_row->GetField(3)));   // EnteredBy
            si.SetWhere(sql_row->GetField(4));        //EnteredStation 
            //si.SetType(atoi(sql_row->GetField(5))); //type
            si.SetPLU(sql_row->GetField(6));
            si.SetQuantity(atoi(sql_row->GetField(7)));
            si.SetType(bhpos::BHSI_SITYPE_PLU);
            TransqPluLookup(transdb, &si);
            c.TransPLU(transID, si);
            trr->total_tax+=(si.GetTax() * si.GetPrice());
            trr->total_discount+=(si.GetDiscount() * si.GetPrice());
            trr->total+=si.GetPrice();
         }
      }
   }
   if(flags & TRANSRETRIEVE_FLAGS_PAY){
      sprintf(buff, transq_qry26, (unsigned int) transID);
		if(sql_result.init(transdb, buff, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
         if(DebugLevel > 2){
            printf("TransFill: pay query failed: %s\n", transdb->ErrorMsg());
         }
#endif
         return 1;			
		}
      if(sql_result.NumRows()!=0){
			trr->total_payment_types=sql_result.NumRows();
#ifdef _DEBUGGERY
	      if(DebugLevel > 2){
   	      printf("Fill Trans pay items: %d\n", trr->total_payment_types);
			}
#endif		
         
         while((sql_row=sql_result.GetRow())){
            si.SetType(atoi(sql_row->GetField(3)));
            pi.SetAmount(atof(sql_row->GetField(4)));
            BhDateTime dt(BHTSFORMAT_MYSQL_TS14, sql_row->GetField(5));
            pi.SetTimestamp(dt);
            pi.SetUser(atoi(sql_row->GetField(6)));
            pi.SetWhere(sql_row->GetField(7));
            c.TransPay(transID, pi);
         }
      }
   }
   return 0;
}

int PluRetrieve(bhpos::CTransItems *ct, BhPosMsg *msg)
{
	bhpos::CSaleItem	  si;
	BhTransPluRetrieve *pluret;
	BhRetPLUreply      *plur;
	BhPosMsg			     rmsg;
	int                 tmp_numplu=ct->NumItems();
	
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	pluret=(BhTransPluRetrieve *) &rmsg.data;
	plur=pluret->items;
	
	pluret->transID=ct->GetID();
	rmsg.hdr.Ex=BHPOS_GRANT;
	rmsg.hdr.size=sizeof(BhTransPluRetrieve) - sizeof(BhRetPLUreply);

#ifdef _DEBUGGERY
   if(DebugLevel > 2){
      printf("PluRet: num items: %d\n", tmp_numplu);
   }
#endif
	
	if(tmp_numplu > 0){
      while((float) (1024 / (sizeof(BhRetPLUreply) * tmp_numplu)) < 1.0) tmp_numplu--;
      pluret->num_items=tmp_numplu;
      while(tmp_numplu){
         ct->Pop(si);
         
         strcpy(plur->PLU, si.GetPLU().data());
         strcpy(plur->displayText, si.GetDispText().data());
         strcpy(plur->recuText, si.GetRecText().data());
         plur->tax=si.GetTax();
         plur->discounttype=0;
         plur->discount=si.GetDiscount();
         plur->deposittype=0;
         plur->deposit=si.GetDeposit();
         plur->quantity=si.GetQuantity();
         plur->site=si.GetSite();
         plur->dept=si.GetDept();
         plur->group=si.GetGroup();
         plur->station=si.GetStation();
         plur->user=si.GetUser();
         si.GetTimestamp(&plur->when);
         plur->flags=si.GetFlags();
         plur->type=si.GetType();

         plur++;
         tmp_numplu--;
         rmsg.hdr.size+=sizeof(BhPLUreply);
      }
   }
   if(OutQ) OutQ->Add(&rmsg);
	return 0;
}

int PayRetrieve(bhpos::CTransItems *ct, BhPosMsg *msg)
{
	bhpos::CPayItem         pi;
	BhPosMsg			       rmsg;
	BhTransRetPayment      *rpi;
	BhTransPaymentRetrieve *rp;
	int                     tmp_numpay=ct->NumPayment();
   
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	rp=(BhTransPaymentRetrieve *) &rmsg.data;
	rpi=rp->payment;
	
	rp->transID=ct->GetID();
	
	rmsg.hdr.Ex=BHPOS_GRANT;
	rmsg.hdr.size=sizeof(BhTransPaymentRetrieve) - sizeof(BhTransRetPayment);	
	
	if(tmp_numpay > 0){      
      while((float) (1024 / (sizeof(BhTransRetPayment) * tmp_numpay)) < 1.0) tmp_numpay--;
      rp->num_items=tmp_numpay;
      while(tmp_numpay){
         ct->PopPayment(pi);
         
         rpi->type=pi.GetID();
         rpi->flags=pi.GetFlags();
         rpi->site=pi.GetSite();
         rpi->dept=pi.GetDept();
         rpi->group=pi.GetGroup();
         rpi->station=pi.GetStation();
         rpi->user=pi.GetUser();
         pi.GetTimestamp(&rpi->when);
         rpi->amount=pi.GetAmount();
         
         rpi++;
         tmp_numpay--;
         rmsg.hdr.size+=sizeof(BhTransRetPayment);
      }
   }
   if(OutQ) OutQ->Add(&rmsg);
   
	return 0;
}

int TransqPluLookup(bhpos::BhSQL *transdb, bhpos::CSaleItem *si)
{
   bhpos::BhSQLResult  sql_result;
   char                sql_query[256];
   bhpos::BhSQLRow    *sql_row;
   
   sprintf(sql_query, "select * from plu_cache where plu='%s'", si->GetPLU().data());
   if(transdb)
      sql_result.init(transdb, sql_query, true);
   if(sql_result.NumRows() > 0){
      sql_row=sql_result.GetRow();
      if(!sql_row){
         return 1;
      }
      si->SetDispText(sql_row->GetField(2));
      si->SetRecText(sql_row->GetField(3));
      si->SetPrice(atof(sql_row->GetField(4)));
      si->SetDeposit(atof(sql_row->GetField(6)));
      si->SetTax(atof(sql_row->GetField(7)));
      si->SetFlags(atoi(sql_row->GetField(8)));
   }
   return 0;
}

int ChequeCommit(bhpos::BhSQL *db, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   int                 pid, status, postID=0, userID=0;
   BhTransPayCheque    *ptc=(BhTransPayCheque *) msg->data;
   char                buff[256], eaid[21];
   bhpos::BhSQLResult  sql_result;
   bhpos::BhSQLRow    *sql_row;
   BhPosMsg 	        rmsg;
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	rmsg.hdr.Ex=BHPOS_REFUSE;
   rmsg.hdr.size=0;
	
   c.GetParameter(BHCI_PARAM_STATIONID, &postID);
	c.GetParameter(BHCI_PARAM_USERID, &userID);
	memset(eaid, 0, sizeof(eaid));
	c.GetEnteredAt(eaid);
   
   
   sprintf(buff, "INSERT into paymentitem VALUES ('','0','%d','%d','%f',NOW(),'%d','%s','0')", 
      ptc->transID, bhpos::PTYPE_CHEQUE, ptc->amount, userID, eaid);
   
   status=db->SimpleQuery(buff);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
         printf("Failed To insert cheque payment: %s\n", db->ErrorMsg());
		}
#endif
   	if(OutQ) OutQ->Add(&rmsg);
		return 1;
   }
   
   if(sql_result.init(db, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
      if(DebugLevel > 2){
         printf("Failed to query 145: %s\n", db->ErrorMsg());
      }
#endif
      //db->SimpleQuery(sqlq_qry146);
      rmsg.hdr.Ex=BHPOS_REFUSE;
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      return 2;
   }
   sql_row=sql_result.GetRow();
   pid=atoi(sql_row->GetField(0));
   //pid=sql_result.GetLastID(db);
   
   sprintf(buff, "UPDATE paymentitem SET PayID='%d' WHERE transID='%d'", pid, ptc->transID);
   status=db->SimpleQuery(buff);
   if(status!=0){
#ifdef _DEBUGGERY				
   	if(DebugLevel > 2){
			printf("Failed To update trans pay ID: %s\n", db->ErrorMsg());
      }
#endif
      rmsg.hdr.Ex=BHPOS_REFUSE;
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      return 1;
   }
   
   sprintf(buff, "INSERT into paymentdetails VALUES ('','%d','','%s')", pid, ptc->data.raw);
   status=db->SimpleQuery(buff);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
         printf("Failed To insert cheque payment: %s\n", db->ErrorMsg());
		}
#endif
   	if(OutQ) OutQ->Add(&rmsg);
		return 2;
   }
   rmsg.hdr.Ex=BHPOS_GRANT;
   if(OutQ) OutQ->Add(&rmsg);
   
   return 0;
}

int ChargeCommit(bhpos::BhSQL *db, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   /*
   int                 pid, status, postID=0, userID=0;
   BhTransPayCheque    *ptc=(BhTransPayCheque *) msg->data;
   char                buff[256], eaid[21];
   bhpos::BhSQLResult  sql_result;
   bhpos::BhSQLRow    *sql_row;
   BhPosMsg 	        rmsg;
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	rmsg.hdr.Ex=BHPOS_REFUSE;
   rmsg.hdr.size=0;
	
   c.GetParameter(BHCI_PARAM_STATIONID, &postID);
	c.GetParameter(BHCI_PARAM_USERID, &userID);
	memset(eaid, 0, sizeof(eaid));
	c.GetEnteredAt(eaid);
   
   sprintf(buff, "INSERT into paymentitem VALUES ('','0','%d','%d','%f',NOW(),'%d','%s','0')", 
      ptc->transID, bhpos::PTYPE_CHEQUE, ptc->amount, userID, eaid);
   
   status=db->SimpleQuery(buff);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
         printf("Failed To insert cheque payment: %s\n", db->ErrorMsg());
		}
#endif
   	if(OutQ) OutQ->Add(&rmsg);
		return 1;
   }
   
   if(sql_result.init(db, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
      if(DebugLevel > 2){
         printf("Failed to query 145: %s\n", db->ErrorMsg());
      }
#endif
      rmsg.hdr.Ex=BHPOS_REFUSE;
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      return 2;
   }
   sql_row=sql_result.GetRow();
   pid=atoi(sql_row->GetField(0));
   
   sprintf(buff, "UPDATE paymentitem SET PayID='%d' WHERE transID='%d'", pid, ptc->transID);
   status=db->SimpleQuery(buff);
   if(status!=0){
#ifdef _DEBUGGERY				
   	if(DebugLevel > 2){
			printf("Failed To update trans pay ID: %s\n", db->ErrorMsg());
      }
#endif
      rmsg.hdr.Ex=BHPOS_REFUSE;
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      return 1;
   }
   
   sprintf(buff, "INSERT into chargeitem VALUES ('','%d','','%s')", pid, ptc->data.raw);
   status=db->SimpleQuery(buff);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
         printf("Failed To insert cheque payment: %s\n", db->ErrorMsg());
		}
#endif
   	if(OutQ) OutQ->Add(&rmsg);
		return 2;
   }
   rmsg.hdr.Ex=BHPOS_GRANT;
   if(OutQ) OutQ->Add(&rmsg);
   */
   return 0;
}

int CreditCommit(bhpos::BhSQL *db, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	return 0;
}

int CreditDone(bhpos::BhSQL *db, bhpos::CClientInfo& c, BhPosMsg *msg)
{/*
   int                 pid, status, postID=0, userID=0;
   BhCreditReply      *ptc=(BhTransPayCheque *) msg->data;
   char                buff[256], eaid[21];
   bhpos::BhSQLResult  sql_result;
   bhpos::BhSQLRow    *sql_row;
   BhPosMsg 	        rmsg;
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	rmsg.hdr.Ex=BHPOS_REFUSE;
   rmsg.hdr.size=0;
	
   c.GetParameter(BHCI_PARAM_STATIONID, &postID);
	c.GetParameter(BHCI_PARAM_USERID, &userID);
	memset(eaid, 0, sizeof(eaid));
	c.GetEnteredAt(eaid);
   
   
   sprintf(buff, "INSERT into paymentitem VALUES ('','0','%d','%d','%f',NOW(),'%d','%s','0')", 
      ptc->transID, bhpos::PTYPE_CREDIT, ptc->amount, userID, eaid);
   
   status=db->SimpleQuery(buff);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
         printf("Failed To insert credit payment: %s\n", db->ErrorMsg());
		}
#endif
   	if(OutQ) OutQ->Add(&rmsg);
		return 1;
   }
   
   pid=sql_result.GetLastID(db);
   
   sprintf(buff, "UPDATE paymentitem SET PayID='%d' WHERE transID='%d'", pid, ptc->transID);
   status=db->SimpleQuery(buff);
   if(status!=0){
#ifdef _DEBUGGERY				
   	if(DebugLevel > 2){
			printf("Failed To update trans pay ID: %s\n", db->ErrorMsg());
      }
#endif
      rmsg.hdr.Ex=BHPOS_REFUSE;
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      return 1;
   }
   
   sprintf(buff, "INSERT into paymentdetails VALUES ('','%d','','%s')", pid, ptc->data.raw);
   status=db->SimpleQuery(buff);
   if(status!=0){
#ifdef _DEBUGGERY				
      if(DebugLevel > 2){
         printf("Failed To insert credit payment: %s\n", db->ErrorMsg());
		}
#endif
   	if(OutQ) OutQ->Add(&rmsg);
		return 2;
   }
   rmsg.hdr.Ex=BHPOS_GRANT;
   if(OutQ) OutQ->Add(&rmsg);
*/   
   return 0;
}
