/*
udp_thr.cpp - udp threading gear 
Copyright (C)2001 Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Don't write to me at: kingsley@bananahead.com 

   Part of:
   BhPos, Copyright (C) 2001-2005 Al Kingsley
   BhPos comes with ABSOLUTELY NO WARRANTY.
*/
#include <pthread.h>
#include <cstdio>
#include <cstring>
#include <cerrno>
#include <unistd.h>

#include "pos.h"
#include "plusd.h"
#include "msgq.h"

#include "bhmsgq.cpp"

#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

extern CMessageQueue<BhPosMsg>   *InQ;


void *UDPThread(void *p)
{
   PLUServer   *plus=(PLUServer *) p;
   int          ret;
   int          sock;
   socklen_t    len=0;
   BhPosMsg     pmsg;
   socklen_t    opt=1;
   unsigned short port;
   char         host[33];

   memset(&pmsg, 0, sizeof(pmsg));

   sockaddr_in tsin, from;
   memset(&from, 0, sizeof(from));
   memset(&tsin, 0, sizeof(tsin));
   tsin.sin_family=AF_INET;
   
   plus->GetOption(BH_PLUS_OPTION_PORT, &port);
   plus->GetOption(BH_PLUS_OPTION_HOST, host);
   
   tsin.sin_addr.s_addr=inet_addr(host);
   tsin.sin_port=htons(port);
   
   sock=socket(AF_INET, SOCK_DGRAM, 0);
   if(sock < 0){
#ifdef _DEBUGGERY
      printf("no socket creation");
#endif   
      ret=THREAD_DEAD;
      plus->SetThread(UDP_THREADID, THREAD_PARAM_STATE, &ret);
      return NULL;
   }
   
   if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(int))<0){
#ifdef _DEBUGGERY
      if(DebugLevel>1)
         printf("no udp socket reuse");
#endif
   }
   if((ret=bind(sock, (sockaddr *) &tsin, sizeof(tsin))) < 0){
#ifdef _DEBUGGERY
      if(DebugLevel>1)
         printf("no host socket bind: %d\n", ret);
#endif
      ret=THREAD_DEAD;
      plus->SetThread(UDP_THREADID, THREAD_PARAM_STATE, &ret);
      return NULL;
   }
#ifdef _DEBUGGERY
   if(DebugLevel>2) printf("UDP Thread: Entering Worker Loop\n");
#endif
   
   ret=THREAD_RUNNING;
   plus->SetThread(UDP_THREADID, THREAD_PARAM_STATE, &ret);
   
   while(1){
      memset(&pmsg, 0, sizeof(pmsg));
      ret=recvfrom(sock, (void *) &pmsg.hdr, 1042, 0, (sockaddr *) &from, &len); 

      if(ret < 0){
#ifdef _DEBUGGERY
         if(DebugLevel>2)
            printf("rx error: %d: %s \naddr len:%d\n", errno, strerror(errno), len);
#endif
      }
      else{
         plus->AdjParameter(BH_PLUS_PARM_MSGCOUNT);
         plus->AdjParameter(BH_PLUS_PARM_TOTALMSG);
         memcpy(&pmsg.return_addr, &from, sizeof(from));
         if(strcmp(inet_ntoa((struct in_addr) pmsg.return_addr.sin_addr), "0.0.0.0")==0) continue;
         /*
         if(pmsg.hdr.type==BHPOS_TYPE_POST && pmsg.hdr.cmd==BHMSG_POST_TRANSACTION && pmsg.hdr.subcmd==BHMSG_POST_TRANSACTION_COMMIT_PLU){
            BhTransPluCommit *pc=(BhTransPluCommit *) pmsg.data;
            int tmp_numplu=pc->num_items;
            //BhTransItem *plu_c=(BhTransItem *) &pmsg.data[8];
            BhTransItem *plu_c=(BhTransItem *) pmsg.data + (sizeof(BhTransPluCommit) - sizeof(BhTransItem));
            
            printf("udp: PLU data addr:0x%X\n plu points to: 0x%X\n size=%d\n", &pc, &plu_c, pmsg.hdr.size);
            printf("udp: TransCommit commiting %d PLUs\n", tmp_numplu);
            printf("udp: PLU Commit\nPLU\t%s\nflags\t%d\nqty\t%d\n", plu_c->PLU, plu_c->flags, plu_c->quantity);
         
         }*/
         
#ifdef _DEBUGGERY
         if(DebugLevel>2)
         printf("\nUDP Thread: rx packet\nfrom:%s:%d\ntype:%02d\ncmd:%02d\nsubcmd:%02d\nmsgsize:%02d\npktsize:%02d\n", inet_ntoa((struct in_addr) pmsg.return_addr.sin_addr), ntohs(pmsg.return_addr.sin_port), pmsg.hdr.type, pmsg.hdr.cmd, pmsg.hdr.subcmd, pmsg.hdr.size,len);
#endif
         if(InQ!=NULL){
#ifdef _DEBUGGERY
         if(DebugLevel>2)
            printf("InQ has %d messages\n", InQ->NumInQueue());
#endif            
            InQ->Add(&pmsg);
            
         }
         else{
#ifdef _DEBUGGERY
            if(DebugLevel) printf("InQ NULL\n");
#endif
         }
         
         // check cancellation
         plus->GetThread(UDP_THREADID, THREAD_PARAM_DEATH, &ret);
         if(ret) break;
         
      }
   }
   close(sock);
   ret=THREAD_DEAD;
   plus->SetThread(UDP_THREADID, THREAD_PARAM_STATE, &ret); 
   return NULL;
}
