/*
sqlq.cpp 
Copyright (C)2001 Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Part of:
	BhPos, Copyright (C) 2001-2004 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY.
*/
#include <pthread.h>
#include <cstdio>

#include "pos.h"
#include "msgq.h"
#include "plusd.h"
#include "sql.h"
#include "plucache.h"

#include "bhmsgq.cpp"

CMessageQueue<BhPosMsg>        *SqlQ;

extern CMessageQueue<BhPosMsg> *InQ;
extern CMessageQueue<BhPosMsg> *OutQ;
extern CMessageQueue<BhPosMsg> *CtrlQ;
extern CMessageQueue<BhPosMsg> *ErrorQ;

#ifdef _CCVENGINEMCVE
extern CMessageQueue<BhPosMsg> *CcvQ;
#endif

static int SqlSysMessage(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlSysMsgConnect(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlSysMsgShutdown(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlSysMsgInterface(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlSysMsgLogin(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);

static int SqlServiceMessage(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache);
static int SqlserviceSite(bhpos::BhSQL *db, PLUServer *p, BhPosMsg *msg); 
static int SqlserviceDb(bhpos::BhSQL *db, PLUServer *p, BhPosMsg *msg, CPluCache *pluCache);
static int SqlserviceCustomer(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);

static int SqlPostService(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);

static int SqlPostMessage(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache);
static int SqlPlu(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache);
static int SqlPluInternalWeight(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache);
static int SqlPluInternalQuantity(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache);
static int SqlPluInternalPrice(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache);
static int SqlPluInternalInterface(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache);
static int SqlPluToken(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache);
static int SqlPluVoucher(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache);
static int SqlPluInStore(bhpos::BhSQL *db, char *, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache);
static double SqlPluDiscount(bhpos::BhSQL *db, double *price, int qty, int pid);
static double SqlPluPromotion(bhpos::BhSQL *db, double *price, int qty, int pid);
static int SqlPluLookup(bhpos::BhSQL *db, char *, int, int, BhPosMsg *msg, CPluCache *pluCache);
static int BarcodeGetNumber(char c);
static int BarcodeLunBun(char *barcode);
static int SqlPostNoSale(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlPostList(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlPostSignature(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlPostStocktake(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlPostBlindBalance(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlPostPayIn(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlPostPayOut(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlPostDelivery(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlPostRental(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlPostCustomer(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlPostEphemeral(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlPostAdjust(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlPostDiscrepancy(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlPostAttendance(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlMenuLoad(bhpos::BhSQL *db, PLUServer *plusd, BhPosMsg	*rmsg);
static int SqlTransaction(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlTransNew(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlTransReturnsGet(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlTransPayAuth(bhpos::BhSQL *db, PLUServer *p, BhPosMsg *msg);
static int SqlGoods(bhpos::BhSQL *, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);
static int SqlUserLookup(bhpos::BhSQL *db, int, void *, int *);
static int SqlUserNameLookup(bhpos::BhSQL *, int, char *, char *);
static int SqlUserModifiedWhoLookup(bhpos::BhSQL *, int, char *, char *);
static int SqlCreatedWhenLookup(bhpos::BhSQL *, int, BhTimestamp *);
static int SqlModifiedWhenLookup(bhpos::BhSQL *, int, BhTimestamp *);
static int SqlErrorMsgFill(bhpos::CClientInfo& c, BhErrorMsg *errmsg);
static int SqlDrawerQuery(bhpos::BhSQL *db, bhpos::CClientInfo& c, char *start, char *end, double *opening, double *closing);
static int SqlPayQuery(bhpos::BhSQL *db, char *start, char *end, int type, int postID, int *num, double *amount);

int ScalesMessage(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg);


#ifdef _DEBUGGERY
extern int DebugLevel;
#endif

static bool bCachedPlu;

void *SqlQMonitor(void *arg)
{
	PLUServer			 *plu=(PLUServer *) arg;
	BhPosMsg			     msg;
	int					  status;
	bhpos::CClientInfo  c;
	CPluCache			 *pluCache=NULL;
   bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
	
	bhpos::BhSQL	    *db;

	bool bPluandPlay=false;
   bool bServTargs=false;

	int  cachelimit;
	bool bPreCached;
	char pctable[32];

	char dbhost[32];
	char dbname[32];
	char dbuser[32];
	char dbpass[32];
	
	SqlQ=new CMessageQueue<BhPosMsg>();
	if(!SqlQ){
#ifdef _DEBUGGERY
		if(DebugLevel) 
			printf("No SqlQ Allocation\n");
#endif			
		return NULL;
	}
	
	if(SqlQ->GetError()){
#ifdef _DEBUGGERY
		if(DebugLevel) 
			printf("SqlQ instansiation error\n");
#endif		
	}
	
	status=::pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	if(status!=0){
#ifdef _DEBUGGERY
		if(DebugLevel)	 
			printf("sql thread: setcanceltype failed.\n");
#endif			
	}

	if(SqlQ->GetError()!=0){
#ifdef _DEBUGGERY
		if(DebugLevel)	 
			printf("SqlQ Msg Queue Error: %d\n", SqlQ->GetError());
#endif
		
		SqlQ=NULL;  
		status=THREAD_DEAD;
		plu->SetThread(SQLQ_THREADID, THREAD_PARAM_STATE, &status);		 
		return NULL;
	}  
	
	//plu->Lock();
	plu->GetOption(BH_PLUS_OPTION_CACHEPLU, &bCachedPlu);
	plu->GetOption(BH_PLUS_OPTION_CACHELIMIT, &cachelimit);
	plu->GetOption(BH_PLUS_OPTION_PRECACHED, &bPreCached);
	plu->GetOption(BH_PLUS_OPTION_PRECACHETABLE, pctable);

	plu->GetOption(BH_PLUS_OPTION_DBHOSTPRIME, dbhost);
	plu->GetOption(BH_PLUS_OPTION_DBDATABASE, dbname);
	plu->GetOption(BH_PLUS_OPTION_DBUSER, dbuser);
	plu->GetOption(BH_PLUS_OPTION_DBPASSWORD, dbpass);
	plu->GetOption(BH_PLUS_OPTION_PLUANDPLAY, &bPluandPlay);
   plu->GetOption(BH_PLUS_OPTION_SERVERTARGETS, &bServTargs);
	//plu->Release();

#ifdef _DEBUGGERY
	if(DebugLevel > 2){
		printf("Sql Thread : %s, %s, %s, %s %s - %d\n",dbhost, dbname, dbuser, dbpass, pctable, cachelimit);
	}
#endif
	
	db=new bhpos::BhSQL(dbhost, dbname, 0, dbuser, dbpass);
	
	if(!db){
#ifdef _DEBUGGERY
		if(DebugLevel)	 
			printf("PLUServer - Sql Thread : failed to create SQL object\n");
#endif
		delete SqlQ;
		SqlQ=NULL;  
		status=THREAD_DEAD;
		plu->SetThread(SQLQ_THREADID, THREAD_PARAM_STATE, &status);		 
		return NULL;
	}

	if((status=db->Connect())){
#ifdef _DEBUGGERY
		if(DebugLevel)	 
			printf("PLUServer - Sql Thread : failed to connect to SQL database - \n%s - %d\n", db->ErrorMsg(), status);
#endif			
		delete SqlQ;
		SqlQ=NULL;  
		status=THREAD_DEAD;
		plu->SetThread(SQLQ_THREADID, THREAD_PARAM_STATE, &status);
		return NULL;
	}
	
	if(bCachedPlu && cachelimit){
		pluCache=new CPluCache();
		if(!pluCache){
#ifdef _DEBUGGERY
			if(DebugLevel)			 
				printf("No SqlQ pluCache allocation\n");
#endif
			bCachedPlu=false;
		}
		else{
			if(pluCache->Reload(db, cachelimit, pctable)){
#ifdef _DEBUGGERY
			if(DebugLevel)			 
				printf("SqlQ pluCache failed to load\n");
#endif            
				bCachedPlu=false;
				delete pluCache;
				pluCache=NULL;				
			}
		}
	}
#ifdef _DEBUGGERY
	if(DebugLevel > 2){
		if(bCachedPlu) printf("PLUs are cached\n");
	}
#endif
   
   if(bServTargs){
      if(sql_result.init(db, "SELECT id,name,capacity FROM tabler", true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
			if(DebugLevel > 2){
				printf("Sqlq: bad target query: %s\n", db->ErrorMsg());
			}
#endif
		}
      else{
         while((sql_row=sql_result.GetRow())){
            bhpos::CServSideTarget sst(atoi(sql_row->GetField(0)),sql_row->GetField(1), atoi(sql_row->GetField(2)));
            plu->ssts.AddTarget(sst);
		   }
      }
   }
   
   plu->GetOption(BH_PLUS_OPTION_SERVERMENU, &bServTargs);
   if(bServTargs){
   }

	// inital lock out
	status=SqlQ->Grab();
	if(status != 0) {
#ifdef _DEBUGGERY
		if(DebugLevel)	 
			printf("SqlQ failed to Grab Mutex, %d (%s)\n", status, strerror(status));
#endif
		delete SqlQ;
		SqlQ=NULL;  
		status=THREAD_DEAD;
		plu->SetThread(SQLQ_THREADID, THREAD_PARAM_STATE, &status); 
		return NULL;
	}
	
	status=THREAD_RUNNING;
	plu->SetThread(SQLQ_THREADID, THREAD_PARAM_STATE, &status);
	
	while(1){
		status=SqlQ->WaitOnMsg();
		if(status != 0) {
#ifdef _DEBUGGERY
			if(DebugLevel)		 
				printf("SqlQ thread wait failed, %d (%s)\n", status, strerror(status));
#endif			
			SqlQ->Release();
			delete SqlQ;
			SqlQ=NULL;  
			status=THREAD_DEAD;
			plu->SetThread(SQLQ_THREADID, THREAD_PARAM_STATE, &status);
			return NULL;
		}
		if(!SqlQ->HasMsg()){
#ifdef _DEBUGGERY
			if(DebugLevel)		 
				printf("SqlQ thread no msg\n");
#endif		
			continue;
		}
		if(SqlQ->GetNext(&msg)){
#ifdef _DEBUGGERY
			if(DebugLevel)		 
				printf("SqlQ msgq pop error\n");
#endif			
			continue;
		}
		SqlQ->DeleteLast();
		SqlQ->Release();
#ifdef _DEBUGGERY
		if(DebugLevel > 2)	
		printf("SqlQ popped type:%02d\ncmd:%02d\nsubcmd:%02d\n", msg.hdr.type, msg.hdr.cmd, msg.hdr.subcmd);
#endif
#ifdef _DEBUGGERY
		if(DebugLevel > 2)
			printf("\nSqlQpacket from:%s:%d\n\n", inet_ntoa((struct in_addr) msg.return_addr.sin_addr), ntohs(msg.return_addr.sin_port));
#endif
		
		if(plu->ClientExists(c, &msg)){
#ifdef _DEBUGGERY		 
			printf("client exists\n");
#endif			
		}
		else{
			/*
			msg.hdr.Ex=BHPOS_REFUSE;
			if(OutQ)
				OutQ->Add(&msg);*/
#ifdef _DEBUGGERY
			printf("couldn't acquire client\n");
#endif				
			status=SqlQ->Grab();
			continue;
		}
		switch(msg.hdr.type){
		case BHPOS_TYPE_NULL:
			break;
		
		case BHPOS_TYPE_SYSTEM:
			SqlSysMessage(db, plu, c, &msg);
			break;

		case BHPOS_TYPE_POST:
			SqlPostMessage(db, plu, c, &msg, pluCache);
			break;
			
		case BHPOS_TYPE_SCALES:
			ScalesMessage(db, plu, c, &msg);
			break;
		
		case BHPOS_TYPE_SERVICE:
			SqlServiceMessage(db, plu, c, &msg, pluCache); 
			break;
			
		case BHPOS_TYPE_INFO:
		case BHPOS_TYPE_RPRINTER:
		case BHPOS_TYPE_ATTENDANCE:
		case BHPOS_TYPE_CUSTOM:
		case BHPOS_TYPE_ERROR:
			break;
      
      case BHPOS_TYPE_GOODS:
         SqlGoods(db, plu, c, &msg);
         break;

		default:
			break;
		}
      
      plu->UpdateClient(c);
		
		plu->GetThread(SQLQ_THREADID, THREAD_PARAM_DEATH, &status);
		if(status && !SqlQ->HasMsg()) break;
		
		status=SqlQ->Grab();
		if(status != 0) {
#ifdef _DEBUGGERY
			if(DebugLevel)		 
				printf("SqlQ failed to Grab Mutex, %d (%s)\n", status, strerror(status));
#endif				
			delete SqlQ;
			SqlQ=NULL;  
			status=THREAD_DEAD;
			plu->SetThread(SQLQ_THREADID, THREAD_PARAM_STATE, &status);
			return NULL;
		}
#ifdef _DEBUGGERY	 
		printf("SQL mutex grabbed");
#endif		
	}
	SqlQ->Release();
	delete SqlQ;
	SqlQ=NULL;  
	status=THREAD_DEAD;
	plu->SetThread(SQLQ_THREADID, THREAD_PARAM_STATE, &status);
	return NULL;
}


/*****************************************************************************************************/

int SqlSysMessage(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{  
	int ret=0;
	
#ifdef _DEBUGGERY
	if(DebugLevel > 2)			
		printf("SqlSysMessage\n");
#endif	

	switch(msg->hdr.cmd){
	case BHMSG_SYSTEM_CONNECTION:
#ifdef _DEBUGGERY
	if(DebugLevel > 2)			
		printf("SqlSysMessage: Connection\n");
#endif			
		ret=SqlSysMsgConnect(db, p, c, msg);
		break;
		
	case BHMSG_SYSTEM_LOGIN:
		ret=SqlSysMsgLogin(db, p, c, msg);
		break;
		
	case BHMSG_SYSTEM_SHUTDOWN:
		ret=SqlSysMsgShutdown(db, p, c, msg);
		break;
		
	case BHMSG_SYSTEM_INTERFACE:
		ret=SqlSysMsgInterface(db, p, c, msg);
		break;	
		
	default:
		break;
	}
	
	return ret;
}
int SqlSysMsgShutdown(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	BhPosMsg			  rmsg;
	int					 temp;
	char				  sql_query[128];
	bhpos::CStation	 station=c.GetStation();
	int					 postID=station.GetID();
	bhpos::BhSQLResult sql_result;

#ifdef _DEBUGGERY
	if(DebugLevel > 2)			
		printf("SqlQ: SqlSysMsgShutdown\n");
#endif
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	rmsg.hdr.Ex=BHPOS_GRANT;
	rmsg.hdr.size=0;
	
	switch(msg->hdr.subcmd){
	case BHMSG_SYSTEM_SHUTDOWN_SHUTDOWN:
      db->SimpleQuery(sqlq_qry70);
		sprintf(sql_query, sqlq_qry1, SQL_P_SHUTDOWN, postID);
		temp=db->SimpleQuery(sql_query);
		if(temp!=0){
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("sqlshutdown - Failed log shutdown message for post %d - %d - %s\n", postID, temp, db->ErrorMsg());
            
			}
#endif
		}
      db->SimpleQuery(sqlq_qry77);
		temp=POST_STATUS_DISCONNECTED;
		c.SetParameter(BHCI_PARAM_STATUS, &temp);
		if(OutQ) OutQ->Add(&rmsg);
		break;
	
	// permissions come later...  
	case BHMSG_SYSTEM_SHUTDOWN_REBOOT:
	case BHMSG_SYSTEM_SHUTDOWN_RESTART:
	case BHMSG_SYSTEM_SHUTDOWN_SLEEP:
		if(OutQ) OutQ->Add(&rmsg);
		break;
	}
	return 0;
}

int SqlSysMsgInterface(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   BhPosMsg	    rmsg;
   
   memcpy(&rmsg, msg, sizeof(rmsg));
   
	switch(msg->hdr.cmd){
	case BHMSG_SYSTEM_INTERFACE_LOGIN:
	case BHMSG_SYSTEM_INTERFACE_LOGOUT:
      break;
   case BHMSG_SYSTEM_INTERFACE_BARCODE:
   {
   	char struser[8];
	   char strif[4];
	   int  user=0;
	   int  iif=0;

   	strcpy(strif, "000");
      strcpy(struser, "0000000");
	   //memcpy(strif, &plu.PLU[2], 3);
      //memcpy(struser, &plu.PLU[6], 7);
   }
	   break;
	case BHMSG_SYSTEM_INTERFACE_MAGSWIPE:
	case BHMSG_SYSTEM_INTERFACE_SMARTCARD:
		break;
	}
	return 0;
}

int SqlSysMsgConnect(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	int					 status;
	BhPosMsg			    rmsg;
	BhPosConnect		 conn;
	int					 postid=0, groupid=0, deptid=0, siteid=0, ifid=0, locid=0, currid=0, crndtype=0;
	char					 sql_query[255];
	char					 temp[64];
	bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow	*sql_row;
	bhpos::CStation	 station;
	
	memcpy(&conn, msg->data, sizeof(BhPosConnect));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
   BhPosSConnectReply *sconn=(BhPosSConnectReply *) rmsg.data;
	BhPosConnectReply  *rconn=(BhPosConnectReply *) rmsg.data;
	
   p->GetOption(BH_PLUS_OPTION_CURRRNDTYPE, &crndtype);
   
	switch(msg->hdr.subcmd){
   case BHMSG_SYSTEM_SCONNECT:
      sprintf(temp, "%02X%02X%02X%02X%02X%02X", conn.mac[0], conn.mac[1], conn.mac[2], conn.mac[3], conn.mac[4], conn.mac[5]);	
		sprintf(sql_query, sqlq_qry2, temp);
		
#ifdef _DEBUGGERY
	if(DebugLevel > 2)
		printf("SqlConn: SConnection with MAC - %s\n", temp);
#endif		

		if((status=sql_result.init(db, sql_query, true))!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
			if(DebugLevel > 2){
				printf("Sqlsconnect: mac failed: %s - %d\n", db->ErrorMsg(), status);
			}
#endif			
			return 1;
		}
		if(sql_result.NumRows()==0){
			if(p->IsPlugAndPlay()){
				rmsg.hdr.Ex=BHPOS_GRANT;				
				sprintf(sql_query, sqlq_qry3, temp);			 
			}
			else{
#ifdef _DEBUGGERY
				if(DebugLevel > 2)
					printf("System Message: Connection Request refused - no db entry\n");
#endif				
				rmsg.hdr.Ex=BHPOS_REFUSE;
			}
		}
		else{
			rmsg.hdr.Ex=BHPOS_GRANT;
			sql_row=sql_result.GetRow();
			if(!sql_row){
#ifdef _DEBUGGERY
				if(DebugLevel > 2)
					printf("System Message: NULL Row\n");
            return 2;
#endif
			}
			postid=atoi(sql_row->GetField(0));	// the ID field
			groupid=atoi(sql_row->GetField(3)); // the group field
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("SqlQ: SConnect: Post:%d grp:%d\n",postid, groupid);
			}
#endif			
			sprintf(sql_query, sqlq_qry4, groupid);
         if(sql_result.init(db, sql_query, true)==DB_COMMAND_OK){
			   if(sql_result.NumRows()==0){
				   deptid=0;
			   }
			   else{
				   sql_row=sql_result.GetRow();
   				deptid=atoi(sql_row->GetField(0));
            }
         }
         sprintf(sql_query, sqlq_qry5, deptid);
         if(sql_result.init(db, sql_query, true)==DB_COMMAND_OK){
			   if(sql_result.NumRows()==0){
				   siteid=0;
			   }
			   else{
				   sql_row=sql_result.GetRow();
   				siteid=atoi(sql_row->GetField(0));
            }
         }
			station.SetID(postid);
			station.SetGroup(groupid, NULL);
			station.SetDept(deptid, NULL);
			station.SetSite(siteid, NULL);
			c.SetStation(station);
         
			rmsg.hdr.size=sizeof(BhPosSConnectReply);
			
#ifdef _DEBUGGERY
			if(DebugLevel > 2){
				char n[21];
				c.GetEnteredAt(n);
				printf("SqlQ: Connect: Entered At - %s\n", n);
			}
#endif
			status=BH_POST_CONNNECTED;
			c.SetParameter(BHCI_PARAM_STATUS, &status);
			c.SetParameter(BHCI_PARAM_MAC, &conn.mac);
			
			if(sconn->mode==0){
				status=POST_MODE_NOUSER;
				c.SetParameter(BHCI_PARAM_MODE, &status);
			}
         
         sconn->SiteID=siteid;
			sconn->DeparmentID=deptid;
			sconn->GroupID=groupid;
			sconn->PostID=postid;
         
			sconn->status=0;
			sconn->mode=0;
			sconn->logintypes=0;
			sconn->pktmax=1436;
			sconn->pktEncryption=0;
			sconn->pktCompress=0;
		}
		status=-1;

		if(OutQ)
			status=OutQ->Add(&rmsg);
		
      break;
   
	case BHMSG_SYSTEM_CONNECT:
		// if no EM_STORE is present, dump all old transactions
		if(c.NumTrans()!=0){
			c.ClearAllTrans();
		}
		
	/*
	  c.GetParameter(BHCI_PARAM_SESSID, &status);
		// see if reconnect... this needs state recovery in client...
		c.GetParameter(BHCI_PARAM_STATUS, &status);
		switch(status){
		case POST_STATUS_DISCONNECTED:
			rconn->status=0;
			rconn->mode=0;
			break;
		case POST_STATUS_CONNECTED:
		case POST_STATUS_LOGOUT:
		case POST_STATUS_LOGIN:
		case POST_STATUS_IDLE:
		case POST_STATUS_TRANSACTION:
		case POST_STATUS_EMSHUTDOWN:
			rmsg.hdr.subcmd=BHMSG_SYSTEM_RECONNECT;
			rconn->status=status;
			c.GetParameter(BHCI_PARAM_MODE, &status);
			rconn->status=status;
			break;
		}
		*/
			
		// support version check
		if(conn.postver.MajorVersion < 1 && conn.postver.MinorVersion <= 1 && conn.postver.BuildVersion < 1){
		}
			
		// need dbID, department, group from IP	
		sprintf(temp, "%02X%02X%02X%02X%02X%02X", conn.mac[0], conn.mac[1], conn.mac[2], conn.mac[3], conn.mac[4], conn.mac[5]);	
		sprintf(sql_query, sqlq_qry2, temp);
		
#ifdef _DEBUGGERY
	if(DebugLevel > 2)
		printf("SqlConn: Connection with MAC - %s\n", temp);
#endif		

		if((status=sql_result.init(db, sql_query, true))!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
			if(DebugLevel > 2){
				printf("SqlLogin: mac failed: %s - %d\n", db->ErrorMsg(), status);
			}
#endif			
			return 1;
		}
		if(sql_result.NumRows()==0){
			if(p->IsPlugAndPlay()){
				rmsg.hdr.Ex=BHPOS_GRANT;				
				sprintf(sql_query, sqlq_qry3, temp);			 
			}
			else{
#ifdef _DEBUGGERY
				if(DebugLevel > 2)
					printf("System Message: Connection Request refused - no db entry\n");
#endif				
				rmsg.hdr.Ex=BHPOS_REFUSE;
			}
		}
		else{
			rmsg.hdr.Ex=BHPOS_GRANT;
			sql_row=sql_result.GetRow();
			if(!sql_row){
#ifdef _DEBUGGERY
				if(DebugLevel > 2)
					printf("System Message: NULL Row\n");
            return 2;
#endif
			}
			postid=atoi(sql_row->GetField(0));	// the ID field
			groupid=atoi(sql_row->GetField(3)); // the group field
			ifid=atoi(sql_row->GetField(4));	 // the i/f field
			locid=atoi(sql_row->GetField(5));	  // the locale
			currid=atoi(sql_row->GetField(6));	// the currency
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("SqlQ: Connect: Post-%d: grp:%d locale:%d currency:%d\n",postid, groupid,locid, currid);
			}
#endif			
			sprintf(sql_query, sqlq_qry4, groupid);
         if(sql_result.init(db, sql_query, true)==DB_COMMAND_OK){
			   if(sql_result.NumRows()==0){
				   deptid=0;
			   }
			   else{
				   sql_row=sql_result.GetRow();
   				deptid=atoi(sql_row->GetField(0));
	   			if(locid==0 || currid==0){ // no redefined locale/currency in post - check groups
		   			locid=atoi(sql_row->GetField(1));	
			   		currid=atoi(sql_row->GetField(2));
#ifdef _DEBUGGERY			 
				   	if(DebugLevel > 2){
					   	printf("SqlQ: Connect: GroupQry:locale:%d currency:%d\n",locid, currid);
					   }
#endif
					   if(locid==0 || currid==0){
						   sprintf(sql_query, sqlq_qry5, deptid);
                     if(sql_result.init(db, sql_query, true)==DB_COMMAND_OK){
						      if(sql_result.NumRows()!=0){
                           sql_row=sql_result.GetRow();
							      siteid=atoi(sql_row->GetField(0));
   							   if(locid==0){  // see if the site has default locale					  
	   							   locid=atoi(sql_row->GetField(1));
		   				   	}
			   				   if(currid==0){
				   				   currid=atoi(sql_row->GetField(2));
					   		   }
#ifdef _DEBUGGERY			 
						   	   if(DebugLevel > 2){
							   	   printf("SqlQ: Connect: DeptQry:locale:%d currency:%d\n",locid, currid);
							      }
#endif										
						      }
   						   // look in sites
	   					   if(locid==0 || currid==0){
		   					   sprintf(sql_query, sqlq_qry6, siteid);
                           if(sql_result.init(db, sql_query, true)==DB_COMMAND_OK){
				   			      if(sql_result.NumRows()!=0){
                                 sql_row=sql_result.GetRow();
					   			      if(locid==0){  // see if the site has default locale					  
      				   					locid=atoi(sql_row->GetField(0));
		      			   			}
				      		   		if(currid==0){
						      	   		currid=atoi(sql_row->GetField(1));
   								      }
#ifdef _DEBUGGERY			 
         								if(DebugLevel > 2){
		         							printf("SqlQ: Connect: SiteQry:locale:%d currency:%d\n",locid, currid);
			   	      				}
#endif										
				   		      	}				  
                           }
					      	}
				      	}
				      }
			      }
            }
         }
			station.SetID(postid);
			station.SetGroup(groupid, NULL);
			station.SetDept(deptid, NULL);
			station.SetSite(siteid, NULL);
			c.SetParameter(BHCI_PARAM_CURRENCYID, &currid);
			c.SetParameter(BHCI_PARAM_LOCALEID, &locid);
				
			rmsg.hdr.size=(sizeof(BhPosConnectReply)-sizeof(BhPosMonetaryUnit));
			
			sprintf(sql_query, sqlq_qry7, currid);
			sql_result.init(db, sql_query, true);
			
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("SqlQ: Connect: Qry:MonetaryUnitID returned %d rows for ID %d\n",sql_result.NumRows(), currid);
			}
#endif			
			
			if(sql_result.NumRows()!=0){
				sql_row=sql_result.GetRow();
				strncpy(rconn->currency_majname, sql_row->GetField(1), 31);
				strncpy(rconn->currency_minname, sql_row->GetField(2), 30);
				rconn->roundtype=crndtype;
            
				int muid=atoi(sql_row->GetField(0));
				sql_result.free();
				sprintf(sql_query, sqlq_qry8, muid);
				sql_result.init(db, sql_query, true);
#ifdef _DEBUGGERY			 
				if(DebugLevel > 2){
					printf("SqlQ: Connect: Qry:MonetaryUnits id:%d returned %d rows\n",muid, sql_result.NumRows());
				}
#endif				
				if(sql_result.NumRows()!=0){
					double v; 
					int	 u;
					int	 a;
					bool	usedef=false;
					p->GetOption(BH_PLUS_OPTION_USEDEFCASHD, &usedef);
					bhpos::CMoney m;
					BhPosMonetaryUnit *mu=rconn->unit;
					while((sql_row=sql_result.GetRow())){
						u=atoi(sql_row->GetField(2));
						v=atof(sql_row->GetField(3));
						a=atoi(sql_row->GetField(4));
						if(usedef)
							m.Set(u, v, a);
						else
							m.Set(u, v, 0);
					   
						c.AddMonetaryUnit(m);
						mu->unit=u;
						mu->value=v;
						if(usedef)
							mu->amount=a;
						else 
							mu->amount=0;
							mu->operation=BHPOS_CASHOPERATION_SET;
						
#ifdef _DEBUGGERY
						if(DebugLevel > 2){
							printf("SqlQ: Connect: Adding Monetary Unit: %d - %f - %d\n",mu->unit, mu->value, mu->amount);
						}
#endif
                  if(strcmp(sql_row->GetField(5), "Y")==0){
                     rconn->roundunit=v;
                  }
						mu++;	 
						rmsg.hdr.size+=sizeof(BhPosMonetaryUnit);
					}
				}
			}
			
			c.SetStation(station);
#ifdef _DEBUGGERY
			if(DebugLevel > 2){
				char n[21];
				c.GetEnteredAt(n);
				printf("SqlQ: Connect: Entered At - %s\n", n);
			}
#endif
			status=BH_POST_CONNNECTED;
			c.SetParameter(BHCI_PARAM_STATUS, &status);
			c.SetParameter(BHCI_PARAM_MAC, &conn.mac);
			
			if(rconn->mode==0){
				status=POST_MODE_NOUSER;
				c.SetParameter(BHCI_PARAM_MODE, &status);
			}
			
			//p->UpdateClient(c);
			rconn->numtaxitems=0;
			rconn->numpriceitems=0; 
			
			sql_result.init(db, sqlq_qry9, true);
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("SqlQ: Connect: Qry:Num Tax On Menu: %d\n", sql_result.NumRows());
			}
#endif	
					  
			if(sql_result.NumRows()!=0){
				rconn->numtaxitems=sql_result.NumRows();
			}
			
			sql_result.init(db, sqlq_qry10, true);
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("SqlQ: Connect: Qry: Num Price On Menu: %d\n", sql_result.NumRows());
			}
#endif	
					  
			if(sql_result.NumRows()!=0){
				rconn->numpriceitems=sql_result.NumRows();
			}			
		
			rconn->DeparmentID=deptid;
			rconn->GroupID=groupid;
			rconn->PostID=postid;
			rconn->baseCurrency=currid;
			rconn->locale=locid;
			rconn->status=0;
			rconn->mode=0;
			rconn->logintypes=0;
			rconn->pktmax=1436;
			rconn->pktEncryption=0;
			rconn->pktCompress=0;
		}
		status=-1;

		if(OutQ)
			status=OutQ->Add(&rmsg);
		
		break;
	}
	
	// log connect / disconnect etc...
	if(rmsg.hdr.subcmd==BHMSG_SYSTEM_CONNECT || rmsg.hdr.subcmd==BHMSG_SYSTEM_SCONNECT){
		if(rmsg.hdr.Ex==BHPOS_GRANT){ 
			sprintf(sql_query, sqlq_qry1, postid, SQL_P_CONNECT);
		}
		else{ 
			sprintf(sql_query, sqlq_qry1, postid, SQL_P_CONNECTREFUSE);
		}
	}
	else if(rmsg.hdr.subcmd==BHMSG_SYSTEM_DISCONNECT){
		if(rmsg.hdr.Ex==BHPOS_GRANT){
			c.Reset();
			sprintf(sql_query, sqlq_qry1, postid, SQL_P_DISCONNECT);
		}
		else{
			sprintf(sql_query, sqlq_qry1, postid, SQL_P_DISCONNECTREFUSE);
		}
	}
	db->SimpleQuery(sqlq_qry70);
	status=db->SimpleQuery(sql_query);
   db->SimpleQuery(sqlq_qry77);
	if(status!=0){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("Failed log connect message for post %d\n", postid);
		}
#endif
	}
	sql_result.free();
	return 0;
}

char * SqlSysFormatLoginSQL(int loginmethod, int type, char *ident, char *data)
{
	static char buf[2048];

	if(type==SQL_LOGINMETHOD_DATA && ident != 0) {
		sprintf(buf, sqlq_qry12, loginmethod, ident);
		return buf;
	}
	else if(type==SQL_LOGINMETHOD_PTR && data != 0) {
		sprintf(buf, sqlq_qry13, loginmethod, data);
		return buf;
	}
	else if(type==SQL_LOGINMETHOD_PAIR && data != 0 && ident != 0) {
		sprintf(buf, sqlq_qry143, loginmethod, ident, data);
		return buf;
	}
	return 0;
}

int SqlSysMsgLogin(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	BhPosMsg					   rmsg;
	BhLogout					   bhlout;
	BhLogonRequestEx		  *lreq=(BhLogonRequestEx *) msg->data;
	BhLogonRequest			  *l=(BhLogonRequest *) msg->data;
	BhLogonRequestBaroode  *lb=(BhLogonRequestBaroode *) msg->data;
	BhLogonRequestMagswipe *lm=(BhLogonRequestMagswipe *) msg->data;
	BhLogonReply				lr;
	int							InterfaceID=0, empID=0, LoginID, AuthID, PostID, LoginMethodID;
	char							sql_query[1024];
	bhpos::BhSQLResult		sql_result;
	bhpos::BhSQLRow		  *sql_row;
	char						   temp[16];

#ifdef _DEBUGGERY
		if(DebugLevel > 2)			
			printf("SqlQ: SqlSysMsgLogin\n");
#endif
	
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	if(msg->hdr.subcmd < 128){
		rmsg.hdr.flags=0;
		rmsg.hdr.size=0;
		rmsg.hdr.Ex=BHPOS_REFUSE;
		memset(&lr, 0, sizeof(BhLogonReply));
			 
		// New user query, uses new auth system.
		// First use the subcmd as an index into the loginmethods table
		// to first find if this is a valid loginmethod, and second
		// to determine the next action to take.
      if(msg->hdr.subcmd==BHMSG_SYSTEM_LOGIN_LOGINEXT){
         sprintf(sql_query, sqlq_qry14, lreq->type);
      }
      else{
		   sprintf(sql_query, sqlq_qry14, msg->hdr.subcmd);
      }
		if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
			if(DebugLevel > 2){
				printf("SqlLogin: loginmethod query1 failed: %s\n", db->ErrorMsg());
			}
#endif
			if(OutQ) OutQ->Add(&rmsg);
			return 1;
		}
		if(sql_result.NumRows() == 1){  // Should be either one row or none.
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("SqlLogin: found loginmethod : %s\n", sql_query);
			}
#endif			
			// Depending upon the data type we use either the full Data
			// field, or the Ident field or both to get a valid logintypes record,
			// or error on the pair for non-valid types
			// Note: need to do something about magswipe only using track 1
			//		 we can either cast to full mag data and let the db settle, 
			//		 or go to the pub and mentioning "maybe later".
			sql_row = sql_result.GetRow();
			LoginMethodID = atoi(sql_row->GetField(0));
			char *query_tmp = 0;
			int ltype=atoi(sql_row->GetField(1));
			char *method=0;
			switch (msg->hdr.subcmd) {
				case BHMSG_SYSTEM_LOGIN_LOGINEXT:
               switch(lreq->type){
               case BHPOS_LOGIN_TYPES_KEYBOARD:
                  method = "LOGIN";
                  query_tmp = SqlSysFormatLoginSQL(LoginMethodID, ltype, lreq->data.login.login, lreq->data.login.password);
					   break;
               case BHPOS_LOGIN_TYPES_PASSIVE:
                  method = "PASSIVE";
                  query_tmp = SqlSysFormatLoginSQL(LoginMethodID, ltype, lreq->data.raw, 0);
					   break;
               case BHPOS_LOGIN_TYPES_BARCODE:
                  method = "BARCODE";
                  query_tmp = SqlSysFormatLoginSQL(LoginMethodID, ltype, lreq->data.raw, 0);
                  break;
               case BHPOS_LOGIN_TYPES_MAGSWIPE:
                  method = "MAGSWIPE";
                  //sprintf(query_tmp, sqlq_qry666, loginmethod, lreq->data.magswipe.track1.pan, lreq->data.magswipe.track1.name);
                  query_tmp = SqlSysFormatLoginSQL(LoginMethodID, ltype, lreq->data.magswipe.track1.track1_details.pan, 0);
					   break;
               case BHPOS_LOGIN_TYPES_SMARTCARD:
               case BHPOS_LOGIN_TYPES_FINGERPRINT:
               case BHPOS_LOGIN_TYPES_RETINA:
               case BHPOS_LOGIN_TYPES_VOICE:
                  break;
               }
               break;
               
            case BHMSG_SYSTEM_LOGIN_LOGIN:
					method = "LOGIN";
               query_tmp = SqlSysFormatLoginSQL(LoginMethodID, ltype, l->name, l->password);
					break;
				case BHMSG_SYSTEM_LOGIN_LIPASSIVE:
					method = "PASSIVE";
               query_tmp = SqlSysFormatLoginSQL(LoginMethodID, ltype, l->name, 0);
					break;
				case BHMSG_SYSTEM_LOGIN_BARCODE:
					method = "BARCODE";
               query_tmp = SqlSysFormatLoginSQL(LoginMethodID, ltype, lb->barcode, 0);
					break;
				case BHMSG_SYSTEM_LOGIN_MAGSWIPE:
					method = "MAGSWIPE";
               query_tmp = SqlSysFormatLoginSQL(LoginMethodID, ltype, lm->track1, 0);
					break;
				default:
					method = "UNKNOWN";
					break;
			}
			if(query_tmp)
				strcpy(sql_query, query_tmp);
			else{
#ifdef _DEBUGGERY		 
				if(DebugLevel > 2) {
					printf("SqlLogin: Invalid %s loginmethod search type %d\n", method, ltype);
			   }
#endif
   		   if(OutQ)
			      OutQ->Add(&rmsg);
   		   return 2;
   		}

			if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK || sql_result.NumRows() != 1){
#ifdef _DEBUGGERY
				if(DebugLevel > 2){
					printf("SqlLogin: loginmethod query2 failed: %s - %d - %s\n", db->ErrorMsg(), sql_result.NumRows(), sql_query);
				}
#endif
				if (OutQ)
					OutQ->Add(&rmsg);
				return 3;
			}
			sql_row = sql_result.GetRow();
			LoginID = atoi(sql_row->GetField(1));

			// This gives us a user id that we then use to lookup the users table.
			sprintf(sql_query, sqlq_qry15, LoginID);
			if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
				if(DebugLevel > 2){
					printf("SqlLogin: user query failed: %s\n", db->ErrorMsg());
				}
#endif
				if(OutQ) OutQ->Add(&rmsg);
				return 4;
			}
			if(sql_result.NumRows() != 1){
				rmsg.hdr.Ex=BHPOS_REFUSE;
#ifdef _DEBUGGERY		 
				if(DebugLevel > 2){
					printf("SqlLogin: no user with login:%s and password:%s\n", l->name, l->password);
				}
#endif
				goto loopout;
			}
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("SqlLogin: found user\n");
			}
#endif			
			sql_row=sql_result.GetRow();
			lr.userID=atoi(sql_row->GetField(0));
			c.SetParameter(BHCI_PARAM_USERID, & lr.userID);
			empID=atoi(sql_row->GetField(6));
			// LogintypeID=atoi(sql_row->GetField(5));
				
			// From users we get the employee id and the full details for the user.
         sprintf(sql_query, sqlq_qry160, empID);
			if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
				if(OutQ) OutQ->Add(&rmsg);
				return 5;
			}
					
			if(sql_result.NumRows() == 1){
#ifdef _DEBUGGERY			 
				if(DebugLevel > 2){
					printf("SqlLogin: found employee record for: %d\n", empID);
				}
#endif
            sql_row=sql_result.GetRow();
            sprintf(sql_query, sqlq_qry161, atoi(sql_row->GetField(0)));
   			if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
	   			if(OutQ) OutQ->Add(&rmsg);
		   		return 6;
			   }
				sql_row=sql_result.GetRow();
				strcpy(temp, sql_row->GetField(0));
				strcat(temp, " ");
				strcat(temp, sql_row->GetField(1));
				if(strlen(temp) > 31){
					temp[32]=0;
				}
				strcpy(lr.userdispname, temp);
			}

			// Also from users we get an index into the userauth table to see what
			// interfaces they are allowed.  We need to check this against the
			// posts interface table to see where we are at.
         if(msg->hdr.subcmd==BHMSG_SYSTEM_LOGIN_LOGINEXT){
            sprintf(sql_query, sqlq_qry17, LoginID, lreq->type, l->postID);
         }
         else{
			   sprintf(sql_query, sqlq_qry17, LoginID, msg->hdr.subcmd, l->postID);
         }
			// This will give us either no common interface, or an authority record
			// for each common interface.  For the common interfaces, we check that
			// the perm_login flag is set to Y before we allow a login.
#ifdef _DEBUGGERY
			if(DebugLevel > 3)
				printf("SqlLogin: %s\n", sql_query);
#endif
			if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
				if(OutQ) OutQ->Add(&rmsg);
				return 7;
			}
			if(sql_result.NumRows() == 0){
#ifdef _DEBUGGERY		 
				if(DebugLevel > 2){
					printf("SqlLogin: No common interfaces\n");
				}
#endif
				goto loopout;
			}
			bool login_ok = false;
			while((sql_row = sql_result.GetRow())){
				InterfaceID = atoi(sql_row->GetField(0));
				AuthID = atoi(sql_row->GetField(1));
#ifdef _DEBUGGERY
				if(DebugLevel > 3)
					printf("SqlLogin: IF=%d Auth=%d PermLogin=%s\n", InterfaceID, AuthID, sql_row->GetField(4));
#endif
				if(strncmp(sql_row->GetField(4), "Y", 1) == 0){
					login_ok = true;
					break;
				}
			}
			if(!login_ok){
#ifdef _DEBUGGERY		 
				if(DebugLevel > 2){
					printf("SqlLogin: No Authority to login\n");
				}
#endif
				goto loopout;
			}
         rmsg.hdr.Ex=BHPOS_GRANT;
         rmsg.hdr.size=sizeof(BhLogonReply);
			lr.interfaceID=InterfaceID;
			memcpy(rmsg.data, &lr, sizeof(BhLogonReply));
         
#ifdef _DEBUGGERY
			if(DebugLevel > 2){
				printf("granting login\n");
			}
#endif									
			goto loopout;	  
		}
		else{
			rmsg.hdr.Ex=BHPOS_REFUSE;
#ifdef _DEBUGGERY		 
			if(DebugLevel > 2){
				printf("SqlLogin: no loginmethod %d\n", msg->hdr.subcmd);
			}
#endif
			if (OutQ)
				OutQ->Add(&rmsg);
			return 8;
		}	  
		return 0;
	}
	
	else switch(msg->hdr.subcmd){
	case BHMSG_SYSTEM_LOGIN_LOPASSIVE:
	case BHMSG_SYSTEM_LOGIN_LOGOUT:
	{
		memcpy(&bhlout, msg->data, sizeof(BhLogout));

		rmsg.hdr.type=BHPOS_TYPE_SYSTEM;
		rmsg.hdr.cmd=BHMSG_SYSTEM_LOGIN;
		rmsg.hdr.subcmd=BHMSG_SYSTEM_LOGIN_LOGOUT;
		
		rmsg.hdr.Ex=BHPOS_GRANT;
		rmsg.hdr.size=0;
      
      db->SimpleQuery(sqlq_qry70);
		sprintf(sql_query, sqlq_qry18, bhlout.postID, SQL_P_USERLOGOUT, bhlout.userID);
		sql_result.init(db, sql_query, true);
      db->SimpleQuery(sqlq_qry77);
		PostID=POST_STATUS_LOGOUT;
		c.SetParameter(BHCI_PARAM_STATUS, &PostID);
		c.SetParameter(BHCI_PARAM_LOGOUT, & bhlout.userID);
		//p->UpdateClient(c);
		if(OutQ) OutQ->Add(&rmsg);
	}
	}
	return 0;
	
loopout:
	if(msg->hdr.subcmd==BHMSG_SYSTEM_LOGIN_LOGIN || msg->hdr.subcmd==BHMSG_SYSTEM_LOGIN_LIPASSIVE ||
			msg->hdr.subcmd==BHMSG_SYSTEM_LOGIN_BARCODE || msg->hdr.subcmd==BHMSG_SYSTEM_LOGIN_MAGSWIPE ||
			msg->hdr.subcmd==BHMSG_SYSTEM_LOGIN_SMARTCARD || msg->hdr.subcmd==BHMSG_SYSTEM_LOGIN_LOGINEXT){
		if(rmsg.hdr.Ex==BHPOS_GRANT){
			if(c.UserExists(empID)){
				c.UserStartClock(empID, true);
			}
			else{
            db->SimpleQuery(sqlq_qry70);
				c.AddUser(empID);
				c.UserStartClock(empID, true);
				sprintf(sql_query, sqlq_qry19, empID, SQL_U_LOGIN);
				db->SimpleQuery(sql_query);
            db->SimpleQuery(sqlq_qry77);
            if(sql_result.init(db, sqlq_qry20, true)!=DB_COMMAND_OK){
	         }
            if(sql_result.NumRows()==0){
            }
            char start[32], nowt[32];
	         sql_row=sql_result.GetRow();
            if(sql_row==NULL){
            }
		      strcpy(nowt, sql_row->GetField(0));
		      strcpy(start, sql_row->GetField(0));
		      strcpy(&start[11], "00:01:01");
            db->SimpleQuery(sqlq_qry70);
            sprintf(sql_query, sqlq_qry21, start, nowt, SQL_U_CLOCKIN, empID);
	         if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
            }
            if(sql_result.NumRows() == 0) {
               sprintf(sql_query, sqlq_qry19, empID, SQL_U_CLOCKIN);
               sql_result.init(db, sql_query, true);
            }
			}
			sprintf(sql_query, sqlq_qry18, l->postID, SQL_P_USERLOGIN, lr.userID);
			PostID=POST_STATUS_LOGIN;
			c.SetParameter(BHCI_PARAM_STATUS, &PostID);
			c.SetParameter(BHCI_PARAM_LOGIN, &lr.userID);
			//p->UpdateClient(c);
		}
		else{
			sprintf(sql_query, sqlq_qry18, l->postID, SQL_P_USERLOGINREFUSE, lr.userID);
		}
		sql_result.init(db, sql_query, true);
      db->SimpleQuery(sqlq_qry77);
	}
	else if(msg->hdr.subcmd==BHMSG_SYSTEM_LOGIN_LOGOUT){
      db->SimpleQuery(sqlq_qry70);
		if(rmsg.hdr.Ex==BHPOS_GRANT){
			if(c.UserExists(empID)){
				c.UserStartClock(empID, false);
            sprintf(sql_query, sqlq_qry19, empID, SQL_U_LOGOUT);
				sql_result.init(db, sql_query, true);
			}
			sprintf(sql_query, sqlq_qry18, l->postID, SQL_P_USERLOGOUT, lr.userID);
			PostID=POST_STATUS_LOGIN;
			c.SetParameter(BHCI_PARAM_STATUS, &PostID);
			c.SetParameter(BHCI_PARAM_LOGIN, &lr.userID);
			//p->UpdateClient(c);
		}
		else{
			sprintf(sql_query, sqlq_qry18, l->postID, SQL_P_USERLOGOUTREFUSE, lr.userID);
		}
		sql_result.init(db, sql_query, true);
      db->SimpleQuery(sqlq_qry77);
	}
	
	//p->UpdateClient(c);
	if(OutQ)
		OutQ->Add(&rmsg);

	return 0;
}

/*****************************************************************************/

int SqlServiceMessage(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache)
{  
#ifdef _DEBUGGERY
	if(DebugLevel > 2)			
		printf("SqlServiceMessage\n");
#endif	

	switch(msg->hdr.cmd){
	case BHMSG_SERVICE_SYSTEM:
	case BHMSG_SERVICE_CLIENT: break;
	case BHMSG_SERVICE_SITE: return SqlserviceSite(db, p, msg);
	case BHMSG_SERVICE_MESSAGE: break;
	case BHMSG_SERVICE_DATABASE: return SqlserviceDb(db, p, msg, pluCache);
	case BHMSG_SERVICE_CUSTOMER: return SqlserviceCustomer(db, p, c, msg);
	case BHMSG_SERVICE_FINANCIAL:
	case BHMSG_SERVICE_PNP:
	case BHMSG_SERVICE_CCV:
	case BHMSG_SERVICE_KEYMAP:
		break;
	}
	
	return 0;
}

int SqlserviceSite(bhpos::BhSQL *db, PLUServer *p, BhPosMsg *msg)
{
	char				  query[255];
	int					num, *temp;
	BhServiceSite	  ss;
	BhPosMsg		  rmsg;
	
	bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow	*sql_row;
	
	memcpy(&rmsg, msg, sizeof(rmsg));
	BhServiceSiteNum *sn=(BhServiceSiteNum *) &rmsg.data;

	switch(msg->hdr.subcmd){
	case BHMSG_SERVICE_SITE_SITE_NUM:
		if(sn->flags==BHPOS_SERVSITEFLAG_ALL){
			sql_result.init(db, sqlq_qry22, true);
		}
		else if(sn->flags==BHPOS_SERVSITEFLAG_ACTIVE){
			sql_result.init(db, sqlq_qry23, true);
		}
		else if(sn->flags==BHPOS_SERVSITEFLAG_TARGETALL){
		}
		else if(sn->flags==BHPOS_SERVSITEFLAG_TARGETACTIVE){
		}
		else{
			return 1;
		}
		num=sn->num=sql_result.NumRows();
		temp=sn->ID;
		rmsg.hdr.size=sizeof(BhServiceSiteNum) - sizeof(int);
		while(num){
			sql_row=sql_result.GetRow();
			*temp=atoi(sql_row->GetField(1));
			rmsg.hdr.size+=sizeof(int);
			temp++;
			num--;
		}
		if(OutQ) OutQ->Add(&rmsg);
		break;
	case BHMSG_SERVICE_SITE_SITE_DETAILS:			
		sprintf(query, sqlq_qry24, sn->targetID);
		sql_result.init(db, query, true);
		ss.numDepartments=sql_result.NumRows();
		
		sprintf(query, sqlq_qry25, sn->targetID);
		sql_result.init(db, query, true);
		ss.numActiveDepartments=sql_result.NumRows();
		
		sprintf(query, sqlq_qry26, sn->targetID);
		sql_result.init(db, query, true);
		
		if(sql_result.NumRows()!=0){
			sql_row=sql_result.GetRow();
			strcpy(ss.name, sql_row->GetField(2));
		}
		memcpy(msg->data, &ss, sizeof(ss));
		msg->hdr.size=sizeof(ss);
		if(OutQ) OutQ->Add(&rmsg);
		break;
	case BHMSG_SERVICE_SITE_SITE_ACTIVE:
		break;
	
	case BHMSG_SERVICE_SITE_DEPT_NUM:
		if(sn->flags==BHPOS_SERVSITEFLAG_ALL){
			sql_result.init(db, sqlq_qry27, true);
		}
		else if(sn->flags==BHPOS_SERVSITEFLAG_ACTIVE){
			sql_result.init(db, sqlq_qry28, true);
		}
		else if(sn->flags==BHPOS_SERVSITEFLAG_TARGETALL){
			sprintf(query, sqlq_qry29, sn->targetID);
			sql_result.init(db, query, true);
		}
		else if(sn->flags==BHPOS_SERVSITEFLAG_TARGETACTIVE){
			sprintf(query, sqlq_qry30, sn->targetID);
			sql_result.init(db, query, true);
		}
		else{
			return 1;
		}
		num=sn->num=sql_result.NumRows();
		temp=sn->ID;
		rmsg.hdr.size=sizeof(BhServiceSiteNum) - sizeof(int);
		while(num){
			sql_row=sql_result.GetRow();
			*temp=atoi(sql_row->GetField(1));
			rmsg.hdr.size+=sizeof(int);
			temp++;
			num--;
		}
		if(OutQ) OutQ->Add(&rmsg);
		break;
		
	case BHMSG_SERVICE_SITE_DEPT_DETAILS:
	{
		BhServiceDept dd;
		
		sprintf(query, sqlq_qry31, sn->targetID);
		sql_result.init(db, query, true);
		dd.numGroups=sql_result.NumRows();
		
		sprintf(query, sqlq_qry32, sn->targetID);
		sql_result.init(db, query, true);
		dd.numActiveGroups=sql_result.NumRows();
		
		sprintf(query, sqlq_qry33, sn->targetID);
		sql_result.init(db, query, true);
		
		if(sql_result.NumRows()!=0){
			sql_row=sql_result.GetRow();
			strcpy(dd.name, sql_row->GetField(2));
		}
		memcpy(msg->data, &dd, sizeof(dd));
		msg->hdr.size=sizeof(dd);
		if(OutQ) OutQ->Add(&rmsg);
	}
		break;
		
	case BHMSG_SERVICE_SITE_DEPT_ACTIVE:
		break;
		
	case BHMSG_SERVICE_SITE_GROUP_NUM:
		if(sn->flags==BHPOS_SERVSITEFLAG_ALL){
			sql_result.init(db, sqlq_qry34, true);
		}
		else if(sn->flags==BHPOS_SERVSITEFLAG_ACTIVE){
			sql_result.init(db, sqlq_qry35, true);
		}
		else if(sn->flags==BHPOS_SERVSITEFLAG_TARGETALL){
			sprintf(query, sqlq_qry36, sn->targetID);
			sql_result.init(db, query, true);
		}
		else if(sn->flags==BHPOS_SERVSITEFLAG_TARGETACTIVE){
			sprintf(query, sqlq_qry37, sn->targetID);
			sql_result.init(db, query, true);
		}
		else{
			return 1;
		}
		num=sn->num=sql_result.NumRows();
		temp=sn->ID;
		rmsg.hdr.size=sizeof(BhServiceSiteNum) - sizeof(int);
		while(num){
			sql_row=sql_result.GetRow();
			*temp=atoi(sql_row->GetField(1));
			rmsg.hdr.size+=sizeof(int);
			temp++;
			num--;
		}
		if(OutQ) OutQ->Add(&rmsg);
		break;
		
	case BHMSG_SERVICE_SITE_GROUP_DETAILS:
	{
		BhServiceGroup gg;
		
		sprintf(query, sqlq_qry38, sn->targetID);
		sql_result.init(db, query, true);
		gg.numStations=sql_result.NumRows();
		
		sprintf(query, sqlq_qry39, sn->targetID);
		sql_result.init(db, query, true);
		gg.numActiveStations=sql_result.NumRows();
		
		sprintf(query, sqlq_qry40, sn->targetID);
		sql_result.init(db, query, true);
		
		if(sql_result.NumRows()!=0){
			sql_row=sql_result.GetRow();
			strcpy(gg.name, sql_row->GetField(2));
		}
		memcpy(msg->data, &gg, sizeof(gg));
		msg->hdr.size=sizeof(gg);
		if(OutQ) OutQ->Add(&rmsg);
	}
		break;

	case BHMSG_SERVICE_SITE_GROUP_ACTIVE:
		break;
		
	case BHMSG_SERVICE_SITE_STATION_NUM:
		if(sn->flags==BHPOS_SERVSITEFLAG_ALL){
			sql_result.init(db, sqlq_qry41, true);
		}
		else if(sn->flags==BHPOS_SERVSITEFLAG_ACTIVE){
			sql_result.init(db, sqlq_qry42, true);
		}
		else if(sn->flags==BHPOS_SERVSITEFLAG_TARGETALL){
			sprintf(query, sqlq_qry43, sn->targetID);
			sql_result.init(db, query, true);
		}
		else if(sn->flags==BHPOS_SERVSITEFLAG_TARGETACTIVE){
			sprintf(query, sqlq_qry44, sn->targetID);
			sql_result.init(db, query, true);
		}
		else{
			return 1;
		}
		num=sn->num=sql_result.NumRows();
		temp=sn->ID;
		rmsg.hdr.size=sizeof(BhServiceSiteNum) - sizeof(int);
		while(num){
			sql_row=sql_result.GetRow();
			*temp=atoi(sql_row->GetField(1));
			rmsg.hdr.size+=sizeof(int);
			temp++;
			num--;
		}
		if(OutQ) OutQ->Add(&rmsg);
		break;
		
	case BHMSG_SERVICE_SITE_STATION_DETAILS:
	{
		BhServiceStation s;
		
		sprintf(query, sqlq_qry45, sn->targetID);
		sql_result.init(db, query, true);
		
		if(sql_result.NumRows()!=0){
			sql_row=sql_result.GetRow();
			strcpy(s.name, sql_row->GetField(2));
		}
		memcpy(msg->data, &s, sizeof(s));
		msg->hdr.size=sizeof(s);
		if(OutQ) OutQ->Add(&rmsg);
	}
		break;
		/*
		int	status;
		int	userID;
		char  username[64];
		*/
	case BHMSG_SERVICE_SITE_STATION_ACTIVE:
		break;
	}
	
	return 0;
}

int SqlSysUser(bhpos::BhSQL *db, PLUServer *p, BhPosMsg *msg, CPluCache *pluCache)
{
	/*
	int					 status;
	char					buff[255];
	bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow	*sql_row;*/
   
   return 0;
}

int SqlserviceDb(bhpos::BhSQL *db, PLUServer *p, BhPosMsg *msg, CPluCache *pluCache)
{
	/*
	int					 status;
	char					buff[255];
	bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow	*sql_row;*/
	
	switch(msg->hdr.subcmd){
	case BHMSG_SERVICE_DATABASE_PLUEADD:
/*
	pluCacheItem p;

	ephemeral_plu (
	ID int(11) NOT NULL auto_increment,
	refID int(11) NOT NULL,
	entryat timestamp(14) NOT NULL,
	PLU varchar(128) NOT NULL default '',
	PLUType enum('EAN-13','EAN-8','UPC-A','UPC-E','ISBN','EAN-13W','EAN-13Q','EAN-13P','EAN-13I','BH_SC6') NOT NULL default 'EAN-13',
	Name varchar(64) NOT NULL default 'No Name',
	UnitPrice double NOT NULL default '0',
	Discount float NOT NULL default '0',
	Tax float NOT NULL default '0',
	type varchar(32) NOT NULL,
  
		sprintf(buff, "INSERT INTO ephemeral_plu VALUES(0,0,NOW(),'%s','EAN-13','%s','%f','0.00','%f')", 
			si->GetPLU(), si->GetDispText(), si->GetPrice(), si->GetTax());
		status=transdb->SimpleQuery(buff);
		
#ifdef _DEBUGGERY
	if(status!=0){
		if(DebugLevel > 2){
			printf("Failed To insert ephemeral plu: %s\n", transdb->ErrorMsg());
		}
	}
#endif

	pluCache->Add(p);

*/
		break;
		
	case BHMSG_SERVICE_DATABASE_PLUEDELETE:
		break;
		
	case BHMSG_SERVICE_DATABASE_PLUEGET:
	/*
		sql_result.init(db, "SELECT * FROM ephemeral_plu", true);
		
		if(sql_result.NumRows() == 0){
			rmsg.hdr.Ex=BHPOS_REFUSE;
			rmsg.hdr.size=0;
			if(OutQ) OutQ->Add(&rmsg);
				return 1;
			}
				
		sql_row=sql_result.GetRow();
		AccountID=atoi(sql_row->GetField(0));
	*/	
		break;
	}
	return 0;
}

int SqlserviceCustomer(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{  
	return 0;
}


/*****************************************************************************************************/

int SqlPostMessage(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache)
{
#ifdef _DEBUGGERY			 
	if(DebugLevel > 2){
		printf("SqlPostMsg\n");
	}
#endif
	switch(msg->hdr.cmd){
	case BHMSG_POST_PLU:          return SqlPlu(db, p, c, msg, pluCache); 
	case BHMSG_POST_TRANSACTION:  return SqlTransaction(db, p, c, msg);
	case BHMSG_POST_SERVICE:		return SqlPostService(db, p, c, msg); 
	case BHMSG_POST_STOCKTAKE:	   return SqlPostStocktake(db, p, c, msg);
	case BHMSG_POST_LIST:			return SqlPostList(db, p, c, msg);
	case BHMSG_POST_SIGNATURE:	   return SqlPostSignature(db, p, c, msg);
	case BHMSG_POST_BLINDBALANCE: return SqlPostBlindBalance(db, p, c, msg);
	case BHMSG_POST_DELIVERY:	   return SqlPostDelivery(db, p, c, msg);
	case BHMSG_POST_RENTAL:		   return SqlPostRental(db, p, c, msg);
	case BHMSG_POST_CUSTOMER:	   return SqlPostCustomer(db, p, c, msg);
	case BHMSG_POST_DISCREPANCY:  return SqlPostDiscrepancy(db, p, c, msg);
	case BHMSG_POST_EPHEMERAL:    return SqlPostEphemeral(db, p, c, msg);
	case BHMSG_POST_ATTENDANCE:	return SqlPostAttendance(db, p, c, msg);
   case BHMSG_POST_ADJUST:       return SqlPostAdjust(db, p, c, msg);
	default: break;
	}
	return 0;
}

int SqlPostService(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	int		           id;
	BhPosMsg	           rmsg;
	bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
	char                query[512];
	char                start[20];
	char                nowt[20];
   char                eaid[21];
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	rmsg.hdr.Ex=BHPOS_GRANT;

	c.GetParameter(BHCI_PARAM_STATIONID, &id);
   c.GetEnteredAt(eaid);
   
#ifdef _DEBUGGERY			 
   if(DebugLevel > 2){
	   printf("SqlPostService\n");
   }
#endif   
	
	switch(msg->hdr.subcmd){
	case BHMSG_POST_SERVICE_SUMMARY_TRANS:
   case BHMSG_POST_SERVICE_SUMMARY_XTAPE:
   case BHMSG_POST_SERVICE_SUMMARY_ZTAPE:
	{
	   BhServiceTransSum	  *ts= (BhServiceTransSum *) rmsg.data;
		BhServiceTransSumItem *tsi=ts->item;

	   if(sql_result.init(db, sqlq_qry20, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to select now: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 1;
	   }

	   sql_row=sql_result.GetRow();
		strcpy(nowt, sql_row->GetField(0));
		strcpy(start, sql_row->GetField(0));
		strcpy(&start[11], "00:01:01");
		
		sprintf(query, sqlq_qry46, start, nowt, SQL_P_USERLOGIN, id);

		if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 1: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 2;
	   }
	   if(sql_result.NumRows()!=0){
   		sql_row=sql_result.GetRow();
   		ts->num_users=atoi(sql_row->GetField(0));
   	}
		
		sprintf(query, sqlq_qry47, start, nowt, eaid);
		   
		if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 2: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 3;
	   }
	   
	   if(sql_result.NumRows()!=0){
   		sql_row=sql_result.GetRow();
   		ts->num_trans=atoi(sql_row->GetField(0));
	   }
	   
	   sprintf(query, sqlq_qry48, start, nowt, eaid);
	  if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 3: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 4;
	   }
	   if(sql_result.NumRows()!=0){
   		sql_row=sql_result.GetRow();
   		ts->num_items=atoi(sql_row->GetField(0));
   	}
	   
	   sprintf(query, sqlq_qry49, start, nowt, SQL_P_NOSALE, id);
	   if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 5: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 5;
	   }
	   if(sql_result.NumRows()!=0){
   		sql_row=sql_result.GetRow();
   		ts->num_nosale=atoi(sql_row->GetField(0));
   	}

	   sprintf(query, sqlq_qry50, start, nowt, SQL_P_VOIDSALE, id);
	   if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 6: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 6;
	   }
	   if(sql_result.NumRows()!=0){
   		sql_row=sql_result.GetRow();
   		ts->num_voidsale=atoi(sql_row->GetField(0));
		}
      
		sprintf(query, sqlq_qry51, start, nowt, SQL_A_PAYIN, eaid);

		if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 7: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 7;
	   }
	   if(sql_result.NumRows()!=0){
   		sql_row=sql_result.GetRow();
         ts->num_payin=atoi(sql_row->GetField(0));
         if(sql_row->GetField(1)!=NULL)
   		   ts->total_payin=atof(sql_row->GetField(1));
         else 
            ts->total_sold=0.0;
		}
  
      sprintf(query, sqlq_qry52, start, nowt, SQL_A_PAYOUT, eaid);
		if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 8: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 8;
	   }
	   if(sql_result.NumRows()!=0){
   		sql_row=sql_result.GetRow();
         ts->num_payout=atoi(sql_row->GetField(0));
         if(sql_row->GetField(1)!=NULL)
   		   ts->total_payout=atof(sql_row->GetField(1));
         else 
            ts->total_sold=0.0;
		}

		sprintf(query, sqlq_qry53, start, nowt, eaid);
		if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 8: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 8;
	   }
	   if(sql_result.NumRows()!=0){
   		sql_row=sql_result.GetRow();
         ts->num_trans=atoi(sql_row->GetField(0));
         if(sql_row->GetField(1)!=NULL)
   		   ts->total_sold=atof(sql_row->GetField(1));
         else 
            ts->total_sold=0.0;
   	}
      
      SqlDrawerQuery(db, c, start, nowt, &ts->opening_balance, &ts->closing_balance);

		sprintf(query, sqlq_qry54, start, nowt, SQL_PAY_CASH, eaid);
		if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 9: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 9;
	   }
	   if(sql_result.NumRows()!=0){
   		sql_row=sql_result.GetRow();
   		tsi->type=bhpos::PTYPE_CASH;
   		tsi->num=atoi(sql_row->GetField(0));
         if(sql_row->GetField(1)!=NULL)
   		   tsi->amount=atof(sql_row->GetField(1));
		   tsi++;
		}

		sprintf(query, sqlq_qry55, start, nowt, SQL_PAY_CREDIT, eaid);
		if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 10: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 10;
	   }
	   if(sql_result.NumRows()!=0){
   		sql_row=sql_result.GetRow();
   		tsi->type=bhpos::PTYPE_CREDIT;
   		tsi->num=atoi(sql_row->GetField(0));
         if(sql_row->GetField(1)!=NULL)
   		   tsi->amount=atof(sql_row->GetField(1));
   		tsi++;
		}

		sprintf(query, sqlq_qry56, start, nowt, SQL_PAY_CHEQUE, eaid);
		if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 11: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 11;
	   }
	   if(sql_result.NumRows()!=0){
   		sql_row=sql_result.GetRow();
   		tsi->type=bhpos::PTYPE_CHEQUE;
   		tsi->num=atoi(sql_row->GetField(0));
         if(sql_row->GetField(1)!=NULL)
   		   tsi->amount=atof(sql_row->GetField(1));
   		tsi++;
   	}

	   sprintf(query, sqlq_qry57, start, nowt, eaid);
	   if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 12: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 1;
	   }
	   if(sql_result.NumRows()!=0){
   		sql_row=sql_result.GetRow();
   		tsi->type=bhpos::PTYPE_OTHER;
   		tsi->num=atoi(sql_row->GetField(0));
         if(sql_row->GetField(1)!=NULL)
   		   tsi->amount=atof(sql_row->GetField(1));
		}
		rmsg.hdr.size=sizeof(BhServiceTransSum) + (sizeof(BhServiceTransSumItem) * 3);
		if(OutQ) OutQ->Add(&rmsg);
	}
		break;
	
	case BHMSG_POST_SERVICE_SUMMARY_USER:
	{
      bhpos::BhSQLResult  sql_res;
	   bhpos::BhSQLRow	 *sql_r;
      BhServiceUserSum *us= (BhServiceUserSum *) rmsg.data;
      BhServiceUserSumItem *usi=us->user;
      
      rmsg.hdr.size=sizeof(BhServiceUserSum) -sizeof(BhServiceUserSumItem);
      
      if(sql_result.init(db, sqlq_qry20, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to select now: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 1;
	   }
      
	   sql_row=sql_result.GetRow();
		strcpy(nowt, sql_row->GetField(0));
		strcpy(start, sql_row->GetField(0));
		strcpy(&start[11], "00:01:01");
      
      sprintf(query, sqlq_qry58,
		   start, nowt);
      if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 1: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 1;
	   }
	   if(sql_result.NumRows()!=0){
   		sql_row=sql_result.GetRow();
   		us->num_users=atoi(sql_row->GetField(0));
		}
      sprintf(query, sqlq_qry59, start, nowt, SQL_U_LOGIN);
      if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to query 2: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 1;
	   }
	   if(sql_result.NumRows()!=0){
         char fname[32], lname[32];
         // for each user...
   		while((sql_row=sql_result.GetRow())){
   		   usi->ID=atoi(sql_row->GetField(0));
            SqlUserNameLookup(db, usi->ID, fname, lname);
            fname[1]=0;
            strcat(lname, ", ");
            strcat(lname, fname);
            strcpy(usi->name, lname);
            sprintf(query, sqlq_qry60, start, nowt, eaid, usi->ID);
            if(sql_res.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		      if(DebugLevel > 2){
			         printf("Failed to query 3: %s\n", db->ErrorMsg());
   		      }
#endif
      		   rmsg.hdr.Ex=BHPOS_REFUSE;
      		   rmsg.hdr.size=0;
   	   	   if(OutQ) OutQ->Add(&rmsg);
   		      return 2;
	         }
	         if(sql_res.NumRows()!=0){
   		      sql_r=sql_res.GetRow();
               usi->num_trans=atoi(sql_r->GetField(0));
            }
            sprintf(query, sqlq_qry61, start, nowt, SQL_P_NOSALE, usi->ID);
         	if(sql_res.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		      if(DebugLevel > 2){
         	      printf("Failed to query 4: %s\n", db->ErrorMsg());
   		      }
#endif
            	rmsg.hdr.Ex=BHPOS_REFUSE;
   		      rmsg.hdr.size=0;
            	if(OutQ) OutQ->Add(&rmsg);
   		      return 3;
	         }
         	if(sql_res.NumRows()!=0){
   		      sql_r=sql_res.GetRow();
               usi->num_nosale=atoi(sql_r->GetField(0));
   	      }

	         sprintf(query, sqlq_qry62, start, nowt, SQL_P_VOIDSALE, usi->ID);
	         if(sql_res.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
         	   if(DebugLevel > 2){
		            printf("Failed to query 5: %s\n", db->ErrorMsg());
   		      }
#endif
            	rmsg.hdr.Ex=BHPOS_REFUSE;
   		      rmsg.hdr.size=0;
               if(OutQ) OutQ->Add(&rmsg);
   		      return 4;
	         }
         	if(sql_res.NumRows()!=0){
   		      sql_r=sql_res.GetRow();
               usi->num_voidsale=atoi(sql_r->GetField(0));
		      }

            sprintf(query, sqlq_qry63, start, nowt, SQL_A_PAYIN, usi->ID);
		      if(sql_res.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
            	if(DebugLevel > 2){
			         printf("Failed to query 6: %s\n", db->ErrorMsg());
            	}
#endif
   		      rmsg.hdr.Ex=BHPOS_REFUSE;
            	rmsg.hdr.size=0;
   		      if(OutQ) OutQ->Add(&rmsg);
            	return 5;
	         }
	         if(sql_res.NumRows()!=0){
            	sql_r=sql_res.GetRow();
               if(sql_r->GetField(0)!=NULL)
   		         usi->total_payin=atof(sql_r->GetField(0));
		      }

            sprintf(query, sqlq_qry64, start, nowt, SQL_A_PAYOUT, usi->ID);
         	if(sql_res.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		      if(DebugLevel > 2){
			         printf("Failed to query 7: %s\n", db->ErrorMsg());
            	}
#endif
   		      rmsg.hdr.Ex=BHPOS_REFUSE;
   		      rmsg.hdr.size=0;
            	if(OutQ) OutQ->Add(&rmsg);
   		      return 6;
	         }
         	if(sql_res.NumRows()!=0){
   		      sql_r=sql_res.GetRow();
               if(sql_r->GetField(0)!=NULL)
   		         usi->total_payout=atof(sql_r->GetField(0));
		      }

            sprintf(query, sqlq_qry65, start, nowt, usi->ID);
         	if(sql_res.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		      if(DebugLevel > 2){
      		      printf("Failed to query 8: %s\n", db->ErrorMsg());
   	        	}
#endif
            	rmsg.hdr.Ex=BHPOS_REFUSE;
   		      rmsg.hdr.size=0;
            	if(OutQ) OutQ->Add(&rmsg);
   		      return 1;
         	}
         	if(sql_res.NumRows()!=0){
   		      sql_r=sql_res.GetRow();
               if(sql_r->GetField(0)!=NULL)
   		         usi->total_transacted=atof(sql_r->GetField(0));
		      }

            sprintf(query, sqlq_qry66, start, nowt, SQL_U_LOGIN, usi->ID);
	         if(sql_res.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
         	if(DebugLevel > 2){
		         printf("Failed to query 9: %s\n", db->ErrorMsg());
   		   }
#endif
            	rmsg.hdr.Ex=BHPOS_REFUSE;
   		      rmsg.hdr.size=0;
            	if(OutQ) OutQ->Add(&rmsg);
   		      return 4;
	         }
         	if(sql_res.NumRows()!=0){
   		      sql_r=sql_res.GetRow();
            	usi->num_login=atoi(sql_r->GetField(0));
		      }

            sprintf(query, sqlq_qry67, start, nowt, SQL_U_CLOCKIN, usi->ID);
	         if(sql_res.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
         	   if(DebugLevel > 2){
		      	   printf("Failed to query 10: %s\n", db->ErrorMsg());
   		      }
#endif
            	rmsg.hdr.Ex=BHPOS_REFUSE;
   		      rmsg.hdr.size=0;
            	if(OutQ) OutQ->Add(&rmsg);
   		      return 4;
	         }
         	if(sql_res.NumRows()!=0){
   		      sql_r=sql_res.GetRow();
               strcpy(query, sql_r->GetField(0));
            	strncpy(usi->start, &query[11], 5);
               printf("poobert - %s %s\n", query, usi->start);
		      }
            
            sprintf(query, sqlq_qry68, start, nowt, SQL_U_CLOCKOUT, usi->ID);
	         if(sql_res.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
         		if(DebugLevel > 2){
		      	   printf("Failed to query 11: %s\n", db->ErrorMsg());
   		      }
#endif
            	rmsg.hdr.Ex=BHPOS_REFUSE;
   		      rmsg.hdr.size=0;
            	if(OutQ) OutQ->Add(&rmsg);
   		      return 4;
	         }
         	if(sql_res.NumRows()!=0){
   		      sql_r=sql_res.GetRow();
               strcpy(query, sql_r->GetField(0));
            	strncpy(usi->start, &query[11], 5);
            }
            
            sprintf(query, sqlq_qry69, start, nowt, SQL_U_LOGIN, SQL_U_LOGIN, usi->ID);
            // TO DO: finding matching IN/OUT pairs with validation on time diff 
            // positive, sum seconds diff, convert back to timestamp
               
            usi++;
            rmsg.hdr.size+=sizeof(BhServiceUserSumItem);
         }
      }
      
		if(OutQ) OutQ->Add(&rmsg);
	}
		break;
	}
	return 0;
}

int SqlPlu(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache)
{
	BhPosMsg     rmsg;
	BhPLUrequest plu;
	char			 strplu[129];
	int			 temp=0,type=0, int_type=0, qty, nameID=1;

	memset(strplu, 0, sizeof(strplu)); 
	memcpy(&rmsg, msg, sizeof(rmsg));
	memcpy(&plu, msg->data, sizeof(plu));

	c.GetParameter(BHCI_PARAM_MODE, &temp);
	if(temp==POST_MODE_TRANSNEW){	
		temp=POST_MODE_SALE;
		c.SetParameter(BHCI_PARAM_MODE, &temp);
		//p->UpdateClient(c);
	}
#ifdef _DEBUGGERY			 
	if(DebugLevel > 2){
		printf("SqlQ:: Num Trans: %d this trans items: %d\n", c.NumTrans(), c.NumItems(plu.transID, bhpos::BHITEM_TYPE_ALL));
	}
#endif

   nameID=c.GetNameID(plu.transID);
   
	memcpy(strplu, &plu.PLU, 2);
	strplu[2]=0;
	type=atoi(strplu);
	qty=plu.qty;

	p->GetOption(BH_PLUS_OPTION_WEIGHTBCPREFIX, &int_type);
	if(type==int_type){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2)
			printf("SqlPlu: request for weight\n");
#endif		
		return SqlPluInternalWeight(db, p, c, msg, pluCache);
	}
	p->GetOption(BH_PLUS_OPTION_QTYBCPREFIX, &int_type);
	if(type==int_type){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2)
			printf("SqlPlu: request for quantity\n");
#endif		
		return SqlPluInternalQuantity(db, p, c, msg, pluCache);
	}
	p->GetOption(BH_PLUS_OPTION_IFBCPREFIX, &int_type);
	if(type==int_type){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2)
			printf("SqlPlu: request for interface\n");
#endif				
		return SqlPluInternalInterface(db, p, c, msg, pluCache);
	}
	p->GetOption(BH_PLUS_OPTION_PRICEBCPREFIX, &int_type);
	if(type==int_type){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2)
			printf("SqlPlu: request for price\n");
#endif				
		return SqlPluInternalPrice(db, p, c, msg, pluCache);
	}
	p->GetOption(BH_PLUS_OPTION_TOKENBCPREFIX, &int_type);
	if(type==int_type){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2)
			printf("SqlPlu: request for token\n");
#endif				
		return SqlPluToken(db, p, c, msg, pluCache);
	}
	p->GetOption(BH_PLUS_OPTION_VOUCHERBCPREFIX, &int_type);
	if(type==int_type){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2)
			printf("SqlPlu: request for voucher\n");
#endif				
		return SqlPluVoucher(db, p, c, msg, pluCache);
	}
	
	memcpy(strplu, plu.PLU, 13);
	strplu[14]=0;
	
	p->GetOption(BH_PLUS_OPTION_INTBCPREFIX, &int_type);
	if(type==int_type){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2)
			printf("SqlPlu: request for in-store\n");
#endif				
		return SqlPluInStore(db, strplu, p, c, msg, pluCache);
	}
	
#ifdef _DEBUGGERY 
	if(DebugLevel > 2){
		printf("SqlPlu: request for standard lookup: %s\n", strplu);
	}
#endif	
	
	if(SqlPluLookup(db, strplu, qty, nameID, &rmsg, pluCache)){
		BhPosMsg   emsg;
      BhErrorMsg err;
		emsg.hdr.type=BHPOS_TYPE_SYSTEM;
		emsg.hdr.cmd=BHMSG_ERROR_OPERATION;
		emsg.hdr.subcmd=BHMSG_ERROR_OPERATION_PLUNOTFOUND;
      
      SqlErrorMsgFill(c, &err);
      err.type=BHMSG_ERROR_OPERATION_PLUNOTFOUND;
      strcpy(err.more, strplu);
      
		emsg.hdr.size=sizeof(BhErrorMsg);		 
		memcpy(emsg.data, &err, sizeof(BhErrorMsg));
		if(ErrorQ) ErrorQ->Add(&emsg);
	}
	if(OutQ)
		OutQ->Add(&rmsg);

	return 0;
}

/////////////////////////////////////////////////////////////////////////////////
//
//
//
//
////////////////////////////////////////////////////////////////////////////////

int SqlTransaction(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	switch(msg->hdr.subcmd){
	case BHMSG_POST_TRANSACTION_NEW: 
	case BHMSG_POST_TRANSACTION_NEWEX:
	case BHMSG_POST_TRANSACTION_RETURNSNEW:
		return SqlTransNew(db, p, c, msg);

	case BHMSG_POST_TRANSACTION_RETURNSGET:
		return SqlTransReturnsGet(db, p, c, msg);

	case BHMSG_POST_TRANSACTION_PAYAUTH:
		return SqlTransPayAuth(db, p, msg);

	case BHMSG_POST_TRANSACTION_NOSALE:
		return SqlPostNoSale(db, p, c, msg);
	}
	return 0;
}

static int SqlTransNew(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
	BhTransactionNew	  nt;
	BhTransactionNew	 *nt_in=NULL;
	BhPosMsg				  rmsg;
	
	int					  status, mode, link_id=0;
	char					 *ch, buff[256];
	int					  AccountID=0, *nameID;
	char					  tp[8];
	int					  trans_prefix=0;
	unsigned int		  transid;

	memset(&rmsg, 0, sizeof(rmsg));
	memset(&nt, 0, sizeof(nt));
	memcpy(&rmsg, msg, sizeof(rmsg));

	p->GetOption(BH_PLUS_OPTION_TRANSPREFIX, &tp);
	
	if(msg->hdr.subcmd==BHMSG_POST_TRANSACTION_NEWEX){
		nt_in=(BhTransactionNew *) msg->data;
	}
   nameID=(int *) msg->data;
	
#ifdef _DEBUGGERY			 
	if(DebugLevel > 2){
		printf("SqlTransNew - nameID:%d\n", *nameID);
	}
#endif
	c.GetParameter(BHCI_PARAM_STATUS, &status);
	c.GetParameter(BHCI_PARAM_MODE, &mode);
	if(status==POST_STATUS_TRANSACTION && mode==POST_MODE_TRANSNEW){
		// The client has fucked up
		//c->Add()
		//c->GetParameter(BHCI_PARAM_TRANSID, &status);
		//nt.transID=status;
		rmsg.hdr.Ex=BHPOS_GRANT;
		rmsg.hdr.size=sizeof(BhTransactionNew);
		memcpy(rmsg.data, &nt, sizeof(nt));
		
		if(OutQ)
			OutQ->Add(&rmsg);
				
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("Resending Transaction ID=%d\n", nt.transID);
		}
#endif		
		return 0;
	}
	
	status=db->SimpleQuery(sqlq_qry70);
	if(status!=0){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("Failed to lock trans table for new insert: %s - %d\n", db->ErrorMsg(), status);
		}
#endif
		rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
	
	if(msg->hdr.subcmd==BHMSG_POST_TRANSACTION_NEWEX){
		switch(nt_in->flags){
		case TRANSACTIONFLAG_TYPE_SALE:
			break;
		case TRANSACTIONFLAG_TYPE_RETURN:
		case TRANSACTIONFLAG_TYPE_ADJUSTMENT:
		case TRANSACTIONFLAG_TYPE_LINK:
			sprintf(buff, sqlq_qry71, nt_in->transID);
			if(db)
				sql_result.init(db, buff, true);
		
			if(sql_result.NumRows() == 0){
            db->SimpleQuery(sqlq_qry146);
				rmsg.hdr.Ex=BHPOS_REFUSE;
				rmsg.hdr.size=0;
				if(OutQ) OutQ->Add(&rmsg);
				return 1;
			}
				
			sql_row=sql_result.GetRow();
			link_id=atoi(sql_row->GetField(0));		 
			break;
		case TRANSACTIONFLAG_TYPE_RECONCILE:
		case TRANSACTIONFLAG_TYPE_IN:
		case TRANSACTIONFLAG_TYPE_OUT:
			break;
		case TRANSACTIONFLAG_TYPE_MULTI:
			sprintf(buff, sqlq_qry72, nt_in->sid);
			if(db)
				sql_result.init(db, buff, true);
		
			if(sql_result.NumRows() == 0){
            db->SimpleQuery(sqlq_qry146);
				rmsg.hdr.Ex=BHPOS_REFUSE;
				rmsg.hdr.size=0;
				if(OutQ) OutQ->Add(&rmsg);
				return 1;
			}
				
			sql_row=sql_result.GetRow();
			AccountID=atoi(sql_row->GetField(0));
			break;
		}
	}

	sql_result.free();	
	status=db->SimpleQuery(sqlq_qry73);

	if(status!=0){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("Failed To insert new transaction: %s\n", db->ErrorMsg());
		}
#endif
      db->SimpleQuery(sqlq_qry146);
		rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
   
   status=db->SimpleQuery(sqlq_qry77);
   if(status!=0){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("Failed To commit new transaction: %s\n", db->ErrorMsg());
		}
#endif
		rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
   
   // last insert id
   if(sql_result.init(db, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
      if(DebugLevel > 2){
			printf("Failed to query 145: %s\n", db->ErrorMsg());
      }
#endif
      db->SimpleQuery(sqlq_qry146);
      rmsg.hdr.Ex=BHPOS_REFUSE;
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
      return 2;
   }
   if(sql_result.NumRows()!=0){
      sql_row=sql_result.GetRow();
      nt.transID=atoi(sql_row->GetField(0));
   }
   
   if(trans_prefix!=0){
		sprintf(buff, "%d", trans_prefix);
		transid=10-strlen(tp)-status;
		strcpy(buff, tp);
		ch=&buff[transid];
		for(unsigned int i=0; i<=transid; i++){
			*ch='0';
			ch++;
		}
		transid=atoi(buff);
		nt.transID+=transid;
	}
	sql_result.free();
   
   status=db->SimpleQuery(sqlq_qry70);
   if(status!=0){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("Failed To lock for new transaction: %s\n", db->ErrorMsg());
		}
#endif
		rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
   
	sprintf(buff, sqlq_qry74, nt.transID, nt.transID);
	status=db->SimpleQuery(buff);
	if(status!=0){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("Failed To update new transaction: %s\n", db->ErrorMsg());
		}
#endif


		if(msg->hdr.subcmd==BHMSG_POST_TRANSACTION_NEWEX){
			switch(nt_in->flags){
			case TRANSACTIONFLAG_TYPE_SALE:
				break;
			case TRANSACTIONFLAG_TYPE_RETURN:
			case TRANSACTIONFLAG_TYPE_ADJUSTMENT:
			case TRANSACTIONFLAG_TYPE_LINK:
				sprintf(buff, sqlq_qry75, link_id, nt.transID);
				status=db->SimpleQuery(buff);
				if(status!=0){
#ifdef _DEBUGGERY			 
					if(DebugLevel > 2){
						printf("Failed To update linked transID: %s\n", db->ErrorMsg());
					}
#endif
				}
				break;
			case TRANSACTIONFLAG_TYPE_RECONCILE:
			case TRANSACTIONFLAG_TYPE_IN:
			case TRANSACTIONFLAG_TYPE_OUT:
				break;
			case TRANSACTIONFLAG_TYPE_MULTI:
				sprintf(buff, sqlq_qry76, AccountID, nt.transID);
				status=db->SimpleQuery(buff);
				if(status!=0){
#ifdef _DEBUGGERY			 
					if(DebugLevel > 2){
						printf("Failed To update Account ID in multiprice sale: %s\n", db->ErrorMsg());
					}
#endif
				}
				break;
			}
		}

		status=db->SimpleQuery(sqlq_qry77);
		if(status!=0){
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("Failed to unlock trans table for new insert: %s\n", db->ErrorMsg());
			}
#endif
		}
			
		rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
	
	status=db->SimpleQuery(sqlq_qry77);
	if(status!=0){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("Failed to unlock trans table for new insert: %s\n", db->ErrorMsg());
		}
#endif
	}
	
#ifdef _DEBUGGERY			 
	if(DebugLevel > 2){
		printf("SqlQ: New Transaction ID=%d\n", nt.transID);
	}
#endif
	
	memcpy(rmsg.data, &nt, sizeof(nt));

	// add a new transaction
	status=c.NewTrans(nt.transID, *nameID);
   
#ifdef _DEBUGGERY			 
	if(DebugLevel > 2){
		printf("SqlQ::NewTrans: Num Trans in stack now: %d\n", status);
	}
#endif

	if(msg->hdr.subcmd==BHMSG_POST_TRANSACTION_RETURNSNEW){
	}
	
	status=POST_STATUS_TRANSACTION;
	c.SetParameter(BHCI_PARAM_STATUS, &status);
	mode=POST_MODE_TRANSNEW;
	c.SetParameter(BHCI_PARAM_MODE, &mode);
	//p->UpdateClient(c);
	
	rmsg.hdr.Ex=BHPOS_GRANT;
	rmsg.hdr.size=sizeof(BhTransactionNew);
	
	if(OutQ)
		OutQ->Add(&rmsg);
	
	return 0;
}

static int SqlTransReturnsGet(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	BhPosMsg				rmsg;
	bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
	char					 buff[256];
	BhTransactionGet	*tid=(BhTransactionGet *) msg->data;

	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));

#ifdef _DEBUGGERY			 
	if(DebugLevel > 2){
		printf("SqlTransReturnsGet\n");
	}
#endif
	sprintf(buff, sqlq_qry71, tid->transID);
	if(db)
		sql_result.init(db, buff, true);
		
	if(sql_result.NumRows() == 0){
		rmsg.hdr.Ex=BHPOS_REFUSE;
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
	
	rmsg.hdr.Ex=BHPOS_GRANT;
	rmsg.hdr.size=0;
	if(OutQ) OutQ->Add(&rmsg);

	return 0;
}

static int SqlTransPayAuth(bhpos::BhSQL *db, PLUServer *p, BhPosMsg *msg)
{
	BhPosMsg				 rmsg;
	BhTransPayAuth		 pa;
	BhTransPayAuthReply  par;

	//BhTransPayCredit	cc;
	BhTransPayCheque  ch;
	char				  rm[64];
	int					 msgsize=0;
	
	bhpos::BhSQLResult sql_result;
	char					sql_query[256];
	bhpos::BhSQLRow	*sql_row;

	memset(&pa, 0, sizeof(pa));
	memset(&par, 0, sizeof(par)); 
	memcpy(&rmsg, msg, sizeof(rmsg));

	memcpy(&pa, msg->data, sizeof(pa));
	
	par.transID=pa.transID;

	if(pa.type==bhpos::PTYPE_CHEQUE){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("SqlTransPayAuth: cheque\n");
		}
#endif
		/*
		sprintf(sql_query, "SELECT * FROM badcheque WHERE bankSort='%s'", ch.bank);
		if(db)
			sql_result.init(db, sql_query, true);
		
		if(sql_result.NumRows() > 0){	 // we have a bad bank entry entry
			sql_row=sql_result.GetRow();

			// look up account number...
			sprintf(sql_query, "SELECT * FROM badcheque WHERE bankAccountNum='%s'", ch.account);
					
			rmsg.hdr.Ex=BHPOS_REFUSE;
			par.bAccepted=0;
			par.bRequireSupervisor=0;
			par.flags=0;
		}
		else{
			// could do name search...
			// sprintf(sql_query, "SELECT * FROM badcheque WHERE name='%s'", accountname);
			par.bRequireSupervisor=0;
			par.bAccepted=1;
			par.flags=0;
			rmsg.hdr.Ex=BHPOS_GRANT;
		}
		sql_result.free();*/
	}
	else if(pa.type==bhpos::PTYPE_CREDIT){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("SqlTransPayAuth: credit\n");
		}
#endif
#ifdef _CCVENGINEMCVE				
		if(CcvQ){
			CcvQ->Add(msg);
			return 0;
		}
		else{
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("SqlTransPayAuth: credit - no ccvq\n");
			}
#endif			
			rmsg.hdr.Ex=BHPOS_GRANT;
			par.bAccepted=1;
			par.bRequireSupervisor=0;
			par.flags=0;
		}
		
#endif
	}
	rmsg.hdr.size=sizeof(BhTransPayAuthReply) + msgsize;
	memcpy(rmsg.data, &par, sizeof(par));
	if(msgsize)
		memcpy(rmsg.data + sizeof(par), rm, msgsize);

	if(OutQ)
		OutQ->Add(&rmsg);
		
	return 0;
}


int SqlPluInternalWeight(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache)
{
	BhPosMsg	    rmsg;
	BhPLUrequest plu;
	char			 strplu[129];
	char			 strweight[8];
	int			 value=0, nameID=1;

	memcpy(&rmsg, msg, sizeof(rmsg));
	memcpy(&plu, msg->data, sizeof(plu));
	
	strcpy(strplu, "000000000000");
	memcpy(&strplu[7], &plu.PLU[2], 5);
	value=BarcodeLunBun(strplu);
	sprintf(strweight, "%i", value);
	strcat(strplu, strweight);
	
	strcpy(strweight, "000000");
	memcpy(strweight, &plu.PLU[7], 5);
	value=atoi(strweight);
	
	rmsg.hdr.subcmd=BHMSG_POST_PLU_WEIGHT;
	if(SqlPluLookup(db, strplu, plu.qty, nameID, &rmsg, pluCache)){
		BhPosMsg emsg;
		emsg.hdr.type=BHPOS_TYPE_ERROR;
		emsg.hdr.cmd=BHMSG_ERROR_OPERATION;
		emsg.hdr.subcmd=BHMSG_ERROR_OPERATION_PLUNOTFOUND;
		emsg.hdr.size=sizeof(BhPLUrequest);		 
		memcpy(emsg.data, msg->data, sizeof(BhPLUrequest));
		if(ErrorQ) ErrorQ->Add(&emsg);
	}
	if(OutQ)
		OutQ->Add(&rmsg);
		
	return 0;
}

int SqlPluInternalQuantity(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache)
{
	BhPosMsg	    rmsg;
	BhPLUrequest plu;
	char			 strplu[129];
	char			 strqty[8];
	int			 value=0, nameID=1;

	memcpy(&rmsg, msg, sizeof(rmsg));
	memcpy(&plu, msg->data, sizeof(plu));
	
	strcpy(strplu, "000000000000");
	memcpy(&strplu[7], &plu.PLU[2], 5);
	value=BarcodeLunBun(strplu);
	sprintf(strqty, "%i", value);
	strcat(strplu, strqty);
	
	strcpy(strqty, "000000");
	memcpy(strqty, &plu.PLU[7], 5);
	value=atoi(strqty);
	
	rmsg.hdr.subcmd=BHMSG_POST_PLU_QTY;
	if(SqlPluLookup(db, strplu, plu.qty, nameID, &rmsg, pluCache)){
		BhPosMsg emsg;
		emsg.hdr.type=BHPOS_TYPE_ERROR;
		emsg.hdr.cmd=BHMSG_ERROR_OPERATION;
		emsg.hdr.subcmd=BHMSG_ERROR_OPERATION_PLUNOTFOUND;
		emsg.hdr.size=sizeof(BhPLUrequest);		 
		memcpy(emsg.data, msg->data, sizeof(BhPLUrequest));
		if(ErrorQ) ErrorQ->Add(&emsg);
	}
	if(OutQ)
		OutQ->Add(&rmsg);
		
	return 0;
}

int SqlPluInternalPrice(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache)
{
	BhPosMsg	    rmsg;
	BhPLUrequest plu;
	char			 strplu[129];
	char			 strprix[8];
	int			 value=0, nameID=1;

	memcpy(&rmsg, msg, sizeof(rmsg));
	memcpy(&plu, msg->data, sizeof(plu));
	
	strcpy(strplu, "000000000000");
	memcpy(&strplu[7], &plu.PLU[2], 5);
	value=BarcodeLunBun(strplu);
	sprintf(strprix, "%i", value);
	strcat(strplu, strprix);
	
	strcpy(strprix, "000000");
	memcpy(strprix, &plu.PLU[7], 5);
	value=atoi(strprix);
	
	rmsg.hdr.subcmd=BHMSG_POST_PLU_PRICE;
	if(SqlPluLookup(db, strplu, plu.qty, nameID, &rmsg, pluCache)){
		BhPosMsg emsg;
		emsg.hdr.type=BHPOS_TYPE_ERROR;
		emsg.hdr.cmd=BHMSG_ERROR_OPERATION;
		emsg.hdr.subcmd=BHMSG_ERROR_OPERATION_PLUNOTFOUND;
		emsg.hdr.size=sizeof(BhPLUrequest);		 
		memcpy(emsg.data, msg->data, sizeof(BhPLUrequest));
		if(ErrorQ) ErrorQ->Add(&emsg);
	}
	if(OutQ)
		OutQ->Add(&rmsg);
	
	return 0;
}

int SqlPluInternalInterface(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache)
{
	BhPosMsg	    rmsg;
	BhPLUrequest plu;
	char			 struser[8];
	char			 strif[4];
	int			 user=0;
	int			 iif=0;

	memcpy(&rmsg, msg, sizeof(rmsg));
	memcpy(&plu, msg->data, sizeof(plu));
	
	strcpy(strif, "000");
   strcpy(struser, "0000000");
	memcpy(strif, &plu.PLU[2], 3);
   memcpy(struser, &plu.PLU[6], 7);
	//user=BarcodeLunBun(strif);
	//sprintf(struser, "%i", user);
	//strcat(strif, struser);
	
	iif=atoi(strif);
	
	strcpy(struser, "000000");
	memcpy(struser, &plu.PLU[7], 5);
	user=atoi(struser);
	
	rmsg.hdr.subcmd=BHMSG_POST_PLU_IF;
	/*
	if(SqlPluLookup(db, strplu, plu.qty, &rmsg, pluCache)){
		BhPosMsg emsg;
		emsg.hdr.type=BHPOS_TYPE_ERROR;
		emsg.hdr.cmd=BHMSG_ERROR_OPERATION;
		emsg.hdr.subcmd=BMSG_ERROR_OPERATION_PLUNOTFOUND;
		emsg.hdr.size=sizeof(BhPLUrequest);		 
		memcpy(emsg.data, msg->data, sizeof(BhPLUrequest));
		if(ErrorQ) ErrorQ->Add(&emsg);
	}
	if(OutQ)
		OutQ->Add(&rmsg);*/
	
   rmsg.hdr.Ex=BHPOS_REFUSE;
   rmsg.hdr.size=0;
   if(OutQ) OutQ->Add(&rmsg);
   //return 1;	
	return 0;
}

int SqlPluToken(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache)
{
	BhPosMsg		 rmsg;
	BhPLUrequest	plu;
	char			  strtoken[14];
	char			  strtemp[4];
	int				value;

	memcpy(&rmsg, msg, sizeof(rmsg));
	memcpy(&plu, msg->data, sizeof(plu));
	
	strcpy(strtoken, "000000000000");
	memcpy(&strtoken[2], &plu.PLU[2], 10);
	value=BarcodeLunBun(strtoken);
	sprintf(strtemp, "%i", value);
	strcat(strtoken, strtemp);
	
	rmsg.hdr.subcmd=BHMSG_POST_PLU_TOKEN;
	/*
	if(SqlPluLookup(db, strplu, plu.qty, &rmsg, pluCache)){
		BhPosMsg emsg;
		emsg.hdr.type=BHPOS_TYPE_ERROR;
		emsg.hdr.cmd=BHMSG_ERROR_OPERATION;
		emsg.hdr.subcmd=BMSG_ERROR_OPERATION_PLUNOTFOUND;
		emsg.hdr.size=sizeof(BhPLUrequest);		 
		memcpy(emsg.data, msg->data, sizeof(BhPLUrequest));
		if(ErrorQ) ErrorQ->Add(&emsg);
	}
	if(OutQ)
		OutQ->Add(&rmsg);*/
		
	return 0;
}

int SqlPluVoucher(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache)
{
	BhPosMsg	  rmsg;
	BhPLUrequest plu;
	char			strvoucher[14];
	char			strtemp[4];
	int			 value;
	
	memcpy(&rmsg, msg, sizeof(rmsg));
	memcpy(&plu, msg->data, sizeof(plu));
	
	strcpy(strvoucher, "000000000000");
	memcpy(&strvoucher[2], &plu.PLU[2], 10);
	value=BarcodeLunBun(strvoucher);
	sprintf(strtemp, "%i", value);
	strcat(strvoucher, strtemp);
	
	rmsg.hdr.subcmd=BHMSG_POST_PLU_VOUCHER;
	/*
	if(SqlPluLookup(db, strplu, plu.qty, &rmsg, pluCache)){
		BhPosMsg emsg;
		emsg.hdr.type=BHPOS_TYPE_ERROR;
		emsg.hdr.cmd=BHMSG_ERROR_OPERATION;
		emsg.hdr.subcmd=BHMSG_ERROR_OPERATION_PLUNOTFOUND;
		emsg.hdr.size=sizeof(BhPLUrequest);		 
		memcpy(emsg.data, msg->data, sizeof(BhPLUrequest));
		if(ErrorQ) ErrorQ->Add(&emsg);
	}
	if(OutQ)
		OutQ->Add(&rmsg);*/
		
	return 0;
}

int SqlPluInStore(bhpos::BhSQL *db, char *strplu, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg, CPluCache *pluCache)
{
	BhPosMsg				rmsg;
	BhPLUreply			 plur;
	bhpos::BhSQLResult  sql_result;
	char					 sql_query[256];
	bhpos::BhSQLRow	 *sql_row;
	int					  rxid=0;

	memcpy(&rmsg, msg, sizeof(rmsg));
	memset(&plur, 0, sizeof(plur));

#ifdef _DEBUGGERY			 
	if(DebugLevel > 2){
		printf("Ephemeral lookup\n");
	}
#endif
	
	sprintf(sql_query, sqlq_qry79, strplu);
	if(db)
		sql_result.init(db, sql_query, true);
	if(sql_result.NumRows() > 0){
		sql_row=sql_result.GetRow();
		if(!sql_row){
			rmsg.hdr.Ex=BHPOS_REFUSE;
			rmsg.hdr.size=0;
			if(OutQ) OutQ->Add(&rmsg);
			return 1;
		}
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("found ephemeral\n");
		}
#endif		
		rmsg.hdr.Ex=BHPOS_GRANT;
		rmsg.hdr.size=sizeof(plur);
		rxid=atoi(sql_row->GetField(1));
		strcpy(plur.PLU, strplu);
		strcpy(plur.displayText, sql_row->GetField(5));
		strcpy(plur.recuText, sql_row->GetField(7));
		plur.unitPrice=atof(sql_row->GetField(7));
		plur.tax=atof(sql_row->GetField(9));
		plur.deposit=0.0;
		plur.tax=0.0;
		/*
		sprintf(sql_query, "SELECT * FROM prescription WHERE prescriptionID='%d'", rxid);
		if(db)
			sql_result.init(db, sql_query, true);
		if(sql_result.NumRows() > 0){
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("found prescription\n");
			}
#endif			  
			sql_row=sql_result.GetRow();
			if(!sql_row){
				rmsg.hdr.Ex=BHPOS_REFUSE;
				rmsg.hdr.size=0;
				if(OutQ) OutQ->Add(&rmsg);
				return 1;
			}
				
			if(strcmp(sql_row->GetField(17), "Y")==0){
				plur.flags=0;
			}
			else{
				plur.flags=1;
			}
		}*/
		memcpy(rmsg.data, &plur, sizeof(plur));
		if(OutQ) OutQ->Add(&rmsg);
	}
	else{
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("No row found for PLU: %s\n", strplu);
		}
#endif		
		if(OutQ) OutQ->Add(&rmsg);
		return 1;
	}
	/*
	if(SqlPluLookup(db, strplu, plu.qty, &rmsg, pluCache)){
		BhPosMsg emsg;
		emsg.hdr.type=BHPOS_TYPE_ERROR;
		emsg.hdr.cmd=BHMSG_ERROR_OPERATION;
		emsg.hdr.subcmd=BHMSG_ERROR_OPERATION_PLUNOTFOUND;
		emsg.hdr.size=sizeof(BhPLUrequest);		 
		memcpy(emsg.data, msg->data, sizeof(BhPLUrequest));
		if(ErrorQ) ErrorQ->Add(&emsg);
	}
  */
		
	return 0;
}

int SqlPluLookup(bhpos::BhSQL *db, char *strplu, int qty, int nameID, BhPosMsg *rmsg, CPluCache *pluCache)
{
	BhPLUreply			  plur;
	bhpos::BhSQLResult  sql_result;
	char					  sql_query[256];
	bhpos::BhSQLRow	 *sql_row;
	double				  prix=0.0, discount;
	PluCacheItem		  pitem;
   bool                bFound=false;

#ifdef _DEBUGGERY			 
	if(DebugLevel > 2){
		printf("SqlPluLookup: nameID: %d\n", nameID);
	}
#endif

	if(bCachedPlu){
		pitem.SetPLU(strplu);
      pitem.SetNameID(nameID);

      if(!pluCache->Find(pitem)){
         pitem.SetNameID(1);
      	if(pluCache->Find(pitem)){ // found on default nameID
            bFound=true;
         }
      }
      else{ // found with nameID
         bFound=true;
      }
      if(bFound){
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("SqlPluLookup: found plu in cache\n");
			}
#endif			
			rmsg->hdr.Ex=BHPOS_GRANT;
			strcpy(plur.displayText, pitem.GetDispText());
			strcpy(plur.recuText, pitem.GetRecuText());
			plur.tax=pitem.GetTax();
			prix=pitem.GetPrice();
         plur.discount=pitem.GetDiscount();
         // compound discounts...
         //plur.discount+=SqlPluDiscount(db, &prix, qty, atoi(pitem.GetProductID()));
			plur.deposittype=pitem.GetDepositType();
			plur.deposit=pitem.GetDeposit();
			plur.unitPrice=prix;
			plur.flags=pitem.GetFlags();
         memcpy(rmsg->data, &plur, sizeof(plur));
			if(plur.flags!=ITEMFLAG_TYPE_NONE){ // flag type...
				sprintf(sql_query, sqlq_qry80, plur.flags);
				if(db)
					sql_result.init(db, sql_query, true);
				if(sql_result.NumRows() > 0){
					sql_row=sql_result.GetRow();
					strcpy(sql_query, sql_row->GetField(3));
					char	 *iflag_ptr=rmsg->data+sizeof(plur);
					
					if(atoi(sql_row->GetField(2))==SQL_ITEMFLAGS_MSG){
						plur.flags=ITEMFLAG_TYPE_MSG;
					}
					else if(atoi(sql_row->GetField(2))==SQL_ITEMFLAGS_AUTH){
						plur.flags=ITEMFLAG_TYPE_AUTH;
					}
					else if(atoi(sql_row->GetField(2))==SQL_ITEMFLAGS_RECORD){
						plur.flags=ITEMFLAG_TYPE_RECORD;
					}
					else{
					}
					strcpy(iflag_ptr, sql_query);
					rmsg->hdr.size=sizeof(plur)+strlen(sql_query);
				}
			}
         else{
            rmsg->hdr.size=sizeof(plur);
			}
			return 0;
		}
	}

#ifdef _DEBUGGERY			 
	if(DebugLevel > 2){
		printf("SqlPluLookup: none cached check\n");
	}
#endif

	sprintf(sql_query, sqlq_qry81, strplu, nameID);
	if(db)
		sql_result.init(db, sql_query, true);
	if(sql_result.NumRows()==0){
      sprintf(sql_query, sqlq_qry82, strplu);
	   if(db)
		   sql_result.init(db, sql_query, true);
      if(sql_result.NumRows() > 0){
         bFound=true;
      }
   }
   else{
      bFound=true;
   }
   if(bFound){
		sql_row=sql_result.GetRow();
		if(!sql_row){
			rmsg->hdr.Ex=BHPOS_REFUSE;
			rmsg->hdr.size=0;
			return 1;
		}
		rmsg->hdr.Ex=BHPOS_GRANT;
		strcpy(plur.PLU, strplu);
		strcpy(plur.displayText, sql_row->GetField(4));
		strcpy(plur.recuText, sql_row->GetField(5));
		plur.unitPrice=atof(sql_row->GetField(6));
		prix=plur.unitPrice;
      plur.discount=atof(sql_row->GetField(7));
      // compound discounts...
      //plur.discount+=SqlPluDiscount(db, &prix, qty, atoi(sql_row->GetField(1)));
		plur.unitPrice=prix;
		plur.deposittype=atoi(sql_row->GetField(8));
		plur.deposit=atof(sql_row->GetField(9));
		plur.tax=atof(sql_row->GetField(10));
		plur.flags=atoi(sql_row->GetField(11));
		memcpy(rmsg->data, &plur, sizeof(plur));
		if(plur.flags!=ITEMFLAG_TYPE_NONE){ // flag type...
			sprintf(sql_query, sqlq_qry80, plur.flags);
			if(db)
				sql_result.init(db, sql_query, true);
			if(sql_result.NumRows() > 0){
				sql_row=sql_result.GetRow();
				strcpy(sql_query, sql_row->GetField(3));
					
				char *iflag_ptr=rmsg->data+sizeof(plur);
					
				if(atoi(sql_row->GetField(2))==SQL_ITEMFLAGS_MSG){
						plur.flags=ITEMFLAG_TYPE_MSG;
					}
					else if(atoi(sql_row->GetField(2))== SQL_ITEMFLAGS_AUTH){
						plur.flags=ITEMFLAG_TYPE_AUTH;
					}
					else if(atoi(sql_row->GetField(2))==SQL_ITEMFLAGS_RECORD){
						plur.flags=ITEMFLAG_TYPE_RECORD;
					}
				else{
				}
				strcpy(iflag_ptr, sql_query);
				rmsg->hdr.size=sizeof(plur)+strlen(sql_query);
			}
		}
		else{
			rmsg->hdr.size=sizeof(plur);
		}
		return 0;
	}
	
	// last, but not least, a peek in the unknown plu table
#ifdef _DEBUGGERY			 
	if(DebugLevel > 2){
		printf("SqlPluLookup: checking u_plu table\n");
	}
#endif
	
	sprintf(sql_query, sqlq_qry83, strplu);
	if(db)
		sql_result.init(db, sql_query, true);
	if(sql_result.NumRows() > 0){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("SqlPluLookup: found unknown plu in u_plu table\n");
		}
#endif		
		sql_row=sql_result.GetRow();
		rmsg->hdr.Ex=BHPOS_GRANT;
		strcpy(plur.PLU, strplu);
		strcpy(plur.displayText, sql_row->GetField(4));
		strcpy(plur.recuText, sql_row->GetField(4));
		plur.unitPrice=atof(sql_row->GetField(5));
		
		plur.discount=0.00;
		
		plur.tax=atof(sql_row->GetField(7));
		plur.flags=0;		  
		memcpy(rmsg->data, &plur, sizeof(plur));		  
		rmsg->hdr.size=sizeof(plur);
		return 0;
	}
	else{
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("SqlPluLookup: not found\n");
		}
#endif
		rmsg->hdr.Ex=BHPOS_REFUSE;
		rmsg->hdr.size=0;
		return 1;
	}
	
	return 0;
}

int BarcodeGetNumber(char c)
{
	char buff[2]={c, 0};
	if(c < 48 || c > 57) return -1;
	return atoi(buff);
}

int BarcodeLunBun(char *barcode)
{
	int i, value=0; 
	int len=0;
	
	if(barcode==NULL) return -1;
	len=strlen(barcode);
	
	int uwstart=len%2;
	
	for(i=0; i<=len; i++){
		if(uwstart){	
			value+=BarcodeGetNumber(barcode[i]);
			uwstart=1;
		}
		else{	// even num digits - weight on odd dig
			value+=BarcodeGetNumber(barcode[i]) * 3;
			uwstart=0;
		}
	}
	return (10-(value%10)); 
}

double SqlPluDiscount(bhpos::BhSQL *db, double *price, int qty, int pid)
{
	bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
	char					  sql_query[255];
	int					  linkID;
	double				  margin=0.0;

   // static discount
	sprintf(sql_query, sqlq_qry84, pid);

	if(db){
      if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
		   if(DebugLevel > 2){
			  	printf("Discount query failed: %s\n", db->ErrorMsg());
		   }
#endif
        return 0.00;
      }
            
		if((sql_row=sql_result.GetRow())){
			margin=atof(sql_row->GetField(0));
		}
	}
#ifdef _DEBUGGERY			 
	if(DebugLevel > 2){
		printf("static discount found: %0.2f\n", margin);
	}
#endif

	return margin;
}

double SqlPluPromotion(bhpos::BhSQL *db, double *price, int qty, int pid)
{
	bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
	char					  sql_query[255];
	int					  linkID;
	double				  margin;

	sprintf(sql_query, sqlq_qry85, pid);

	if(db)
		sql_result.init(db, sql_query, true);
	if(sql_result.NumRows() > 0){
		while((sql_row=sql_result.GetRow())){
			margin=atof(sql_row->GetField(5));
			linkID=atoi(sql_row->GetField(7));

			if(atoi(sql_row->GetField(2))==SQL_PROMOTION_PRICEABSOLUTE){
				*price=margin;
				return 0.0;
			}
			else if(atoi(sql_row->GetField(2))==SQL_PROMOTION_PRICEINCREMENTAL){
				return 0.0;
			}
			else if(atoi(sql_row->GetField(2))==SQL_PROMOTION_PRICEMARGIN){
				return margin;
			}
			else if(atoi(sql_row->GetField(2))==SQL_PROMOTION_ITEMABSOLUTE){
				return margin;
			}
			else if(atoi(sql_row->GetField(2))==SQL_PROMOTION_ITEMINCREMENTAL){
				return 0.0;
			}
			else if(atoi(sql_row->GetField(2))==SQL_PROMOTION_ITEMMARGIN){
				return margin;
			}
		}
	}
			
	return 0.0;
}

int SqlPostNoSale(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	int  status, postid, userid;
	char buff[255];
	
	c.SetParameter(BHCI_PARAM_NOSALE, NULL);
	c.GetParameter(BHCI_PARAM_STATIONID, &postid);
	c.GetParameter(BHCI_PARAM_USERID, &userid);
	sprintf(buff, sqlq_qry1, postid, SQL_P_NOSALE);
	status=db->SimpleQuery(buff);
	if(status!=0){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("sqlnosale - Failed log connect message for post %d\n", postid);
		}
#endif
	}
   
   sprintf(buff, sqlq_qry19, postid, SQL_U_NOSALE);
	status=db->SimpleQuery(buff);
	if(status!=0){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("sqlnosale - Failed log connect message for user %d\n", userid);
		}
#endif
	}
	return 0;
}

int SqlPostList(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	BhTaxList			*tl;
	BhTaxListItem	   *tli;
	BhPriceListItem   *pli;
	BhPriceList		   *pl;
	bhpos::BhSQLResult sql_result, a_res;
	bhpos::BhSQLRow	*sql_row, *a_row;
	char					 sql_query[255];
	BhPosMsg			    rmsg;

#ifdef _DEBUGGERY
	printf("SqlPostList message\n");
#endif

	memcpy(&rmsg, msg, sizeof(rmsg));
	rmsg.hdr.Ex=BHPOS_GRANT;
   rmsg.hdr.size=0;
	
	switch(msg->hdr.subcmd){
	case BHMSG_POST_LIST_TAX:
#ifdef _DEBUGGERY
	   printf("SqlPostList message: tax list\n");
#endif
      
		rmsg.hdr.size=sizeof(BhTaxList);
		tl=(BhTaxList *) rmsg.data;
		tli=tl->item;
		if(db)
			sql_result.init(db, sqlq_qry86, true);
		else 
			return 1;

		if((tl->numitems=sql_result.NumRows()) > 0){
			while((sql_row=sql_result.GetRow())){
				sprintf(sql_query, sqlq_qry87, sql_row->GetField(1));
				a_res.init(db, sql_query, true);
				if(a_res.NumRows() > 0){
					a_row=a_res.GetRow();
					tli->ID=atoi(sql_row->GetField(0));
					tli->rate=atof(sql_row->GetField(2));
					strncpy(tli->name, a_row->GetField(0), 31);
					tli++;
					rmsg.hdr.size+=sizeof(BhTaxListItem);
				}
				else{ // no account? - why?
               tli->ID=atoi(sql_row->GetField(0));
					tli->rate=atof(sql_row->GetField(2));
               strcpy(tli->name, "");
               tli++;
					rmsg.hdr.size+=sizeof(BhTaxListItem);
				}
#ifdef _DEBUGGERY
	         printf("SqlPostList Tax List Item: %s, %0.2f\n", sql_row->GetField(2), atof(sql_row->GetField(2)));
#endif            
			}
		}
		else{
			rmsg.hdr.Ex=BHPOS_REFUSE;
			if(OutQ) OutQ->Add(&rmsg);
			return 1;
		}
		break;

	case BHMSG_POST_LIST_PRICE:
#ifdef _DEBUGGERY
	   printf("SqlPostList message: price list\n");
#endif
      
		rmsg.hdr.size=sizeof(BhPriceList);
		pl=(BhPriceList *) rmsg.data;
		pli=pl->item;
		if(db)
			sql_result.init(db, sqlq_qry88, true);
		else 
			return 1;

		if((pl->numitems=sql_result.NumRows()) > 0){
			while((sql_row=sql_result.GetRow())){
				pli->ID=atoi(sql_row->GetField(0));
				strncpy(pli->name, sql_row->GetField(1), 31);
            pli->AccountID=atoi(sql_row->GetField(2));				
            pli++;
				rmsg.hdr.size+=sizeof(BhPriceListItem);
			}
		}
		else{
			rmsg.hdr.Ex=BHPOS_REFUSE;
			if(OutQ) OutQ->Add(&rmsg);
			return 1;
		}
		break;
      
   case BHMSG_POST_LIST_RETURNREASON:
   {
#ifdef _DEBUGGERY
	printf("SqlPostList return reason\n");
#endif      
      BhReturnReasonList *rl=(BhReturnReasonList *) rmsg.data;
      BhReturnReasonItem *rli=rl->item;
   
      rmsg.hdr.size=sizeof(BhReturnReasonList) - sizeof(BhReturnReasonItem);
   
      if(db){
         if(sql_result.init(db, sqlq_qry89, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
		   	if(DebugLevel > 2){
			   	printf("SqlPostList return reason query failed: %s\n", db->ErrorMsg());
		   	}
#endif
            rmsg.hdr.Ex=BHPOS_REFUSE;
			   if(OutQ) OutQ->Add(&rmsg);
			   return 1;
         }
      }
      if((rl->numitems=sql_result.NumRows()) > 0){
			while((sql_row=sql_result.GetRow())){
            rli->ID=atoi(sql_row->GetField(0));
            strncpy(rli->name, sql_row->GetField(1), 63);
            rli++;
            rmsg.hdr.size+=sizeof(BhReturnReasonItem);
         }
      }
      else{
#ifdef _DEBUGGERY
	printf("SqlPostList return reason: no rows\n");
#endif           
			rmsg.hdr.Ex=BHPOS_REFUSE;
			if(OutQ) OutQ->Add(&rmsg);
			return 2;
		}
   }
      break;
   
   case BHMSG_POST_LIST_PAYTYPES:
   {
#ifdef _DEBUGGERY
	printf("SqlPostList pay types\n");
#endif
      BhExPaytypeList *pil=(BhExPaytypeList *) rmsg.data;
      BhExPaytype     *pi=pil->item;
   
      rmsg.hdr.size=sizeof(BhExPaytypeList) - sizeof(BhExPaytype);
   
      if(db){
         if(sql_result.init(db, sqlq_qry162, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
		   	if(DebugLevel > 2){
			   	printf("SqlPostList pay item query failed: %s\n", db->ErrorMsg());
		   	}
#endif
            rmsg.hdr.Ex=BHPOS_REFUSE;
			   if(OutQ) OutQ->Add(&rmsg);
			   return 1;
         }
      }
      if((pil->num_types=sql_result.NumRows()) > 0){
			while((sql_row=sql_result.GetRow())){
            pi->ID=atoi(sql_row->GetField(0));
            strncpy(pi->name, sql_row->GetField(2), 31);
            if(strcmp(sql_row->GetField(3), "Y")==0){
               pi->cashback=true;
            }
            else{
               pi->cashback=false;
            }
            pi->limit=atof(sql_row->GetField(4));
            pi++;
            rmsg.hdr.size+=sizeof(BhExPaytype);
         }
      }
      else{
#ifdef _DEBUGGERY
	printf("SqlPostList pay item: no rows\n");
#endif           
			rmsg.hdr.Ex=BHPOS_REFUSE;
			if(OutQ) OutQ->Add(&rmsg);
			return 2;
		}
   }
      break;
	}
	if(OutQ) OutQ->Add(&rmsg);
	return 0;
}

int SqlPostSignature(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	bhpos::BhSQLResult sql_result;
   bhpos::BhSQLRow   *sql_row;
	BhSignatureCommit *sc=(BhSignatureCommit *) msg->data;
	BhSignatureData	*sd;
	BhSignatureEnd	   *se;
	static int			 left;
	int					 status, id;	
	static int			 temp_file;
	static char		    temp_name[32];
	BhPosMsg			    rmsg;
	char					 buff[256];

	memcpy(&rmsg, msg, sizeof(rmsg));
	rmsg.hdr.size=0;
	rmsg.hdr.Ex=BHPOS_GRANT;

	strcpy(temp_name, "/tmp/bhsigXXXXXXX");

	switch(msg->hdr.subcmd){
	case BHMSG_POST_SIGNATURE_COMMIT:
		c.GetParameter(BHCI_PARAM_USERID, &status);
		sprintf(buff, sqlq_qry90,  status, status);
		sql_result.init(db, buff, true);
   
      // last insert id
      if(sql_result.init(db, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
         if(DebugLevel > 2){
			   printf("Failed to query 145: %s\n", db->ErrorMsg());
         }
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
         rmsg.hdr.size=0;
         if(OutQ) OutQ->Add(&rmsg);
         return 2;
      }
      if(sql_result.NumRows()!=0){
         sql_row=sql_result.GetRow();
         id=atoi(sql_row->GetField(0));
      }
   
		//sprintf(buff, "INSERT into signature VALUES (%d,%d,%d,%d,%d,%d,'')", sc->id, status, sc->xdim, sc->ydim, sc->numstrokes, sc->size);
		//status=db->SimpleQuery(buff);
		left=sc->size;
		temp_file=mkstemp(temp_name);
		if(temp_file){
			status=msg->hdr.size-(sizeof(BhSignatureCommit) - sizeof(int));
			write(temp_file, sc->data, status);
		}
		else{
			rmsg.hdr.Ex=BHPOS_REFUSE;
		}
		break;
	
	case BHMSG_POST_SIGNATURE_DATA:
		sd=(BhSignatureData *) msg->data;
		if(temp_file){
			write(temp_file, sc->data, sd->size);
		}
		else{
			rmsg.hdr.Ex=BHPOS_REFUSE;
		}
		break;
		
	case BHMSG_POST_SIGNATURE_END:
		se=(BhSignatureEnd *) msg->data;
		close(temp_file);
		//sprintf(buff, "UPDATE signature SET data=LOAD_FILE(temp_name) WHERE ID='%d'", sc->id);
		//status=db->SimpleQuery(buff);
		unlink(temp_name);
		break;
		
	case BHMSG_POST_SIGNATURE_RETRIEVE:
	case BHMSG_POST_SIGNATURE_RETDATA:
	case BHMSG_POST_SIGNATURE_RETEND:
		break;
	}
	
	if(OutQ) OutQ->Add(&rmsg);
			
	return 0;
}

int SqlPostStocktake(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	BhStockTakeItem	 *sti=(BhStockTakeItem *) msg->data;
	bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
	int					  num;
	char					 sql_query[255];
	BhPosMsg				rmsg;

	memcpy(&rmsg, msg, sizeof(rmsg));
	rmsg.hdr.size=0;
	rmsg.hdr.Ex=BHPOS_GRANT;

	switch(msg->hdr.subcmd){
	case BHMSG_POST_STOCKTAKE_PLU: 
		break;
	case BHMSG_POST_STOCKTAKE_ITEM:
		sprintf(sql_query, sqlq_qry91, sti->PLU);
		if(db)
			sql_result.init(db, sql_query, true);

		if(sql_result.NumRows() > 0){
			sql_row=sql_result.GetRow();
			num=atoi(sql_row->GetField(0));
			num+=sti->qty;
			sprintf(sql_query, sqlq_qry92, num, sti->PLU);
			if(db)
				sql_result.init(db, sql_query, true);
		}
		else{
			sprintf(sql_query, sqlq_qry93, sti->PLU, sti->qty);
			if(db)
				sql_result.init(db, sql_query, true);
		}
		if(OutQ)
			OutQ->Add(&rmsg);
		break;
	}

	return 0;
}

int SqlPostDelivery(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	bhpos::BhSQLResult  sql_result, sql_res, sql_redt;
	bhpos::BhSQLRow	 *sql_row, *sql_r;
	int					  status, cd_id;
	char					 buff[255];
	BhPosMsg				rmsg;
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	rmsg.hdr.Ex=BHPOS_GRANT;
	
	switch(msg->hdr.subcmd){
	case BHMSG_POST_DELIVERY_ADD:
	{
		BhDeliveryOutList *dl;
		BhDeliveryOut	  *di;
		
		rmsg.hdr.size=sizeof(BhDeliveryOutList) - sizeof(BhDeliveryOut);
	
		dl=(BhDeliveryOutList *) msg->data;
		di=dl->item;
		
		for(int i=0; i < dl->num_items; i++){
			sprintf(buff, sqlq_qry94, SQL_DELIVERY_INTRANSIT, di->delivererID, di->deliveryID);
			if(db){
				status=db->SimpleQuery(buff);
				if(status!=0){
#ifdef _DEBUGGERY				
					if(DebugLevel > 2){
						printf("Failed To update delivery item: %s\n", db->ErrorMsg());
					}
#endif
					rmsg.hdr.Ex=BHPOS_REFUSE;
					rmsg.hdr.size=0;
					if(OutQ) OutQ->Add(&rmsg);
					return 1;
				}
			}
		}
	}
		break;
		
	case BHMSG_POST_DELIVERY_LIST:
	{
		int				 cid, aid; // tid;
		//int				 uid=(int) msg->data;
		BhDeliveryList *dl;
		BhDeliveryItem *di;
	
		rmsg.hdr.size=sizeof(BhDeliveryList) - sizeof(BhDeliveryItem);
	
		dl=(BhDeliveryList *) rmsg.data;
		di=dl->item;

#ifdef _DEBUGGERY
		printf("SqlPostDelivery list\n");
#endif

		strcpy(buff, sqlq_qry95);

		if(db){
			sql_result.init(db, buff, true);
			if(sql_result.NumRows() == 0){
				rmsg.hdr.Ex=BHPOS_REFUSE;
				rmsg.hdr.size=0;
				if(OutQ) OutQ->Add(&rmsg);
				return 1;
			}
			else{
				while((sql_row=sql_result.GetRow())){
					cd_id=atoi(sql_row->GetField(3));
					di->deliveryID=atoi(sql_row->GetField(0));
					di->total=atoi(sql_row->GetField(6));
#ifdef _DEBUGGERY					
					printf("find on custid=%d\n", cd_id);
#endif					
					sprintf(buff, sqlq_qry96, cd_id);
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
#ifdef _DEBUGGERY						
						printf("no customer found\n");
#endif						
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					sql_r=sql_res.GetRow();
					di->customerID=atoi(sql_r->GetField(0));
					cid=atoi(sql_r->GetField(1));
					aid=atoi(sql_r->GetField(2));
					
					sprintf(buff, sqlq_qry97, cid);
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
#ifdef _DEBUGGERY						
						printf("no contact found for id: %d\n", cid);
#endif						
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					sql_r=sql_res.GetRow();
					strncpy(di->fname, sql_r->GetField(0), 32);
					strncpy(di->lname, sql_r->GetField(1), 32);

					sprintf(buff, sqlq_qry98, aid);
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
#ifdef _DEBUGGERY						
						printf("no address found for id: %d\n", aid);
#endif						
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					sql_r=sql_res.GetRow();
					strcpy(buff, sql_r->GetField(0));
					strcat(buff, ", ");
					strcat(buff, sql_r->GetField(1));
					strcat(buff, ", ");
					strcat(buff, sql_r->GetField(2));
					strncpy(di->address, buff, 64);
					/*					
					sprintf(buff, "SELECT transactionID FROM transaction WHERE ID='%d'", atoi(sql_row->GetField(4)));
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
						printf("no transaction found\n");
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					sql_r=sql_res.GetRow();
					t_id=
					
					printf(buff, "SELECT ID FROM transactionitems WHERE transactionID='%d'", tid);
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
						printf("no transaction found\n");
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					di->num_items=sql_res.NumRows();
					
					printf(buff, "SELECT ID FROM transactionitems WHERE transactionID='%d'", tid);
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
						printf("no transaction found\n");
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					 */		  
			 
					rmsg.hdr.size+=sizeof(BhDeliveryItem);
				}			
				if(OutQ) OutQ->Add(&rmsg);
			}
		}
		else{
			rmsg.hdr.Ex=BHPOS_REFUSE;
			rmsg.hdr.size=0;
			if(OutQ) OutQ->Add(&rmsg);
			return 1;
		}
	}
		return 0;

	case BHMSG_POST_DELIVERY_STATUS:
		break;
	case BHMSG_POST_DELIVERY_RETURNLIST:
	{
		int				 cid, aid; //tid;
		BhDeliveryList *dl;
		BhDeliveryItem *di;
	
		rmsg.hdr.size=sizeof(BhDeliveryList) - sizeof(BhDeliveryItem);
	
		dl=(BhDeliveryList *) rmsg.data;
		di=dl->item;

#ifdef _DEBUGGERY
		printf("SqlPostDelivery list\n");
#endif

		sprintf(buff, "SELECT * FROM delivery WHERE status='%d'", SQL_DELIVERY_INTRANSIT);

		if(db){
			sql_result.init(db, buff, true);
			if(sql_result.NumRows() == 0){
				rmsg.hdr.Ex=BHPOS_REFUSE;
				rmsg.hdr.size=0;
				if(OutQ) OutQ->Add(&rmsg);
				return 1;
			}
			else{
				while((sql_row=sql_result.GetRow())){
					cd_id=atoi(sql_row->GetField(3));
					di->deliveryID=atoi(sql_row->GetField(0));
					di->total=atoi(sql_row->GetField(6));
					
					sprintf(buff, sqlq_qry96, cd_id);
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
						//printf("no customer found\n");
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					sql_r=sql_res.GetRow();
					di->customerID=atoi(sql_r->GetField(0));
					cid=atoi(sql_r->GetField(1));
					aid=atoi(sql_r->GetField(2));
					
					sprintf(buff, sqlq_qry97, cid);
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
						//printf("no contact found\n");
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					sql_r=sql_res.GetRow();
					strncpy(di->fname, sql_r->GetField(0), 32);
					strncpy(di->lname, sql_r->GetField(1), 32);

					sprintf(buff, sqlq_qry98, aid);
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
						// printf("no address found\n");
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					sql_r=sql_res.GetRow();
					strcpy(buff, sql_r->GetField(0));
					strcat(buff, ", ");
					strcat(buff, sql_r->GetField(1));
					strcat(buff, ", ");
					strcat(buff, sql_r->GetField(2));
					strncpy(di->address, buff, 64);
					/*					
					sprintf(buff, "SELECT transactionID FROM transaction WHERE ID='%d'", atoi(sql_row->GetField(4)));
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
						printf("no transaction found\n");
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					sql_r=sql_res.GetRow();
					t_id=
					
					printf(buff, "SELECT ID FROM transactionitems WHERE transactionID='%d'", tid);
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
						printf("no transaction found\n");
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					di->num_items=sql_res.NumRows();
					
					printf(buff, "SELECT ID FROM transactionitems WHERE transactionID='%d'", tid);
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
						printf("no transaction found\n");
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					 */		  
			 
					rmsg.hdr.size+=sizeof(BhDeliveryItem);
				}			
				if(OutQ) OutQ->Add(&rmsg);
			}
		}
		else{
			rmsg.hdr.Ex=BHPOS_REFUSE;
			rmsg.hdr.size=0;
			if(OutQ) OutQ->Add(&rmsg);
			return 1;
		}
	}
		return 0;

	case BHMSG_POST_DELIVERY_DUE:
		// retreive transaction
		
		break;
	case BHMSG_POST_DELIVERY_PAYMENT:
		break;
	case BHMSG_POST_DELIVERY_UNDELIVERED:
		break;
	}
	
	return 0;
}

int SqlPostRental(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	bhpos::BhSQLResult  sql_result, sql_res;
	bhpos::BhSQLRow	 *sql_row, *sql_r;
	char					  buff[255];//, buff2[64], temp[16];
	BhRentalList		 *rl;
	BhRentalItem		 *ri;
	BhPosMsg				  rmsg;
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	rmsg.hdr.Ex=BHPOS_GRANT;
	
	switch(msg->hdr.subcmd){
	case BHMSG_POST_RENTAL_ADD:
		break;
	case BHMSG_POST_RENTAL_LIST:
	{
		int tid, coid, aid;
		
		rl=(BhRentalList *) msg->data;
		ri=rl->item;
		
		rmsg.hdr.size=sizeof(BhRentalList) - sizeof(BhRentalItem);
	
#ifdef _DEBUGGERY
		printf("SqlPostRental list\n");
#endif

		strcpy(buff, sqlq_qry103);
		
		/*
		ri->transactionID;
		ri->outat;
		ri->outfrom[32];
		ri->duration
		ri->description[32];
		*/

		if(db){
			sql_result.init(db, buff, true);
			if(sql_result.NumRows() == 0){
				rmsg.hdr.Ex=BHPOS_REFUSE;
				rmsg.hdr.size=0;
				if(OutQ) OutQ->Add(&rmsg);
				return 1;
			}
			else{
				while((sql_row=sql_result.GetRow())){
					ri->rentalID=atoi(sql_row->GetField(0));
					tid=atoi(sql_row->GetField(1));
					ri->customerID=atoi(sql_row->GetField(2));
					// pull up the customer
					sprintf(buff, sqlq_qry104, ri->customerID);
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
#ifdef _DEBUGGERY
						printf("no customer found\n");
#endif
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					sql_r=sql_res.GetRow();
					coid=atoi(sql_r->GetField(0));
					aid=atoi(sql_r->GetField(1));
					
					sprintf(buff, sqlq_qry97, coid);
					sql_res.init(db, buff, true);
					if(sql_res.NumRows() == 0){
#ifdef _DEBUGGERY
						printf("no contact found\n");
#endif
						rmsg.hdr.Ex=BHPOS_REFUSE;
						rmsg.hdr.size=0;
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
					sql_r=sql_res.GetRow();
					strncpy(ri->fname, sql_r->GetField(0), 32);
					strncpy(ri->lname, sql_r->GetField(1), 32);
					
					// calculate duration
					// "SELECT (TO_DAYS(NOW()) - TO_DAYS(out))"
					
					// get rate from prodpriceqty
					
					// amount due (minus deposit)
				}
			}
		}
	}
		break;
	case BHMSG_POST_RENTAL_STATUS:
	case BHMSG_POST_RENTAL_RETURN:
	case BHMSG_POST_RENTAL_DUE:
		break;
	case BHMSG_POST_RENTAL_PAYMENT:
		// change rental status
		// add payment item to transaction
		// close transaction
		break;
	}
	
	return 0;
}

int SqlPostCustomer(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	bhpos::BhSQLResult   sql_result, sql_res, sql_resp;
	bhpos::BhSQLRow	   *sql_row, *sql_r;
	BhCustomerLookup	   *cu=(BhCustomerLookup *) msg->data;
	int					   tfname, tlname, cd_id, cdd_id; //tphone;
	char					   buff[255], buff2[64];//, temp[16];
	BhCustomerList	      *cl;
	BhCustomerItem	      *ci;
   BhCustomerCharBal    *ccb=(BhCustomerCharBal *) msg->data;
   BhCustomerCharBalRet *ccbr;
	BhPosMsg				    rmsg;
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	rmsg.hdr.Ex=BHPOS_GRANT;

#ifdef _DEBUGGERY
	printf("SqlPostCustomer message\n");
#endif	
	
	switch(msg->hdr.subcmd){
	case BHMSG_POST_CUSTOMER_NEW:
   {
      ci=(BhCustomerItem *) msg->data;
      // insert into address
      if(db){
         sprintf(buff, sqlq_qry147, 0, 0, ci->address1,ci->address2,ci->address3,ci->city,ci->state,ci->zip,ci->country);
			sql_result.init(db, buff, true);
         if(sql_result.init(db, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
           if(DebugLevel > 2){
	            printf("Failed to query sql_qry145: %s\n", db->ErrorMsg());
            }
#endif
           rmsg.hdr.Ex=BHPOS_REFUSE;
            rmsg.hdr.size=0;
            if(OutQ) OutQ->Add(&rmsg);
            return 2;
         }
         sql_row=sql_result.GetRow();
         int addressid=atoi(sql_row->GetField(0));

         // insert contact details (telephone)
         sprintf(buff, sqlq_qry149,0,1,ci->phone);
			sql_result.init(db, buff, true);
         
         if(sql_result.init(db, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
           if(DebugLevel > 2){
	            printf("Failed to query sql_qry145: %s\n", db->ErrorMsg());
            }
#endif
            rmsg.hdr.Ex=BHPOS_REFUSE;
            rmsg.hdr.size=0;
            if(OutQ) OutQ->Add(&rmsg);
            return 2;
         }
         sql_row=sql_result.GetRow();
         int contactdetailsid=atoi(sql_row->GetField(0));
         sprintf(buff, sqlq_qry152,contactdetailsid);
			sql_result.init(db, buff, true);
         
         // insert contact
         sprintf(buff, sqlq_qry148,0,contactdetailsid,0,addressid,0,"",ci->fname,ci->lname,"","","",0,"N",0);         
			sql_result.init(db, buff, true);
         if(sql_result.init(db, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
           if(DebugLevel > 2){
	            printf("Failed to query sql_qry145: %s\n", db->ErrorMsg());
            }
#endif
            rmsg.hdr.Ex=BHPOS_REFUSE;
            rmsg.hdr.size=0;
            if(OutQ) OutQ->Add(&rmsg);
            return 2;
         }
         sql_row=sql_result.GetRow();
         int contactid=atoi(sql_row->GetField(0));

         // insert charge account (if required)
         sprintf(buff, sqlq_qry150,0,0,0,0.0,"",0,"Y","N","N","1","N",0,0,"");
         sql_result.init(db, buff, true);
         if(sql_result.init(db, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
           if(DebugLevel > 2){
	            printf("Failed to query sql_qry145: %s\n", db->ErrorMsg());
            }
#endif
            rmsg.hdr.Ex=BHPOS_REFUSE;
            rmsg.hdr.size=0;
            if(OutQ) OutQ->Add(&rmsg);
            return 2;
         }
         sql_row=sql_result.GetRow();
         int chargeaccid=atoi(sql_row->GetField(0));

         // get a cmid
         c.GetParameter(BHCI_PARAM_USERID, &cd_id);
   		sprintf(buff, sqlq_qry90,  cd_id, cd_id);
	   	sql_result.init(db, buff, true);
         if(sql_result.init(db, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
            if(DebugLevel > 2){
			      printf("Failed to query 145: %s\n", db->ErrorMsg());
            }
#endif
            rmsg.hdr.Ex=BHPOS_REFUSE;
            rmsg.hdr.size=0;
            if(OutQ) OutQ->Add(&rmsg);
            return 2;
         }
         int cmid=0;
         if(sql_result.NumRows()!=0){
            sql_row=sql_result.GetRow();
            cmid=atoi(sql_row->GetField(0));
         }

         // insert customer
         sprintf(buff, sqlq_qry151,cmid,contactid,0,0,0,chargeaccid,0,"");
         sql_result.init(db, buff, true);
         if(sql_result.init(db, sqlq_qry145, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
           if(DebugLevel > 2){
	            printf("Failed to query sql_qry145: %s\n", db->ErrorMsg());
            }
#endif
            rmsg.hdr.Ex=BHPOS_REFUSE;
            rmsg.hdr.size=0;
            if(OutQ) OutQ->Add(&rmsg);
            return 2;
         }
         sql_row=sql_result.GetRow();
         int customerid=atoi(sql_row->GetField(0));

         // return customerID
         memcpy(rmsg.data, &customerid, sizeof(int));
         rmsg.hdr.size=sizeof(int);
         if(OutQ) OutQ->Add(&rmsg);
      }
   }
		break;
		
	case BHMSG_POST_CUSTOMER_LIST:
	{
		rmsg.hdr.size=sizeof(BhCustomerList) - sizeof(BhCustomerItem);
	
		cl=(BhCustomerList *) rmsg.data;
		ci=cl->item;

#ifdef _DEBUGGERY
		printf("SqlPostCustomer list: %s, %s - %s\n", cu->fname, cu->lname, cu->phone);
#endif
	
		strcpy(buff, "SELECT * FROM contacts WHERE ");
		if(strcmp(cu->fname, "")==0){
			tfname=0;
		}
		else{
			tfname=1;
			//if(strtok(cu->fname, "*")){
			//	sprintf(buff2, "FirstName=SOUNDEX(%s) ", cu->fname);
			//}
			//else{
				sprintf(buff2, "FirstName='%s' ", cu->fname);
			//}
			strcat(buff, buff2);
		}
	
		if(strcmp(cu->lname, "")==0){
			tlname=0;
		}
		else{
			tlname=1;
			//if(strtok(cu->fname, "*")){
			//	sprintf(buff2, "AND LastName=SOUNDEX(%s) ", cu->lname);
			//}
			//else{
				sprintf(buff2, "AND LastName='%s' ", cu->lname);
			//}
			strcat(buff, buff2);
		}
		
		//printf("query: %s\n", buff);
		
		if(db){
			sql_result.init(db, buff, true);
			if(sql_result.NumRows() == 0){
				rmsg.hdr.Ex=BHPOS_REFUSE;
				rmsg.hdr.size=0;
				if(OutQ) OutQ->Add(&rmsg);
				return 1;
			}
			else if(sql_result.NumRows() == 1){ // ok - lookup customer
#ifdef _DEBUGGERY			
				printf("1 row found\n");
#endif
				sql_row=sql_result.GetRow();
				cd_id=atoi(sql_row->GetField(1));
#ifdef _DEBUGGERY				
				printf("found on contactid=%d\n", cd_id);
#endif
				sprintf(buff, sqlq_qry106, cd_id);
				sql_res.init(db, buff, true);
				if(sql_res.NumRows() == 0){
#ifdef _DEBUGGERY					
					printf("no customer found\n");
#endif
					rmsg.hdr.Ex=BHPOS_REFUSE;
					rmsg.hdr.size=0;
					if(OutQ) OutQ->Add(&rmsg);
					return 1;
				}
				//memset(ci, 0, sizeof(BhCustomerLookup));
				sql_r=sql_res.GetRow();
				ci->customerID=atoi(sql_r->GetField(0));
				strncpy(ci->fname, cu->fname, 32);
				strncpy(ci->lname, cu->lname, 32);
				strncpy(ci->phone, cu->phone, 32);
				strcpy(ci->address1, "");
				strcpy(ci->zip, "");
				rmsg.hdr.size+=sizeof(BhCustomerItem);
				if(OutQ) OutQ->Add(&rmsg);
				return 0;
			}
			else{ // more than one row
				if(strcmp(cu->phone, "")==0){
#ifdef _DEBUGGERY					
					printf("No phone\n");
#endif
					// send off list
					cl->num_items=sql_result.NumRows();
					while((sql_row=sql_result.GetRow())){
						cd_id=atoi(sql_row->GetField(1));
						cdd_id=atoi(sql_row->GetField(2));
#ifdef _DEBUGGERY						
						printf("cd_id: %d\n", cd_id);
#endif
						sprintf(buff, sqlq_qry106, cd_id);
						sql_res.init(db, buff, true);
						if(sql_res.NumRows() == 0){
#ifdef _DEBUGGERY							
							printf("no customer rows\n");
#endif
							rmsg.hdr.Ex=BHPOS_REFUSE;
							rmsg.hdr.size=0;
							if(OutQ) OutQ->Add(&rmsg);
							return 1;
						}
						sql_r=sql_res.GetRow();
						ci->customerID=atoi(sql_r->GetField(0));
						strncpy(ci->fname, cu->fname, 32);
						strncpy(ci->lname, cu->lname, 32);
						//printf("cust - %d: %s, %s\n", ci->customerID, ci->fname, ci->lname);
						
						// need a third lookup for phone
						sprintf(buff, sqlq_qry108, cdd_id);
						sql_res.init(db, buff, true);
						if(sql_res.NumRows() != 0){
							sql_r=sql_res.GetRow();
							strncpy(ci->phone, sql_r->GetField(0), 32);
						}
						
						//strncpy(ci->phone, cu->phone, 32);
						strcpy(ci->address1, "");
						strcpy(ci->zip, "");
						ci++;
						rmsg.hdr.size+=sizeof(BhCustomerItem);
					}
					if(OutQ) OutQ->Add(&rmsg);
					return 1;
				}
				else{
					sprintf(buff, sqlq_qry109, cu->phone);
					sql_result.init(db, buff, true);
					if(sql_result.NumRows() == 0){
						if(strcmp(cu->phone, "")==0){
							rmsg.hdr.Ex=BHPOS_REFUSE;
							rmsg.hdr.size=0;
							if(OutQ) OutQ->Add(&rmsg);
							return 1;
						}
					}
					else{
						cl->num_items=sql_result.NumRows();
						while((sql_row=sql_result.GetRow())){
							cd_id=atoi(sql_row->GetField(0));
			
							sprintf(buff, sqlq_qry106, cd_id);
							sql_res.init(db, buff, true);
							if(sql_res.NumRows() == 0){
							rmsg.hdr.Ex=BHPOS_REFUSE;
								rmsg.hdr.size=0;
								if(OutQ) OutQ->Add(&rmsg);
								return 1;
							}
							memset(ci, 0, sizeof(BhCustomerLookup));
							sql_r=sql_res.GetRow();
							ci->customerID=atoi(sql_r->GetField(0));
							strncpy(ci->fname, cu->fname, 32);
							strncpy(ci->lname, cu->lname, 32);
							strncpy(ci->phone, cu->phone, 32);
							strcpy(ci->address1, "");
							strcpy(ci->zip, "");
							
							rmsg.hdr.size+=sizeof(BhCustomerItem);
						}
						if(OutQ) OutQ->Add(&rmsg);
						return 1;
					}
				}
			}
		}
		else{
			rmsg.hdr.Ex=BHPOS_REFUSE;
			rmsg.hdr.size=0;
			if(OutQ) OutQ->Add(&rmsg);
			return 1;
		}
	}
	break;
   
   case BHMSG_POST_CUSTOMER_DISCOUNT:
   case BHMSG_POST_CUSTOMER_PRODUCTS:
      break;
   
   case BHMSG_POST_CUSTOMER_CHARGEBAL:
      //sprintf(buff, transq_qry34, ccb->customerID);
   
      sprintf(buff, sqlq_qry153, ccb->customerID);
      if(db){
			if(sql_result.init(db, buff, true)!=DB_COMMAND_OK){
				rmsg.hdr.Ex=BHPOS_REFUSE;
				rmsg.hdr.size=0;
				if(OutQ) OutQ->Add(&rmsg);
				return 1;
			}
         if(sql_result.NumRows() == 0){
            rmsg.hdr.Ex=BHPOS_REFUSE;
				rmsg.hdr.size=0;
				if(OutQ) OutQ->Add(&rmsg);
				return 1;
         }
         
         ccbr->customerID=ccb->customerID;
         
         sql_row=sql_result.GetRow();
         cd_id=atoi(sql_row->GetField(0));
         ccbr->limit=atof(sql_row->GetField(1));
         
         // charge amounts...
         sprintf(buff, sqlq_qry154, cd_id);
         if(sql_result.init(db, buff, true)!=DB_COMMAND_OK){
				rmsg.hdr.Ex=BHPOS_REFUSE;
				rmsg.hdr.size=0;
				if(OutQ) OutQ->Add(&rmsg);
				return 1;
			}
         if(sql_result.NumRows() == 0){
            ccbr->old_balance=0.0;
            ccbr->new_balance=ccb->amount;
				return 1;
         }
         else{
            sql_row=sql_result.GetRow();
            ccbr->old_balance=atof(sql_row->GetField(0));
            ccbr->new_balance=atof(sql_row->GetField(0)) + ccb->amount;
         }
      }
      break;
	}	  
	
	return 0;
}

int SqlPostBlindBalance(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	BhPosMsg					    rmsg;
	char							 sql_query[255], start[32], end[32];
	bhpos::BhSQLResult		 sql_result;
	bhpos::BhSQLRow			*sql_row;
	int							 currid, postid;
	BhBlindBalance			   *bb=(BhBlindBalance *) msg->data;
	BhBlindBalanceReply	   *bbrep=(BhBlindBalanceReply *) rmsg.data;
	BhBlindBalanceFullReply *bbr;
	BhPosMonetaryUnit		   *mu;
	double						 a;
		
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
   
   c.GetParameter(BHCI_PARAM_STATIONID, &postid);
	
	switch(msg->hdr.subcmd){
	case BHMSG_POST_BLINDBALANCE_TRANSAMOUNT:
	{
      bhpos::BhSQLResult  sql_result;
	   bhpos::BhSQLRow	 *sql_row;
      double              opening=0.0, closing=0.0;
      
      if(sql_result.init(db, sqlq_qry20, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   		if(DebugLevel > 2){
			   printf("Failed to select now: %s\n", db->ErrorMsg());
   		}
#endif
   		rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 1;
	   }

	   sql_row=sql_result.GetRow();
		strcpy(end, sql_row->GetField(0));
		strcpy(start, sql_row->GetField(0));
		strcpy(&start[11], "00:01:01");
      if(SqlDrawerQuery(db, c, start, end, &opening, &closing)){
         rmsg.hdr.Ex=BHPOS_REFUSE;
   		rmsg.hdr.size=0;
   		if(OutQ) OutQ->Add(&rmsg);
   		return 2;
      }
      bbrep->opening_balance=opening;
      bbrep->closing_balance=closing;
      if(!SqlPayQuery(db, start, end, 1, postid, &currid, &a)){
			rmsg.hdr.Ex=BHPOS_GRANT;
		   rmsg.hdr.size=sizeof(BhBlindBalanceReply);
		   if(OutQ) OutQ->Add(&rmsg);
      }
	}	
		break;
		
	case BHMSG_POST_BLINDBALANCE_CONTENTS:
		bbr=(BhBlindBalanceFullReply *) &rmsg.data;
		mu=bbr->unit;
	
		c.GetParameter(BHCI_PARAM_CURRENCYID, &currid);
	
		rmsg.hdr.size=(sizeof(BhBlindBalanceReply)-sizeof(BhPosMonetaryUnit));
		rmsg.hdr.Ex=BHPOS_GRANT;
	
		sprintf(sql_query, sqlq_qry111, currid);
		sql_result.init(db, sql_query, true);
			
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("SqlQ:Blind Balance: Qry:MonetaryUnitID returned %d rows for ID %d\n",sql_result.NumRows(), currid);
		}
#endif			
	
		bbr->num_units=sql_result.NumRows();
	
		if(sql_result.NumRows()!=0){
			sql_row=sql_result.GetRow();
			int muid=atoi(sql_row->GetField(0));
			sql_result.free();
			sprintf(sql_query, sqlq_qry8, muid);
			sql_result.init(db, sql_query, true);
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("SqlQ Blind Balance: MonetaryUnits id:%d returned %d rows\n",muid, sql_result.NumRows());
			}
#endif				
			if(sql_result.NumRows()!=0){
				double v; 
				int	 u;
				int	 a;
				while((sql_row=sql_result.GetRow())){
					u=atoi(sql_row->GetField(2));
					v=atof(sql_row->GetField(3));
					a=atoi(sql_row->GetField(4));
					mu->unit=u;
					mu->value=v;
					mu->amount=a;
					mu->operation=BHPOS_CASHOPERATION_SET;
						
#ifdef _DEBUGGERY
					if(DebugLevel > 2){
						printf("SqlQ Blind Balance: Adding Monetary Unit: %d - %f - %d\n",mu->unit, mu->value, mu->amount);
					}
#endif				
					mu++;	 
					rmsg.hdr.size+=sizeof(BhPosMonetaryUnit);
				}
			}
			else{
				rmsg.hdr.Ex=BHPOS_REFUSE;
				rmsg.hdr.size=0;
			}
		}
		else{
			rmsg.hdr.Ex=BHPOS_REFUSE;
			rmsg.hdr.size=0;
		}
		if(OutQ) OutQ->Add(&rmsg);
		
		break;
	}
	return 0;
}


int SqlPostEphemeral(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	bhpos::BhSQLResult  sql_result, sql_res;
	bhpos::BhSQLRow	 *sql_row;
	int					  rx_id, uid;
	char					  buff[512];
	BhPosMsg				  rmsg;
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
   
   c.GetParameter(BHCI_PARAM_USERID, &uid);
	
	rmsg.hdr.Ex=BHPOS_GRANT;

#ifdef _DEBUGGERY
	printf("SqlPostEphemeral message\n");
#endif

	switch(msg->hdr.subcmd){
	case BHMSG_POST_EPHEMERAL_LOOKUP:
	{
#ifdef _DEBUGGERY
	if(DebugLevel > 2) printf("SqlPostEphemeral: lookup\n");
#endif	
		BhTransEphemLookup       *ep=(BhTransEphemLookup *) msg->data;
		BhTransEphemLookupReturn *er=(BhTransEphemLookupReturn *) rmsg.data;
      BhTransEphemLookupError  *eerr=(BhTransEphemLookupError *) rmsg.data;
		er->flags=0;
		
		sprintf(buff, sqlq_qry113, ep->PLU);
			
		if(db){
			if(sql_result.init(db, buff, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			
				if(DebugLevel > 2) printf("SqlEphemeral: query failed: %s\n", db->ErrorMsg());
#endif
				rmsg.hdr.Ex=BHPOS_REFUSE;
            eerr->error=0;
				rmsg.hdr.size=rmsg.hdr.size=sizeof(BhTransEphemLookupError);
				if(OutQ) OutQ->Add(&rmsg);
				return 1;
			}
			if(sql_result.NumRows() == 0){
#ifdef _DEBUGGERY			
				if(DebugLevel > 2) printf("SqlEphemeral: no eplu rows for: %s\n", ep->PLU);
#endif
            eerr->error=0;
				rmsg.hdr.Ex=BHPOS_REFUSE;
				rmsg.hdr.size=rmsg.hdr.size=sizeof(BhTransEphemLookupError);
				if(OutQ) OutQ->Add(&rmsg);
				return 2;
			}
			else if(sql_result.NumRows() == 1){ // ok - have single item
				sql_row=sql_result.GetRow();
            int z=atoi(sql_row->GetField(3));
            if(z!=SQL_PRESCRIPTION_FILLED){
               if(z==SQL_PRESCRIPTION_GENERATED){
                  eerr->error=1;
               }
               if(z==SQL_PRESCRIPTION_READY){
                  eerr->error=2;
               }
               if(z==SQL_PRESCRIPTION_COLLECTED){
                  eerr->error=3;
               }
               rmsg.hdr.Ex=BHPOS_REFUSE;
				   rmsg.hdr.size=sizeof(BhTransEphemLookupError);
				   if(OutQ) OutQ->Add(&rmsg);
               return 3;
            }
            
				er->rx_id=atoi(sql_row->GetField(1));
				er->price=atoi(sql_row->GetField(4));
				strncpy(er->rxnumber, sql_row->GetField(2), 32);
				SqlUserModifiedWhoLookup(db, atoi(sql_row->GetField(0)), er->fname, er->lname);
				
				sprintf(buff, sqlq_qry114, er->rx_id);
				if(sql_result.init(db, buff, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			
					if(DebugLevel > 2) printf("SqlEphemeral: qty query failed: %s\n", db->ErrorMsg());
#endif
					rmsg.hdr.Ex=BHPOS_REFUSE;
					rmsg.hdr.size=0;
					if(OutQ) OutQ->Add(&rmsg);
					return 3;
				}
				sql_row=sql_result.GetRow();
				er->price*=atof(sql_row->GetField(0));
				
				rmsg.hdr.Ex=BHPOS_GRANT;
				rmsg.hdr.size=sizeof(BhTransEphemLookupReturn);
				if(OutQ) OutQ->Add(&rmsg);
			}
		}
	}
		break;
		
	case BHMSG_POST_EPHEMERAL_COMMIT:
      rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
		break;
		
	case BHMSG_POST_EPHEMERAL_PLU:
	{
		BhTransEphemPluCommit *ec=(BhTransEphemPluCommit *) msg->data;
		BhTransEphemItem		 *eci=ec->items;
		
      int num=ec->num_items;
		
      while(num){
         //eci->PLU[16];
         //eci->flags;
		   sprintf(buff, sqlq_qry115, eci->rxid);
         if(db){
			   if(sql_result.init(db, buff, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			
				   if(DebugLevel > 2) printf("SqlEphemeral: commit error: %s\n", db->ErrorMsg());
#endif
   				rmsg.hdr.Ex=BHPOS_REFUSE;
	   			rmsg.hdr.size=0;
		   		if(OutQ) OutQ->Add(&rmsg);
			   	return 1;
            }
			}
         eci++;
         num--;
      }
		rmsg.hdr.size=0;
		if(OutQ) OutQ->Add(&rmsg);
	}
		break;
   
	case BHMSG_POST_EPHEMERAL_END:
      rmsg.hdr.size=0;
      if(OutQ) OutQ->Add(&rmsg);
		break;
	}
	
	return 0;	
}

int SqlPostDiscrepancy(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   char sql_query[255], eaid[32];
   int  userid;

#ifdef _DEBUGGERY			 
	if(DebugLevel > 2){
		printf("SqlPostDiscrepancy\n");
	}
#endif   
   
   c.GetParameter(BHCI_PARAM_USERID, &userid);
   c.GetEnteredAt(eaid);
   
   bhpos::BhSQLResult  sql_result;
   
	switch(msg->hdr.subcmd){
	case BHMSG_POST_DISCREPANCY_BLINDBALANCE:
   {
      BhDiscrepancy *dis=(BhDiscrepancy *) msg->data;
      db->SimpleQuery(sqlq_qry70);
      sprintf(sql_query, sqlq_qry116, eaid, userid, SQL_D_BLINDBAL, dis->amount);
      if(db->SimpleQuery(sql_query)){
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("discrep blind bal commit failed\n");
			}
#endif
         db->SimpleQuery(sqlq_qry146);
		}
      db->SimpleQuery(sqlq_qry77);
   }         

      break;
   }
	return 0;
}

int SqlPostAttendance(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	char		 sql_query[255];
	int		  userid=0, status;
	BhClockIn *clkin=(BhClockIn *) msg->data;
	BhPosMsg	rmsg;
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
	
	rmsg.hdr.Ex=BHPOS_GRANT;
	rmsg.hdr.size=0;

	userid=SqlUserLookup(db, clkin->type, clkin->data, &status);

	if(userid==0){
		rmsg.hdr.Ex=BHPOS_REFUSE;
		if(OutQ) OutQ->Add(&rmsg);
	}

   db->SimpleQuery(sqlq_qry70);
	switch(msg->hdr.subcmd){
	case BHMSG_POST_ATTENDANCE_CLOCKIN:
		if(c.UserExists(userid)){
			sprintf(sql_query, sqlq_qry19, userid, SQL_U_CLOCKOUT);
			c.RemoveUser(userid);
		}
		else{
			c.UserStartClock(userid, true);
			sprintf(sql_query, sqlq_qry19, userid, SQL_U_CLOCKIN);
		}
		
	case BHMSG_POST_ATTENDANCE_CLOCKOUT:
		sprintf(sql_query, sqlq_qry19, userid, SQL_U_CLOCKOUT);
		c.RemoveUser(userid);
		break;
	}
	
	status=db->SimpleQuery(sql_query);
	if(status!=0){
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("sqlattendance - Failed log clock message for user %d\n", userid);
		}
#endif
      db->SimpleQuery(sqlq_qry146);
	}
   db->SimpleQuery(sqlq_qry77);
	if(OutQ) OutQ->Add(&rmsg);
	return 0;
}

int SqlPostAdjust(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
	int 		           i, userid;
	BhPosMsg            rmsg;
   char					  sql_query[256], eaid[32];
	bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
	
	memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
   rmsg.hdr.size=0;
   rmsg.hdr.Ex=BHPOS_GRANT;
   
   c.GetParameter(BHCI_PARAM_USERID, &userid);
   c.GetEnteredAt(eaid);

#ifdef _DEBUGGERY
			printf("SqlPostAdjust\n");
#endif
   
   db->SimpleQuery(sqlq_qry70);
   
	switch(msg->hdr.subcmd){
	case BHMSG_POST_ADJUST_CASHDRAWER:
	{
		bhpos::CMoney 		m;
      bhpos::CCash 		adj;
      double            t;
		BhCashDrawerAdjust *cda=(BhCashDrawerAdjust *) msg->data;
		BhPosMonetaryUnit  *mu=cda->unit;
   
   	for(i=0; i < cda->num_units; i++){
#ifdef _DEBUGGERY
			printf("cash adjust: %d - %0.2f - %d\n", mu->unit, mu->value, mu->amount);
#endif
			m.Set(mu->unit, mu->value, mu->amount);
			adj.Add(m);
			mu++;
		}
      t=adj.GetTotal();
      sprintf(sql_query, sqlq_qry120, eaid, userid, SQL_A_CDEDIT, t);
      if((db->SimpleQuery(sql_query))){
#ifdef _DEBUGGERY          
         if(DebugLevel > 2){
            printf("Failed insert adj\n");
         }
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
         db->SimpleQuery(sqlq_qry146);
	      if(OutQ) OutQ->Add(&rmsg);
      }
	}
      break;
   
   case BHMSG_POST_ADJUST_PAYIN:
   {
      BhPayinAdjust *pi=(BhPayinAdjust *)msg->data;
   
      sprintf(sql_query, sqlq_qry120, eaid, userid, SQL_A_PAYIN, pi->amount);
      if((db->SimpleQuery(sql_query))){
#ifdef _DEBUGGERY          
         if(DebugLevel > 2){
            printf("Failed payin insert adj\n");
         }
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
         db->SimpleQuery(sqlq_qry146);
	      if(OutQ) OutQ->Add(&rmsg);
      }
   }
      break;
   
   case BHMSG_POST_ADJUST_PAYOUT:
   {
      BhPayoutAdjust *po=(BhPayoutAdjust *)msg->data;
      
      sprintf(sql_query, sqlq_qry120, eaid, userid, SQL_A_PAYOUT, po->amount);
      if((db->SimpleQuery(sql_query))){
#ifdef _DEBUGGERY          
         if(DebugLevel > 2){
            printf("Failed payout insert adj \n");
         }
#endif
         rmsg.hdr.Ex=BHPOS_REFUSE;
         db->SimpleQuery(sqlq_qry146);
	      if(OutQ) OutQ->Add(&rmsg);
      }
   }
      break;
	}
   
   db->SimpleQuery(sqlq_qry77);
   
	if(OutQ) OutQ->Add(&rmsg);
	
	return 0;
}

int SqlMenuLoad(bhpos::BhSQL *db, PLUServer *p, BhPosMsg	*rmsg)
{
   int  id, mid, menuid;
   char buff[256];
   bhpos::BhSQLResult  sql_result, sql_res, sql_resu;
   bhpos::BhSQLRow	 *sql_row, *sql_r, *sql_ro;
	
   sprintf(buff, sqlq_qry123, menuid);
   if(sql_result.init(db, buff, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
      if(DebugLevel > 2){
         printf("Sqlq: bad menu query: %s\n", db->ErrorMsg());
      }
#endif			
   }
   else if(sql_result.NumRows()==1){
      sql_row=sql_result.GetRow();
      menuid=atoi(sql_row->GetField(0));
      bhpos::CServSideMenu menu(menuid);
      
      sprintf(buff, sqlq_qry124, id);
      if(sql_res.init(db, buff, true)==DB_COMMAND_OK){
         while((sql_r=sql_res.GetRow())){
            mid=atoi(sql_row->GetField(0));
            bhpos::CServSideProduct menuitem(mid, sql_row->GetField(1), sql_row->GetField(2), sql_row->GetField(3), atoi(sql_row->GetField(4)), atof(sql_row->GetField(5)));
            sprintf(buff, sqlq_qry125, mid);
            if(sql_resu.init(db, buff, true)==DB_COMMAND_OK){
               while((sql_ro=sql_resu.GetRow())){
                  bhpos::CServSideProductOption po(atoi(sql_ro->GetField(0)),sql_ro->GetField(1),sql_ro->GetField(2));
                  menuitem.AddOption(po);
               }
            }
            menu.Add(menuitem);
         }
         p->ssts.AddMenu(menu);
      }       
   }
   else{
      // bum query - more than one menu returned
   }
   return 0;
}

int SqlGoods(bhpos::BhSQL *db, PLUServer *p, bhpos::CClientInfo& c, BhPosMsg *msg)
{
   BhPosMsg           rmsg;
   bhpos::BhSQLResult sql_result, sql_res;
   bhpos::BhSQLRow	*sql_row, *sql_r;
   BhGoodsInList     *gl=(BhGoodsInList *) msg->data;
   BhGoodsInListItem *gli=gl->items;
   char               buff[256];
   int                n, poid;
   
   memset(&rmsg, 0, sizeof(rmsg));
	memcpy(&rmsg, msg, sizeof(rmsg));
   rmsg.hdr.size=0;
   rmsg.hdr.Ex=BHPOS_GRANT;
   rmsg.hdr.size=sizeof(BhGoodsInList) - sizeof(BhGoodsInListItem);
   
   if(msg->hdr.cmd==BHMSG_GOODS_IN){
      switch(msg->hdr.subcmd){
      case BHMSG_GOODS_IN_POLIST:
         sprintf(buff, "select purchaseorderid,purchasenumber,supplierid from purchaseorder where statusid='%d'", 3);
         if(sql_result.init(db, buff, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
            if(DebugLevel > 2) printf("goods module: po list error: %s\n", db->ErrorMsg());
#endif
   	      rmsg.hdr.Ex=BHPOS_REFUSE;
            rmsg.hdr.size=0;
            if(OutQ) OutQ->Add(msg);
            return 1;
         }
         
         n=sql_result.NumRows();
         gl->num_items=n;
         if(n==0){
            rmsg.hdr.Ex=BHPOS_GRANT;
            if(OutQ) OutQ->Add(msg);
            return 0;
         }
   
         while((sql_row=sql_result.GetRow())){
            strncpy(gli->ponumber, sql_row->GetField(1), 63);
            gli->supplierid=atoi(sql_row->GetField(2));
            sprintf(buff, "select name from supplier where id='%d'", gli->supplierid);
            if(sql_res.init(db, buff, true)!=DB_COMMAND_OK){
            }
            sql_r=sql_res.GetRow();
            strncpy(gli->suppliername, sql_r->GetField(0), 63);
            poid=atoi(sql_r->GetField(0));
            sprintf(buff, "select count(*) from purchaseorderitems where purchaseorderid='%d'", poid);
            if(sql_res.init(db, buff, true)!=DB_COMMAND_OK){
            }
            sql_r=sql_res.GetRow();
            gli->num_items=atoi(sql_r->GetField(0));
            rmsg.hdr.size+=sizeof(BhGoodsInListItem);
            gli++;
         }
         if(OutQ) OutQ->Add(msg);
         return 0;
         
      case BHMSG_GOODS_IN_POFETCH:
      case BHMSG_GOODS_IN_POCOMMIT:
	      break;
      }
   }
   else if(msg->hdr.cmd==BHMSG_GOODS_OUT){
      switch(msg->hdr.subcmd){
      case BHMSG_GOODS_OUT_ORDERLIST:
      case BHMSG_GOODS_OUT_ORDERFETCH:
      case BHMSG_GOODS_OUT_ORDERCOMMIT:
	      break;
      }
   }
   else if(msg->hdr.cmd==BHMSG_GOODS_TRANSFERIN){
      switch(msg->hdr.subcmd){
      case BHMSG_GOODS_TRANSFERIN_LIST:
      case BHMSG_GOODS_TRANSFERIN_FETCH:
      case BHMSG_GOODS_TRANSFERIN_COMMIT:
	      break;
      }
   }
   else if(msg->hdr.cmd==BHMSG_GOODS_TRANSFEROUT){
      switch(msg->hdr.subcmd){
      case BHMSG_GOODS_TRANSFEROUT_LIST:
      case BHMSG_GOODS_TRANSFEROUT_FETCH:
      case BHMSG_GOODS_TRANSFEROUT_COMMIT:
	      break;
      }
   }
   return 0;
}

///////////////////////////////////////////////////////////////////////////////////////
//
//	A few helper funcs for commonly requested things
//
//
int SqlInterfaceLookup(bhpos::BhSQL *db, int type, int iid, void *data, BhPosMsg	*rmsg)
{
	BhLogonReply				lr;
	int							InterfaceID=0, empID=0, LoginID=0, AuthID=0, PostID=0, LoginMethodID=0;
	char							sql_query[1024];
	bhpos::BhSQLResult		sql_result;
	bhpos::BhSQLRow		  *sql_row;
	char						   temp[16];

#ifdef _DEBUGGERY
	if(DebugLevel > 2)			
		printf("SqlQ: SqlSysMsgInterface\n");
#endif
	
	if(type){	
      empID=SqlUserLookup(db, type, data, &LoginID);
      /*SqlUserNameLookup(db, empID, char *fname, char *lname)
			
		strcpy(temp, sql_row->GetField(0));
		strcat(temp, " ");
		strcat(temp, sql_row->GetField(1));
		if(strlen(temp) > 12){
			temp[12]=0;
		}
		strcpy(lr.userdispname, temp);
		*/
		// Also from users we get an index into the userauth table to see what
		// interfaces they are allowed.  We need to check this against the
		// posts interface table to see where we are at.
		sprintf(sql_query, sqlq_qry17, LoginID, type, PostID);
		// This will give us either no common interface, or an authority record
		// for each common interface.  For the common interfaces, we check that
		// the perm_login flag is set to Y before we allow a login.
#ifdef _DEBUGGERY
		if (DebugLevel > 3)
			printf("SqlLogin: %s\n", sql_query);
#endif
		if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
			return 6;
		}
		if (sql_result.NumRows() == 0) {
#ifdef _DEBUGGERY		 
			if(DebugLevel > 2){
				printf("SqlLogin: No common interfaces\n");
			}
#endif
			goto loopout;
		}
		bool login_ok = false;
		while ((sql_row = sql_result.GetRow())) {
			InterfaceID = atoi(sql_row->GetField(0));
			AuthID = atoi(sql_row->GetField(1));
#ifdef _DEBUGGERY
			if (DebugLevel > 3)
				printf("SqlLogin: IF=%d Auth=%d PermLogin=%s\n", InterfaceID, AuthID, sql_row->GetField(4));
#endif
			if (strncmp(sql_row->GetField(4), "Y", 1) == 0) {
				login_ok = true;
				break;
			}
		}
		if(!login_ok){
#ifdef _DEBUGGERY		 
			if(DebugLevel > 2){
				printf("SqlLogin: No Authority to login\n");
			}
#endif
			goto loopout;
		}
		rmsg->hdr.Ex=BHPOS_GRANT;
		rmsg->hdr.size=sizeof(BhLogonReply);
		lr.interfaceID=InterfaceID;
		memcpy(rmsg->data, &lr, sizeof(BhLogonReply));
#ifdef _DEBUGGERY
		if(DebugLevel > 2){
			printf("granting login\n");
		}
#endif									
		goto loopout;
	}
   
loopout:   
   
   return 0;
}

int SqlUserLookup(bhpos::BhSQL *db, int type, void *data, int *lid)
{
	int					      empID=0, LoginID, LoginMethodID;
	BhLogonRequestEx		  *lreq=(BhLogonRequestEx *) data;
	BhLogonRequest			  *l=(BhLogonRequest *) data;
	BhLogonRequestBaroode  *lb=(BhLogonRequestBaroode *) data;
	BhLogonRequestMagswipe *lm=(BhLogonRequestMagswipe *) data;
	char					      sql_query[256];
	bhpos::BhSQLResult      sql_result;
	bhpos::BhSQLRow	     *sql_row;
	
	if(type < 128){	 
		sprintf(sql_query, sqlq_qry14, type);
		if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
			if(DebugLevel > 2){
				printf("SqlUserLookup: loginmethod query failed: %s\n", db->ErrorMsg());
			}
#endif
			return 0;
		}
		if(sql_result.NumRows() == 1){  // Should be either one row or none.
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("SqlUserLookup: found loginmethod\n");
			}
#endif			
			sql_row = sql_result.GetRow();
			LoginMethodID = atoi(sql_row->GetField(0));
			char * query_tmp = 0;
			int ltype = atoi(sql_row->GetField(1));
			char * method;
			switch(type){
				case BHMSG_SYSTEM_LOGIN_LOGIN:
            case BHMSG_SYSTEM_LOGIN_LOGINEXT:
					method = "LOGIN";
               query_tmp = SqlSysFormatLoginSQL(LoginMethodID, ltype, l->name, l->password);
					break;
				case BHMSG_SYSTEM_LOGIN_LIPASSIVE:
					method = "PASSIVE";
               query_tmp = SqlSysFormatLoginSQL(LoginMethodID, ltype, l->name, 0);
					break;
				case BHMSG_SYSTEM_LOGIN_BARCODE:
					method = "BARCODE";
               query_tmp = SqlSysFormatLoginSQL(LoginMethodID, ltype, lb->barcode, 0);
					break;
				case BHMSG_SYSTEM_LOGIN_MAGSWIPE:
					method = "MAGSWIPE";
               query_tmp = SqlSysFormatLoginSQL(LoginMethodID, ltype, lm->track1, 0);
					break;
				default:
					method = "UNKNOWN";
					break;
			}
			if(query_tmp)
				strcpy(sql_query, query_tmp);
			else{
#ifdef _DEBUGGERY		 
				if(DebugLevel > 2) {
					printf("SqlUserLookup: Invalid %s loginmethod search type %d\n", method, ltype);
			   }
#endif
   		   return 0;
   		}
			if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK || sql_result.NumRows() != 1){
#ifdef _DEBUGGERY
				if(DebugLevel > 2){
					printf("SqlUserLookup: loginmethod type query failed: %s\n", db->ErrorMsg());
				}
#endif
				return 0;
			}
			sql_row = sql_result.GetRow();
			LoginID = atoi(sql_row->GetField(1));
         *lid=LoginID;
			sprintf(sql_query, sqlq_qry15, LoginID);
			if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY
				if(DebugLevel > 2){
					printf("SqlUserLookup: user query failed: %s\n", db->ErrorMsg());
				}
#endif
				return 0;
			}
			if(sql_result.NumRows() != 1){
#ifdef _DEBUGGERY		 
				if(DebugLevel > 2){
					printf("SqlUserLookup: no user with login:%s and password:%s\n", l->name, l->password);
				}
#endif
            return 0;
			}
#ifdef _DEBUGGERY			 
			if(DebugLevel > 2){
				printf("SqlUserLookup: found user\n");
			}
#endif			
			sql_row=sql_result.GetRow();
			empID=atoi(sql_row->GetField(6));
		}
	}

	return empID;
}

int SqlUserNameLookup(bhpos::BhSQL *db, int empID, char *fname, char *lname)
{
	char					  sql_query[64];
	bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
	
	sprintf(sql_query, sqlq_qry16, empID);
	if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
		return 1;
	}
					
	if(sql_result.NumRows() == 1){
		sql_row=sql_result.GetRow();
		strncpy(fname, sql_row->GetField(0), 32);
		strncpy(lname, sql_row->GetField(1), 32);
	}
	else{
		return 3;
	}
	return 0;
}

int SqlUserCreatedWhoLookup(bhpos::BhSQL *db, int cmID, char *fname, char *lname)
{
	char					  sql_query[64];
	bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
	
	sprintf(sql_query, sqlq_qry129, cmID);
	if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
		return 4;
	}
					
	if(sql_result.NumRows() == 1){
		sql_row=sql_result.GetRow();
		return SqlUserNameLookup(db, atoi(sql_row->GetField(0)), fname, lname);
	}
	else{
		return 5;
	}
	return 0;
}

int SqlUserModifiedWhoLookup(bhpos::BhSQL *db, int cmID, char *fname, char *lname)
{
	char					  sql_query[64];
	bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
	
	sprintf(sql_query, sqlq_qry130, cmID);
	if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
		return 4;
	}
					
	if(sql_result.NumRows() == 1){
		sql_row=sql_result.GetRow();
		return SqlUserNameLookup(db, atoi(sql_row->GetField(0)), fname, lname);
	}
	else{
		return 5;
	}
	return 0;
}

int SqlCreatedWhenLookup(bhpos::BhSQL *db, int cmID, BhTimestamp *ts)
{
	char					  sql_query[64];
	bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
	
	sprintf(sql_query, sqlq_qry131, cmID);
	if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
		return 1;
	}
					
	if(sql_result.NumRows() == 1){
		sql_row=sql_result.GetRow();
		BhDateTime dt(BHTSFORMAT_MYSQL_TS8, sql_row->GetField(0));
		dt.ToBhTimeStamp(ts);
	}
	else{
		return 2;
	}
	return 0;
}

int SqlModifiedWhenLookup(bhpos::BhSQL *db, int cmID, BhTimestamp *ts)
{
	char					  sql_query[64];
	bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
	
	sprintf(sql_query, sqlq_qry132, cmID);
	if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
		return 1;
	}
					
	if(sql_result.NumRows() == 1){
		sql_row=sql_result.GetRow();
		BhDateTime dt(BHTSFORMAT_MYSQL_TS8, sql_row->GetField(0));
		dt.ToBhTimeStamp(ts);
	}
	else{
		return 2;
	}
	return 0;
}

int SqlErrorMsgFill(bhpos::CClientInfo& c, BhErrorMsg *errmsg)
{
	int                 currid;
   char                eaid[32];
   
   if(errmsg==NULL) return 1;
      
   c.GetEnteredAt(errmsg->enteredAt);
   c.GetParameter(BHCI_PARAM_USERID, &errmsg->userID);
   
	return 0;
}


int SqlTimeQuery(bhpos::BhSQL *db, char *start, char *end, int empID, int *num, double *amount)
{
   /*
   char					 query[256];
	bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow	*sql_row;
   
   sprintf(query, "SELECT COUNT(ID), SUM(stamper) FROM employeelog WHERE \
		(UNIX_TIMESTAMP(EnteredAt) > UNIX_TIMESTAMP('%s')) AND \
	   (UNIX_TIMESTAMP(EnteredAt) < UNIX_TIMESTAMP('%s')) AND \
	   payType='%d' AND EnteredStation='%d'", start, end, type, postID);
	if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
  		if(DebugLevel > 2){
		   printf("Failed to query 9: %s\n", db->ErrorMsg());
  		}
#endif
  		return 1;
   }
   if(sql_result.NumRows()!=0){
  		sql_row=sql_result.GetRow();
  		*num=atoi(sql_row->GetField(0));
  		*amount=atof(sql_row->GetField(1));
	}*/
   return 0;
}


int SqlDrawerQuery(bhpos::BhSQL *db, bhpos::CClientInfo& c, char *start, char *end, double *opening, double *closing)
{
   char					  sql_query[512];
   int                 currid;
   char                eaid[32];
   double              payin=0.0, payout=0.0, openadj=0.0, cdedit=0.0, opens=0.0, transt=0.0, returnt=0.0, change=0.0;
   bhpos::BhSQLResult  sql_result;
	bhpos::BhSQLRow	 *sql_row;
   
   c.GetParameter(BHCI_PARAM_STATIONID, &currid);
   c.GetEnteredAt(eaid);
      
  sprintf(sql_query, sqlq_qry133, start, end, SQL_A_PAYIN, eaid);
   if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
 		if(DebugLevel > 0){
		   printf("SqlDrawerQuery: Failed to query 1: %s\n", db->ErrorMsg());
  		}
#endif
  		return 1;
   }
	if(sql_result.NumRows()!=0){
   	sql_row=sql_result.GetRow();
      if(sql_row->GetField(0)!=NULL)
   	   payin=atof(sql_row->GetField(0));
#ifdef _DEBUGGERY			 
 		printf("payin query: %f\n", payin);
#endif      
	}
      
   sprintf(sql_query, sqlq_qry134, start, end, SQL_A_PAYOUT, eaid);
   if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   	if(DebugLevel > 2){
		   printf("SqlDrawerQuery: Failed to query 2: %s\n", db->ErrorMsg());
   	}
#endif
   		return 2;
	}
	if(sql_result.NumRows()!=0){
   	sql_row=sql_result.GetRow();
      if(sql_row->GetField(0)!=NULL)
   	   payout=atof(sql_row->GetField(0));
	}
   
   sprintf(sql_query, sqlq_qry135, start, end, SQL_A_CHANGE, eaid);
   if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   	if(DebugLevel > 2){
		   printf("SqlDrawerQuery: Failed to query 2: %s\n", db->ErrorMsg());
   	}
#endif
   		return 2;
	}
	if(sql_result.NumRows()!=0){
   	sql_row=sql_result.GetRow();
      if(sql_row->GetField(0)!=NULL)
   	   change=atof(sql_row->GetField(0));
	}
   printf("change query: %f\n", change);
   
   sprintf(sql_query, sqlq_qry136, start, end, SQL_A_OPENBAL, eaid);
   if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   	if(DebugLevel > 2){
		   printf("SqlDrawerQuery: Failed to query 3: %s\n", db->ErrorMsg());
   	}
#endif
   	return 3;
	}
	if(sql_result.NumRows()!=0){
   	sql_row=sql_result.GetRow();
      if(sql_row->GetField(0)!=NULL)
   	   openadj=atof(sql_row->GetField(0));
	}
   
   sprintf(sql_query, sqlq_qry137, start, end, eaid);
   if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   	if(DebugLevel > 2){
		   printf("SqlDrawerQuery: Failed to query 4: %s\n", db->ErrorMsg());
   	}
#endif
   	return 4;
	}
	if(sql_result.NumRows()!=0){
   	sql_row=sql_result.GetRow();
      if(sql_row->GetField(0)!=NULL)
   	   transt=atof(sql_row->GetField(0));
	}
   
   sprintf(sql_query, sqlq_qry138, start, end, eaid);
   if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   	if(DebugLevel > 2){
		   printf("SqlDrawerQuery: Failed to query 5: %s\n", db->ErrorMsg());
   	}
#endif
   	return 5;
	}
	if(sql_result.NumRows()!=0){
   	sql_row=sql_result.GetRow();
      if(sql_row->GetField(0)!=NULL)
   	   returnt=atof(sql_row->GetField(0));
	} 

   
   sprintf(sql_query, sqlq_qry139, start, end, SQL_A_CDEDIT, eaid);
   if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   	if(DebugLevel > 2){
		   printf("Failed to query: %s\n", db->ErrorMsg());
   	}
#endif
   	return 4;
	}
   if(sql_result.NumRows()!=0){
  		sql_row=sql_result.GetRow();
      if(sql_row->GetField(0)!=NULL)
  		   cdedit=atof(sql_row->GetField(0));
	}
   c.GetParameter(BHCI_PARAM_CURRENCYID, &currid);		
   sprintf(sql_query, sqlq_qry140, currid);
	if(sql_result.init(db, sql_query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
   	if(DebugLevel > 2){
		   printf("Failed to query: %s\n", db->ErrorMsg());
   	}
#endif
   	return 5;
	}
	
#ifdef _DEBUGGERY			 
	if(DebugLevel > 2){
		printf("Qry:MonetaryUnitID returned %d rows for ID %d\n",sql_result.NumRows(), currid);
	}
#endif			
			
	if(sql_result.NumRows()!=0){
		sql_row=sql_result.GetRow();
		int muid=atoi(sql_row->GetField(0));
		sql_result.free();
		sprintf(sql_query, sqlq_qry8, muid);
		sql_result.init(db, sql_query, true);
#ifdef _DEBUGGERY			 
		if(DebugLevel > 2){
			printf("Qry:MonetaryUnits id:%d returned %d rows\n",muid, sql_result.NumRows());
		}
#endif				
		if(sql_result.NumRows()!=0){
			opens=0.0; 
         double v;
         int a;
			while((sql_row=sql_result.GetRow())){
			   if(sql_row->GetField(3)!=NULL) v=atof(sql_row->GetField(3));
				a=atoi(sql_row->GetField(4));
            opens+=(v * a);
			}
		}
	}
   *opening=opens-openadj;
   *closing=*opening + payin + returnt - payout - change - transt;
   return 0;
}

int SqlPayQuery(bhpos::BhSQL *db, char *start, char *end, int type, int postID, int *num, double *amount)
{
   char					 query[256];
	bhpos::BhSQLResult sql_result;
	bhpos::BhSQLRow	*sql_row;
   
   sprintf(query, sqlq_qry142, start, end, type, postID);
	if(sql_result.init(db, query, true)!=DB_COMMAND_OK){
#ifdef _DEBUGGERY			 
  		if(DebugLevel > 2){
		   printf("Failed to query 9: %s\n", db->ErrorMsg());
  		}
#endif
  		return 1;
   }
   if(sql_result.NumRows()!=0){
  		sql_row=sql_result.GetRow();
  		*num=atoi(sql_row->GetField(0));
      if(sql_row->GetField(1))
  		   *amount=atof(sql_row->GetField(1));
	}
   return 0;
}
